#ifndef __DRIVERS_RS485_HPP__
#define __DRIVERS_RS485_HPP__

#include "stl/stl_config.hpp"

#if defined(STL_ENABLE_UART) && (STL_ENABLE_UART == 1)

#include "drivers/board/gpio.hpp"
#include "drivers/board/uart.hpp"

#include "stl/error.hpp"
#include "stl/fstream.hpp"
#include "stl/functional.hpp"
#include "stl/memory.hpp"
#include "stl/mutex.hpp"
#include "stl/set.hpp"
#include "stl/stddef.hpp"
#include "stl/thread.hpp"
#include "stl/time.hpp"
#include "stl/timer.hpp"

namespace stl::driver::board {
class RS485;
class RS485SlaveBase;

using RS485Callable = void (*)(const stl::RunnableBase::Params* params,
                               RS485* const instance);

using RS485Runnable = stl::Runnable<RS485Callable, RS485*>;

class RS485 // TODO: Find slaves by tag (slave list)? "evse", "rfid"
  : public UART
  , public RS485Runnable
{
public:
  using PollingTimeT = time::Millis;

private:
  using SlaveContainerT = Set<RS485SlaveBase*>;

public:
  RS485()
    : RS485Runnable("RS485")
    , m_lock("rs485")
  {
  }

  RS485(const RS485&) = delete;

  RS485(RS485&& other)
    : UART(stl::move(static_cast<UART&>(other)))
    , RS485Runnable(stl::move(static_cast<RS485Runnable&>(other)))
    , m_slaveContainer(stl::move(other.m_slaveContainer))
  {
  }

  RS485& operator=(const RS485&) = delete;

  RS485& operator=(RS485&& other)
  {
    static_cast<UART&>(*this) = stl::move(static_cast<UART&>(other));
    static_cast<RS485Runnable&>(*this) =
      stl::move(static_cast<RS485Runnable&>(other));
    m_slaveContainer = stl::move(other.m_slaveContainer);

    return *this;
  }

  virtual ~RS485();

  bool start(filesystem::Path path, const UARTConfig& uartConfig);

  bool running() const;

  void stop();

  // TODO: Move
  static UInt16T modbusCRC(const UInt8T* data, const SizeT size)
  {
    Assert(data);

    UInt16T crc = 0xFFFFU;

    for (SizeT i = 0; i < size; i++) {
      crc ^= CU16(data[i]); // XOR byte into least sig. byte of crc

      for (SizeT j = 0; j < 8U; ++j) { // Loop over each bit
        if ((crc & 0x0001U) != 0U) {   // If the LSB is set
          crc >>= 1;                   // Shift right and XOR 0xA001
          crc ^= 0xA001U;
        } else {     // Else LSB is not set
          crc >>= 1; // Just shift right
        }
      }
    }

    // Note, this number has low and high bytes swapped, so use it accordingly
    // (or swap bytes)
    return crc;
  }

  // bool detachSlave(RS485SlaveBase* const slave);

  bool attachSlave(RS485SlaveBase* const slave);

private:
  friend class RS485SlaveBase;

  // bool detachSlaveInternal(RS485SlaveBase* slave);

  bool detachAllSlaveInternal();

  static void worker(const stl::RunnableBase::Params* params,
                     RS485* const instance);

  mutable Mutex m_lock;

  SlaveContainerT m_slaveContainer;
};

class RS485SlaveBase
  : protected virtual stream::IStreamBase
  , protected virtual stream::OStreamBase
{
public:
  RS485SlaveBase(RS485& bus, const RS485::PollingTimeT& pollingTime);

  RS485SlaveBase(const RS485SlaveBase&) = delete;

  RS485SlaveBase(RS485SlaveBase&& other)
    : m_online(other.m_online)
    , m_bus(other.m_bus)
    , m_pollingTimer(stl::move(other.m_pollingTimer))
    , m_hwConfigured(other.m_hwConfigured)
    , m_detach(other.m_detach)
    , m_attached(other.m_attached)
  {
  }

  RS485SlaveBase& operator=(const RS485SlaveBase&) = delete;

  RS485SlaveBase& operator=(RS485SlaveBase&& other)
  {
    m_online = other.m_online;
    m_bus = other.m_bus;
    m_pollingTimer = stl::move(other.m_pollingTimer);
    m_hwConfigured = other.m_hwConfigured;

    return *this;
  }

  virtual ~RS485SlaveBase();

  bool isOpen() const;

  bool online() const
  {
    stl::LockGuard<stl::Mutex> lock(m_lock);

    return m_online;
  }

  bool configured() const {
    stl::LockGuard<stl::Mutex> lock(m_lock); return m_hwConfigured; }

  RS485* bus() { return m_bus; }

  virtual ErrorE configureHardware() = 0;
  virtual void deinitHardware() = 0;
  virtual stl::ErrorE poll() = 0;

  void detach()
  {
    stl::LockGuard<stl::Mutex> lock(m_lock);

    m_detach = true;
  }

  bool detachedSet() const
  {
    bool retv = false;

    {
      stl::LockGuard<stl::Mutex> lock(m_lock);

      retv = !m_attached;
    }

    return retv;
  }

protected:
  virtual SizeT write(const UInt8T* const data, SizeT size) override;

  virtual SizeT read(UInt8T* data, SizeT maxSize) override;

  ErrorE configureHardwareInternal()
  {
    const ErrorE retv = configureHardware();

    {
      stl::LockGuard<stl::Mutex> lock(m_lock);

      if (ErrorE::Ok == retv) {
        m_hwConfigured = true;
      }
    }

    return retv;
  }

  SizeT available() const
  {
    stl::LockGuard<stl::Mutex> lock(m_lock);

    return m_bus->available();
  }

  virtual void flush() override;

  bool m_online;
  RS485* m_bus;

private:
  bool detachSet() const
  {
    bool retv = false;

    {
      stl::LockGuard<stl::Mutex> lock(m_lock);

      retv = m_detach;
    }

    return retv;
  }

  void detached()
  {
    stl::LockGuard<stl::Mutex> lock(m_lock);

    m_attached = false;
  }

  virtual bool seek(const SizeT offset, const SeekModeE seekMode) override
  {
    return false;
  }

  virtual SizeT position() const override { return 0U; }

  virtual SizeT size() const override { return 0U; }

  mutable stl::Mutex m_lock;

  timer::Countdown m_pollingTimer;
  bool m_hwConfigured;
  bool m_detach;
  bool m_attached;

  friend class RS485;
};
}

#endif // defined(STL_ENABLE_UART) && (STL_ENABLE_UART == 1)
#endif // __DRIVERS_RS485_HPP__
