#ifndef __IONT_UART_HPP__
#define __IONT_UART_HPP__

#include "stl/stl_config.hpp"

#if defined(STL_ENABLE_UART) && (STL_ENABLE_UART == 1)

// #define STL_ENABLE_UART_DEBUG 1

#if defined(STL_ENABLE_UART_DEBUG) && (STL_ENABLE_UART_DEBUG == 1)
#include "log/logging.hpp"
#endif

#include "stl/fstream.hpp"
#include "stl/stdint.hpp"
#include "stl/stream_base.hpp"
#include "stl/time.hpp"

namespace stl::driver::board {
enum class UARTRetvE
{
  Ok,
  Error,
  InvalidValue
};

enum class UARTParityE
{
  None,
  Odd,
  Even
};

enum class UARTStopBitsE
{
  One,
  Two
};

enum class UARTCharSizeE
{
  C8Bits,
  C7Bits,
  C6Bits,
  C5Bits,
};

struct UARTConfig
{
  UInt32T baudrate;
  UARTParityE parity;
  UARTStopBitsE stopBits;
  UARTCharSizeE charSize;
};

class UART : public stream::FStream
{
public:
  UART();

  UART(filesystem::Path path,
       const stream::FileFlagsT flags,
       const UARTConfig& config);

  UART(const UART&) = delete;

  UART(UART&& other)
    : stream::FStream(stl::move(static_cast<stream::FStream&>(other)))
    , m_rTimeout(other.m_rTimeout)
    , m_wTimeout(other.m_wTimeout)
  {
  }

  UART& operator=(const UART&) = delete;

  UART& operator=(UART&& other)
  {
    static_cast<stream::FStream&>(*this) =
      stl::move(static_cast<stream::FStream&>(other));
    m_rTimeout = stl::move(other.m_rTimeout);
    m_wTimeout = stl::move(other.m_wTimeout);
    return *this;
  }

  virtual ~UART();

  bool open(filesystem::Path path,
            const stream::FileFlagsT flags,
            const UARTConfig& config);

  SizeT available() const;

  virtual void flush() override;

  SizeT write(const UInt8T* const data, SizeT size)
  {
#if defined(STL_ENABLE_UART_DEBUG) && (STL_ENABLE_UART_DEBUG == 1)
    logging::log << logging::LvlE::Debug << "DATA TO WRITE: " << stl::endl;
    logging::log.dump(data, size, 20U, true);
#endif

    setModeTransmit();

    const stl::SizeT dataWritten = stream::FStream::write(data, size);

#if defined(STL_ENABLE_UART_DEBUG) && (STL_ENABLE_UART_DEBUG == 1)
    logging::log << logging::LvlE::Debug << "DATA WRITTEN: " << dataWritten << stl::endl;
#endif

    return dataWritten;
  }

  SizeT read(UInt8T* data, SizeT maxSize)
  {
    setModeReceive();

    const stl::SizeT dataRead = stream::FStream::read(data, maxSize);

#if defined(STL_ENABLE_UART_DEBUG) && (STL_ENABLE_UART_DEBUG == 1)
    logging::log << logging::LvlE::Debug << "DATA READ: " << stl::endl;
    logging::log.dump(data, dataRead, 20U, true);
#endif

    return dataRead;
  }

  void setRTimeout(time::Millis timeout) { m_rTimeout = move(timeout); }

  void setWTimeout(time::Millis timeout) { m_wTimeout = move(timeout); }

protected:
  bool transmitterInit();

  void setModeReceive();

  void setModeTransmit();

private:
  time::Millis m_rTimeout;
  time::Millis m_wTimeout;
};
}

#endif // defined(STL_ENABLE_UART) && (STL_ENABLE_UART == 1)

#endif
