#ifndef __STL_DRIVER_LED_LED_STRIP_HPP__
#define __STL_DRIVER_LED_LED_STRIP_HPP__

#include "stl/assert.hpp"
#include "stl/compiler.hpp"
#include "stl/error.hpp"
#include "stl/math.hpp"
#include "stl/memory.hpp"
#include "stl/stdint.hpp"
#include "stl/time.hpp"
#include "stl/vector.hpp"

#include "log/logging.hpp"

#undef RGB

namespace stl::color {
using ColorComponentT = UInt8T;

class RGB
{
public:
  enum ColorE
  {
    RED,
    GREEN,
    BLUE
  };

  RGB() { stl::memory::fill<ColorComponentT>(0U, 3U, m_color); }

  RGB(const ColorComponentT r, const ColorComponentT g, const ColorComponentT b)
  {
    m_color[RED] = r;
    m_color[GREEN] = g;
    m_color[BLUE] = b;
  }

  RGB(const RGB& other)
  {
    stl::memory::copy(other.m_color, other.m_color + 3U, m_color);
  }

  RGB(RGB&& other)
  {
    stl::memory::copy(other.m_color, other.m_color + 3U, m_color);
  }

  RGB& operator=(const RGB& other)
  {
    stl::memory::copy(other.m_color, other.m_color + 3U, m_color);

    return *this;
  }

  RGB& operator=(RGB&& other)
  {
    stl::memory::copy(other.m_color, other.m_color + 3U, m_color);

    return *this;
  }

  ColorComponentT red() const { return m_color[RED]; }

  ColorComponentT green() const { return m_color[GREEN]; }

  ColorComponentT blue() const { return m_color[BLUE]; }

  void red(const ColorComponentT value) { m_color[RED] = value; }

  void green(const ColorComponentT value) { m_color[GREEN] = value; }

  void blue(const ColorComponentT value) { m_color[BLUE] = value; }

  RGB& operator/=(ColorComponentT n)
  {
    m_color[RED] /= n;
    m_color[GREEN] /= n;
    m_color[BLUE] /= n;

    return *this;
  }

  RGB operator/(ColorComponentT n)
  {
    return RGB(m_color[RED] / n, m_color[GREEN] / n, m_color[BLUE] / n);
  }

  RGB& operator*=(ColorComponentT n)
  {
    m_color[RED] *= n;
    m_color[GREEN] *= n;
    m_color[BLUE] *= n;

    return *this;
  }

  RGB operator*(ColorComponentT n)
  {
    return RGB(m_color[RED] * n, m_color[GREEN] * n, m_color[BLUE] * n);
  }

  RGB& operator+=(ColorComponentT n)
  {
    m_color[RED] += n;
    m_color[GREEN] += n;
    m_color[BLUE] += n;

    return *this;
  }

  RGB operator+(ColorComponentT n)
  {
    return RGB(m_color[RED] + n, m_color[GREEN] + n, m_color[BLUE] + n);
  }

  RGB& operator-=(ColorComponentT n)
  {
    m_color[RED] -= n;
    m_color[GREEN] -= n;
    m_color[BLUE] -= n;

    return *this;
  }

  RGB operator-(ColorComponentT n)
  {
    return RGB(m_color[RED] - n, m_color[GREEN] - n, m_color[BLUE] - n);
  }

private:
  ColorComponentT m_color[3];
};

class RGBA
{
public:
  enum ColorE
  {
    RED,
    GREEN,
    BLUE,
    ALPHA
  };

  RGBA() { stl::memory::fill<ColorComponentT>(0U, 4U, m_color); }

  RGBA(const ColorComponentT r,
       const ColorComponentT g,
       const ColorComponentT b,
       const ColorComponentT a)
  {
    m_color[RED] = r;
    m_color[GREEN] = g;
    m_color[BLUE] = b;
    m_color[ALPHA] = a;
  }

  RGBA(const RGBA& other)
  {
    stl::memory::copy(other.m_color, other.m_color + 4U, m_color);
  }

  RGBA(RGBA&& other)
  {
    stl::memory::copy(other.m_color, other.m_color + 4U, m_color);
  }

  RGBA& operator=(const RGBA& other)
  {
    stl::memory::copy(other.m_color, other.m_color + 4U, m_color);

    return *this;
  }

  RGBA& operator=(RGBA&& other)
  {
    stl::memory::copy(other.m_color, other.m_color + 4U, m_color);

    return *this;
  }

  RGBA& operator=(const RGB& other)
  {
    m_color[RED] = other.red();
    m_color[GREEN] = other.green();
    m_color[BLUE] = other.blue();
    m_color[ALPHA] = m_color[ALPHA];

    return *this;
  }

  ColorComponentT red() const { return m_color[RED]; }

  ColorComponentT green() const { return m_color[GREEN]; }

  ColorComponentT blue() const { return m_color[BLUE]; }

  ColorComponentT alpha() const { return m_color[ALPHA]; }

  void red(const ColorComponentT value) { m_color[RED] = value; }

  void green(const ColorComponentT value) { m_color[GREEN] = value; }

  void blue(const ColorComponentT value) { m_color[BLUE] = value; }

  void alpha(const ColorComponentT value) { m_color[ALPHA] = value; }

  RGBA& operator/=(ColorComponentT n)
  {
    m_color[RED] /= n;
    m_color[GREEN] /= n;
    m_color[BLUE] /= n;

    return *this;
  }

  RGBA operator/(ColorComponentT n)
  {
    return RGBA(
      m_color[RED] / n, m_color[GREEN] / n, m_color[BLUE] / n, m_color[ALPHA]);
  }

  RGBA& operator*=(ColorComponentT n)
  {
    m_color[RED] *= n;
    m_color[GREEN] *= n;
    m_color[BLUE] *= n;

    return *this;
  }

  RGBA operator*(ColorComponentT n)
  {
    return RGBA(
      m_color[RED] * n, m_color[GREEN] * n, m_color[BLUE] * n, m_color[ALPHA]);
  }

  RGBA& operator+=(ColorComponentT n)
  {
    m_color[RED] += n;
    m_color[GREEN] += n;
    m_color[BLUE] += n;

    return *this;
  }

  RGBA operator+(ColorComponentT n)
  {
    return RGBA(
      m_color[RED] + n, m_color[GREEN] + n, m_color[BLUE] + n, m_color[ALPHA]);
  }

  RGBA& operator-=(ColorComponentT n)
  {
    m_color[RED] -= n;
    m_color[GREEN] -= n;
    m_color[BLUE] -= n;

    return *this;
  }

  RGBA operator-(ColorComponentT n)
  {
    return RGBA(
      m_color[RED] - n, m_color[GREEN] - n, m_color[BLUE] - n, m_color[ALPHA]);
  }

private:
  ColorComponentT m_color[4];
};
}

namespace stl::driver::led {
class LEDStripEffectBase;

class LEDStripBase
{
public:
  static constexpr float BRIGHTNESS_MAX = 1.0f;

  LEDStripBase(const SizeT ledCount)
    : m_ledCount(ledCount) {}

  virtual ~LEDStripBase() {}

  virtual ErrorE configureHardware() = 0;

  virtual ErrorE apply(const color::RGBA* const colorList, const stl::SizeT colorListLength, const bool inverse = false) = 0;

  virtual void clear() = 0;

  void brightness(const float value) { m_brightness = math::min(1.0f, value); }

  void brightness(const UInt32T value)
  {
    m_brightness = CFLT(math::min<UInt32T>(100U, value)) / 100.0f;
  }

  float brightness() const { return m_brightness; }

  stl::SizeT ledCount() const { return m_ledCount; }

private:
  float m_brightness = BRIGHTNESS_MAX;

protected:
  stl::SizeT m_ledCount;
};

class LEDStripEffectBase
{
public:
  LEDStripEffectBase() = default;

  LEDStripEffectBase(const LEDStripEffectBase& other)
    : m_running(other.m_running)
    , m_repeat(other.m_repeat)
    , m_loopStarted(other.m_loopStarted)
    , m_loopEnded(other.m_loopEnded)
    , m_effectDurationTimeMax(other.m_effectDurationTimeMax)
    , m_effectDurationTime(other.m_effectDurationTime)
    , m_lastUpdateTime(other.m_lastUpdateTime)
    , m_snakeDirectionInvert(other.m_snakeDirectionInvert)
    , m_color(other.m_color)
    , m_firstCycle(other.m_firstCycle)
  {
  }

  LEDStripEffectBase(LEDStripEffectBase&& other) = delete;

  LEDStripEffectBase& operator=(const LEDStripEffectBase& other)
  {
    m_running = other.m_running;
    m_repeat = other.m_repeat;
    m_loopStarted = other.m_loopStarted;
    m_loopEnded = other.m_loopEnded;
    m_effectDurationTimeMax = other.m_effectDurationTimeMax;
    m_effectDurationTime = other.m_effectDurationTime;
    m_lastUpdateTime = other.m_lastUpdateTime;
    m_snakeDirectionInvert = other.m_snakeDirectionInvert;
    m_color = other.m_color;
    m_firstCycle = other.m_firstCycle;

    return *this;
  }

  LEDStripEffectBase& operator=(LEDStripEffectBase&& other) = delete;

  static void printLEDStates(const stl::color::RGBA* const ledList,
                             const stl::SizeT ledCount,
                             const float brightness)
  {
    stl::Vector<char> charList(ledCount);

    logging::log << logging::LvlE::Debug << "LED brightness: " << brightness << stl::endl;

    for (stl::SizeT i = 0U; i < ledCount; ++i) {
      const int step = CI32((9.0f / 255.0f) * (CFLT(ledList[i].alpha()) * brightness));
      charList[i] = '0' + step;
    }

    charList.push_back('\0');

    logging::log << logging::LvlE::Debug << "LED states: " << charList.data()
                 << stl::endl;
  }

  ErrorE update(color::RGBA* const ledList, const stl::SizeT ledCount)
  {
    Assert(!stl::math::equal(m_effectDurationTimeMax, 0.0f));

    if (m_running) {
      const time::Millis previousUpdateTime = m_lastUpdateTime;

      m_lastUpdateTime = stl::time::getUpTime();

      const float delta =
        CFLT((m_lastUpdateTime - previousUpdateTime).value()) / 1000.0f;

      m_effectDurationTime += delta;

      if (m_effectDurationTime >= m_effectDurationTimeMax) {
        if (m_repeat) {
          while (m_effectDurationTime >= m_effectDurationTimeMax) {
            m_effectDurationTime -= m_effectDurationTimeMax;
          }

          m_loopStarted = true;
          m_firstCycle = false;
        } else {
          m_effectDurationTime = m_effectDurationTimeMax;
          m_running = false;
        }

        m_loopEnded = true;
      }
    }

    return updateInternal(ledList, ledCount);
  }

  void effectDurationMax(const float duration)
  {
    m_effectDurationTimeMax = duration;
  }

  float effectDurationMax() const { return m_effectDurationTimeMax; }

  void effectDuration(const float duration) { m_effectDurationTime = duration; }

  float effectDuration() const { return m_effectDurationTime; }

  bool running() { return m_running; }

  bool repeat() const { return m_repeat; }

  void repeat(const bool value) { m_repeat = value; }

  void start()
  {
    m_running = true;
    m_lastUpdateTime = time::getUpTime();
  }

  void stop() { m_running = false; }

  void reset()
  {
    m_effectDurationTime = 0.0f;
    m_loopStarted = true;
    m_loopEnded = false;
    m_firstCycle = true;

    resetInternal();
  }

  bool loopStarted() const
  {
    if (m_loopStarted) {
      m_loopStarted = false;
      return true;
    } else {
      return false;
    }
  }

  bool loopEnded() const
  {
    if (m_loopEnded) {
      m_loopEnded = false;
      return true;
    } else {
      return false;
    }
  }

  void snakeDirectionInvert(const bool value)
  {
    m_snakeDirectionInvert = value;
  }

  bool snakeDirectionInvert() const { return m_snakeDirectionInvert; }

  void color(const color::RGB& newColor) { m_color = newColor; }

  const color::RGB& color() const { return m_color; }

  bool firstCycle() const { return m_firstCycle; }

  void clear(color::RGBA* const ledList, const stl::SizeT ledCount)
  {
    for (stl::SizeT i = 0U; i < ledCount; ++i) {
      ledList[i] = color::RGBA();
    }
  }

  void fill(const color::RGB& color, color::RGBA* const ledList, const stl::SizeT ledCount)
  {
    for (stl::SizeT i = 0U; i < ledCount; ++i) {
      ledList[i] = color;
      ledList[i].alpha(255U);
    }
  }

  void fill(color::RGBA* const ledList, const stl::SizeT ledCount, const color::RGB& color)
  {
    for (stl::SizeT i = 0U; i < ledCount; ++i) {
      ledList[i] = color;
      ledList[i].alpha(255U);
    }
  }

protected:
  virtual ErrorE updateInternal(color::RGBA* const ledList, const stl::SizeT ledCount) = 0;
  virtual void resetInternal() = 0;

  void fadeIn(color::RGBA* const ledList,
              const color::RGB color,
              const stl::SizeT ledCount,
              color::ColorComponentT maxBrightness,
              const float maxDuration,
              const float duration)
  {
    const float effectDurationNorm =
      stl::math::min(duration / maxDuration, 1.0f);

    for (stl::SizeT i = 0U; i < ledCount; ++i) {
      const color::ColorComponentT rgbBrightness =
        CU08(CFLT(maxBrightness) * (effectDurationNorm));

      ledList[i] = color;
      ledList[i].alpha(rgbBrightness);
    }
  }

  void fadeInSin(color::RGBA* const ledList,
                 const color::RGB color,
                 const stl::SizeT ledCount,
                 color::ColorComponentT maxBrightness,
                 const float maxDuration,
                 const float duration)
  {
    const float effectDurationNorm =
      stl::math::min(duration / maxDuration, 1.0f);

    const float ledDistance = 1.0f / CFLT(ledCount);

    for (stl::SizeT i = 0U; i < ledCount; ++i) {
      const float ledPosition = ledDistance * CFLT(i);
      const float brightness = CFLT(maxBrightness) * effectDurationNorm;
      const float compensation = ledDistance * ledPosition;
      const float brightnesSined =
        brightness *
        stl::math::sin(stl::math::PI_FLT * (ledPosition + compensation));

      ledList[i] = color;
      ledList[i].alpha(CU08(brightnesSined));
    }
  }

  void fadeOut(color::RGBA* const ledList,
               const color::RGB color,
               const stl::SizeT ledCount,
               color::ColorComponentT maxBrightness,
               const float maxDuration,
               const float duration)
  {
    const float effectDurationNorm =
      stl::math::min(duration / maxDuration, 1.0f);

    for (stl::SizeT i = 0U; i < ledCount; ++i) {
      const color::ColorComponentT rgbBrightness =
        CU08(CFLT(maxBrightness) * (1.0f - effectDurationNorm));

      ledList[i] = color;
      ledList[i].alpha(rgbBrightness);
    }
  }

  void fadeOutSin(color::RGBA* const ledList,
                  const color::RGB color,
                  const stl::SizeT ledCount,
                  color::ColorComponentT maxBrightness,
                  const float maxDuration,
                  const float duration)
  {
    const float effectDurationNorm =
      stl::math::min(duration / maxDuration, 1.0f);

    const float ledDistance = 1.0f / CFLT(ledCount);

    for (stl::SizeT i = 0U; i < ledCount; ++i) {
      const float ledPosition = ledDistance * CFLT(i);
      const float brightness =
        CFLT(maxBrightness) * (1.0f - effectDurationNorm);
      const float compensation = ledDistance * ledPosition;
      const float brightnesSined =
        brightness * (sin(stl::math::PI_FLT * (ledPosition + compensation)));

      ledList[i] = color;
      ledList[i].alpha(CU08(brightnesSined));
    }
  }

  void waveSin(color::RGBA* const ledList,
               const color::RGB color,
               const stl::SizeT ledCount,
               color::ColorComponentT maxBrightness,
               const float maxDuration,
               const float duration,
               const float direction)
  {
    const float effectDurationNorm =
      stl::math::min(duration / maxDuration, 1.0f);
    const float ledDistance = 1.0f / CFLT(ledCount);

    for (stl::SizeT i = 0U; i < ledCount; ++i) {
      const float ledPosition = ledDistance * CFLT(i);
      const float brightness = CFLT(maxBrightness);
      const float compensation = ledDistance * ledPosition;
      const float brightnesSined =
        brightness *
        stl::math::max(
          (stl::math::sin(stl::math::PI_FLT * (ledPosition + compensation) +
                          direction *
                            ((stl::math::PI_FLT) +
                             stl::math::PI_FLT * effectDurationNorm * 2.0f))),
          0.0f);

      ledList[i] = color;
      ledList[i].alpha(CU08(brightnesSined));
    }
  }

  void snake(color::RGBA* const ledList,
             const color::RGB color,
             const stl::SizeT ledCount,
             const stl::SizeT snakeLedCount,
             color::ColorComponentT maxBrightness,
             color::ColorComponentT minBrightness,
             const float maxDuration,
             const float duration,
             const bool invert)
  {
    const float effectDurationNorm =
      stl::math::min(duration / maxDuration, 1.0f);
    const float ledDistance = 1.0f / CFLT(ledCount);
    const float snakeLength = CFLT(snakeLedCount) / CFLT(ledCount);

    for (stl::SizeT i = 0U; i < ledCount; ++i) {
      float ledPosition = ledDistance * CFLT(i);
      if (invert) {
        ledPosition = 1.0f - ledPosition;
      }

      float brightness;

      // duration = headPosition
      const float distanceFromHead =
        (effectDurationNorm - ledPosition) + snakeLength * effectDurationNorm;
      if ((distanceFromHead < 0.0f) || (distanceFromHead > snakeLength)) {
        brightness = 0.0f;
      } else {
        const float distanceFromHeadNorm = distanceFromHead / snakeLength;
        const float currentAttenuation = 1.0f - distanceFromHeadNorm;
        brightness = CFLT(maxBrightness) * currentAttenuation;
      }

      ledList[i] = color;
      ledList[i].alpha(stl::math::max(CU08(brightness), minBrightness));
    }
  }

  void sinSnakeCircular(color::RGBA* const ledList,
                        const color::RGB color,
                        const stl::SizeT ledCount,
                        const stl::SizeT snakeLedCount,
                        color::ColorComponentT maxBrightness,
                        color::ColorComponentT minBrightness,
                        const float maxDuration,
                        const float duration,
                        const bool invert,
                        const bool firstCycle)
  {
    const float ledDistance = 1.0f / CFLT(ledCount);
    const float snakeLength = CFLT(snakeLedCount) / CFLT(ledCount);
    float effectDurationNorm = stl::math::min(duration / maxDuration, 1.0f);

    effectDurationNorm =
      invert ? (1.0f - effectDurationNorm) : effectDurationNorm;
    effectDurationNorm +=
      firstCycle ? (snakeLength * effectDurationNorm) : (snakeLength);

    for (stl::SizeT i = 0U; i < ledCount; ++i) {
      float brightness = 0.0f;
      float ledPosition = 0.0f;

      ledPosition = (CFLT(i) * ledDistance);

      float distance = ledPosition - effectDurationNorm;
      const float toMove = 1.0f + distance;
      if (toMove < snakeLength) {
        distance = toMove;
      }

      stl::UInt8T minBrightnessTemp;

      if ((distance > 0.0f) || (-distance > snakeLength)) {
        brightness = 0.0f;
        minBrightnessTemp = minBrightness;
      } else {
        const float distanceNorm = stl::math::abs(distance / snakeLength);
        brightness =
          stl::math::sin(distanceNorm * stl::math::PI_FLT) * CU08(255U);
        minBrightnessTemp = 0U;
      }

      brightness = (CFLT(maxBrightness) / 255.0f) * brightness;

      ledList[i] = color;
      ledList[i].alpha(stl::math::max(CU08(brightness), minBrightnessTemp));
    }
  }

  void doubleFlashMerge(color::RGBA* const ledList,
                        const color::RGB color,
                        const stl::SizeT ledCount,
                        const stl::SizeT snakeLedCount,
                        color::ColorComponentT maxBrightness,
                        color::ColorComponentT minBrightness,
                        const float maxDuration,
                        const float duration,
                        const bool invert)
  {
    const float ledDistance = 1.0f / CFLT(ledCount);
    const float snakeLength = CFLT(snakeLedCount) / CFLT(ledCount);
    float effectDurationNorm = stl::math::min(duration / maxDuration, 1.0f);

    effectDurationNorm =
      invert ? (1.0f - effectDurationNorm) : effectDurationNorm;
    effectDurationNorm += snakeLength * effectDurationNorm;

    for (stl::SizeT i = 0U; i < ledCount; ++i) {
      float brightness = 0.0f;
      float ledPosition = 0.0f;

      ledPosition = stl::math::sin((CFLT(i) * ledDistance) * stl::math::PI_FLT);

      const float distance = ledPosition - effectDurationNorm;

      stl::UInt8T minBrightnessTemp;

      if ((distance > 0.0f) || (-distance > snakeLength)) {
        brightness = 0.0f;
        minBrightnessTemp = minBrightness;
      } else {
        const float distanceNorm = stl::math::abs(distance / snakeLength);
        brightness =
          stl::math::sin(distanceNorm * stl::math::PI_FLT) * CU08(255U);
        minBrightnessTemp = 0U;
      }

      brightness = (CFLT(maxBrightness) / 255.0f) * brightness;

      ledList[i] = color;
      ledList[i].alpha(stl::math::max(CU08(brightness), minBrightnessTemp));
    }
  }

  void sinSnake(color::RGBA* const ledList,
                const color::RGB color,
                const stl::SizeT ledCount,
                const stl::SizeT snakeLedCount,
                color::ColorComponentT maxBrightness,
                color::ColorComponentT minBrightness,
                const float maxDuration,
                const float duration,
                const bool invert)
  {
    const float ledDistance = 1.0f / CFLT(ledCount);
    const float snakeLength = CFLT(snakeLedCount) / CFLT(ledCount);
    float effectDurationNorm = stl::math::min(duration / maxDuration, 1.0f);

    effectDurationNorm =
      invert ? (1.0f - effectDurationNorm) : effectDurationNorm;
    effectDurationNorm += snakeLength * effectDurationNorm;

    for (stl::SizeT i = 0U; i < ledCount; ++i) {
      float brightness;
      const float ledPosition = CFLT(i) * ledDistance;
      const float distance = ledPosition - effectDurationNorm;

      stl::UInt8T minBrightnessTemp;

      if ((distance > 0.0f) || (-distance > snakeLength)) {
        brightness = 0.0f;
        minBrightnessTemp = minBrightness;
      } else {
        const float distanceNorm = stl::math::abs(distance / snakeLength);
        brightness =
          stl::math::sin(distanceNorm * stl::math::PI_FLT) * CU08(255U);
        minBrightnessTemp = 0U;
      }

      brightness = (CFLT(maxBrightness) / 255.0f) * brightness;

      ledList[i] = color;
      ledList[i].alpha(stl::math::max(CU08(brightness), minBrightnessTemp));
    }
  }

  void bar(color::RGBA* const ledList,
           const color::RGB color,
           const stl::SizeT ledCount,
           color::ColorComponentT maxBrightness,
           const float fill,
           const bool invert)
  {
    const float ledDistance = 1.0f / CFLT(ledCount);

    for (stl::SizeT i = 0U; i < ledCount; ++i) {
      float ledPosition = ledDistance * CFLT(i);
      if (invert) {
        ledPosition = 1.0f - ledPosition;
      }

      const float compensation = ledDistance * fill;

      float brightness;

      // duration = headPosition
      const float distance = (fill - ledPosition) + compensation;
      if (distance <= 0.0f) {
        brightness = 0.0f;
      } else {
        brightness = CFLT(maxBrightness);
      }

      ledList[i] = color;
      ledList[i].alpha(CU08(brightness));
    }
  }

  void blink(color::RGBA* const ledList,
             const color::RGB color,
             const stl::SizeT ledCount,
             const stl::SizeT blinkCount,
             color::ColorComponentT maxBrightness,
             const float maxDuration,
             const float duration)
  {
    const float phaseTime = (maxDuration / CFLT(blinkCount)) * 2.0f;
    const float phase = duration / phaseTime;

    const bool on = (CU08(phase) % 2) == 0;

    for (stl::SizeT i = 0U; i < ledCount; ++i) {
      ledList[i] = color;
      ledList[i].alpha(on ? maxBrightness : 0U);
    }
  }

private:
  bool m_running = false;
  bool m_repeat = false;
  mutable bool m_loopStarted = false;
  mutable bool m_loopEnded = false;
  float m_effectDurationTimeMax = 1.0f;
  float m_effectDurationTime = 0.0f;
  time::Millis m_lastUpdateTime;
  bool m_snakeDirectionInvert = false;
  color::RGB m_color = color::RGB(255U, 255U, 255U);
  bool m_firstCycle = true;
};
}

#endif
