#ifndef __IONT_STL_SK9822_HPP__
#define __IONT_STL_SK9822_HPP__

#include "drivers/board/gpio.hpp"
#include "drivers/led/led_strip.hpp"

#include "stl/array.hpp"
#include "stl/limits.hpp"
#include "stl/stdint.hpp"
#include "stl/stl_config.hpp"
#include "stl/thread.hpp"

#include "log/logging.hpp"

namespace stl::driver::led {
#if (defined STL_ENABLE_GPIO && (STL_ENABLE_GPIO == 1))
class SK9822 : public LEDStripBase
{
public:
  SK9822(const board::GPIOT m_dataPin,
         const board::GPIOT clkPin,
         const stl::SizeT ledCount)
    : LEDStripBase(ledCount)
    , m_dataPin(m_dataPin)
    , m_clkPin(clkPin)
  {
  }

  virtual ~SK9822() {}

  virtual ErrorE configureHardware() override
  {
    init();
    clear();

    return ErrorE::Ok;
  }

  virtual ErrorE apply(const color::RGBA* const colorList,
                       const stl::SizeT colorListLength,
                       const bool inverse = false) override
  {
    if (colorListLength != m_ledCount) {
      return stl::ErrorE::InvalidArgument;
    }

    if(colorListLength == 0U)
    {
      return stl::ErrorE::InvalidArgument;
    }

    startFrame();
    if(inverse)
    {
      for (SizeT i = colorListLength; i > 0U; ++i) {
        sendColor(colorList[i - 1]);
      }
    }
    else
    {
      for (SizeT i = 0; i < colorListLength; ++i) {
        sendColor(colorList[i]);
      }
    }
    endFrame(colorListLength);

    return ErrorE::Ok;
  }

  virtual void clear() override
  {
    startFrame();
    for (SizeT i = 0; i < m_ledCount; ++i) {
      sendColor(stl::color::RGBA());
    }
    endFrame(m_ledCount);
  }

private:
  void startFrame()
  {
    init();
    transfer(0);
    transfer(0);
    transfer(0);
    transfer(0);
  }

  void endFrame(const SizeT count)
  {
    for (SizeT i = 0; i < (count + 14) / 16; i++) {
      transfer(0);
    }

    /* We call init() here to make sure we leave the data line driving low
     * even if count is 0 or 1. */
    init();
  }

  void sendColor(const color::ColorComponentT red,
                 const color::ColorComponentT green,
                 const color::ColorComponentT blue,
                 const color::ColorComponentT alpha)
  {
    static const float brightnessMaxInternal = 31.0f;

    const float brightnessMaxReduced = brightnessMaxInternal * brightness();
    const float alphaToFloat =
      CFLT(alpha) / CFLT(stl::Limits<color::ColorComponentT>::max());
    const float brightnessReduced = brightnessMaxReduced * alphaToFloat;

    transfer(0b11100000 | CU08(brightnessReduced));
    transfer(blue);
    transfer(green);
    transfer(red);
  }

  void sendColor(const color::RGBA& color)
  {
    sendColor(color.red(), color.green(), color.blue(), color.alpha());
  }

protected:
  void init()
  {
#ifdef SK9822_USE_FAST_GPIO
    FastGPIO::Pin<m_dataPin>::setOutputLow();
    FastGPIO::Pin<m_clkPin>::setOutputLow();
#else
    board::GPIO_setLvl(m_dataPin, board::GPIOLvlE::Low);
    board::GPIO_setDir(m_dataPin, board::GPIODirE::Output);
    board::GPIO_setLvl(m_clkPin, board::GPIOLvlE::Low);
    board::GPIO_setDir(m_clkPin, board::GPIODirE::Output);
#endif
  }

  void transfer(const color::ColorComponentT b)
  {
    stl::thread::delayForNs(35);
    board::GPIO_setLvl(m_dataPin, static_cast<board::GPIOLvlE>(b >> 7 & 1));
    stl::thread::delayForNs(35);
    board::GPIO_setLvl(m_clkPin, board::GPIOLvlE::High);
    stl::thread::delayForNs(35);
    board::GPIO_setLvl(m_clkPin, board::GPIOLvlE::Low);
    stl::thread::delayForNs(35);
    board::GPIO_setLvl(m_dataPin, static_cast<board::GPIOLvlE>(b >> 6 & 1));
    stl::thread::delayForNs(35);
    board::GPIO_setLvl(m_clkPin, board::GPIOLvlE::High);
    stl::thread::delayForNs(35);
    board::GPIO_setLvl(m_clkPin, board::GPIOLvlE::Low);
    stl::thread::delayForNs(35);
    board::GPIO_setLvl(m_dataPin, static_cast<board::GPIOLvlE>(b >> 5 & 1));
    stl::thread::delayForNs(35);
    board::GPIO_setLvl(m_clkPin, board::GPIOLvlE::High);
    stl::thread::delayForNs(35);
    board::GPIO_setLvl(m_clkPin, board::GPIOLvlE::Low);
    stl::thread::delayForNs(35);
    board::GPIO_setLvl(m_dataPin, static_cast<board::GPIOLvlE>(b >> 4 & 1));
    stl::thread::delayForNs(35);
    board::GPIO_setLvl(m_clkPin, board::GPIOLvlE::High);
    stl::thread::delayForNs(35);
    board::GPIO_setLvl(m_clkPin, board::GPIOLvlE::Low);
    stl::thread::delayForNs(35);
    board::GPIO_setLvl(m_dataPin, static_cast<board::GPIOLvlE>(b >> 3 & 1));
    stl::thread::delayForNs(35);
    board::GPIO_setLvl(m_clkPin, board::GPIOLvlE::High);
    stl::thread::delayForNs(35);
    board::GPIO_setLvl(m_clkPin, board::GPIOLvlE::Low);
    stl::thread::delayForNs(35);
    board::GPIO_setLvl(m_dataPin, static_cast<board::GPIOLvlE>(b >> 2 & 1));
    stl::thread::delayForNs(35);
    board::GPIO_setLvl(m_clkPin, board::GPIOLvlE::High);
    stl::thread::delayForNs(35);
    board::GPIO_setLvl(m_clkPin, board::GPIOLvlE::Low);
    stl::thread::delayForNs(35);
    board::GPIO_setLvl(m_dataPin, static_cast<board::GPIOLvlE>(b >> 1 & 1));
    stl::thread::delayForNs(35);
    board::GPIO_setLvl(m_clkPin, board::GPIOLvlE::High);
    stl::thread::delayForNs(35);
    board::GPIO_setLvl(m_clkPin, board::GPIOLvlE::Low);
    stl::thread::delayForNs(35);
    board::GPIO_setLvl(m_dataPin, static_cast<board::GPIOLvlE>(b >> 0 & 1));
    stl::thread::delayForNs(35);
    board::GPIO_setLvl(m_clkPin, board::GPIOLvlE::High);
    stl::thread::delayForNs(35);
    board::GPIO_setLvl(m_clkPin, board::GPIOLvlE::Low);
  }

  board::GPIOT m_dataPin;
  board::GPIOT m_clkPin;
};

#endif // (defined STL_ENABLE_GPIO && (STL_ENABLE_GPIO == 1))

} // namespace driver::led

#endif // __IONT_STL_SK9822_HPP__
