#ifndef __STL_DRIVERS_RFID_ELATEC_HPP__
#define __STL_DRIVERS_RFID_ELATEC_HPP__

#include "drivers/rfid/serial/base.hpp"

#include "drivers/rfid/rfid_base.hpp"

#include "network/packet_builder.hpp"
#include "stl/error.hpp"
#include "stl/fstream.hpp"
#include "stl/math.hpp"
#include "stl/stdint.hpp"
#include "stl/string.hpp"
#include "stl/string_conv.hpp"

#include "log/logging.hpp"

// TODO: What if serial port is closed for some reason?

namespace stl::driver::rfid {
class ElatecTWN4Base : public RFIDReaderBase
{
public:
  ElatecTWN4Base(const ConnectionTypeE connectionType)
    : RFIDReaderBase(ManufacturerE::Elatec, ModelE::TWN4, connectionType)
  {
  }

  ElatecTWN4Base(const ElatecTWN4Base&) = delete;

  ElatecTWN4Base(ElatecTWN4Base&& other)
    : RFIDReaderBase(stl::move(static_cast<RFIDReaderBase&>(other)))
  {
  }

  virtual ~ElatecTWN4Base() = default;

  ElatecTWN4Base& operator=(const ElatecTWN4Base&) = delete;

  ElatecTWN4Base& operator=(ElatecTWN4Base&& other)
  {
    static_cast<RFIDReaderBase&>(*this) =
      stl::move(static_cast<RFIDReaderBase&>(other));

    return *this;
  }

  static const ConnectionTypeListT& connectionTypeList()
  {
    return m_connectionTypeList;
  }

private:
  static const ConnectionTypeListT m_connectionTypeList;
};

class ElatecTWN4_Serial
  : public ElatecTWN4Base
  , public RFIDSerialBase
{
public:
  ElatecTWN4_Serial(filesystem::Path serialPath)
    : ElatecTWN4Base(ConnectionTypeE::Serial)
    , RFIDSerialBase(stl::move(serialPath), time::Millis(300))
    , m_speakerVolume(100U)
  {
  }

  virtual ~ElatecTWN4_Serial() = default;

  virtual ErrorE configureHardwareCustom() override
  {
    m_online = false;

    logging::log << logging::LvlE::Debug << "Starting ELATEC thread" << stl::endl;

    if (!start()) {
      logging::log << logging::LvlE::Debug << "ELATEC thread NOT started" << stl::endl;

      return ErrorE::UnknownError;
    }

    logging::log << logging::LvlE::Debug << "ELATEC thread started" << stl::endl;

    m_online = true;

    return stl::ErrorE::Ok;
  }

  virtual void deinitHardwareCustom() override
  {
    if(running())
    {
      stop();
    }
  }

  virtual stl::ErrorE update() override
  {
    stl::network::PacketDataT packet{};

    if (pull(packet)) {
      processCardUIDPacket(packet);
    }

    return stl::ErrorE::Ok;
  }

  void setSpeakerVolume(const stl::UInt32T value)
  {
    Assert(value <= 100U);

    m_speakerVolume = stl::math::min<stl::UInt32T>(value, 100U);
  }

private:
  bool pull(stl::network::PacketDataT& packet)
  {
    packet.clear();
    packet.resize(32U);

    stl::SizeT received = 0U;

    stl::timer::Countdown timeout(stl::time::Millis(200));

    timeout.reset();

    bool gotUID = false;

    do {
      const stl::SizeT toRead =
        stl::math::min(available(), static_cast<stl::SizeT>(1U));
      if (toRead > 0U) {
        const stl::SizeT dataRead = read(packet.data() + received, toRead);

        if(dataRead > 0U)
        {
          timeout.reset();
        }

        received += dataRead;

        if (toRead != dataRead) {
          logging::log << logging::LvlE::Error << "Data read error ["
                       << dataRead << "] expected [" << toRead << "]"
                       << stl::endl;

          if ((m_connectionStatus != RFIDConnectionStatusE::Offline) &&
              m_statusCb) {
            m_statusCb(RFIDConnectionStatusE::Offline, m_arg);
          }

          m_connectionStatus = RFIDConnectionStatusE::Offline;

          // notifyEvent(EVSEEventE::Error); // TODO

          return false;
        }

        if ((received > 1U) && (packet[received - 1U] == '\r')) {
          gotUID = true;
          break;
        }
      }

      stl::thread::sleepFor(10);
    } while (!timeout && (received < 32U)); // Max packet size

    packet.resize(received);

    if ((m_connectionStatus != RFIDConnectionStatusE::Online) && m_statusCb) {
      m_statusCb(RFIDConnectionStatusE::Online, m_arg);
    }

    m_connectionStatus = RFIDConnectionStatusE::Online;

    logging::log << logging::LvlE::Debug << "RFID pull: ";
    logging::log.dump(packet.data(), packet.size(), 20U, true);

    return gotUID;
  }

  void processCardUIDPacket(const stl::network::PacketDataT& packet)
  {
    if(packet.size() > 1U)
    {
      stl::String uid;

      stl::SizeT offset = 0U;

      for(stl::SizeT i = 2U; i < packet.size(); ++i)
      {
        if(static_cast<stl::UInt8T>('0') != packet[i])
        {
          offset = i;
          break;
        }
      }

      for(stl::SizeT i = offset; i < (packet.size() - 1U); ++i) // -1 because of 0x0D at the end
      {
        uid += static_cast<char>(packet[i]);
      }

      Assert(m_cardReadCb);

      if (m_cardReadCb && !uid.empty()) {
        logging::log << logging::LvlE::Info << "ELATEC READ CARD UID: " << uid
                  << stl::endl;

        m_cardReadCb(uid, m_arg);
      }
      else
      {
        logging::log << logging::LvlE::Info << "ELATEC READ CARD UID: <EMPTY> \"" << uid << "\"" << uid
                  << stl::endl;
      }
    }
  }

  stl::UInt32T m_speakerVolume; // TODO: Is it configurable by FW?
};
}

#endif // __STL_DRIVERS_RFID_NETRONIX_HPP__
