#ifndef __STL_DRIVERS_RFID_NETRONIX_HPP__
#define __STL_DRIVERS_RFID_NETRONIX_HPP__

// Brown - Rx (B)
// Red   - Tx (A)
// Blue  - GND
// Pink  - VCC (7-16V)

#include "drivers/board/rs485.hpp"
#include "drivers/rfid/rfid_base.hpp"

#include "network/packet_builder.hpp"
#include "stl/error.hpp"
#include "stl/fstream.hpp"
#include "stl/math.hpp"
#include "stl/stdint.hpp"
#include "stl/string_conv.hpp"

#include "log/logging.hpp"

namespace stl::driver::rfid {
class NetronixD4RBase : public RFIDReaderBase
{
public:
  NetronixD4RBase(const ConnectionTypeE connectionType)
    : RFIDReaderBase(ManufacturerE::Netronix, ModelE::D4R, connectionType)
  {
  }

  NetronixD4RBase(const NetronixD4RBase&) = delete;

  NetronixD4RBase(NetronixD4RBase&& other)
    : RFIDReaderBase(stl::move(static_cast<RFIDReaderBase&>(other)))
  {
  }

  virtual ~NetronixD4RBase() = default;

  NetronixD4RBase& operator=(const NetronixD4RBase&) = delete;

  NetronixD4RBase& operator=(NetronixD4RBase&& other)
  {
    static_cast<RFIDReaderBase&>(*this) =
      stl::move(static_cast<RFIDReaderBase&>(other));

    return *this;
  }

  static const ConnectionTypeListT& connectionTypeList()
  {
    return m_connectionTypeList;
  }

private:
  static const ConnectionTypeListT m_connectionTypeList;
};

class NetronixD4R_RS485
  : public NetronixD4RBase
  , public board::RS485SlaveBase
{
public:
  NetronixD4R_RS485(board::RS485& bus, const stl::UInt8T busId = 0x01U)
    : NetronixD4RBase(ConnectionTypeE::RS485)
    , RS485SlaveBase(bus, time::Millis(200))
    , m_id(busId)
    , m_speakerVolume(100U)
  {
  }

  virtual ~NetronixD4R_RS485() {}

  virtual ErrorE configureHardware() override
  {
    logging::log << logging::LvlE::Info << "[rfid] Setting RFID hardware configuration .. "
                 << stl::endl;
    stl::network::PacketDataT deviceConfigPacket = createConfigPacket();

    if (!push(deviceConfigPacket)) {
      logging::log << logging::StatE::Fail << stl::endl;
      return stl::ErrorE::BusError;
    }

    logging::log << logging::StatE::Success;

    stl::network::PacketDataT deviceConfigPacketResponse;
    if (!pull(deviceConfigPacketResponse)) {
      logging::log << logging::StatE::Fail << stl::endl;
      return stl::ErrorE::BusError;
    }

    logging::log << logging::StatE::Success << stl::endl;
    logging::log << logging::LvlE::Info << "[rfid] Setting RFID transponder type .. ";
    stl::network::PacketDataT transponderTypePacket =
      createSetTransponderTypePacket();
    if (!push(transponderTypePacket)) {
      logging::log << logging::StatE::Fail << stl::endl;
      return stl::ErrorE::BusError;
    }

    logging::log << logging::StatE::Success;

    stl::network::PacketDataT transponderTypePacketResponse;
    if (!pull(transponderTypePacketResponse)) {
      logging::log << logging::StatE::Fail << stl::endl;
      return stl::ErrorE::BusError;
    }

    logging::log << logging::StatE::Success << stl::endl;
    logging::log << logging::LvlE::Info << "[rfid] Setting RFID buzzer configuration .. ";

    stl::network::PacketDataT buzzerConfigPacket =
      createBuzzerVolumeConfigPacket();
    if (!push(buzzerConfigPacket)) {
      logging::log << logging::StatE::Fail << stl::endl;
      return stl::ErrorE::BusError;
    }

    logging::log << logging::StatE::Success;

    stl::network::PacketDataT buzzerConfigPacketResponse;
    if (!pull(buzzerConfigPacketResponse)) {
      logging::log << logging::StatE::Fail << stl::endl;
      return stl::ErrorE::BusError;
    }

    logging::log << logging::StatE::Success << stl::endl;

    m_online = true;

    return stl::ErrorE::Ok;
  }

  virtual void deinitHardware() override
  {
    
  }

  virtual stl::ErrorE configureHardwareCustom() override
  {
    return stl::ErrorE::Ok;
  }

  virtual void deinitHardwareCustom() override
  {
  }

  virtual stl::ErrorE poll() override
  {
    stl::network::PacketDataT packet = createReadCardPacket();
    if (!push(packet)) {
      // TODO: Error?
      return stl::ErrorE::BusError;
    }

    if (!pull(packet)) {
      // TODO: Error?
      return stl::ErrorE::BusError;
    }

    processReadCardPacketResponse(packet);

    return stl::ErrorE::Ok;
  }

  void setSpeakerVolume(const stl::UInt32T value)
  {
    Assert(value <= 100U);

    m_speakerVolume = stl::math::min<stl::UInt32T>(value, 100U);
  }

private:
  static stl::UInt16T calcCRC(const stl::UInt8T* const data,
                              const stl::SizeT size)
  {
    stl::UInt16T crc = 0U;
    stl::UInt16T temp = 0U;

    for (stl::SizeT i = 0U; i < size; ++i) {
      temp = ((crc >> 8) ^ data[i]) << 8;

      for (stl::SizeT b = 0U; b < 8U; ++b) {
        if (temp & 0x8000U) {
          temp = (temp << 1) ^ 0x1021U;
        } else {
          temp <<= 1;
        }
      }

      crc = temp ^ (crc << 8);
    }

    return crc;
  }

  stl::network::PacketDataT createReadCardPacket()
  {
    stl::network::Packet<stl::EndianE::EndianE_Little> packet{};

    // HEAD (1B)
    // - ID
    packet <<= m_id;
    // - Complete packet size
    packet <<= static_cast<stl::UInt8T>(0x05U);
    // - Command
    packet <<= static_cast<stl::UInt8T>(0x12U);
    // CRC (2B)
    packet <<= stl::Endian::hostSync(stl::EndianE::EndianE_Big,
                                     calcCRC(packet.data(), packet.size()));

    return packet;
  }

  void processReadCardPacketResponse(const stl::network::PacketDataT& packet)
  {
    if (packet.size() == 7U) {
      // Probably no transponder used
      Assert(m_cardRemoveCb);
      m_cardRemoveCb(m_arg);
    } else if (packet.size() == 13U) {
      const stl::UInt16T calcCRC =
        this->calcCRC(packet.data(), packet.size() - 2U);
      const stl::UInt16T packCRC = stl::Endian::hostSync(
        stl::EndianE::EndianE_Big,
        *reinterpret_cast<const stl::UInt16T*>(packet.data() + 11));

      if (calcCRC != packCRC) {
        logging::log << logging::LvlE::Info << "[rfid] Invalid packet reponse CRC [" << calcCRC
                     << " x " << packCRC << "]" << stl::endl;
        return;
      }

      if (m_id != packet[0]) {
        logging::log << logging::LvlE::Info << "[rfid] Unexpected packet response id ["
                     << CU32(packet[0]) << " x " << CU32(packet[0]) << "]"
                     << stl::endl;
        return;
      }

      if (packet.size() != packet[1]) {
        logging::log << logging::LvlE::Info << "[rfid] Invalid packet reponse size [" << packet[1]
                     << " x " << packet.size() << "]" << stl::endl;
        return;
      }

      if (packet.size() < 8U) {
        logging::log << logging::LvlE::Info << "[rfid] Invalid UID size [" << packet[1]
                     << " x " << packet.size() << "]" << stl::endl;
        return;
      }

      if (0x13U != packet[2]) {
        logging::log << logging::LvlE::Info << "[rfid] Invalid packet reponse command [" << packet[2]
                     << " x " << 0x13U << "]" << stl::endl;
        return;
      }

      if (0xFFU != packet[packet.size() - 3]) {
        logging::log << logging::LvlE::Info << "[rfid] Command [" << packet[2]
                     << " x " << 0x13U << "] not successful [" << packet[packet.size() - 3] << "]" << stl::endl;
        return;
      }

      // packet[3] // Collision
      // packet[4] // Transponder type

      stl::String uidTemp;

      const stl::SizeT toCopy = packet.size() - 8U;
      for(stl::SizeT i = 0U; i < toCopy; ++i)
      {
        uidTemp += stl::string::nibbleToHex(static_cast<char>(packet[packet.size() - 4 - i]) >> 4, true);
        uidTemp += stl::string::nibbleToHex(static_cast<char>(packet[packet.size() - 4 - i]), true);
      }

      stl::String uid;

      stl::SizeT offset = 0U;

      for(stl::SizeT i = 2U; i < uidTemp.size(); ++i)
      {
        if(static_cast<stl::UInt8T>('0') != uidTemp[i])
        {
          offset = i;
          break;
        }
      }

      for(stl::SizeT i = offset; i < uidTemp.size(); ++i)
      {
        uid += static_cast<char>(uidTemp[i]);
      }

      Assert(m_cardReadCb);

      if (m_cardReadCb && !uid.empty()) {
        logging::log << logging::LvlE::Info << "NETRONIX READ CARD UID: " << uid
                  << stl::endl;

        m_cardReadCb(uid, m_arg);
      }
      else
      {
        logging::log << logging::LvlE::Info << "ELATEC READ CARD UID: <EMPTY> \"" << uid << "\"" << uid
                  << stl::endl;
      }
    } else {
      // TODO
    }
  }

  stl::network::PacketDataT createSetTransponderTypePacket()
  {
    stl::network::Packet<stl::EndianE::EndianE_Little> packet{};

    // HEAD (1B)
    // - ID
    packet <<= m_id;
    // - Complete packet size
    packet <<= static_cast<stl::UInt8T>(0x07U);
    // - Command
    packet <<= static_cast<stl::UInt8T>(0x02U);
    // - TransponderType - UNIQUE
    packet <<= static_cast<stl::UInt8T>(0x01U);
    // - GAIN
    packet <<= static_cast<stl::UInt8T>(0x02U); // 0-3 (1-2 recommended)
    // CRC (2B)
    packet <<= stl::Endian::hostSync(stl::EndianE::EndianE_Big,
                                     calcCRC(packet.data(), packet.size()));

    return packet;
  }

  stl::network::PacketDataT createConfigPacket()
  {
    stl::network::Packet<stl::EndianE::EndianE_Little> packet{};

    // Default: 0x2,0x14,0x1,0x1,0x01,0xff

    // HEAD (1B)
    // - ID
    packet <<= m_id;
    // - Complete packet size
    packet <<= static_cast<stl::UInt8T>(0x0BU);
    // - Command
    packet <<= static_cast<stl::UInt8T>(0x58U);
    // ATrig
    packet <<= static_cast<stl::UInt8T>(0x02U); // Auto read enabled
    // AOfflineTime
    // packet <<= static_cast<stl::UInt8T>(0x1EU); // 0x1E * 100ms
    packet <<= static_cast<stl::UInt8T>(0x14U); // 0x1E * 100ms
    // ASerial
    packet <<=
      static_cast<stl::UInt8T>(0x00U); // Never automaticaly send UID after read
    // AMode
    packet <<= static_cast<stl::UInt8T>(0x20U); // Binary format
    // ABuzz
    packet <<= static_cast<stl::UInt8T>(0x02U); // Indicate all
    // AMulti
    packet <<=
      static_cast<stl::UInt8T>(0xFFU); // Read all known transponder types
    // CRC (2B)
    packet <<= stl::Endian::hostSync(stl::EndianE::EndianE_Big,
                                     calcCRC(packet.data(), packet.size()));

    return packet;
  }

  stl::network::PacketDataT createBuzzerVolumeConfigPacket()
  {
    stl::network::Packet<stl::EndianE::EndianE_Little> packet{};

    // static constexpr stl::UInt8T speakerVolumeMax = 0x0AU; // TODO
    static constexpr stl::UInt8T speakerVolumeMax = 0x00U;
    const stl::UInt32T speakerVolumeLimited =
      stl::math::min(m_speakerVolume, 100U);
    const float speakerVolumeInternal =
      (CFLT(speakerVolumeMax) / 100.0f) * CFLT(speakerVolumeLimited);

    logging::log << logging::LvlE::Info << "[rfid] Buzzer volume set to " << speakerVolumeLimited
                 << " [" << CU32(speakerVolumeInternal) << ']' << stl::endl;

    // HEAD (1B)
    // - ID
    packet <<= m_id;
    // - Complete packet size
    packet <<= static_cast<stl::UInt8T>(0x06U);
    // - Command
    packet <<= static_cast<stl::UInt8T>(0xD8U);
    // - Volume
    packet <<= static_cast<stl::UInt8T>(
      CU08(speakerVolumeInternal)); // Max volume (0x0A)
    // CRC (2B)
    packet <<= stl::Endian::hostSync(stl::EndianE::EndianE_Big,
                                     calcCRC(packet.data(), packet.size()));

    return packet;
  }

  bool push(const stl::network::PacketDataT& packet)
  {
    logging::log << logging::LvlE::Debug << "RFID to push: ";
    logging::log.dump(packet.data(), packet.size(), 20U, true);

    const stl::SizeT written = write(packet.data(), packet.size());
    if (packet.size() != written) {
      logging::log << logging::LvlE::Error << "Data write error [" << written << "] expected [" << packet.size() << "]" << stl::endl;

      if ((m_connectionStatus != RFIDConnectionStatusE::Offline) &&
          m_statusCb) {
        m_statusCb(RFIDConnectionStatusE::Offline, m_arg);
      }
      m_connectionStatus = RFIDConnectionStatusE::Offline;

      // notifyEvent(EVSEEventE::Error); // TODO

      return false;
    }

    logging::log << logging::LvlE::Debug << "RFID push: " << packet.size() << stl::endl;

    if ((m_connectionStatus != RFIDConnectionStatusE::Online) && m_statusCb) {
      m_statusCb(RFIDConnectionStatusE::Online, m_arg);
    }
    m_connectionStatus = RFIDConnectionStatusE::Online;

    return true;
  }

  bool pull(stl::network::PacketDataT& packet)
  {
    packet.clear();
    packet.resize(128);

    stl::SizeT received = 0U;
    stl::SizeT headSize = 2U;

    stl::timer::Countdown timeout(stl::time::Millis(3000));

    timeout.reset();

    do {
      const stl::SizeT toRead =
        stl::math::min(available(), headSize - received);
      if (toRead > 0U) {
        const stl::SizeT dataRead = read(packet.data() + received, toRead);

        received += dataRead;
      }

      if (timeout) {
        if ((m_connectionStatus != RFIDConnectionStatusE::Offline) &&
            m_statusCb) {
          m_statusCb(RFIDConnectionStatusE::Offline, m_arg);
        }
        m_connectionStatus = RFIDConnectionStatusE::Offline;

        // notifyEvent(EVSEEventE::Error); // TODO

        return false;
      }

      stl::thread::sleepFor(10);
    } while (received < headSize);

    timeout.reset();

    const stl::UInt8T expectedPacketSize = packet[1];

    do {
      const stl::SizeT toRead =
        stl::math::min(available(), expectedPacketSize - received);
      if (toRead > 0U) {
        const stl::SizeT dataRead = read(packet.data() + received, toRead);

        received += dataRead;

        if (toRead != dataRead) {
          logging::log << logging::LvlE::Error << "Data read error ["
                       << dataRead << "] expected [" << toRead << "]"
                       << stl::endl;

          if ((m_connectionStatus != RFIDConnectionStatusE::Offline) &&
              m_statusCb) {
            m_statusCb(RFIDConnectionStatusE::Offline, m_arg);
          }
          m_connectionStatus = RFIDConnectionStatusE::Offline;

          // notifyEvent(EVSEEventE::Error); // TODO

          return false;
        }
      }

      stl::thread::sleepFor(10);
    } while (!timeout && (received < expectedPacketSize));

    packet.resize(received);

    if ((m_connectionStatus != RFIDConnectionStatusE::Online) && m_statusCb) {
      m_statusCb(RFIDConnectionStatusE::Online, m_arg);
    }
    m_connectionStatus = RFIDConnectionStatusE::Online;

    logging::log << logging::LvlE::Debug << "RFID pull: ";
    logging::log.dump(packet.data(), packet.size(), 20U, true);

    return packet.size() > 0U;
  }

  stl::UInt8T m_id;
  stl::UInt32T m_speakerVolume;
};
}

#endif // __STL_DRIVERS_RFID_NETRONIX_HPP__
