#ifndef __STL_RFID_BASE_HPP__
#define __STL_RFID_BASE_HPP__

#include "network/packet_builder.hpp"

#include "stl/error.hpp"
#include "stl/stdint.hpp"
#include "stl/vector.hpp"

namespace stl::driver::rfid {
using RFIDUIDT = stl::String;

#define DEF_RFID_CONNECTIONTYPE                                                \
  ENTRY(Unknown, "Unknown")                                                    \
  ENTRY(Serial, "Serial")                                                      \
  ENTRY(RS485, "RS485")

enum class ConnectionTypeE
{
#define ENTRY(en, st) en,
  DEF_RFID_CONNECTIONTYPE
#undef ENTRY
    _Count
};

#define DEF_RFID_MANUFACTURER                                                  \
  ENTRY(Unknown, "Unknown")                                                    \
  ENTRY(Netronix, "Netronix")                                                  \
  ENTRY(Elatec, "Elatec")

enum class ManufacturerE
{
#define ENTRY(en, st) en,
  DEF_RFID_MANUFACTURER
#undef ENTRY
    _Count
};

#define DEF_RFID_MODEL                                                         \
  ENTRY(Unknown, "Unknown")                                                    \
  ENTRY(D4R, "D4R")                                                            \
  ENTRY(TWN4, "TWN4")

enum class ModelE
{
#define ENTRY(en, str) en,
  DEF_RFID_MODEL
#undef ENTRY
    _Count
};

enum class RFIDConnectionStatusE
{
  Offline = 1,
  Online = 2
};

const stl::String&
rfidConnectionTypeToString(stl::driver::rfid::ConnectionTypeE value);

stl::driver::rfid::ConnectionTypeE
stringToRFIDConnectionType(const stl::String& value);

const stl::String&
rfidManufacturerToString(stl::driver::rfid::ManufacturerE value);

stl::driver::rfid::ManufacturerE
stringToRFIDManufacturer(const stl::String& value);

const stl::String&
rfidModelToString(stl::driver::rfid::ModelE value);

stl::driver::rfid::ModelE
stringToRFIDModel(const stl::String& value);

class RFIDReaderBase
{
public:
  enum class CardTypeE
  {
    Unique // Q5
    ,
    HITAG,
    HID
  };

  using ConnectionTypeListT = stl::Vector<ConnectionTypeE>;
  using CardReadCb = void (*)(const RFIDUIDT& uid, void* const arg);
  using CardRemoveCb = void (*)(void* const arg);
  using StatusCb = void (*)(const RFIDConnectionStatusE status,
                            void* const arg);

  RFIDReaderBase(const ManufacturerE manufacturer,
                 const ModelE model,
                 const ConnectionTypeE connectionType)
    : m_manufacturer(manufacturer)
    , m_model(model)
    , m_connectionType(connectionType)
    , m_cardReadCb(nullptr)
    , m_cardRemoveCb(nullptr)
    , m_statusCb(nullptr)
    , m_arg(nullptr)
    , m_connectionStatus(RFIDConnectionStatusE::Offline)
  {
  }

  RFIDReaderBase(RFIDReaderBase&& other)
    : m_manufacturer(other.m_manufacturer)
    , m_model(other.m_model)
    , m_connectionType(other.m_connectionType)
    , m_cardReadCb(other.m_cardReadCb)
    , m_cardRemoveCb(other.m_cardRemoveCb)
    , m_statusCb(other.m_statusCb)
    , m_arg(other.m_arg)
    , m_connectionStatus(other.m_connectionStatus)
  {
  }

  virtual ~RFIDReaderBase() {}

  RFIDReaderBase(const RFIDReaderBase&) = delete;

  RFIDReaderBase& operator=(RFIDReaderBase&& other)
  {
    m_manufacturer = other.m_manufacturer;
    m_model = other.m_model;
    m_connectionType = other.m_connectionType;
    m_cardReadCb = other.m_cardReadCb;
    m_cardRemoveCb = other.m_cardRemoveCb;
    m_statusCb = other.m_statusCb;
    m_arg = other.m_arg;
    m_connectionStatus = other.m_connectionStatus;

    return *this;
  }

  RFIDReaderBase& operator=(const RFIDReaderBase&) = delete;

  void setArg(void* const arg) { m_arg = arg; }

  void setCardReadCb(const CardReadCb cb) { m_cardReadCb = cb; }

  void setCardRemoveCb(const CardRemoveCb cb) { m_cardRemoveCb = cb; }

  void setStatusCb(const StatusCb cb) { m_statusCb = cb; }

  ManufacturerE manufacturer() const { return m_manufacturer; }

  ModelE model() const { return m_model; }

  ConnectionTypeE connectionType() const { return m_connectionType; }

  virtual ErrorE configureHardwareCustom() = 0;
  virtual void deinitHardwareCustom() = 0;

protected:
  ManufacturerE m_manufacturer;
  ModelE m_model;
  ConnectionTypeE m_connectionType;
  CardReadCb m_cardReadCb;
  CardRemoveCb m_cardRemoveCb;
  StatusCb m_statusCb;
  void* m_arg;
  RFIDConnectionStatusE m_connectionStatus;
};
}

#endif // __STL_RFID_BASE_HPP__
