#ifndef __STL_RFID_SERIAL_BASE_HPP__
#define __STL_RFID_SERIAL_BASE_HPP__

#include "drivers/board/uart.hpp"
#include "log/logging.hpp"
#include "stl/error.hpp"
#include "stl/filesystem.hpp"
#include "stl/thread.hpp"
#include "stl/time.hpp"
#include "stl/timer.hpp"
#include "stl/string.hpp"
#include "stl/string_conv.hpp"

namespace stl::driver::rfid {
class RFIDSerialBase;

using PollingTimeT = time::Millis;

using RFIDSerialCallable = void (*)(const RunnableBase::Params* params,
                                    RFIDSerialBase* const instance);

using RFIDSerialRunnable = Runnable<RFIDSerialCallable, RFIDSerialBase*>;

class RFIDSerialBase
  : public RFIDSerialRunnable
  , protected driver::board::UART
{
public:
  RFIDSerialBase(filesystem::Path serialPath, const PollingTimeT pollingTime)
    : RFIDSerialRunnable("RFIDSerialBase")
    , m_pollingTimer(pollingTime)
    , m_serialPath(move(serialPath))
    , m_online(false)
  {
  }

  RFIDSerialBase(const RFIDSerialBase&) = delete;

  RFIDSerialBase(RFIDSerialBase&& other)
    : RFIDSerialRunnable(move(static_cast<RFIDSerialRunnable&>(other)))
    , m_pollingTimer(move(other.m_pollingTimer))
    , m_serialPath(move(other.m_serialPath))
    , m_online(move(other.m_online))
  {
  }

  RFIDSerialBase& operator=(const RFIDSerialBase&) = delete;

  RFIDSerialBase& operator=(RFIDSerialBase&& other)
  {
    static_cast<RFIDSerialRunnable&>(*this) =
      move(static_cast<RFIDSerialRunnable&>(other));
    m_pollingTimer = move(other.m_pollingTimer);
    m_serialPath = move(other.m_serialPath);
    m_online = move(other.m_online);

    return *this;
  }

  virtual ~RFIDSerialBase()
  {
  }

  bool start()
  {
    try {
      if (RFIDSerialRunnable::active()) {
        return false;
      }

      driver::board::UARTConfig uartConfig{};

      uartConfig.baudrate = 9600;

      bool open = false;

      // TODO: Workaround
      if(m_serialPath.ciFind("ttyACM") != String::nPos)
      {
        for(SizeT i = 0U; i < 32U; ++i)
        {
          m_serialPath = String("/dev/ttyACM") + string::toString(i); 

          open = driver::board::UART::open(
            m_serialPath,
            stream::FF_RW | stream::FF_NoBlock |
              stream::FF_NoCTTY | stream::FF_NoDelay,
            uartConfig);

          if (!open) {
            logging::log << logging::LvlE::Warning << "Could not open serial port "
                        << m_serialPath << endl;
          } else {
            break;
          }
        }
      }
      else
      {
        open = driver::board::UART::open(
            m_serialPath,
            stream::FF_RW | stream::FF_NoBlock |
              stream::FF_NoCTTY | stream::FF_NoDelay,
            uartConfig);
      }

      if(!open)
      {
        logging::log << logging::LvlE::Error << "Could not open serial port "
          << m_serialPath << endl;

        #if defined IONT_DEBUG_IGNORE_SERIAL_PORT_OPEN_ERROR
        return true;
        #else
        return false;
        #endif
      }

      RFIDSerialRunnable::start(worker, this);

      return true;
    } catch (...) {
      Assert(false);
      return false;
    }
  }

  bool running() const { return RFIDSerialRunnable::running(); }

  void stop()
  {
    try {
      if (RFIDSerialRunnable::active()) {
        RFIDSerialRunnable::stop();
      }

      if (driver::board::UART::isOpen()) {
        driver::board::UART::close();
      }
    } catch (...) {
      Assert(false);
      // TODO
    }
  }

  PollingTimeT pollingTime() const
  {
    return PollingTimeT(m_pollingTimer.timeoutValue());
  }

  bool online() const { return m_online; }

private:
  static void worker(const RunnableBase::Params* params,
                     RFIDSerialBase* const instance)
  {
    Assert(params);
    Assert(instance);

    logging::log << logging::LvlE::Info << "RFIDSerial worker started."
                 << endl;

    instance->m_pollingTimer.reset();

    // TODO: Thread unsafe

    while (!params->m_stop) {
      {
        if (instance->m_pollingTimer) {
          const ErrorE retv = instance->update();

          if ((!instance->m_online) && (ErrorE::Ok == retv)) {
            logging::log << logging::LvlE::Info << "RFID reader ONLINE"
                         << endl;
          } else if (instance->m_online && (ErrorE::Ok != retv)) {
            logging::log << logging::LvlE::Warning << "RFID reader OFFLINE"
                         << endl;
          }

          instance->m_online = ErrorE::Ok == retv;

          instance->m_pollingTimer.reset();
        }
      }

      thread::sleepFor(100);
    }
  }

  timer::Countdown m_pollingTimer;
  filesystem::Path m_serialPath;

protected:
  virtual ErrorE update() = 0;

  bool m_online;
};
}

#endif // __STL_RFID_SERIAL_BASE_HPP__
