#ifndef __STL_FSTREAM_HPP__
#define __STL_FSTREAM_HPP__

#include "stl/error.hpp"
#include "stl/filesystem.hpp"
#include "stl/istream.hpp"
#include "stl/map.hpp"
#include "stl/memory.hpp"
#include "stl/ostream.hpp"
#include "stl/vector.hpp"

namespace stl::stream {
class FStreamObserver;

using FileFlagsT = UInt32T;

static constexpr UInt32T FF_NoBlock = 0x01U;
static constexpr UInt32T FF_R = 0x02U;
static constexpr UInt32T FF_W = 0x04U;
static constexpr UInt32T FF_RW = FF_R | FF_W;
static constexpr UInt32T FF_NoCTTY = 0x08U;
static constexpr UInt32T FF_NoDelay = 0x10U;
static constexpr UInt32T FF_App = 0x20U;
static constexpr UInt32T FF_Trunc = 0x40U;
static constexpr UInt32T FF_Create = 0x80U;

class FStreamBase
  : public virtual IStreamBase
  , public virtual OStreamBase
{
public:
  using DataContainerT = Vector<ByteT>;

  FStreamBase();

  FStreamBase(filesystem::Path path, const FileFlagsT flags);

  FStreamBase(const FStreamBase&) = delete;

  FStreamBase(FStreamBase&& other)
    : m_error(other.m_error)
    , m_path(stl::move(other.m_path))
    , m_flags(other.m_flags)
    , m_observerMap(stl::move(other.m_observerMap)) {}

  FStreamBase& operator = (const FStreamBase&) = delete;

  FStreamBase& operator = (FStreamBase&& other)
  {
    m_error = other.m_error;
    m_path = stl::move(other.m_path);
    m_flags = other.m_flags;
    m_observerMap = stl::move(other.m_observerMap);

    return *this;
  }

  virtual ~FStreamBase();

  virtual bool open(filesystem::Path path, const FileFlagsT flags) = 0;

  virtual bool isOpen() const = 0;

  virtual void close() = 0;

  ErrorE error() const { return m_error; }

  operator bool() const { return isOpen(); }

  template<typename T>
  FStreamBase& operator <<(const T& value)
  {
    OStreamBase::operator<<(value);

    return *this;
  }

private:
  friend class FStreamObserver;

  using ObserverMapT =
    stl::Map<const FStreamObserver*, ReferenceWrapper<FStreamObserver>>;

  bool registerObserver(FStreamObserver& observer)
  {
    return m_observerMap
      .insert(
        stl::Pair<const FStreamObserver*, ReferenceWrapper<FStreamObserver>>(
          &observer, ReferenceWrapper<FStreamObserver>(observer)))
      .second;
    return false;
  }

  void unregisterObserver(FStreamObserver& observer)
  {
    m_observerMap.erase(&observer);
  }

protected:
  ErrorE m_error;
  filesystem::Path m_path;
  FileFlagsT m_flags;
  ObserverMapT m_observerMap;
};

class FStream : public FStreamBase
{
public:
  FStream();
  virtual ~FStream();

  FStream(const FStream&) = delete;
  FStream& operator=(const FStream&) = delete;

  FStream(FStream&& other);
  FStream& operator=(FStream&& other);

  FStream(filesystem::Path path, const FileFlagsT flags);

  virtual bool open(filesystem::Path path, const FileFlagsT flags) override;

  virtual bool isOpen() const override;

  virtual SizeT write(const UInt8T* const data, SizeT size) override;

  virtual SizeT read(UInt8T* data, SizeT maxSize) override;

  virtual bool seek(const SizeT offset, const SeekModeE seekMode) override;

  virtual SizeT position() const override;

  virtual void flush() override;

  virtual void close() override;

  virtual SizeT size() const override;

  operator bool() const;

  template<typename T>
  FStream& operator <<(const T& value)
  {
    FStreamBase::operator<<(value);

    return *this;
  }

protected:
  class FStreamImpl;

  UniquePtr<FStreamImpl> m_impl;
};

class FStreamObserver
{
public:
  FStreamObserver(stream::FStreamBase& stream);

  FStreamObserver(const FStreamObserver& other) = delete;
  FStreamObserver& operator=(const FStreamObserver&) = delete;

  FStreamObserver(FStreamObserver&& other);

  virtual ~FStreamObserver();

  FStreamObserver& operator=(FStreamObserver&& other);

  bool alive() const;

  stream::FStreamBase& file();

  operator bool() const;

private:
  friend class FStreamBase;

  bool m_alive;
  ReferenceWrapper<stream::FStreamBase> m_file;
};

FStreamBase::DataContainerT
load(FStreamBase& stream);

String
loadString(FStreamBase& stream);
}

#endif // __STL_FSTREAM_HPP__
