#ifndef _STL_FUNCTIONAL_HPP_
#define _STL_FUNCTIONAL_HPP_

#include "stl/stl_config.hpp"

#include "stl/assert.hpp"
#include "stl/limits.hpp"
#include "stl/utility.hpp"
#include "stl/math.hpp"

namespace stl {
template<typename T>
class ReferenceWrapper
{
public:
  ReferenceWrapper() = delete;

  ReferenceWrapper(T& instance)
    : m_ptr(&instance)
  {
  }

  ReferenceWrapper(ReferenceWrapper&& other) noexcept
  {
    Assert(nullptr != other.m_ptr);
    m_ptr = other.m_ptr;
  }

  ReferenceWrapper& operator=(ReferenceWrapper&& other) noexcept
  {
    Assert(nullptr != other.m_ptr);
    m_ptr = other.m_ptr;

    return *this;
  }

  ReferenceWrapper(const ReferenceWrapper& other) noexcept
  {
    Assert(nullptr != other.m_ptr);
    m_ptr = other.m_ptr;
  }

  ReferenceWrapper& operator=(const ReferenceWrapper& other) noexcept
  {
    Assert(nullptr != other.m_ptr);
    m_ptr = other.m_ptr;

    return *this;
  }

  constexpr operator T&() const noexcept { return *m_ptr; }
  operator T&() noexcept { return *m_ptr; }

  constexpr T& get() const noexcept { return *m_ptr; }
  T& get() noexcept { return *m_ptr; }

  const T& operator()() const { return get(); }
  T& operator()() { return get(); }

  bool operator < (const ReferenceWrapper& other) const
  {
    return m_ptr < other.m_ptr;
  }

private:
  T* m_ptr;
};

template<typename T>
class Less
{
public:
  bool operator()(const T& left, const T& right) { return left < right; }

  bool operator()(const T& left, const T& right) const
  {
    return left < right;
  }
};

template<typename T>
class Greater
{
public:
  bool operator()(const T& left, const T& right) { return left > right; }

  bool operator()(const T& left, const T& right) const
  {
    return left > right;
  }
};

template<typename T>
class Equal
{
public:
  bool operator()(const T& left, const T& right) { return left == right; }

  const bool operator()(const T& left, const T& right) const
  {
    return left == right;
  }
};

#if defined(STL_ENABLE_FLOAT) && (STL_ENABLE_FLOAT == 1)
template<>
class Equal<float>
{
public:
  bool operator()(const float& left, const float& right)
  {
    return math::abs(left - right) < Limits<float>::epsilon();
  }

  const bool operator()(const float& left, const float& right) const
  {
    return math::abs(left - right) < Limits<float>::epsilon();
  }
};
#endif // defined(STL_ENABLE_FLOAT) && (STL_ENABLE_FLOAT == 1)

#if defined(STL_ENABLE_DOUBLE) && (STL_ENABLE_DOUBLE == 1)
template<>
class Equal<double>
{
public:
  bool operator()(const double& left, const double& right)
  {
    return math::abs(left - right) < Limits<double>::epsilon();
  }

  const bool operator()(const double& left, const double& right) const
  {
    return math::abs(left - right) < Limits<double>::epsilon();
  }
};
#endif // defined(STL_ENABLE_DOUBLE) && (STL_ENABLE_DOUBLE == 1)
}

#endif // _STL_FUNCTIONAL_HPP_
