#ifndef _HTTP_PARSER_HPP_
#define _HTTP_PARSER_HPP_

#include "stl/string.hpp"
#include "stl/vector.hpp"
#include "stl/list.hpp"
#include "stl/stddef.hpp"
#include "stl/iostream.hpp"

#define WS_DDEBUG

namespace http {
    enum class VersionE {
        HTTP_V0_9
        , HTTP_V1_0
        , HTTP_V1_1
        , HTTP_V2_0
        , _Count
    };

    enum class CodeE {
        CodeE_100_Continue
        , CodeE_103_EarlyHints
        , CodeE_200_OK
        , CodeE_201_Created
        , CodeE_204_NoContent
        , CodeE_206_PartialContent
        , CodeE_301_MovedPermanently
        , CodeE_302_Found
        , CodeE_303_SeeOther
        , CodeE_304_NotModified
        , CodeE_307_TemporaryRedirect
        , CodeE_308_PermanentRedirect
        , CodeE_401_Unauthorized
        , CodeE_403_Forbidden
        , CodeE_404_NotFound
        , CodeE_406_NotAcceptable
        , CodeE_407_ProxyAuthenticationRequired
        , CodeE_409_Conflict
        , CodeE_410_Gone
        , CodeE_412_PreconditionFailed
        , CodeE_416_RangeNotSatisfiable
        , CodeE_418_ImATeapot
        , CodeE_425_TooEarly
        , CodeE_451_UnavailableForLegalReasons
        , CodeE_500_InternalServerError
        , CodeE_501_NotImplemented
        , CodeE_502_BadGateway
        , CodeE_503_ServiceUnavailable
        , CodeE_504_GatewayTimeout
        , _Count
    };

    inline stl::String getVersionString(const VersionE ver) {
        Assert(static_cast<stl::SizeT>(ver) < static_cast<stl::SizeT>(VersionE::_Count));

        static const stl::String lut[] = {
            "0.9"
            , "1.0"
            , "1.1"
            , "2.0"
        };

        return lut[static_cast<stl::SizeT>(ver)];
    }

    inline stl::UInt8T getVersionMajor(const VersionE ver) {
        Assert(static_cast<stl::SizeT>(ver) < static_cast<stl::SizeT>(VersionE::_Count));

        static stl::UInt8T lut[] = {
            0U, 1U, 1U, 2U
        };

        return lut[static_cast<stl::SizeT>(ver)];
    }

    inline stl::UInt8T getVersionMinor(const VersionE ver) {
        Assert(static_cast<stl::SizeT>(ver) < static_cast<stl::SizeT>(VersionE::_Count));

        static stl::UInt8T lut[] = {
            9U, 0U, 1U, 0U
        };

        return lut[static_cast<stl::SizeT>(ver)];
    }

    inline stl::String getCodeString(const CodeE code) {
        Assert(static_cast<stl::SizeT>(code) < static_cast<stl::SizeT>(CodeE::_Count));

        static const stl::String lut[] = {
            "100", "103", "200", "201"
            , "204", "206", "301", "302"
            , "303", "304", "307", "308"
            , "401", "403", "404", "406"
            , "407", "409", "410", "412"
            , "416", "418", "425", "451"
            , "500", "501", "502", "503"
            , "504"
        };

        return lut[static_cast<stl::SizeT>(code)];
    }

    inline stl::UInt16T getCode(const CodeE code) {
        Assert(static_cast<stl::SizeT>(code) < static_cast<stl::SizeT>(CodeE::_Count));

        static const stl::UInt16T lut[] = {
            100U, 103U, 200U, 201U
            , 204U, 206U, 301U, 302U
            , 303U, 304U, 307U, 308U
            , 401U, 403U, 404U, 406U
            , 407U, 409U, 410U, 412U
            , 416U, 418U, 425U, 451U
            , 500U, 501U, 502U, 503U
            , 504U
        };

        return lut[static_cast<stl::SizeT>(code)];
    }

    class HTTPURL {
    public:
        static bool isReservedChar(const char c);

        // TODO: Percent encoded characters

        static bool isUnreservedChar(const char c);

        static bool isAllowedChar(const char c);
    };

    class HTTPHeader {
    public:
        HTTPHeader() = default;

        HTTPHeader(stl::String name, stl::String value);

        stl::String m_name;
        stl::String m_value;
    };

    class HTTPRequest {
    public:
        using HeaderList = stl::List<HTTPHeader>;

        enum class MethodE {
            Get
            , Head
            , Post
            , Put
            , Delete
            , Connect
            , Options
            , Trace
            , Path
        };

        const stl::String& versionMajor() const;

        const stl::String& versionMinor() const;

        void addHeader(stl::String&& name, stl::String&& value);

        const stl::String& uri() const;

        const stl::String& content() const;

#if defined STL_ENABLE_IOSTREAM
        void print() {
            stl::cout << m_method << ' ' << m_uri << ' ' << "HTTP/"
                << m_versionMajor << '.' << m_versionMinor << "\r\n";

            const HeaderList::ListItem* item = m_headerList.front();
            while (nullptr != item) {
                stl::cout << item->item().m_name << ": "
                    << item->item().m_value << "\r\n";

                item = item->next();
            }
        }
#endif

    private:
        stl::String m_versionMinor;
        stl::String m_versionMajor;
        stl::String m_method; // TODO: Map containing text?
        stl::String m_uri; // TODO: Own class
        stl::String m_content;
        HeaderList m_headerList;
        bool m_keepAlive = false;

        friend class HTTPRequestParser;
    };

    class HTTPRequestParser {
    public:
        enum class ReturnE {
            ReturnE_Ok
            , ReturnE_Fail
            , ReturnE_Continue
            , ReturnE_ResultReady
        };

    private:
        using ParserFuncT = ReturnE(*)(HTTPRequestParser& instance
            , const char*& dataCurr
            , const char* const dataEnd);

    public:
        HTTPRequestParser() = default;

        ReturnE consume(const char* const data, const stl::SizeT length);

        void reset();

        bool hasResult() const;

        stl::UniquePtr<HTTPRequest> result();

    private:
        static ReturnE parseSecondLF(HTTPRequestParser& instance
            , const char*& dataCurr
            , const char* const dataEnd);

        static ReturnE parseSecondCROrHeader(HTTPRequestParser& instance
            , const char*& dataCurr
            , const char* const dataEnd);

        static ReturnE parseFirstLF(HTTPRequestParser& instance
            , const char*& dataCurr
            , const char* const dataEnd);

        static ReturnE parseHeaderValue(HTTPRequestParser& instance
            , const char*& dataCurr
            , const char* const dataEnd);

        static ReturnE skipSpacesBeforeHeaderValue(HTTPRequestParser& instance
            , const char*& dataCurr
            , const char* const dataEnd);

        static ReturnE parseDoubleDotAfterHeaderName(HTTPRequestParser& instance
            , const char*& dataCurr
            , const char* const dataEnd);

        static ReturnE parseHeaderName(HTTPRequestParser& instance
            , const char*& dataCurr
            , const char* const dataEnd);

        static ReturnE parseRequestLineVerMinor(HTTPRequestParser& instance
            , const char*& dataCurr
            , const char* const dataEnd);

        static ReturnE parseRequestLineVerMajor(HTTPRequestParser& instance
            , const char*& dataCurr
            , const char* const dataEnd);

        static ReturnE parseRequestLineSlashAfterHTTP(HTTPRequestParser& instance
            , const char*& dataCurr
            , const char* const dataEnd);

        static ReturnE parseRequestLineHTTP(HTTPRequestParser& instance
            , const char*& dataCurr
            , const char* const dataEnd);

        static ReturnE parseRequestLineHTT(HTTPRequestParser& instance
            , const char*& dataCurr
            , const char* const dataEnd);

        static ReturnE parseRequestLineHT(HTTPRequestParser& instance
            , const char*& dataCurr
            , const char* const dataEnd);

        static ReturnE parseRequestLineH(HTTPRequestParser& instance
            , const char*& dataCurr
            , const char* const dataEnd);

        static ReturnE parseRequestLineWSBeforeH(HTTPRequestParser& instance
            , const char*& dataCurr
            , const char* const dataEnd);

        static ReturnE parseRequestLineURI(HTTPRequestParser& instance
            , const char*& dataCurr
            , const char* const dataEnd);

        static ReturnE parseRequestLineWSBeforeURI(HTTPRequestParser& instance
            , const char*& dataCurr
            , const char* const dataEnd);

        static ReturnE parseRequestLineMethod(HTTPRequestParser& instance
            , const char*& dataCurr
            , const char* const dataEnd);

        ParserFuncT m_parserFunc = parseRequestLineMethod;
        stl::UniquePtr<HTTPRequest> m_result;
        bool m_resultReady = false;
        bool m_stringInProgress = false;
    };

    class HTTPResponse {
    public:
        HTTPResponse() = default;

        int consume(const char* const data, const stl::SizeT length);

        void setVersion(const VersionE ver) {
            m_versionMajor = getVersionMajor(ver);
            m_versionMinor = getVersionMinor(ver);
        }

        void setStatusCode(const CodeE code) {
            m_statusCode = getCode(code);
        }

        void setStatus(stl::String status) {
            m_status = stl::move(status);
        }

        void keepAlive() {
            m_keepAlive = true;
        }

        void addHeader(HTTPHeader header) {
            m_headerList.emplaceBack(stl::move(header));
        }

        void addContent(stl::String content) {
            m_content = stl::move(content);
        }

    private:
        stl::UInt8T m_versionMinor = 0U;
        stl::UInt8T m_versionMajor = 0U;
        stl::UInt16T m_statusCode = 0U;
        stl::String m_status;
        stl::String m_content;
        stl::List<HTTPHeader> m_headerList;
        bool m_keepAlive = false;
    };
}

#endif
