#ifndef _LIST_HPP_
#define _LIST_HPP_

#include "stl/assert.hpp"
#include "stl/stddef.hpp"
#include "stl/stl_config.hpp"

#include <list>

namespace stl
{
template<typename T>
using List2 = std::list<T>;

template <typename T>
class List
{
public:
    using BaseT = T;

    class ListItem
    {
    public:
        ListItem() = default;
        ListItem(const ListItem&) = delete;

        ListItem(BaseT&& item)
            : m_next(nullptr), m_item(makeUnique<BaseT>(move(item))),
              m_prev(nullptr)
        {
        }

        ListItem(ListItem&& other)
            : m_next(move(other.m_next)), m_item(move(other.m_item)),
              m_prev(move(other.m_prev))
        {
        }

        ListItem(UniquePtr<ListItem>&& next, BaseT&& item, ListItem* const prev)
            : m_next(move(next)), m_item(makeUnique<BaseT>(move(item))),
              m_prev(prev)
        {
        }

        ListItem& operator=(const ListItem&) = delete;

        ListItem& operator=(ListItem&& other)
        {
            m_next = move(other.m_next);
            m_item = move(other.m_item);
            m_prev = other.m_prev;
            return *this;
        }

        void push(BaseT&& item)
        {
            if (m_next)
            {
                UniquePtr<ListItem> backup = move(m_next);
                m_next = makeUnique<ListItem>(move(item));
                m_next->m_next = move(backup);
            }
            else
            {
                m_next = makeUnique<ListItem>(move(item));
            }
        }

        void pop()
        {
            Assert(m_next);

            m_next.reset();
        }

        ListItem* next() { return m_next.get(); }

        ListItem* prev() { return m_prev; }

        const ListItem* next() const { return m_next.get(); }

        const ListItem* prev() const { return m_prev; }

        BaseT& item() { return *m_item.get(); }

        const BaseT& item() const { return *m_item.get(); }

    private:
        UniquePtr<ListItem> m_next;
        UniquePtr<BaseT> m_item;
        ListItem* m_prev = nullptr;
    };

    class Iterator
    {
    public:
        Iterator() : m_list(nullptr), m_listItem(nullptr) {}
        Iterator(List<BaseT>* const list, List<BaseT>::ListItem* const listItem)
            : m_list(list), m_listItem(listItem)
        {
            Assert(nullptr != m_list);
        }

        Iterator(const Iterator& other)
            : m_list(other.m_list), m_listItem(other.m_listItem)
        {
            Assert(nullptr != m_list);
        }

        Iterator& operator=(const Iterator& other)
        {
            Assert(nullptr != other.m_list);

            m_list = other.m_list;
            m_listItem = other.m_listItem;

            return *this;
        }

        Iterator(Iterator&& other)
            : m_list(move(other.m_list)), m_listItem(move(other.m_listItem))
        {
            Assert(nullptr != m_list);
        }

        Iterator& operator=(Iterator&& other)
        {
            Assert(nullptr != other.m_list);

            m_list = move(other.m_list);
            m_listItem = move(other.m_listItem); // TODO: Exchange with null?

            return *this;
        }

        using ValueType = BaseT;
        using Pointer = BaseT*;
        using Reference = BaseT&;

        // Default Copy/Move Are Fine.
        // Default Destructor fine.

        Reference operator*()
        {
            Assert(nullptr != m_listItem);

            return m_listItem->item();
        }

        const Reference operator*() const
        {
            Assert(nullptr != m_listItem);

            return m_listItem->item();
        }

        Pointer operator->()
        {
            Assert(nullptr != m_listItem);

            return &m_listItem->item();
        }

        const Pointer operator->() const
        {
            Assert(nullptr != m_listItem);

            return &m_listItem->item();
        }

        // Reference operator[](int m) { return (*v)[i + m]; }

        // const Reference operator[](int m) const { return (*v)[i + m]; }

        Iterator& operator++()
        {
            Assert(nullptr != m_listItem);

            m_listItem = m_listItem->next();

            return *this;
        }

        Iterator& operator--()
        {
            Assert(nullptr != m_listItem);

            m_listItem = m_listItem->prev();

            return *this;
        }

        Iterator operator++(int)
        {
            Iterator r(*this);

            Assert(nullptr != m_listItem);

            m_listItem = m_listItem->next();

            return r;
        }

        Iterator operator--(int)
        {
            Iterator r(*this);

            Assert(nullptr != m_listItem);

            m_listItem = m_listItem->prev();

            return r;
        }

        /* Iterator& operator+=(int n)
        {
            i += n;
            return *this;
        }

        Iterator& operator-=(int n)
        {
            i -= n;
            return *this;
        }

        Iterator operator+(int n) const
        {
            Iterator r(*this);
            return r += n;
        }
        Iterator operator-(int n) const
        {
            Iterator r(*this);
            return r -= n;
        } */

        // DifferenceType operator-(Iterator const& r) const {return i - r.i;}

        // Note: comparing iterator from different containers
        //       is undefined behavior so we don't need to check
        //       if they are the same container.

        bool operator<(Iterator const& other) const
        {
            return m_listItem < other.m_listItem;
        }

        bool operator<=(Iterator const& other) const
        {
            return m_listItem <= other.m_listItem;
        }

        bool operator>(Iterator const& other) const
        {
            return m_listItem > other.m_listItem;
        }

        bool operator>=(Iterator const& other) const
        {
            return m_listItem >= other.m_listItem;
        }

        bool operator!=(const Iterator& other) const
        {
            return m_listItem != other.m_listItem;
        }

        bool operator==(const Iterator& other) const
        {
            return m_listItem == other.m_listItem;
        }

    private:
        List* m_list;
        ListItem* m_listItem;
    };

    class ConstIterator
    {
    public:
        ConstIterator() : m_list(nullptr), m_listItem(nullptr) {}
        ConstIterator(const List<BaseT>* const list,
                      const List<BaseT>::ListItem* const listItem)
            : m_list(list), m_listItem(listItem)
        {
            Assert(nullptr != m_list);
        }

        ConstIterator(const ConstIterator& other)
            : m_list(other.m_list), m_listItem(other.m_listItem)
        {
            Assert(nullptr != m_list);
        }

        ConstIterator& operator=(const ConstIterator& other)
        {
            Assert(nullptr != other.m_list);

            m_list = other.m_list;
            m_listItem = other.m_listItem;

            return *this;
        }

        ConstIterator(ConstIterator&& other)
            : m_list(move(other.m_list)), m_listItem(move(other.m_listItem))
        {
            Assert(nullptr != m_list);
        }

        ConstIterator& operator=(ConstIterator&& other)
        {
            Assert(nullptr != other.m_list);

            m_list = move(other.m_list);
            m_listItem = move(other.m_listItem); // TODO: Exchange with null?

            return *this;
        }

        using ValueType = BaseT;
        using Pointer = const BaseT*;
        using Reference = const BaseT&;

        // Default Copy/Move Are Fine.
        // Default Destructor fine.

        Reference operator*() const
        {
            Assert(nullptr != m_listItem);

            return m_listItem->item();
        }

        Pointer operator->() const
        {
            Assert(nullptr != m_listItem);
            
            return &m_listItem->item();
        }

        // const Reference operator[](int m) const { return (*v)[i + m]; }

        ConstIterator& operator++()
        {
            Assert(nullptr != m_listItem);

            m_listItem = m_listItem->next();

            return *this;
        }

        ConstIterator& operator--()
        {
            Assert(nullptr != m_listItem);

            m_listItem = m_listItem->prev();

            return *this;
        }

        ConstIterator operator++(int)
        {
            ConstIterator r(*this);

            Assert(nullptr != m_listItem);

            m_listItem = m_listItem->next();

            return r;
        }

        ConstIterator operator--(int)
        {
            ConstIterator r(*this);

            Assert(nullptr != m_listItem);

            m_listItem = m_listItem->prev();

            return r;
        }

        /* ConstIterator& operator+=(int n)
        {
            i += n;
            return *this;
        }

        ConstIterator& operator-=(int n)
        {
            i -= n;
            return *this;
        }

        ConstIterator operator+(int n) const
        {
            ConstIterator r(*this);
            return r += n;
        }

        ConstIterator operator-(int n) const
        {
            ConstIterator r(*this);
            return r -= n;
        } */

        // DifferenceType operator-(ConstIterator const& r) const {return i -
        // r.i;}

        // Note: comparing iterator from different containers
        //       is undefined behavior so we don't need to check
        //       if they are the same container.

        bool operator<(ConstIterator const& other) const
        {
            return m_listItem < other.m_listItem;
        }

        bool operator<=(ConstIterator const& other) const
        {
            return m_listItem <= other.m_listItem;
        }

        bool operator>(ConstIterator const& other) const
        {
            return m_listItem > other.m_listItem;
        }

        bool operator>=(ConstIterator const& other) const
        {
            return m_listItem >= other.m_listItem;
        }

        bool operator!=(const ConstIterator& other) const
        {
            return m_listItem != other.m_listItem;
        }

        bool operator==(const ConstIterator& other) const
        {
            return m_listItem == other.m_listItem;
        }

    private:
        const List* m_list;
        const ListItem* m_listItem;
    };

    List() = default;

public:
    ListItem* front() { return m_root.get(); }

    const ListItem* front() const { return m_root; }

    ListItem* back() { return m_last; }

    const ListItem* back() const { return m_last; }

    Iterator begin() { return Iterator(this, m_root.get()); }

    ConstIterator begin() const { return ConstIterator(this, m_root.get()); }

    ConstIterator cbegin() const { return ConstIterator(this, m_root.get()); }

    Iterator end() { return Iterator(this, nullptr); }

    ConstIterator end() const { return ConstIterator(this, nullptr); }

    ConstIterator cend() const { return ConstIterator(this, nullptr); }

    SizeT size() const { return m_size; }

    bool empty() const { return m_size == 0U; }

    template <typename... ArgsT>
    void emplaceBack(ArgsT&&... Args)
    {
        this->pushBack(BaseT(forward<ArgsT>(Args)...));
    }

    template <typename... ArgsT>
    void emplaceFront(ArgsT&&... Args)
    {
        this->pushFront(BaseT(forward<T>(Args)...));
    }

    void pushBack(BaseT&& item)
    {
        if (m_last)
        {
            m_last->push(move(item));
            m_last = m_last->next();
        }
        else
        {
            m_root = makeUnique<ListItem>(move(item));
            m_last = m_root.get();
        }

        ++this->m_size;
    }

    void popBack()
    {
        if (m_last)
        {
            if (m_last == m_root.get())
            {
                m_root.reset();
                m_last = nullptr;
            }
            else
            {
                ListItem* const prevBackup = m_last->prev();
                m_last->prev()->pop();
                m_last = prevBackup;
            }
            --this->m_size;
        }
        else
        {
            THROW();
        }
    }

    void pushFront(BaseT item)
    {
        if (m_root)
        {
            UniquePtr<ListItem> temp = move(m_root);
            m_root = makeUnique<ListItem>(move(item));
            m_root->push(move(temp));
        }
        else
        {
            m_root = makeUnique<ListItem>(move(item));
            m_last = m_root.get();
        }
        ++this->m_size;
    }

    void popFront()
    {
        if (m_root)
        {
            if (m_last == m_root.get())
            {
                m_last = nullptr;
            }

            if (m_root->next())
            {
                m_root = move(m_root->next());
            }

            --this->m_size;
        }
        else
        {
            THROW();
        }
    }

    // TODO: Insert
    // TODO: Emplace
    // TODO: Merge
    // TODO: Sort
    // TODO: Swap
    // TODO: Reverse
    // TODO: Splice
    // TODO: Unique
    // TODO: RemoveIf

    // TODO:
    /* void resize(const SizeT count) {
        if(count > m_size) {

        } else if(count < m_size) {

        }
    }

    // TODO:
    void resize(const SizeT count, const BaseT& value) {

    } */

    // TODO:
    /* void erase(ListItem& item) {
        if(&item == m_root.get()) {

        } else if(&item == m_last) {

        } else {

        }
    } */

    void clear()
    {
        m_root.reset();
        this->m_size = 0U;
    }

private:
    UniquePtr<ListItem> m_root;
    ListItem* m_last = nullptr;
    SizeT m_size = 0U;
};
} // namespace stl

#endif
