#ifndef __STL_MAP_HPP__
#define __STL_MAP_HPP__

#include "stl/functional.hpp"
#include "stl/memory.hpp"
#include "stl/utility.hpp"

#include <map>

namespace stl {

template<typename KeyT, typename T, typename CompareT = Less<KeyT>>
using Map = std::map<KeyT, T, CompareT>;

template<typename KeyT, typename T, typename CompareT = Less<KeyT>>
using MultiMap = std::multimap<KeyT, T, CompareT>;

#ifdef HRIC_IMPL // TODO Implement
template<typename KeyT, typename T, typename CompareT = Less<KeyT>>
class Map // TODO: Derive from tree (or use it)
{
private:
  struct Node
  {
    Node(Node* const root, const T& value)
      : m_root(root)
      , m_value(value)
    {
    }

    Node(Node* const root, T&& value)
      : m_root(root)
      , m_value(move(value))
    {
    }

    Node* m_root;
    UniquePtr<Node> m_lower;
    UniquePtr<Node> m_higher;

    T m_value;
  };

  public:

  using SizeTypeT = SizeT;

  class Iterator
  {
  public:
    Iterator()
      : m_set(nullptr)
      , m_setNode(nullptr)
    {
    }

    Iterator(Set<T>* const set, Set<T>::Node* const setNode)
      : m_set(set)
      , m_setNode(setNode)
    {
      Assert(nullptr != m_set);
    }

    Iterator(const Iterator& other)
      : m_set(other.m_set)
      , m_setNode(other.m_setNode)
    {
      Assert(nullptr != m_set);
    }

    Iterator& operator=(const Iterator& other)
    {
      Assert(nullptr != other.m_set);

      m_set = other.m_set;
      m_setNode = other.m_setNode;

      return *this;
    }

    Iterator(Iterator&& other)
      : m_set(exchange(other.m_set, nullptr))
      , m_setNode(exchange(other.m_setNode, nullptr))
    {
      Assert(nullptr != m_set);
    }

    Iterator& operator=(Iterator&& other)
    {
      Assert(nullptr != other.m_set);

      m_set = exchange(other.m_set, nullptr);
      m_setNode = exchange(other.m_setNode, nullptr);

      return *this;
    }

    using ValueType = T;
    using Pointer = T*;
    using Reference = T&;

    // Default Copy/Move Are Fine.
    // Default Destructor fine.

    Reference operator*()
    {
      Assert(nullptr != m_setNode);

      return m_setNode->m_value;
    }

    const Reference operator*() const
    {
      Assert(nullptr != m_setNode);

      return m_setNode->m_value;
    }

    Pointer operator->()
    {
      Assert(nullptr != m_setNode);

      return &m_setNode->m_value;
    }

    const Pointer operator->() const
    {
      Assert(nullptr != m_setNode);

      return &m_setNode->m_value;
    }

    // Reference operator[](int m) { return (*v)[i + m]; }

    // const Reference operator[](int m) const { return (*v)[i + m]; }

    Iterator& operator++()
    {
      Assert(nullptr != m_setNode);
      Assert(nullptr != m_setNode->m_higher);

      m_setNode = m_setNode->m_higher.get();

      return *this;
    }

    Iterator& operator--()
    {
      Assert(nullptr != m_setNode);

      m_setNode = m_setNode->;

      return *this;
    }

    Iterator operator++(int)
    {
      Iterator r(*this);

      Assert(nullptr != m_setNode);

      m_setNode = m_setNode->next();

      return r;
    }

    Iterator operator--(int)
    {
      Iterator r(*this);

      Assert(nullptr != m_setNode);

      m_setNode = m_setNode->prev();

      return r;
    }

    /* Iterator& operator+=(int n)
    {
        i += n;
        return *this;
    }

    Iterator& operator-=(int n)
    {
        i -= n;
        return *this;
    }

    Iterator operator+(int n) const
    {
        Iterator r(*this);
        return r += n;
    }
    Iterator operator-(int n) const
    {
        Iterator r(*this);
        return r -= n;
    } */

    // DifferenceType operator-(Iterator const& r) const {return i - r.i;}

    // Note: comparing iterator from different containers
    //       is undefined behavior so we don't need to check
    //       if they are the same container.

    bool operator<(Iterator const& other) const
    {
      return m_setNode < other.m_setNode;
    }

    bool operator<=(Iterator const& other) const
    {
      return m_setNode <= other.m_setNode;
    }

    bool operator>(Iterator const& other) const
    {
      return m_setNode > other.m_setNode;
    }

    bool operator>=(Iterator const& other) const
    {
      return m_setNode >= other.m_setNode;
    }

    bool operator!=(const Iterator& other) const
    {
      return m_setNode != other.m_setNode;
    }

    bool operator==(const Iterator& other) const
    {
      return m_setNode == other.m_setNode;
    }

  private:
    Set* m_set;
    Node* m_setNode;
  };

  Set() = default;
  virtual ~Set();

  Iterator begin() { return Iterator(this, m_root.get()); }

  ConstIterator cbegin() const { return ConstIterator(this, m_root.get()); }

  Iterator end() { return Iterator(this, nullptr); }

  ConstIterator cend() const { return ConstIterator(this, nullptr); }

  void clear() { m_root.reset(); }

  bool contains(const T& key) const { return cend() != find(key); }

  bool empty() const { return static_cast<SizeTypeT>(0) == m_size; }

  SizeTypeT size() const { return m_size; }

  Iterator erase(ConstIterator it) {}

  Iterator erase(Iterator it) {}

  SizeT erase(const KeyT& value) {} // 1 or 0

  Iterator find(const KeyT& value) {}

  ConstIterator find(const KeyT& value) const {}

  Pair<Iterator, bool> emplace() {}

  template<typename... ArgsT>
  Pair<Iterator, bool> emplace(KeyT key, ArgsT&&... args)
  {
    return insert(Pair<KeyT, T>(move(key), T(forward<ArgsT>(args)...)));
  }

  Pair<Iterator, bool> insert(const Pair<KeyT, T>& pair)
  {
    return insert(T(pair));
  }

  Pair<Iterator, bool> insert(Pair<KeyT, T>&& pair)
  {
    return Pair<Iterator, bool>(Iterator(this, nullptr), true);
  }

  // void merge(...) ...

private:
  UniquePtr<Node> m_root;
  SizeTypeT m_size;
};
#endif // HRIC_IMPL

}

#endif // __STL_SET_HPP__
