#ifndef _MATH_HPP_
#define _MATH_HPP_

#include "stl/limits.hpp"
#include "stl/memory.hpp"
#include "stl/stddef.hpp"
#include "stl/stl_config.hpp"

#include <cmath>

#if defined abs
#undef abs
#endif // abs

namespace stl {
namespace math {
static constexpr float PI_FLT = 3.14159265358979323846f;
static constexpr double PI_DBL = 3.14159265358979323846;
static constexpr long double PI_LDB = 3.141592653589793238462643383279502884L;
// TODO: Specialization for unsigned types to prevent overhead
template<typename T>
inline T
abs(const T a)
{
  return a < 0 ? -a : a;
}

template<typename T>
inline T
min(const T a, const T b)
{
  return a < b ? a : b;
}

template<typename T>
inline T
max(const T a, const T b)
{
  return a > b ? a : b;
}

template<typename T>
inline bool
equal(const T a, const T b)
{
  return a == b ? a : b;
}

#ifdef STL_ENABLE_FLOAT
template<>
inline bool
equal<float>(const float a, const float b)
{
  return abs(a - b) < Limits<float>::epsilon();
}

#ifdef STL_ENABLE_DOUBLE
template<>
inline bool
equal<double>(const double a, const double b)
{
  return abs(a - b) < Limits<double>::epsilon();
}
#endif

template<typename T>
inline T
sin(const T value)
{
  return std::sin(value);
}

template<typename T>
inline T
cos(const T value)
{
  return std::cos(value);
}
#endif

template<typename T>
T
adjust(const T& min, const T& max, const T& value)
{
  if (equal(min, value)) {
    return value;
  } else if (equal(max, value)) {
    return value;
  } else if (value < min) {
    return min;
  } else {
    return max;
  }
}

template<typename T>
T
sign(const T& value)
{
  if (equal(static_cast<T>(0), value)) {
    return static_cast<T>(0);
  } else if (value > static_cast<T>(0)) {
    return static_cast<T>(1);
  } else {
    return static_cast<T>(-1);
  }
}
}

namespace math::vec {
struct Vec2
{
  Vec2(float x, float y)
    : x(x)
    , y(y)
  {
  }

  float x, y;
};

Vec2
operator+(Vec2 a, Vec2 b);

Vec2
operator-(Vec2 a, Vec2 b);

Vec2
operator*(float s, Vec2 a);
}
}

#endif
