#ifndef __MEMORY_HPP__
#define __MEMORY_HPP__

#include "stl/allocate.hpp"
#include "stl/assert.hpp"
#include "stl/limits.hpp"
#include "stl/new.hpp"
#include "stl/stddef.hpp"
#include "stl/stl_config.hpp"
#include "stl/utility.hpp"

namespace stl {
using ByteT = UInt8T;

namespace memory {
template<typename T>
SizeT
firstActiveBitPos(const T value)
{
  const SizeT bitCount = sizeof(T) * static_cast<SizeT>(8);

  for (SizeT i = 0U; i < bitCount; ++i) {
    if (0 != ((value >> i) & 0x01U)) {
      return i;
    }
  }

  return Limits<T>::max();
}

template<typename T>
T*
copy(const T* begin, const T* const end, T* const output)
{
  Assert(nullptr != begin);
  Assert(nullptr != end);
  Assert(nullptr != output);

  T* outputPtr = output;

  while (begin != end) {
    *(outputPtr++) = *(begin++);
  }

  return output;
}

template<typename T>
bool
compare(const T* a, const T* b, stl::SizeT count)
{
  Assert(nullptr != a);
  Assert(nullptr != b);

  while (count--) {
    if (*(a++) != *(b++)) {
      return false;
    }
  }

  return true;
}

template<typename T>
T*
fill(const T fillWith, stl::SizeT count, T* const output)
{
  Assert(nullptr != output);

  T* outputPtr = output;

  while (count--) {
    *(outputPtr++) = fillWith;
  }

  return output;
}

template<typename T>
T*
zero(stl::SizeT count, T* const output)
{
  Assert(nullptr != output);

  ::stl::UInt8T* outputPtr = reinterpret_cast<stl::UInt8T*>(&output[0]);

  ::stl::SizeT size = sizeof(T) * count;

  while (size--) {
    *(outputPtr++) = static_cast<::stl::UInt8T>(0x00U);
  }

  return output;
}

template<typename T>
T*
zero(T* const output)
{
  Assert(nullptr != output);

  ::stl::UInt8T* outputPtr = reinterpret_cast<stl::UInt8T*>(&output[0]);
  ::stl::SizeT size = sizeof(T);

  while (size--) {
    *(outputPtr++) = static_cast<::stl::UInt8T>(0x00U);
  }

  return output;
}
} // namespace memory

template<typename T>
class DefaultDeleter
{
public:
  DefaultDeleter() noexcept = default;

  DefaultDeleter(const DefaultDeleter&) {}

  void operator()(T* const ptr) const noexcept
  {
    // memory::MemoryManager<T>::deallocate(ptr);
    delete ptr;
  }
};

template<typename T>
class DefaultDeleter<T[]>
{
public:
  DefaultDeleter() noexcept = default;

  DefaultDeleter(const DefaultDeleter&) {}

  void operator()(T* const ptr) const noexcept
  {
    // memory::MemoryManager<T[]>::deallocate(ptr);
    delete [] ptr;
  }
};

template<typename T, typename DeleterT>
class UniquePtr
{
public:
  using DeleterType = DeleterT;

  UniquePtr() = default;
  UniquePtr(const UniquePtr&) = delete;

  UniquePtr(UniquePtr&& other) noexcept {
    reset(other.m_ptr);
    other.m_ptr = nullptr;
  }

  UniquePtr(T* const ptr)
    : m_ptr(ptr)
  {
  }

  ~UniquePtr() { reset(); }

  UniquePtr& operator=(const UniquePtr& other) = delete;

  UniquePtr& operator=(UniquePtr&& other) noexcept
  {
    reset(other.m_ptr);
    other.m_ptr = nullptr;

    return *this;
  }

  UniquePtr& operator=(T* const newPtr)
  {
    reset(newPtr);

    return *this;
  }

  void reset(T* const newPtr = nullptr)
  {
    if (m_ptr) {
      m_deleter(m_ptr);
    }

    m_ptr = newPtr;
  }

  T* get() { return m_ptr; }

  const T* get() const { return m_ptr; }

  T* operator()() { return m_ptr; }

  const T* operator()() const { return m_ptr; }

  T* release()
  {
    T* temp = m_ptr;
    m_ptr = nullptr;
    return temp;
  }

  operator bool() const { return m_ptr != nullptr; }

  T& operator*() { return *m_ptr; }

  const T& operator*() const { return *m_ptr; }

  T* operator->() { return m_ptr; }

  const T* operator->() const { return m_ptr; }

  bool operator==(const T* const ptr) const { return m_ptr == ptr; }

  bool operator==(const decltype(nullptr)) const { return m_ptr == nullptr; }

private:
  T* m_ptr = nullptr;
  DeleterT m_deleter;
};

template<typename T, typename DeleterT>
inline bool
operator==(decltype(nullptr), const UniquePtr<T, DeleterT>& ptr)
{
  return ptr == nullptr;
}

template<typename T, typename DeleterT>
class UniquePtr<T[], DeleterT>
{
public:
  using DeleterType = DeleterT;

  UniquePtr() = default;
  UniquePtr(const UniquePtr&) = delete;

  UniquePtr(UniquePtr&& other)
  {
    reset(other.m_ptr);
    other.m_ptr = nullptr;
  }

  UniquePtr(T* const ptr)
    : m_ptr(ptr)
  {
  }

  ~UniquePtr() { reset(); }

  UniquePtr& operator=(const UniquePtr& other) = delete;

  UniquePtr& operator=(UniquePtr&& other) noexcept
  {
    reset(other.m_ptr);
    other.m_ptr = nullptr;

    return *this;
  }

  UniquePtr& operator=(T* const newPtr)
  {
    reset(newPtr);

    return *this;
  }

  void reset(T* const newPtr = nullptr)
  {
    if (m_ptr) {
      m_deleter(m_ptr);
    }

    m_ptr = newPtr;
  }

  T* get() { return m_ptr; }

  const T* get() const { return m_ptr; }

  T* operator()() { return m_ptr; }

  const T* operator()() const { return m_ptr; }

  T* release()
  {
    T* temp = m_ptr;
    m_ptr = nullptr;
    return temp;
  }

  operator bool() const { return m_ptr != nullptr; }

  T& operator*() { return *m_ptr; }

  const T& operator*() const { return *m_ptr; }

  T* operator->() { return m_ptr; }

  const T* operator->() const { return m_ptr; }

  bool operator==(const T* const ptr) const { return m_ptr == ptr; }

  bool operator==(const decltype(nullptr)) const { return m_ptr == nullptr; }

private:
  T* m_ptr = nullptr;
  DeleterT m_deleter;
};

template<class T, class DeleterT = DefaultDeleter<T>>
class UniquePtr;

template<typename T>
inline bool
operator==(decltype(nullptr), const UniquePtr<T[]>& ptr)
{
  return ptr == nullptr;
}

template<typename T, EnableIfT<!IsArrayV<T>, int> = 0>
inline UniquePtr<T>
makeUnique()
{
  return UniquePtr<T>(memory::MemoryManager<T>::allocate());
  // return UniquePtr<T>(new T());
}

template<typename T, typename... TOthers, EnableIfT<!IsArrayV<T>, int> = 0>
inline UniquePtr<T>
makeUnique(TOthers&&... Args)
{
  return UniquePtr<T>(
    memory::MemoryManager<T>::allocate(forward<TOthers>(Args)...));
  // return UniquePtr<T>(new T(forward<TOthers>(Args)...));
}
// #include <memory>
template<typename T, EnableIfT<IsArrayV<T> && ExtentV<T> == 0, int> = 0>
inline UniquePtr<T>
makeUnique(const SizeT count)
{
  // using BareT = RemoveExtentT<T>;
  return UniquePtr<T>(memory::MemoryManager<T>::allocate(count));
  // return UniquePtr<T>(new BareT[count]());
}
} // namespace stl

#endif
