#ifndef __STL_NETWORK_HPP__
#define __STL_NETWORK_HPP__

#include "stl/stl_config.hpp"

#if defined STL_ENABLE_NETWORKING && (STL_ENABLE_NETWORKING == 1)

#include "stl/compiler.hpp"
#include "stl/functional.hpp"
#include "stl/memory.hpp"
#include "stl/stddef.hpp"
#include "stl/stdint.hpp"
#include "stl/string.hpp"
#include "stl/string_conv.hpp"
#include "stl/time.hpp"
#include "stl/timer.hpp"



#define OPENSSL_NO_DEPRECATED

#include "openssl/crypto.h"
#include "openssl/err.h"
#include "openssl/ssl.h"

#include "stl/iostream.hpp"



#define NETWORK_AF_LUT_SIZE 2
#define NETWORK_COMTYPE_LUT_SIZE 3
#define NETWORK_IPPROTO_LUT_SIZE 3
#define NETWORK_SHDIR_LUT_SIZE 3

#define NETWORK_IO_STATUS_LIST                                                 \
  ENTRY(Error)                                                                 \
  ENTRY(SSLError)                                                              \
  ENTRY(SSLInitError)                                                          \
  ENTRY(SSLNotReady)                                                           \
  ENTRY(SSLMethodError)                                                        \
  ENTRY(SSLContextError)                                                       \
  ENTRY(SSLCipherError)                                                        \
  ENTRY(SSLCertError)                                                          \
  ENTRY(SSLFDError)                                                            \
  ENTRY(AllocError)                                                            \
  ENTRY(NotInitialized)                                                        \
  ENTRY(SocketInErrorState)                                                    \
  ENTRY(InvalidParam)                                                          \
  ENTRY(InvalidSocket)                                                         \
  ENTRY(UnsupportedParamCombination)                                           \
  ENTRY(UnsupportedParam)                                                      \
  ENTRY(ConnectionClosed)                                                      \
  ENTRY(OpTimeout)                                                             \
  ENTRY(ConnectionForciblyClosed)                                              \
  ENTRY(OpWouldBlock)                                                          \
  ENTRY(NoNetwork)                                                             \
  ENTRY(OK)                                                                    \
  ENTRY(SocketAlreadyExist)                                                    \
  ENTRY(UnsupportedFunction)                                                   \
  ENTRY(InconsistentAddress)                                                   \
  ENTRY(NoSocket)

namespace network {
enum class IPProtoE
{
  TCP,
  UDP,
  RAW,
  _Count
};

enum class ComTypeE
{
  Stream,
  Datagram,
  Raw,
  _Count
};

enum class IPVerE
{
  IPv4,
  IPv6,
  _Count
};

enum class SockDirE
{
  Download,
  Upload,
  Both,
  _Count
};

enum class IOStatusE
{
#ifdef ENTRY
#undef ENTRY
#endif

#define ENTRY(errorName) IOStatusE_##errorName,
  NETWORK_IO_STATUS_LIST
#undef ENTRY
};

using PortT = stl::UInt16T;
using SysErrorT = stl::Int32T;
#if STL_ENABLE_INT64
using SharedSocketT = stl::Int64T;
#else
using SharedSocketT = stl::Int32T;
#endif

IOStatusE
init();
IOStatusE
deinit();

class IPv4
{
public:
  IPv4();

  IPv4(const IPv4& other);

  IPv4& operator=(const IPv4& other);

  IPv4(IPv4&& other) noexcept;

  IPv4& operator=(IPv4&& other) noexcept;

  IPv4(stl::UInt32T ip);

  stl::UInt32T netAddr() const;

  const char* ipStr() const;

  stl::SizeT ipStrLen() const;

  bool parse(const char* ipStr);
  bool parse(const stl::String& ipStr);

  void set(const stl::UInt32T netAddr);

private:
  bool parseInternal(const char* ipStr);

  void reset();

  stl::UInt32T m_netAddr;
  char m_ipStr[16];
  stl::SizeT m_ipStrLen;
};

class IPv6
{
public:
  IPv6();

  IPv6(const IPv6& other);

  IPv6& operator=(const IPv6& other);

  IPv6(IPv6&& other) noexcept;

  IPv6& operator=(IPv6&& other) noexcept;

  IPv6(const stl::UInt8T* const ip);

  const stl::UInt8T* netAddr() const;

  const char* ipStr() const;

  stl::SizeT ipStrLen() const;

  bool parse(const char* ipStr);

  void set(const stl::UInt8T* const netAddr);

private:
  void reset();

  stl::UInt32T m_netAddr;
  char m_ipStr[40];
  stl::SizeT m_ipStrLen;
};

class NetworkAddress
{
public:
  NetworkAddress();

  NetworkAddress(const char* const ipStr, const PortT port);

  NetworkAddress(NetworkAddress&& other) noexcept;

  NetworkAddress& operator=(NetworkAddress&& other) noexcept;

  NetworkAddress(const NetworkAddress& other);

  NetworkAddress& operator=(const NetworkAddress& other);

  const stl::UInt8T* sockAddr() const;

  stl::UInt8T* sockAddr();

  bool set(const char* ipStr, const stl::UInt16T port);

  bool set(const IPv4& ipV4, const stl::UInt16T port);

  bool set(const IPv6& ipV6, const stl::UInt16T port);

  IPVerE getIPVer() const;

  void getAsIPv4(IPv4& ipV4) const;

  IPv4 getAsIPv4() const;

  void getAsIPv6(IPv6& ipV6) const;

  IPv6 getAsIPv6() const;

  stl::UInt32T getIPv4NetworkAddr() const;

  const stl::UInt8T* getIPv6NetworkAddr() const;

  stl::String getIPStr() const;

  PortT getPort() const;

  stl::SizeT size() const;

  static stl::SizeT sizeMax();

  bool valid() const
  {
    return m_valid;
  }

private:
  stl::UInt8T m_sockAddr[28];
  bool m_valid;
};

class SocketBase
{
public:
  struct SocketDataContainer
  {
    void* m_data;
  };

  using ActiveDirT = stl::UInt8T;
  using SocketContainerT = stl::UniquePtr<SocketDataContainer>;

  static const char* ioStatusString(const IOStatusE status);

  SocketBase();

  virtual ~SocketBase() {
    if(isOpen())
    {
      shutdownInternal(SockDirE::Both);
      closeInternal();
    }
  }

  SocketBase(SocketContainerT&& socketContainer);

  SocketBase(const SocketBase&) = delete;

  SocketBase(SocketBase&& other) noexcept;

  SocketBase& operator=(const SocketBase&) = delete;

  SocketBase& operator=(SocketBase&& other) noexcept;

  SharedSocketT fd() const;

  IOStatusE shutdown(const SockDirE dir);

  IOStatusE close();

  IOStatusE bind(const NetworkAddress& networkAddress);

  IOStatusE blocking(const bool blocking);

  virtual IOStatusE readReady();

  SysErrorT getLastSysError() const;

  bool isOpen() const;

protected:
  IOStatusE select();

  IOStatusE poll();

  IOStatusE shutdownInternal(const SockDirE dir);
  IOStatusE closeInternal();
  IOStatusE bindInternal();

  void setSysError(const char* filename, const stl::UInt32T line);

  IOStatusE translateSysError() const;

  SocketContainerT m_socketContainer;
  IPVerE m_ipVer;
  ComTypeE m_comType;
  IPProtoE m_ipProto;
  ActiveDirT m_activeDirections;
  SysErrorT m_lastSysError;
  NetworkAddress m_boundAddress; // TODO: Not moving

  bool m_blocking;

  static const ActiveDirT m_dirBitLUT[3];
};

class SocketOpenerBase : virtual public SocketBase
{
public:
  SocketOpenerBase();

  virtual ~SocketOpenerBase() {}

  SocketOpenerBase(const SocketOpenerBase&) = delete;
  SocketOpenerBase& operator=(const SocketOpenerBase&) = delete;

  SocketOpenerBase(SocketOpenerBase&& other) noexcept;

  SocketOpenerBase& operator=(SocketOpenerBase&& other) noexcept;

  SocketOpenerBase(SocketContainerT&& socketContainer);

  SocketOpenerBase& operator=(SocketContainerT&& socketContainer);

  IOStatusE open(const IPVerE ipVer,
                 const ComTypeE comType,
                 const IPProtoE proto);

private:
  IOStatusE openInternal();

protected:
  virtual IOStatusE postOpenAction() { return IOStatusE::IOStatusE_OK; };
};

class ClientSocketBase : virtual public SocketBase
{
public:
  ClientSocketBase();

  ClientSocketBase(SocketContainerT&& socketContainer);

  virtual ~ClientSocketBase();

  ClientSocketBase(const ClientSocketBase&) = delete;

  ClientSocketBase(ClientSocketBase&& other) noexcept;

  ClientSocketBase& operator=(const ClientSocketBase&) = delete;

  ClientSocketBase& operator=(ClientSocketBase&& other) noexcept;

  virtual IOStatusE postConnectAction();

  stl::SizeT available();

  virtual IOStatusE send(const stl::UInt8T* const data,
                         const stl::SizeT size,
                         stl::SizeT& sent);

  virtual IOStatusE receive(stl::UInt8T* const data,
                            const stl::SizeT maxSize,
                            stl::SizeT& received);

  virtual IOStatusE sendTo(const NetworkAddress& networkAddress,
                           const stl::UInt8T* const data,
                           const stl::SizeT size,
                           stl::SizeT& sent);

  virtual IOStatusE receiveFrom(stl::UInt8T* const data,
                                const stl::SizeT maxSize,
                                NetworkAddress& networkAddress,
                                stl::SizeT& received);
};

class ClientSocket
  : virtual public ClientSocketBase
  , public SocketOpenerBase
{
public:
  ClientSocket();

  virtual ~ClientSocket() {}

  ClientSocket(const ClientSocket&) = delete;

  ClientSocket(ClientSocket&& other) noexcept;

  ClientSocket& operator=(const ClientSocket&) = delete;

  ClientSocket& operator=(ClientSocket&& other) noexcept;

  IOStatusE connect(const NetworkAddress& networkAddress);

  IOStatusE connect(const IPv4& ip, const PortT port);

  IOStatusE connect(const IPv6& ip, const PortT port);

  IOStatusE connect(const char* host, const PortT port);

  void setInfo(const stl::String2& info)
  {
    LOGEXT << logging::LvlE::Debug << "Serial number " << info << " set for communication" << stl::endl;
    m_info = info;
  }

protected:
  stl::String2 m_info;
};

class Session
{
public:
  Session() = default;

  Session(stl::UniquePtr<ClientSocketBase>&& clientSocket,
          NetworkAddress&& clientAddress);

  Session(const Session&) = delete;
  Session& operator=(const Session&) = delete;

  Session(Session&& other) noexcept;

  Session& operator=(Session&& other) noexcept;

  stl::timer::Countdown& timeout();

  ClientSocketBase& socket();

  const NetworkAddress& remoteAddress() const;

private:
  stl::UniquePtr<ClientSocketBase> m_clientSocket = nullptr;
  NetworkAddress m_clientAddress;
  stl::timer::Countdown m_timeout;
};

class ServerSocket : public SocketOpenerBase
{
public:
  ServerSocket();

  virtual ~ServerSocket();

  ServerSocket(const ServerSocket&) = delete;
  ServerSocket& operator=(const ServerSocket&) = delete;

  ServerSocket(ServerSocket&& other) noexcept;

  ServerSocket& operator=(ServerSocket&& other) noexcept;

  ServerSocket(SocketContainerT&& socketContainer);

  ServerSocket& operator=(SocketContainerT&& socketContainer);

  IOStatusE listen(const stl::SizeT backlog);

  IOStatusE accept(Session& newSocketContainer);

  virtual IOStatusE postAcceptAction(Session& newSocketContainer);
};

#if (defined STL_ENABLE_SSL) && (STL_ENABLE_SSL != 0) &&                       \
  ((defined STL_ENABLE_SSL_SERVER && (STL_ENABLE_SSL_SERVER == 1)) ||          \
   (defined STL_ENABLE_SSL_CLIENT && STL_ENABLE_SSL_CLIENT == 1))
struct SSLDataContainer
{
  void* m_data;
};

using SSLStructContainerT = stl::UniquePtr<SSLDataContainer>;

IOStatusE
initSecurityLayer();

class SSLSocketBase
{
public:
  SSLSocketBase(); // TODO: Create template to enable / disable SSL .. with
                   // socket base with common methods

  SSLSocketBase(SSLStructContainerT&& socketContainer)
    : m_sslCtxContainer(stl::move(socketContainer))
  {}

  virtual ~SSLSocketBase() {}

  SSLSocketBase(const SSLSocketBase&) = delete;
  SSLSocketBase& operator=(const SSLSocketBase&) = delete;

  SSLSocketBase(SSLSocketBase&& other) noexcept;
  SSLSocketBase& operator=(SSLSocketBase&& other) noexcept;

  int getLastSSLCtxError() const;

  void setLastSSLCtxError(const int error,
                          const char* filename,
                          const unsigned int line);

// protected:
public: // TODO: Remove
  SSLStructContainerT m_sslCtxContainer;
};
#endif

#if (defined STL_ENABLE_SSL) && (STL_ENABLE_SSL != 0) &&                       \
  ((defined STL_ENABLE_SSL_SERVER && (STL_ENABLE_SSL_SERVER == 1)) ||          \
   (defined STL_ENABLE_SSL_CLIENT && STL_ENABLE_SSL_CLIENT == 1))
class SSLClientSocketBase
  : public SSLSocketBase
  , virtual public ClientSocketBase
{
public:
  SSLClientSocketBase();

  virtual ~SSLClientSocketBase() {}

  SSLClientSocketBase(SocketContainerT&& socketContainer,
                      SSLStructContainerT&& sslStructContainer) noexcept;

  SSLClientSocketBase(ClientSocketBase&& clientSocketBase,
                      SSLStructContainerT&& sslStructContainer) noexcept;

  SSLClientSocketBase(const SSLClientSocketBase&) = delete;
  SSLClientSocketBase& operator=(const SSLClientSocketBase&) = delete;

  SSLClientSocketBase(SSLClientSocketBase&& other) noexcept;
  SSLClientSocketBase& operator=(SSLClientSocketBase&& other) noexcept;

  IOStatusE shutdown(const SockDirE dir);

  virtual IOStatusE postConnectAction() override;

  virtual IOStatusE readReady() override;

  virtual IOStatusE send(const stl::UInt8T* const data,
                 const stl::SizeT size,
                 stl::SizeT& sent) override;

  virtual IOStatusE receive(stl::UInt8T* const data,
                    const stl::SizeT maxSize,
                    stl::SizeT& received) override;

  int getLastSSLError() const;

protected:
  void setLastSSLError(const int error,
                       const char* filename,
                       const unsigned int line);
public: // TODO: Remove
  SSLStructContainerT m_sslContainer;
  NetworkAddress m_networkAddr;

  friend class SSLServerSocket; // TODO: Remove
  friend class DeviceSession; // TODO: Remove
};
#endif

#if (defined STL_ENABLE_SSL) && (STL_ENABLE_SSL != 0) &&                       \
  (defined STL_ENABLE_SSL_CLIENT && (STL_ENABLE_SSL_CLIENT == 1))
class SSLClientSocket
  : public SSLClientSocketBase
  , virtual public ClientSocket
{
public:
  SSLClientSocket();

  virtual ~SSLClientSocket();

  SSLClientSocket(const SSLClientSocket&) = delete;
  SSLClientSocket& operator=(const SSLClientSocket&) = delete;

  SSLClientSocket(SSLClientSocket&& other) noexcept;
  SSLClientSocket& operator=(SSLClientSocket&& other) noexcept;

  IOStatusE postConnectAction();

  IOStatusE connect(const NetworkAddress& networkAddress);

  IOStatusE connect(const IPv4& ip, const PortT port);

  IOStatusE connect(const IPv6& ip, const PortT port);

  IOStatusE connect(const char* host, const PortT port);

private:
  IOStatusE connectInternal();
  virtual IOStatusE postOpenAction() override;
};
#endif

#if (defined STL_ENABLE_SSL) && (STL_ENABLE_SSL != 0) &&                       \
  (defined STL_ENABLE_SSL_SERVER && (STL_ENABLE_SSL_SERVER == 1))
class SSLServerSocket
  : public SSLSocketBase
  , public ServerSocket
{
public:
  SSLServerSocket(); // TODO: Create template to enable / disable SSL .. with
                     // socket base with common methods

  virtual ~SSLServerSocket() {}

  SSLServerSocket(SocketContainerT&& socketContainer);

  SSLServerSocket(const SSLServerSocket&) = delete;
  SSLServerSocket& operator=(const SSLServerSocket&) = delete;

  SSLServerSocket(SSLServerSocket&& other) noexcept;
  SSLServerSocket& operator=(SSLServerSocket&& other) noexcept;

  virtual IOStatusE postAcceptAction(Session& newSocketContainer) override;

private:
  virtual IOStatusE postOpenAction() override;
};
#endif

class TCPServer
{
public:
  TCPServer();

  virtual ~TCPServer() = default;

  IOStatusE begin(const NetworkAddress& bindAddress);

  IOStatusE update();

  IOStatusE end();

protected:
  void setEndpoint(ServerSocket& endPoint);

  virtual void handleConnection(Session&& clientSocket);

private:
  ServerSocket* m_socketPtr;
};
} // namespace network

#endif // defined STL_ENABLE_NETWORKING && (STL_ENABLE_NETWORKING == 1)

#endif
