#ifndef __STL_OSTREAM_HPP__
#define __STL_OSTREAM_HPP__

#include "stl/mutex.hpp"
#include "stl/ostream_types.hpp"
#include "stl/stddef.hpp"
#include "stl/stl_config.hpp"
#include "stl/stream_base.hpp"
#include "stl/string.hpp"

namespace stl {
enum class IntegerModeE
{
  Dec,
  Hex,
  Oct
};

struct Precision
{
  Precision(SizeT floatPrecision)
    : m_floatPrecision(floatPrecision)
  {
  }

  SizeT m_floatPrecision;
};

enum class ClrE
{
  Black,
  Gray,
  Blue,
  LightBlue,
  Green,
  LightGreen,
  Aqua,
  LightAqua,
  Red,
  LightRed,
  Purple,
  LightPurple,
  Yellow,
  LightYellow,
  White,
  BrightWhite,
  _Count
};

enum class FmtE
{
  Normal,
  Hex
};

enum class OutputE
{
  File,
  Terminal
};
}

namespace stl::stream {
class OStreamBase : virtual public StreamBase
{
public:
  static constexpr SizeT FLOAT_PRECISION_DEFAULT = 10U;

  OStreamBase()
    : m_integerMode(IntegerModeE::Dec)
    , m_floatPrecision(FLOAT_PRECISION_DEFAULT)
    , m_currentFormat(FmtE::Normal)
    , m_output(OutputE::Terminal)
    , m_fgColor(ClrE::_Count)
  {
  }

  virtual SizeT write(const UInt8T* const data, SizeT size) = 0;

  OStreamBase& write(const char* const str, SizeT length);

  OStreamBase& write(char* const str, SizeT length);

  template<typename T>
  OStreamBase& operator<<(T* const value);

  template<typename T>
  OStreamBase& operator<<(const T& value);

  void dump(const UInt8T* data,
            const SizeT size,
            const SizeT bytesPerLine,
            const bool upperCase);

  void fill(const char c, const SizeT count);

  OutputE output() const { return m_output; }

  ClrE fgColor() const { return m_fgColor; }

  void fgColor(const ClrE color) { m_fgColor = color; }

protected:
  IntegerModeE m_integerMode;
  SizeT m_floatPrecision;
  FmtE m_currentFormat;
  OutputE m_output;
  ClrE m_fgColor;
};

class OStream : public stream::OStreamBase
{
public:
  OStream()
    : m_lock("ostream")
  {
  }

  template<typename T>
  OStream& operator<<(T* const value)
  {
    OStreamBase::operator<<(value);
    return *this;
  }

  template<typename T>
  OStream& operator<<(const T& value)
  {
    OStreamBase::operator<<(value);
    return *this;
  }

  OStream& operator[](const char* message);

  virtual void flush() override;

  virtual SizeT write(const UInt8T* const data, SizeT size) override;

  virtual bool seek(const SizeT offset, const SeekModeE seekMode) override;

  virtual SizeT position() const override { return 0U; }

  virtual SizeT size() const override { return 0U; }

private:
  Mutex m_lock;
};

}

#endif // __STL_STREAM_HPP__