#ifndef __STL_PARAMETER_HPP__
#define __STL_PARAMETER_HPP__

#include "stl/utility.hpp"

namespace stl
{
template <typename T>
class Parameter
{
public:
    Parameter() : m_latestValue(), m_updateValue(), m_updated(false), m_initialRead(false) {}

    explicit Parameter(T value)
        : m_latestValue(move(value)), m_updateValue(m_latestValue),
          m_updated(false), m_initialRead(false)
    {
    }

    explicit Parameter(const T& value, const bool initialUpdated)
        : m_latestValue(value), m_updateValue(m_latestValue),
          m_updated(initialUpdated), m_initialRead(!initialUpdated)
    {
    }

    explicit Parameter(T value, const bool initialUpdated)
        : m_latestValue(move(value)), m_updateValue(m_latestValue),
          m_updated(initialUpdated), m_initialRead(!initialUpdated)
    {
    }

    explicit Parameter(const Parameter& other)
        : m_latestValue(other.m_latestValue), m_updateValue(m_latestValue),
          m_updated(other.m_updated), m_initialRead(other.m_initialRead)
    {
    }

    explicit Parameter(Parameter&& other)
        : m_latestValue(move(other.m_latestValue)),
          m_updateValue(move(other.m_updateValue)),
          m_updated(move(other.m_updated)), m_initialRead(move(other.m_initialRead))
    {
    }

    Parameter& operator=(const Parameter& other)
    {
        m_latestValue = other.m_latestValue;
        m_updateValue = other.m_updateValue;
        m_updated = other.m_updated;
        m_initialRead = other.m_initialRead;

        return *this;
    }

    Parameter& operator=(Parameter&& other)
    {
        m_latestValue = move(other.m_latestValue);
        m_updateValue = move(other.m_updateValue);
        m_updated = move(other.m_updated);
        m_initialRead = move(other.m_initialRead);

        return *this;
    }

    Parameter& operator=(T value)
    {
        if (!m_updated && (m_latestValue != value))
        {
            m_latestValue = move(value);
            m_updateValue = m_latestValue;
            m_updated = true;
        }
        else
        {
            m_latestValue = move(value);
        }

        return *this;
    }

    T& get() { return m_latestValue; }

    const T& get() const { return m_latestValue; }

    void set() {
        m_updated = true;
    }

    T& getUpdate() { return m_updateValue; }

    const T& getUpdate() const { return m_updateValue; }

    bool updatedPeek() const
    {
        return m_updated || !m_initialRead;
    }

    bool updated() const
    {
        if (m_updated)
        {
            m_initialRead = true;
            m_updated = false;
            return true;
        }

        if(!m_initialRead)
        {
            m_initialRead = true;
            return true;
        }

        return false;
    }

    operator bool() const { return updated(); }

private:
    T m_latestValue;
    T m_updateValue;
    mutable bool m_updated;
    mutable bool m_initialRead;
};
} // namespace stl

#endif // __STL_PARAMETER_HPP__
