#ifndef _STL_THREAD_HPP_
#define _STL_THREAD_HPP_

#include "stl/stl_config.hpp"

#if defined(STL_ENABLE_THREADING) && (STL_ENABLE_THREADING == 1)

#include "stl/compile_info.hpp"

#include "stl/atomic.hpp"
#include "stl/stdint.hpp"
#include "stl/string.hpp"
#include "stl/time.hpp"

#include "log/logging.hpp"

#if (defined CI_SUPPORT_STD) && (CI_SUPPORT_STD == 1)
#include <thread>
#endif // (defined CI_SUPPORT_STD) && (CI_SUPPORT_STD == 1)

namespace stl::thread {
void
yield();
void
sleepForNs(const time::STimeT nanos);
void
sleepForUs(const time::STimeT micros);
void
sleepFor(const time::STimeT millis);

void
delayForNs(const time::STimeT nanos);
} // namespace stl::thread

namespace stl {
class Thread
{
public:
  Thread() = default;
  virtual ~Thread() = default;

  template<typename CallableT, typename... ArgsT>
  Thread(CallableT&& callable, ArgsT&&... args)
  {

#if (defined CI_SUPPORT_STD) && (CI_SUPPORT_STD == 1)
    m_thread =
      std::thread(forward<CallableT>(callable), std::forward<ArgsT>(args)...);
#endif
  }

  Thread(const Thread&) = delete;
  Thread& operator=(const Thread&) = delete;

  Thread(Thread&& other)
    : m_thread(::stl::move(other.m_thread))
  {
  }

  Thread& operator=(Thread&& other)
  {
#if (defined CI_SUPPORT_STD) && (CI_SUPPORT_STD == 1)
    m_thread = ::stl::move(other.m_thread);
#endif

    return *this;
  }

  void join() { m_thread.join(); }

  void detach() { m_thread.detach(); }

  bool joinable() const { return m_thread.joinable(); }

private:
#if (defined CI_SUPPORT_STD) && (CI_SUPPORT_STD == 1)
  std::thread m_thread;
#else
#error Unsupported platform
#endif
};

class RunnableBase
{
public:
  struct Params
  {
    Params() = default;
    Params(const Params& other) = delete;
    Params& operator=(const Params& other) = delete;

    Params(Params&& other)
      : m_stop(other.m_stop.load())
    {
    }

    Params& operator=(Params&& other)
    {
      m_stop.store(other.m_stop.load());
      return *this;
    }

    Atomic<bool> m_stop;
  };

  RunnableBase() = default;
  RunnableBase(String name)
    : m_name(move(name))
  {
  }

  RunnableBase(const RunnableBase&) = delete;

  RunnableBase(RunnableBase&& other) { *this = move(other); }

  RunnableBase& operator=(const RunnableBase&) = delete;

  RunnableBase& operator=(RunnableBase&& other)
  {
    m_thread = move(other.m_thread);
    m_active.store(other.m_active.load());
    m_running.store(other.m_running.load());
    m_params = move(other.m_params);
    m_name = move(other.m_name);

    return *this;
  }

  virtual ~RunnableBase() = default;

  void stop()
  {
    if (m_active) {
      m_params.m_stop.store(true);

      m_thread.join();

      m_params.m_stop.store(false);
      m_active.store(false);
    }
  }

  bool active() const { return m_active.load(); }

  bool running() const { return m_running.load(); }

  const String& name() const { return m_name; }

protected:
  Thread m_thread;
  Atomic<bool> m_active;
  Atomic<bool> m_running;
  RunnableBase::Params m_params;
  String m_name;
};

template<typename CallableT, typename... ArgsT>
class Runnable : public RunnableBase
{
public:
  Runnable() = default;

  Runnable(const Runnable&) = delete;
  Runnable& operator=(const Runnable&) = delete;

  Runnable(Runnable&& other)
    : RunnableBase(move(static_cast<RunnableBase&>(other)))
  {
  }

  Runnable& operator=(Runnable&& other)
  {
    RunnableBase::operator=(move(static_cast<RunnableBase&>(other)));
    return *this;
  }

  Runnable(String name)
    : RunnableBase(move(name))
  {
  }

  virtual ~Runnable() = default;

  void start(CallableT&& callable, ArgsT&&... args)
  {
    Assert(callable);
    Assert(!m_active);

    try {
      m_running.store(true);
      m_active.store(true);

      m_thread = Thread(privateWorker,
                        this,
                        forward<CallableT>(callable),
                        forward<ArgsT>(args)...);
      logging::log << logging::LvlE::Info << "Thread " << name() << " created!"
                   << stl::endl;
    } catch (...) {
      logging::log << logging::LvlE::Exception << "Thread " << name()
                   << " cannot be started!" << stl::endl;
      m_running.store(false);
      m_active.store(false);
      Assert(false);
    }
  }

private:
  static void privateWorker(Runnable* const instance,
                            CallableT&& callable,
                            ArgsT&&... args)
  {
    logging::log << logging::LvlE::Info << "Thread " << instance->name()
                 << " started!" << stl::endl;

    Assert(instance);
    Assert(callable);

    try {
      callable(&instance->m_params, forward<ArgsT>(args)...);
    } catch (...) {
      logging::log << logging::LvlE::Exception << "Thread " << instance->name()
                   << " exception occured!" << stl::endl;
    }

    logging::log << logging::LvlE::Warning << "Thread \"" << instance->name()
                 << "\" stopped." << stl::endl;

    instance->m_running.store(false);
  }
};
}

#endif // defined(STL_ENABLE_THREADING) && (STL_ENABLE_THREADING == 1)
#endif // _STL_THREAD_HPP_
