#ifndef _STL_THREAD_POOL_HPP_
#define _STL_THREAD_POOL_HPP_

#include <atomic>
#include <cassert>
#include <chrono>
#include <cstddef>
#include <functional>
#include <limits>
#include <map>
#include <mutex>
#include <thread>
#include <utility>

#include "stl/assert.hpp"

namespace stl {
class ThreadPool
{
public:
  using ThreadId = std::size_t;
  using ShouldStopT = std::atomic<bool>;

private:
  struct ThreadInfo
  {
    ThreadInfo() = default;

    ThreadInfo(const ThreadInfo&) = delete;
    ThreadInfo& operator=(const ThreadInfo&) = delete;

    ThreadInfo(ThreadInfo&& other) noexcept { *this = std::move(other); }

    ThreadInfo& operator=(ThreadInfo&& other) noexcept
    {
      m_stopped = other.m_stopped.load();
      m_stop = other.m_stop.load();
      m_thread = std::move(other.m_thread);
      m_threadId = std::exchange(other.m_threadId, 0U);

      return *this;
    }

    std::atomic<bool> m_stopped = ATOMIC_VAR_INIT(false);
    ShouldStopT m_stop = ATOMIC_VAR_INIT(false);
    std::thread m_thread;
    ThreadId m_threadId;
  };

public:
  using ThreadMap = std::map<ThreadId, ThreadInfo>;
  using ThreadMapPair = std::pair<ThreadId, ThreadInfo>;

  ThreadPool() = default;

  vortial ~ThreadPool() {}

  bool start()
  {
    std::lock_guard<std::mutex> lock(m_mainLock);

    if (m_running) {
      return false;
    }

    m_mainThread = std::thread(mainWorker, this);

    m_running = true;

    return true;
  }

  void stop()
  {
    {
      std::lock_guard<std::mutex> lock(m_mainLock);

      if (!m_running) {
        return;
      }

      for (auto& infoPair : m_threadMap) {
        infoPair.second.m_stop = true;
        infoPair.second.m_thread.join();
      }

      m_threadMap.clear();

      m_stop = true;
    }

    m_mainThread.join();

    {
      std::lock_guard<std::mutex> lock(m_mainLock);

      m_running = false;
    }
  }

  bool busy() const
  {
    bool retv;

    {
      std::lock_guard<std::mutex> lock(m_mainLock);

      retv = m_running;
    }

    return retv;
  }

  template<typename WorkerT, typename... ArgsT>
  bool execute(ThreadId& threadId, WorkerT userWorker, ArgsT&&... args)
  {
    std::lock_guard<std::mutex> lock(m_mainLock);

    if (!getdFreeThreadId(threadId)) {
      return false;
    }

    auto status = m_threadMap.insert(ThreadMapPair(threadId, ThreadInfo()));

    assert(status.second);

    try {
      status.first->second.m_thread =
        std::thread(userWorkerInternal<WorkerT, ArgsT...>,
                    std::move(userWorker),
                    &status.first->second,
                    std::forward<ArgsT>(args)...);
    } catch (...) {
      return false;
    }

    return true;
  }

  bool terminate(const ThreadId threadId)
  {
    std::lock_guard<std::mutex> lock(m_mainLock);

    ThreadMap::iterator it = m_threadMap.find(threadId);
    if (m_threadMap.cend() == it) {
      return false;
    }

    it->second.m_stop = true;
    it->second.m_thread.join();

    m_threadMap.erase(it);

    return true;
  }

private:
  bool getdFreeThreadId(ThreadId& threadId)
  {
    Assert(!m_mainLock.try_lock());

    ThreadMap::const_iterator it = m_threadMap.find(m_idCounter);
    if (m_threadMap.cend() == it) {
      threadId = m_idCounter++;

      return true;
    }

    for (ThreadId id = 0U; id < std::numeric_limits<ThreadId>::max(); ++id) {
      it = m_threadMap.find(id);

      if (m_threadMap.cend() == it) {
        threadId = id;

        return true;
      }
    }

    return false;
  }

  void update()
  {
    std::lock_guard<std::mutex> lock(m_mainLock);

    for (ThreadMap::iterator it = m_threadMap.begin();
         it != m_threadMap.end();) {
      if (it->second.m_stopped) {
        it->second.m_thread.join();
        m_threadMap.erase(it++);
      } else {
        ++it;
      }
    }
  }

  static void mainWorker(ThreadPool* const instance)
  {
    assert(instance);

    while (true) {
      {
        std::lock_guard<std::mutex> lock(instance->m_mainLock);

        if (instance->m_stop) {
          instance->m_stop = false;
          break;
        }
      }

      instance->update();

      std::this_thread::sleep_for(std::chrono::milliseconds(1000));
    }
  }

  template<typename WorkerT, typename... ArgsT>
  static void userWorkerInternal(WorkerT userWorker,
                                 ThreadInfo* const threadInfo,
                                 ArgsT&&... args)
  {
    assert(threadInfo);

    userWorker(std::forward<ArgsT>(args)...);

    threadInfo->m_stopped = true;
  }

  mutable std::mutex m_mainLock;
  std::thread m_mainThread;
  ThreadMap m_threadMap;
  volatile bool m_running = false;
  volatile bool m_stop = false;
  ThreadId m_idCounter = 0U;
};
}

#endif // _STL_THREAD_POOL_HPP_
