#ifndef __STL_TIMER_HPP__
#define __STL_TIMER_HPP__

#include "stl/math.hpp"
#include "stl/stdint.hpp"
#include "stl/time.hpp"
#include "stl/vector.hpp"

#if defined STL_ENABLE_INT64
namespace stl {
namespace timer {
using namespace time;

class Countdown
{
public:
  Countdown() = default;

  Countdown(const TimeValue timeout)
    : m_start(0)
    , m_timeout(timeout)
    , m_running(false)
  {
  }

  Countdown(const Countdown& other)
    : m_start(other.m_start)
    , m_timeout(other.m_timeout)
    , m_running(other.m_running)
  {
  }

  Countdown& operator=(const Countdown& other)
  {
    m_start = other.m_start;
    m_timeout = other.m_timeout;
    m_running = other.m_running;

    return *this;
  }

  Countdown(Countdown&& other) noexcept
    : m_start(other.m_start)
    , m_timeout(other.m_timeout)
    , m_running(other.m_running)
  {
  }

  Countdown& operator=(Countdown&& other) noexcept
  {
    m_start = other.m_start;
    m_timeout = other.m_timeout;
    m_running = other.m_running;

    return *this;
  }

  operator bool() const { return timeout(); }

  void set(const TimeValue& timeout)
  {
    m_timeout = timeout;

    reset();
  }

  void stop() { m_running = false; }

  bool timeout() const
  {
    if (!m_running) {
      return false;
    } else {
      m_lastDiff = (getUpTime() - Millis(m_start)).baseValue();
      m_running = !(m_lastDiff >= Millis(m_timeout).baseValue());

      return !m_running;
    }
  }

  void reset()
  {
    m_start = getUpTime();
    m_running = true;
  }

  bool running() const { return m_running; }

  const TimeValue& timeoutValue() const { return m_timeout; }

  const TimeValue lastDiff() const { return m_lastDiff; }

private:
  mutable STimeT m_lastDiff = 0;
  TimeValue m_start = {};
  TimeValue m_timeout = {};
  mutable bool m_running = false;
};

class TimeInterval
{
public:
  TimeInterval() = default;

  TimeInterval(const TimeValue& startTime,
               const TimeValue& stopTime,
               stl::Vector<stl::UInt32T> dayOfWeek)
    : m_startTime(startTime)
    , m_stopTime(stopTime)
    , m_dayOfWeekList(stl::move(dayOfWeek))
  {
  }

  bool active(const DateTime& dateTime) const
  {
    const TimeValue sinceMidnight = Hours(dateTime.hours()) +
                                    Mins(dateTime.minutes()) +
                                    Sec(dateTime.seconds());

    if (m_stopTime <= m_startTime) {
      // LOGEXT << logging::LvlE::Warning << "Timer stop time " << m_stopTime << " is lower than start time " << m_startTime << stl::endl;
      return false;
    }

    // LOGEXT << logging::LvlE::Debug << "Searching day of week [" << m_dayOfWeekList.size() << ']' << stl::endl;

    for(stl::SizeT i = 0U; i < m_dayOfWeekList.size(); ++i)
    {
      // LOGEXT << logging::LvlE::Debug << "Comparing day of week [" << i << "] " << dateTime.weekDay() << " vs " << m_dayOfWeekList[i] << stl::endl;
      if (dateTime.weekDay() == m_dayOfWeekList[i]) {
        // LOGEXT << logging::LvlE::Debug << "Day of week compare: " << m_startTime << " vs " << sinceMidnight << " and " << m_stopTime << " vs " << sinceMidnight << stl::endl;
        if ((m_startTime <= sinceMidnight) && (m_stopTime > sinceMidnight)) {
          return true;
        }
      }
    }

    return false;
  }

  const TimeValue& startTime() const { return m_startTime; }

  const TimeValue& stopTime() const { return m_stopTime; }

  const stl::Vector<stl::UInt32T>& dayOfWeek() const { return m_dayOfWeekList; }

private:
  TimeValue m_startTime = {};
  TimeValue m_stopTime = {};
  stl::Vector<stl::UInt32T> m_dayOfWeekList = {};
};
} // namespace time
} // namespace stl
#else
#error Time library not supported
#endif

#endif // __STL_TIMER_HPP__
