#ifndef _VECTOR_HPP_
#define _VECTOR_HPP_

#include "stl/assert.hpp"
#include "stl/memory.hpp"
#include "stl/stddef.hpp"
#include "stl/stl_config.hpp"

#include <vector>

namespace stl {
template<typename T>
using Vector = std::vector<T>;

#ifdef HRIC_IMPL // TODO Implement
template<typename T>
class VectorDefaultDeleter
{
public:
  VectorDefaultDeleter() noexcept = default;

  VectorDefaultDeleter(const VectorDefaultDeleter&) {}

  void operator()(T* const ptr) const noexcept
  {
    Assert(nullptr != ptr);
    delete[] ptr; // Use stl::memory::MemoryManager<>::deallocate
  }
};

template<typename T>
class Vector
{
public:
  using BaseT = T;

  Vector() = default;

  Vector(Vector&& other) { *this = move(other); }

  Vector& operator=(Vector&& other)
  {
    m_ptr = exchange(other.m_ptr, nullptr);
    m_size = exchange(other.m_size, static_cast<SizeT>(0U));
    m_reserved = exchange(other.m_reserved, static_cast<SizeT>(0U));

    return *this;
  }

  Vector(const Vector& other) { *this = other; }

  Vector& operator=(const Vector& other)
  {
    resize(other.size());
    memory::copy(other.data(), other.data() + other.size(), m_ptr.get());

    return *this;
  }

  virtual ~Vector() { clear(); }

  BaseT& front()
  {
    Assert(nullptr != m_ptr);
    Assert(this->m_size > 0U);
    return m_ptr[0U];
  }

  BaseT& back()
  {
    Assert(nullptr != m_ptr);
    Assert(this->m_size > 0U);
    return m_ptr[this->m_size - 1U];
  }

  const BaseT& front() const
  {
    Assert(nullptr != m_ptr);
    Assert(this->m_size > 0U);
    return m_ptr[0U];
  }

  const BaseT& back() const
  {
    Assert(nullptr != m_ptr);
    Assert(this->m_size > 0U);
    return m_ptr[this->m_size - 1U];
  }

  BaseT& operator[](const SizeT index)
  {
    Assert(nullptr != m_ptr);
    Assert(this->m_size > 0U);
    return m_ptr[index];
  }

  const BaseT& operator[](const SizeT index) const
  {
    Assert(nullptr != m_ptr);
    Assert(this->m_size > 0U);
    return m_ptr[index];
  }

  BaseT* data() { return m_ptr.get(); }

  const BaseT* data() const { return m_ptr.get(); }

  SizeT capacity() const { return m_reserved; }

  SizeT size() const { return m_size; }

  bool empty() const { return m_size == 0U; }

  template<typename... ArgsT>
  void emplaceBack(ArgsT&&... Args)
  {
    this->pushBack(BaseT(forward<ArgsT>(Args)...));
  }

  void pushBack(BaseT&& item)
  {
    reserve(this->m_size + 1U);

    m_ptr[this->m_size++] = move(item);
  }

  void popBack()
  {
    Assert(nullptr != m_ptr);
    Assert(this->m_size > 0U);

    m_ptr[--this->m_size].~BaseT();
  }

  // TODO: Insert
  // TODO: Emplace
  // TODO: Swap

  // TODO:
  /* void resize(const SizeT count) {
      if(count > m_size) {

      } else if(count < m_size) {

      }
  }

  // TODO:
  void erase(ListItem& item) {

  } */

  void clear()
  {
    for (SizeT i = 0U; i < m_size; ++i) {
      m_ptr.get()[i].~BaseT();
    }

    this->m_size = 0U;
  }

  void reserve(const SizeT count)
  {
    if (count > m_reserved) {
      T* const newPtr = new T[count]; // TODO: Use allocator
                                      // stl::memory::MemoryManager<>::allocate

      m_reserved = count;

      for (SizeT i = 0U; i < this->m_size; ++i) {
        newPtr[i] = move(m_ptr.get()[i]);
      }

      for (SizeT i = this->m_size; i < m_reserved; ++i) {
        new (&newPtr[i]) T();
      }

      m_ptr.reset(newPtr);
    }
  }

  void resize(const SizeT count)
  {
    if (count > m_reserved) {
      reserve(count);
    }

    this->m_size = count;

    // TODO: Resize array if count is much lower than current size
  }

  // TODO void shrinkToFit() {}

private:
  UniquePtr<T[], VectorDefaultDeleter<T>> m_ptr = nullptr;
  SizeT m_size = 0U;
  SizeT m_reserved = 0U;
};
#endif // HRIC_IMPL // TODO Implement
}

#endif
