#include "crypto/sha1.h"
#include "crypto/bitwise.h"

#include "stl/assert.hpp"
#include "stl/memory.hpp"
#include "stl/endian.hpp"

static inline void appl64(stl::UInt8T* const output, const size64_t length) {
#if CRYPTO_ENDIAN == CRYPTO_LITTLE_ENDIAN
    output[56] = (stl::UInt8T)(length >> 56);
    output[57] = (stl::UInt8T)(length >> 48);
    output[58] = (stl::UInt8T)(length >> 40);
    output[59] = (stl::UInt8T)(length >> 32);
    output[60] = (stl::UInt8T)(length >> 24);
    output[61] = (stl::UInt8T)(length >> 16);
    output[62] = (stl::UInt8T)(length >> 8);
    output[63] = (stl::UInt8T)(length);
#else
    * ((stl::UInt64T*)&output[56]) = length;
#endif
}

static const stl::UInt32T c_sha1_word[] = {
    0x5A827999U
    , 0x6ED9EBA1U
    , 0x8F1BBCDCU
    , 0xCA62C1D6U
};

static const stl::UInt32T c_sha1_init[] = {
    0x67452301U
    , 0xEFCDAB89U
    , 0x98BADCFEU
    , 0x10325476U
    , 0xC3D2E1F0U
};

// SHA1 (160-bit output = 20 Bytes)

void sha1_init(struct sha1_ctx* const ctx) {
    Assert(NULL != ctx);

    ctx->_hash[0] = pgm_read_u32(c_sha1_init);
    ctx->_hash[1] = pgm_read_u32(c_sha1_init + 1);
    ctx->_hash[2] = pgm_read_u32(c_sha1_init + 2);
    ctx->_hash[3] = pgm_read_u32(c_sha1_init + 3);
    ctx->_hash[4] = pgm_read_u32(c_sha1_init + 4);

    ctx->_data_len = 0U;
    ctx->_bit_len = 0U;
}

static void sha1_process_block(struct sha1_ctx* const ctx) {
    stl::UInt32T a, b, c, d, e, temp, w[80];
    size64_t i, di;

    // Prepare message schedule
    for (i = 0U, di = 0U; i < 16U; ++i, di += 4U) {
        w[i] = ((stl::UInt32T)ctx->_data[di] << 24)
            + ((stl::UInt32T)ctx->_data[di + 1] << 16)
            + ((stl::UInt32T)ctx->_data[di + 2] << 8)
            + ((stl::UInt32T)ctx->_data[di + 3]);
    }

    for (; i < 80U; ++i) {
        w[i] = rotl32(w[i - 3] ^ w[i - 8] ^ w[i - 14] ^ w[i - 16], 1U);
    }

    // Initialize five working variables with i-1 (previous block hash) hash value
    a = ctx->_hash[0];
    b = ctx->_hash[1];
    c = ctx->_hash[2];
    d = ctx->_hash[3];
    e = ctx->_hash[4];

    // Perform transformations
    for (i = 0U; i < 20U; ++i) {
        temp = rotl32(a, 5) + ch32(b, c, d) + e + pgm_read_u32(c_sha1_word) + w[i];
        e = d;
        d = c;
        c = rotl32(b, 30);
        b = a;
        a = temp;
    }

    for (; i < 40U; ++i) {
        temp = rotl32(a, 5) + par32(b, c, d) + e + pgm_read_u32(c_sha1_word + 1) + w[i];
        e = d;
        d = c;
        c = rotl32(b, 30);
        b = a;
        a = temp;
    }

    for (; i < 60U; ++i) {
        temp = rotl32(a, 5) + maj32(b, c, d) + e + pgm_read_u32(c_sha1_word + 2) + w[i];
        e = d;
        d = c;
        c = rotl32(b, 30);
        b = a;
        a = temp;
    }

    for (; i < 80U; ++i) {
        temp = rotl32(a, 5) + par32(b, c, d) + e + pgm_read_u32(c_sha1_word + 3) + w[i];
        e = d;
        d = c;
        c = rotl32(b, 30);
        b = a;
        a = temp;
    }

    // Compute intermediate hash
    ctx->_hash[0] += a;
    ctx->_hash[1] += b;
    ctx->_hash[2] += c;
    ctx->_hash[3] += d;
    ctx->_hash[4] += e;
}

void sha1_update(struct sha1_ctx* const ctx, const stl::UInt8T* input
    , size64_t length) {
    while (length--) {
        ctx->_data[ctx->_data_len++] = *(input++);

        Assert(SIZEOF(struct sha1_ctx, _data) >= ctx->_data_len);

        if (SIZEOF(struct sha1_ctx, _data) == ctx->_data_len) {
            sha1_process_block(ctx);
            ctx->_bit_len += SIZEOF(struct sha1_ctx, _data) * 8U;
            ctx->_data_len = 0U;
        }
    }
}

const stl::UInt8T* sha1_finish(struct sha1_ctx* const ctx) {
#if CRYPTO_ENDIAN == CRYPTO_LITTLE_ENDIAN
    size64_t i;
#endif

    ctx->_data[ctx->_data_len] = 0x80U; // _data_len is always <= 64

    if (ctx->_data_len < 56U) {
        ctx->_bit_len += ctx->_data_len * 8ULL;

        while (++ctx->_data_len < 56U) {
            ctx->_data[ctx->_data_len] = 0x00U;
        }
    } else {
        ctx->_bit_len += ctx->_data_len * 8ULL;

        while (++ctx->_data_len < SIZEOF(struct sha1_ctx, _data)) {
            ctx->_data[ctx->_data_len] = 0x00U;
        }

        sha1_process_block(ctx);
        stl::memory::fill(static_cast<stl::UInt8T>(0x00U), 56U, ctx->_data);
    }

    appl64(ctx->_data, ctx->_bit_len);
    sha1_process_block(ctx);

#if CRYPTO_ENDIAN == CRYPTO_LITTLE_ENDIAN
    for (i = 0U; i < 5U; ++i) {
        ctx->_hash[i] = stl::Endian::swap(ctx->_hash[i]);
    }
#endif
    return (const stl::UInt8T*)ctx->_hash;
}

#include "crypto/hmac_private.h"

void sha1_hmac_init(struct sha1_hmac_ctx* const ctx, const stl::UInt8T* key
    , const size64_t key_length) {
    // If key is longer than block_size bytes reset it to block_size
    if (key_length > SIZEOF(struct sha1_ctx, _data)) {
        sha1_init(&ctx->_sha_ctx);
        sha1_update(&ctx->_sha_ctx, (const stl::UInt8T*)key, key_length);
        const stl::UInt8T* hash = sha1_finish(&ctx->_sha_ctx);
        stl::memory::copy(hash, hash + SIZEOF(struct sha1_ctx, _hash), ctx->_key);
        ctx->_key_length = SIZEOF(struct sha1_ctx, _data);
    } else {
        stl::memory::copy(key, key + (stl::SizeT)key_length, ctx->_key);
        ctx->_key_length = key_length;
    }
}

const stl::UInt8T* sha1_hmac_finish(struct sha1_hmac_ctx* const ctx
    , const char* message, const size64_t message_length) {
    stl::UInt8T k_ipad[SIZEOF(struct sha1_ctx, _data)]; // inner padding - key XORd with ipad
    stl::UInt8T k_opad[SIZEOF(struct sha1_ctx, _data)]; // outer padding - key XORd with opad
    size64_t i;

    // XOR key with ipad and opad values
    for (i = 0U; i < ctx->_key_length; ++i) {
        k_ipad[i] = I_PAD ^ ctx->_key[i];
        k_opad[i] = O_PAD ^ ctx->_key[i];
    }

    for (; i < SIZEOF(struct sha1_ctx, _data); ++i) {
        k_ipad[i] = 0U ^ I_PAD;
        k_opad[i] = 0U ^ O_PAD;
    }

    // Perform inner hash
    sha1_init(&ctx->_sha_ctx);
    sha1_update(&ctx->_sha_ctx, k_ipad, sizeof(k_ipad));
    sha1_update(&ctx->_sha_ctx, (const stl::UInt8T*)message, message_length);
    const stl::UInt8T* hash = sha1_finish(&ctx->_sha_ctx);
    stl::memory::copy(hash, hash + SIZEOF(struct sha1_hmac_ctx, _digest), ctx->_digest);

    // Perform outer hash
    sha1_init(&ctx->_sha_ctx);
    sha1_update(&ctx->_sha_ctx, k_opad, sizeof(k_opad));
    sha1_update(&ctx->_sha_ctx, (const stl::UInt8T*)ctx->_digest, SIZEOF(struct sha1_hmac_ctx, _digest));
    hash = sha1_finish(&ctx->_sha_ctx);
    stl::memory::copy(hash, hash + SIZEOF(struct sha1_hmac_ctx, _digest), ctx->_digest);

    return ctx->_digest;
}
