#include "crypto/sha256.h"
#include "crypto/bitwise.h"

#include "stl/assert.hpp"
#include "stl/memory.hpp"
#include "stl/endian.hpp"

static inline stl::UInt32T ep0_256(const stl::UInt32T x) {
    return rotr32(x, 2) ^ rotr32(x, 13) ^ rotr32(x, 22);
}

static inline stl::UInt32T ep1_256(const stl::UInt32T x) {
    return rotr32(x, 6) ^ rotr32(x, 11) ^ rotr32(x, 25);
}

static inline stl::UInt32T sig0_256(const stl::UInt32T x) {
    return rotr32(x, 7) ^ rotr32(x, 18) ^ shr32(x, 3);
}

static inline stl::UInt32T sig1_256(const stl::UInt32T x) {
    return rotr32(x, 17) ^ rotr32(x, 19) ^ shr32(x, 10);
}

static inline void appl64(stl::UInt8T* const output, const stl::SizeT length) {
#if CRYPTO_ENDIAN == CRYPTO_LITTLE_ENDIAN
    output[56] = (stl::UInt8T)(length >> 56);
    output[57] = (stl::UInt8T)(length >> 48);
    output[58] = (stl::UInt8T)(length >> 40);
    output[59] = (stl::UInt8T)(length >> 32);
    output[60] = (stl::UInt8T)(length >> 24);
    output[61] = (stl::UInt8T)(length >> 16);
    output[62] = (stl::UInt8T)(length >> 8);
    output[63] = (stl::UInt8T)(length);
#else
    * ((stl::UInt64T*)&output[56]) = length;
#endif
}

static const stl::UInt32T c_sha256_word[] = {
    0x428A2F98U, 0x71374491U, 0xB5C0FBCFU, 0xE9B5DBA5U
    , 0x3956C25BU, 0x59F111F1U, 0x923F82A4U, 0xAB1C5ED5U
    , 0xD807AA98U, 0x12835B01U, 0x243185BEU, 0x550C7DC3U
    , 0x72BE5D74U, 0x80DEB1FEU, 0x9BDC06A7U, 0xC19BF174U
    , 0xE49B69C1U, 0xEFBE4786U, 0x0FC19DC6U, 0x240CA1CCU
    , 0x2DE92C6FU, 0x4A7484AAU, 0x5CB0A9DCU, 0x76F988DAU
    , 0x983E5152U, 0xA831C66DU, 0xB00327C8U, 0xBF597FC7U
    , 0xC6E00BF3U, 0xD5A79147U, 0x06CA6351U, 0x14292967U
    , 0x27B70A85U, 0x2E1B2138U, 0x4D2C6DFCU, 0x53380D13U
    , 0x650A7354U, 0x766A0ABBU, 0x81C2C92EU, 0x92722C85U
    , 0xA2BFE8A1U, 0xA81A664BU, 0xC24B8B70U, 0xC76C51A3U
    , 0xD192E819U, 0xD6990624U, 0xF40E3585U, 0x106AA070U
    , 0x19A4C116U, 0x1E376C08U, 0x2748774CU, 0x34B0BCB5U
    , 0x391C0CB3U, 0x4ED8AA4AU, 0x5B9CCA4FU, 0x682E6FF3U
    , 0x748F82EEU, 0x78A5636FU, 0x84C87814U, 0x8CC70208U
    , 0x90BEFFFAU, 0xA4506CEBU, 0xBEF9A3F7U, 0xC67178F2U
};

static const stl::UInt32T c_sha256_init[] = {
    0x6A09E667U
    , 0xBB67AE85U
    , 0x3C6EF372U
    , 0xA54FF53AU
    , 0x510E527FU
    , 0x9B05688CU
    , 0x1F83D9ABU
    , 0x5BE0CD19U
};

// SHA-256

void sha256_init(struct sha256_ctx* const ctx) {
    Assert(NULL != ctx);

    ctx->_hash[0] = *(c_sha256_init);
    ctx->_hash[1] = *(c_sha256_init + 1);
    ctx->_hash[2] = *(c_sha256_init + 2);
    ctx->_hash[3] = *(c_sha256_init + 3);
    ctx->_hash[4] = *(c_sha256_init + 4);
    ctx->_hash[5] = *(c_sha256_init + 5);
    ctx->_hash[6] = *(c_sha256_init + 6);
    ctx->_hash[7] = *(c_sha256_init + 7);

    ctx->_data_len = 0U;
    ctx->_bit_len = 0U;
}

static void sha256_process_block(struct sha256_ctx* const ctx) {
    stl::UInt32T a, b, c, d, e, f, g, h, temp1, temp2, w[64];
    stl::SizeT i, di;

    // Prepare message schedule
    for (i = 0U, di = 0U; i < 16U; ++i, di += 4U) {
        w[i] = ((stl::UInt32T)ctx->_data[di] << 24)
            + ((stl::UInt32T)ctx->_data[di + 1] << 16)
            + ((stl::UInt32T)ctx->_data[di + 2] << 8)
            + ((stl::UInt32T)ctx->_data[di + 3]);
    }

    for (; i < 64U; ++i) {
        w[i] = sig1_256(w[i - 2U]) + w[i - 7U]
            + sig0_256(w[i - 15U]) + w[i - 16U];
    }

    // Initialize five working variables with i-1 (previous block hash) hash value
    a = ctx->_hash[0];
    b = ctx->_hash[1];
    c = ctx->_hash[2];
    d = ctx->_hash[3];
    e = ctx->_hash[4];
    f = ctx->_hash[5];
    g = ctx->_hash[6];
    h = ctx->_hash[7];

    // Main iteration
    for (i = 0U; i < 64U; ++i) {
        temp1 = h + ep1_256(e) + ch32(e, f, g) + *(c_sha256_word + i) + w[i];
        temp2 = ep0_256(a) + maj32(a, b, c);
        h = g;
        g = f;
        f = e;
        e = d + temp1;
        d = c;
        c = b;
        b = a;
        a = temp1 + temp2;
    }

    // Compute intermediate hash value
    ctx->_hash[0] += a;
    ctx->_hash[1] += b;
    ctx->_hash[2] += c;
    ctx->_hash[3] += d;
    ctx->_hash[4] += e;
    ctx->_hash[5] += f;
    ctx->_hash[6] += g;
    ctx->_hash[7] += h;
}

void sha256_update(struct sha256_ctx* const ctx, const stl::UInt8T* input
    , stl::SizeT length) {
    while (length--) {
        ctx->_data[ctx->_data_len++] = *(input++);

        Assert(SIZEOF(struct sha256_ctx, _data) >= ctx->_data_len);

        if (SIZEOF(struct sha256_ctx, _data) == ctx->_data_len) {
            sha256_process_block(ctx);
            ctx->_bit_len += SIZEOF(struct sha256_ctx, _data) * 8U;
            ctx->_data_len = 0U;
        }
    }
}

const stl::UInt8T* sha256_finish(struct sha256_ctx* const ctx) {
#if CRYPTO_ENDIAN == CRYPTO_LITTLE_ENDIAN
    stl::SizeT i;
#endif

    ctx->_data[ctx->_data_len] = 0x80U; // _data_len is always <= 64

    if (ctx->_data_len < 56U) {
        ctx->_bit_len += ctx->_data_len * 8ULL;

        while (++ctx->_data_len < 56U) {
            ctx->_data[ctx->_data_len] = 0x00U;
        }
    } else {
        ctx->_bit_len += ctx->_data_len * 8ULL;

        while (++ctx->_data_len < SIZEOF(struct sha256_ctx, _data)) {
            ctx->_data[ctx->_data_len] = 0x00U;
        }

        sha256_process_block(ctx);
        stl::memory::fill(static_cast<stl::UInt8T>(0x00U), 56U, ctx->_data);
    }

    appl64(ctx->_data, ctx->_bit_len);
    sha256_process_block(ctx);

#if CRYPTO_ENDIAN == CRYPTO_LITTLE_ENDIAN
    for (i = 0U; i < 8U; ++i) { // 0xb410ff61, 0xf20015ad
        ctx->_hash[i] = stl::Endian::swap(ctx->_hash[i]);
    }
#endif
    return (const stl::UInt8T*)ctx->_hash;
}

#include "crypto/hmac_private.h"

void sha256_hmac_init(struct sha256_hmac_ctx* const ctx, const stl::UInt8T* key
    , const stl::SizeT key_length) {
    // If key is longer than block_size bytes reset it to block_size
    if (key_length > SIZEOF(struct sha256_ctx, _data)) {
        sha256_init(&ctx->_sha_ctx);
        sha256_update(&ctx->_sha_ctx, (const stl::UInt8T*)key, key_length);
        const stl::UInt8T* hash = sha256_finish(&ctx->_sha_ctx);
        stl::memory::copy(hash, hash + SIZEOF(struct sha256_ctx, _hash), ctx->_key);
        ctx->_key_length = SIZEOF(struct sha256_ctx, _data);
    } else {
        stl::memory::copy(key, key + (stl::SizeT)key_length, ctx->_key);
        ctx->_key_length = key_length;
    }
}

const stl::UInt8T* sha256_hmac_finish(struct sha256_hmac_ctx* const ctx
    , const char* message, const stl::SizeT message_length) {
    stl::UInt8T k_ipad[SIZEOF(struct sha256_ctx, _data)]; // inner padding - key XORd with ipad
    stl::UInt8T k_opad[SIZEOF(struct sha256_ctx, _data)]; // outer padding - key XORd with opad
    stl::SizeT i;

    // XOR key with ipad and opad values
    for (i = 0U; i < ctx->_key_length; ++i) {
        k_ipad[i] = I_PAD ^ ctx->_key[i];
        k_opad[i] = O_PAD ^ ctx->_key[i];
    }

    for (; i < SIZEOF(struct sha256_ctx, _data); ++i) {
        k_ipad[i] = 0U ^ I_PAD;
        k_opad[i] = 0U ^ O_PAD;
    }

    // Perform inner hash
    sha256_init(&ctx->_sha_ctx);
    sha256_update(&ctx->_sha_ctx, k_ipad, sizeof(k_ipad));
    sha256_update(&ctx->_sha_ctx, (const stl::UInt8T*)message, message_length);
    const stl::UInt8T* hash = sha256_finish(&ctx->_sha_ctx);
    stl::memory::copy(hash, hash + SIZEOF(struct sha256_hmac_ctx, _digest), ctx->_digest);

    // Perform outer hash
    sha256_init(&ctx->_sha_ctx);
    sha256_update(&ctx->_sha_ctx, k_opad, sizeof(k_opad));
    sha256_update(&ctx->_sha_ctx, (const stl::UInt8T*)ctx->_digest, SIZEOF(struct sha256_hmac_ctx, _digest));
    hash = sha256_finish(&ctx->_sha_ctx);
    stl::memory::copy(hash, hash + SIZEOF(struct sha256_hmac_ctx, _digest), ctx->_digest);

    return ctx->_digest;
}

// SHA-224

static const stl::UInt32T c_sha224_init[] = {
    0xC1059ED8U
    , 0x367CD507U
    , 0x3070DD17U
    , 0xF70E5939U
    , 0xFFC00B31U
    , 0x68581511U
    , 0x64F98FA7U
    , 0xBEFA4FA4U
};

void sha224_init(struct sha256_ctx* ctx) {
    Assert(NULL != ctx);

    ctx->_hash[0] = c_sha224_init[0];
    ctx->_hash[1] = c_sha224_init[1];
    ctx->_hash[2] = c_sha224_init[2];
    ctx->_hash[3] = c_sha224_init[3];
    ctx->_hash[4] = c_sha224_init[4];
    ctx->_hash[5] = c_sha224_init[5];
    ctx->_hash[6] = c_sha224_init[6];
    ctx->_hash[7] = c_sha224_init[7];

    ctx->_data_len = 0U;
}

void sha224_update(struct sha256_ctx* const ctx, const stl::UInt8T* input
    , stl::SizeT length) {
    sha256_update(ctx, input, length);
}

const stl::UInt8T* sha224_finish(struct sha256_ctx* const ctx) {
    sha256_finish(ctx);

    ctx->_hash[7U] = 0x00000000U;

    return (const stl::UInt8T*)ctx->_hash;
}
