#include "stl/compile_info.hpp"
#include "stl/stl_config.hpp"

#if (defined STL_ENABLE_GPIO && (STL_ENABLE_GPIO == 1))

#include "drivers/board/gpio.hpp"

#include "stl/assert.hpp"
#include "stl/stddef.hpp"

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

const stl::SizeT g_gpioNum = 32U;

enum GPIOFuncE : stl::driver::board::GPIOFuncT
{
  GPIOFuncE_NOT_AVAILABLE = 0,
  GPIOFuncE_GPIO = 1 << 0,
  GPIOFuncE_PWM = 1 << 1,
  GPIOFuncE_EEPROM_DATA = 1 << 2,
  GPIOFuncE_EEPROM_CLK = 1 << 3,
  GPIOFuncE_I2C_DATA = 1 << 4,
  GPIOFuncE_I2C_CLK = 1 << 5,
  GPIOFuncE_SPI_MOSI = 1 << 6,
  GPIOFuncE_SPI_MISO = 1 << 7,
  GPIOFuncE_SPI_CLK = 1 << 8,
  GPIOFuncE_SPI_CE = 1 << 9,
  GPIOFuncE_UART_RX = 1 << 10,
  GPIOFuncE_UART_TX = 1 << 11
};

// Wait few ticks after pull set

#if (defined(CI_REAL_GPIO) && (CI_REAL_GPIO == 1)) && (defined(CI_PLATFORM_RPI) && (CI_PLATFORM_RPI == 1)) 
#include "./private/wiringPi.h"

namespace stl::driver::board {
#if (defined(CI_PLATFORM_RPI_ZEROW) && (CI_PLATFORM_RPI_ZEROW == 1))
static const GPIOFuncT g_gpioFlags[] = {
  GPIOFuncE_GPIO | GPIOFuncE_EEPROM_DATA // GPIO00
  ,
  GPIOFuncE_GPIO | GPIOFuncE_EEPROM_CLK // GPIO01
  ,
  GPIOFuncE_GPIO | GPIOFuncE_I2C_DATA // GPIO02
  ,
  GPIOFuncE_GPIO | GPIOFuncE_I2C_CLK // GPIO03
  ,
  GPIOFuncE_GPIO // GPIO04
  ,
  GPIOFuncE_GPIO // GPIO05
  ,
  GPIOFuncE_GPIO // GPIO06
  ,
  GPIOFuncE_GPIO | GPIOFuncE_SPI_CE // GPIO07
  ,
  GPIOFuncE_GPIO | GPIOFuncE_SPI_CE // GPIO08
  ,
  GPIOFuncE_GPIO | GPIOFuncE_SPI_MISO // GPIO09
  ,
  GPIOFuncE_GPIO | GPIOFuncE_SPI_MOSI // GPIO10
  ,
  GPIOFuncE_GPIO | GPIOFuncE_SPI_CLK // GPIO11
  ,
  GPIOFuncE_GPIO | GPIOFuncE_PWM // GPIO12
  ,
  GPIOFuncE_GPIO | GPIOFuncE_PWM // GPIO13
  ,
  GPIOFuncE_GPIO | GPIOFuncE_UART_TX // GPIO14
  ,
  GPIOFuncE_GPIO | GPIOFuncE_UART_RX // GPIO15
  ,
  GPIOFuncE_GPIO | GPIOFuncE_SPI_CE // GPIO16
  ,
  GPIOFuncE_GPIO | GPIOFuncE_SPI_CE // GPIO17
  ,
  GPIOFuncE_GPIO | GPIOFuncE_SPI_CE // GPIO18
  ,
  GPIOFuncE_GPIO | GPIOFuncE_SPI_MISO | GPIOFuncE_PWM // GPIO19
  ,
  GPIOFuncE_GPIO | GPIOFuncE_SPI_MISO // GPIO20
  ,
  GPIOFuncE_GPIO | GPIOFuncE_SPI_CLK // GPIO21
  ,
  GPIOFuncE_GPIO // GPIO22
  ,
  GPIOFuncE_GPIO // GPIO23
  ,
  GPIOFuncE_GPIO // GPIO24
  ,
  GPIOFuncE_GPIO // GPIO25
  ,
  GPIOFuncE_GPIO // GPIO26
  ,
  GPIOFuncE_GPIO // GPIO27
  ,
  GPIOFuncE_NOT_AVAILABLE // GPIO28
  ,
  GPIOFuncE_NOT_AVAILABLE // GPIO29
  ,
  GPIOFuncE_NOT_AVAILABLE // GPIO30
  ,
  GPIOFuncE_NOT_AVAILABLE // GPIO31
};
#elif (defined(CI_PLATFORM_RPI_ZERO3B) && (CI_PLATFORM_RPI_ZERO3B == 1))
static const GPIOFuncT g_gpioFlags[] = {
  GPIOFuncE_GPIO | GPIOFuncE_EEPROM_DATA // GPIO00
  ,
  GPIOFuncE_GPIO | GPIOFuncE_EEPROM_CLK // GPIO01
  ,
  GPIOFuncE_GPIO | GPIOFuncE_I2C_DATA // GPIO02
  ,
  GPIOFuncE_GPIO | GPIOFuncE_I2C_CLK // GPIO03
  ,
  GPIOFuncE_GPIO // GPIO04
  ,
  GPIOFuncE_GPIO // GPIO05
  ,
  GPIOFuncE_GPIO // GPIO06
  ,
  GPIOFuncE_GPIO | GPIOFuncE_SPI_CE // GPIO07
  ,
  GPIOFuncE_GPIO | GPIOFuncE_SPI_CE // GPIO08
  ,
  GPIOFuncE_GPIO | GPIOFuncE_SPI_MISO // GPIO09
  ,
  GPIOFuncE_GPIO | GPIOFuncE_SPI_MOSI // GPIO10
  ,
  GPIOFuncE_GPIO | GPIOFuncE_SPI_CLK // GPIO11
  ,
  GPIOFuncE_GPIO | GPIOFuncE_PWM // GPIO12
  ,
  GPIOFuncE_GPIO | GPIOFuncE_PWM // GPIO13
  ,
  GPIOFuncE_GPIO | GPIOFuncE_UART_TX // GPIO14
  ,
  GPIOFuncE_GPIO | GPIOFuncE_UART_RX // GPIO15
  ,
  GPIOFuncE_GPIO | GPIOFuncE_SPI_CE // GPIO16
  ,
  GPIOFuncE_GPIO | GPIOFuncE_SPI_CE // GPIO17
  ,
  GPIOFuncE_GPIO | GPIOFuncE_SPI_CE // GPIO18
  ,
  GPIOFuncE_GPIO | GPIOFuncE_SPI_MISO | GPIOFuncE_PWM // GPIO19
  ,
  GPIOFuncE_GPIO | GPIOFuncE_SPI_MISO // GPIO20
  ,
  GPIOFuncE_GPIO | GPIOFuncE_SPI_CLK // GPIO21
  ,
  GPIOFuncE_GPIO // GPIO22
  ,
  GPIOFuncE_GPIO // GPIO23
  ,
  GPIOFuncE_GPIO // GPIO24
  ,
  GPIOFuncE_GPIO // GPIO25
  ,
  GPIOFuncE_GPIO // GPIO26
  ,
  GPIOFuncE_GPIO // GPIO27
  ,
  GPIOFuncE_NOT_AVAILABLE // GPIO28
  ,
  GPIOFuncE_NOT_AVAILABLE // GPIO29
  ,
  GPIOFuncE_NOT_AVAILABLE // GPIO30
  ,
  GPIOFuncE_NOT_AVAILABLE // GPIO31
};
#else
#error No pin layout defined for current platform
#endif

GPIORetvE
GPIO_setDir(const GPIOT gpio, const GPIODirE dir)
{
  if (gpio >= g_gpioNum) {
    return GPIORetvE::IndexOutOfRange;
  }

  switch (dir) {
    case GPIODirE::Input:
      pinMode(gpio, INPUT);
      return GPIORetvE::Ok;
    case GPIODirE::Output:
      pinMode(gpio, OUTPUT);
      return GPIORetvE::Ok;
  }

  Assert(false);

  return GPIORetvE::InvalidValue;
}

GPIORetvE
GPIO_getDir(const GPIOT gpio, GPIODirE& dir)
{
  if (gpio >= g_gpioNum) {
    return GPIORetvE::IndexOutOfRange;
  }

  Assert(false);

  return GPIORetvE::InvalidValue;
}

GPIORetvE
GPIO_setLvl(const GPIOT gpio, const GPIOLvlE lvl)
{
  if (gpio >= g_gpioNum) {
    return GPIORetvE::IndexOutOfRange;
  }

  switch (lvl) {
    case GPIOLvlE::High:
      digitalWrite(gpio, HIGH);
      return GPIORetvE::Ok;
    case GPIOLvlE::Low:
      digitalWrite(gpio, LOW);
      return GPIORetvE::Ok;
    default:
      break;
  }

  Assert(false);

  return GPIORetvE::InvalidValue;
}

GPIORetvE
GPIO_getLvl(const GPIOT gpio, GPIOLvlE& lvl)
{
  if (gpio >= g_gpioNum) {
    return GPIORetvE::IndexOutOfRange;
  }

  Assert(false);

  return GPIORetvE::InvalidValue;
}

GPIORetvE
GPIO_setFunc(const GPIOT gpio, const GPIOFuncT func)
{
  if (gpio >= g_gpioNum) {
    return GPIORetvE::IndexOutOfRange;
  }

  switch (func) {
    default:
      break;
  }

  Assert(false);

  return GPIORetvE::InvalidValue;
}

GPIORetvE
GPIO_getFunc(const GPIOT gpio, GPIOFuncT& func)
{
  if (gpio >= g_gpioNum) {
    return GPIORetvE::IndexOutOfRange;
  }

  switch (func) {
    default:
      break;
  }

  Assert(false);

  return GPIORetvE::InvalidValue;
}

GPIORetvE
GPIO_watchISR(const GPIOT gpio, const GPIOEdgeE mode, void(*func)())
{
  Assert(func);
  
  if (gpio >= g_gpioNum) {
    return GPIORetvE::IndexOutOfRange;
  }

  int modeInternal = INT_EDGE_SETUP;
  switch(mode)
  {
    case GPIOEdgeE::Setup:
      break;
    case GPIOEdgeE::Rising:
      modeInternal = INT_EDGE_RISING;
      break;
    case GPIOEdgeE::Falling:
      modeInternal = INT_EDGE_FALLING;
      break;
    case GPIOEdgeE::Both:
      modeInternal = INT_EDGE_BOTH;
      break;
    default:
      Assert(false);
      return GPIORetvE::InvalidValue;
  }

  return wiringPiISR(gpio, modeInternal, func) == 0 ? GPIORetvE::Ok : GPIORetvE::Error;
}

GPIORetvE
GPIO_unwatchISR(const GPIOT gpio)
{
  if (gpio >= g_gpioNum) {
    return GPIORetvE::IndexOutOfRange;
  }

  return GPIORetvE::Ok;
}

// TODO: Verify that access to registers is OK (root?)

bool
GPIO_init()
{
  if (wiringPiSetupGpio() < 0) { // use BCM2835 Pin number table
    return false;
  }

  return true;
}
}
#else

namespace stl::driver::board {
GPIORetvE
GPIO_setDir(const GPIOT gpio, const GPIODirE dir)
{
  if (gpio >= g_gpioNum) {
    return GPIORetvE::IndexOutOfRange;
  }

  return GPIORetvE::Ok;
}

GPIORetvE
GPIO_getLvl(const GPIOT gpio, GPIOLvlE& lvl)
{
  if (gpio >= g_gpioNum) {
    return GPIORetvE::IndexOutOfRange;
  }

  lvl = GPIOLvlE::Low;

  return GPIORetvE::Ok;
}

GPIORetvE
GPIO_setLvl(const GPIOT gpio, const GPIOLvlE lvl)
{
  if (gpio >= g_gpioNum) {
    return GPIORetvE::IndexOutOfRange;
  }

  return GPIORetvE::Ok;
}

GPIORetvE
GPIO_getDir(const GPIOT gpio, GPIODirE& dir)
{
  if (gpio >= g_gpioNum) {
    return GPIORetvE::IndexOutOfRange;
  }

  dir = GPIODirE::Input;

  return GPIORetvE::Ok;
}

GPIORetvE
GPIO_setFunc(const GPIOT gpio, const GPIOFuncT func)
{
  if (gpio >= g_gpioNum) {
    return GPIORetvE::IndexOutOfRange;
  }

  return GPIORetvE::Ok;
}

GPIORetvE
GPIO_getFunc(const GPIOT gpio, GPIOFuncT& func)
{
  if (gpio >= g_gpioNum) {
    return GPIORetvE::IndexOutOfRange;
  }

  func = GPIOFuncE_GPIO;

  return GPIORetvE::Ok;
}

GPIORetvE
GPIO_watchISR(const GPIOT gpio, const GPIOEdgeE mode, void(*func)())
{
  Assert(func);

  if (gpio >= g_gpioNum) {
    return GPIORetvE::IndexOutOfRange;
  }

  return GPIORetvE::Ok;
}

GPIORetvE
GPIO_unwatchISR(const GPIOT gpio)
{
  if (gpio >= g_gpioNum) {
    return GPIORetvE::IndexOutOfRange;
  }

  return GPIORetvE::Ok;
}

bool
GPIO_init()
{
  return true;
}
}

#endif // RPI_ZERO2W
#endif // (defined STL_ENABLE_GPIO && (STL_ENABLE_GPIO == 1))
