#include "drivers/board/rs485.hpp"

#include "log/logging.hpp"

#include "stl/error.hpp"
#include "stl/mutex.hpp"

namespace stl::driver::board {
RS485::~RS485() {}

bool
RS485::attachSlave(RS485SlaveBase* const slave)
{
  LockGuard<Mutex> lock(m_lock);

  stl::Pair<SlaveContainerT::iterator, bool> result =
    m_slaveContainer.insert(slave);
  if (result.second) {
    (*result.first)->m_pollingTimer.reset();
    (*result.first)->m_attached = true;
    logging::log << logging::LvlE::Info << "RS485 slave attached" << stl::endl;
  } else {
    logging::log << logging::LvlE::Error << "RS485 slave NOT attached (" << m_slaveContainer.size() << ")"
                 << stl::endl;
  }

  return result.second;
}

/* bool
RS485::detachSlave(RS485SlaveBase* const slave)
{
  LockGuard<Mutex> lock(m_lock);

  return detachSlaveInternal(slave);
}

bool
RS485::detachSlaveInternal(RS485SlaveBase* const slave)
{
  SlaveContainerT::iterator it = m_slaveContainer.find(slave);
  if (m_slaveContainer.end() == it) {
    logging::log << logging::LvlE::Error << "RS485 slave NOT detached"
                 << stl::endl;
    return false;
  }

  slave->deinitHardware();

  m_slaveContainer.erase(it);

  logging::log << logging::LvlE::Info << "RS485 slave detached" << stl::endl;

  return true;
} */

bool
RS485::detachAllSlaveInternal()
{
  for (stl::Set<RS485SlaveBase*>::iterator it = m_slaveContainer.begin(); it != m_slaveContainer.end(); ) {
    (*it)->deinitHardware();
    RS485SlaveBase* const temp = *it;
    it = m_slaveContainer.erase(it);
    temp->detached();
  }

  return true;
}

bool
RS485::start(filesystem::Path path, const UARTConfig& uartConfig)
{
  try {
    if (!UART::open(move(path),
                    stl::stream::FF_RW | stl::stream::FF_NoBlock |
                      stl::stream::FF_NoCTTY | stl::stream::FF_NoDelay,
                    uartConfig)) {
#if (!defined CI_PLATFORM_PC_TEST) &&                                          \
  (!defined IONT_DEBUG_IGNORE_SERIAL_PORT_OPEN_ERROR)
      logging::log << logging::LvlE::Error << "Could not open serial port "
                   << path << stl::endl;
      return false;
#else
      logging::log << logging::LvlE::Debug << "Ignoring serial open error :)"
                   << stl::endl;
#endif // !defined CI_PLATFORM_PC_TEST
    }

    RS485Runnable::start(worker, this);
    return true;
  } catch (...) {
    Assert(false);
    return false;
  }

  return true;
}

bool
RS485::running() const
{
  return RS485Runnable::running();
}

void
RS485::stop()
{
  try {
    RS485Runnable::stop();

    if (UART::isOpen()) {
      UART::close();
    }
  } catch (...) {
    // TODO
  }
}

void
RS485::worker(const stl::RunnableBase::Params* params, RS485* const instance)
{
  Assert(params);
  Assert(instance);

  logging::log << "RS485 worker started with "
               << instance->m_slaveContainer.size() << " slaves attached."
               << stl::endl;

  while (!params->m_stop) {
    {
      LockGuard<Mutex> lock(instance->m_lock);

      for (SlaveContainerT::iterator it = instance->m_slaveContainer.begin(); it != instance->m_slaveContainer.end();) {
        if((*it)->detachSet())
        {
          RS485SlaveBase* const temp = (*it);
          it = instance->m_slaveContainer.erase(it);
          temp->detached();
        }
        else
        {
          if ((*it)->m_pollingTimer.timeout()) {
            if (!(*it)->configured()) {
              logging::log << logging::LvlE::Info
                          << "Device not configured yet. Configuring.."
                          << stl::endl;
              const ErrorE configRetv = (*it)->configureHardwareInternal();
              if (ErrorE::Ok != configRetv) {
                // TODO: How to handle config errors?
              }

              (*it)->m_pollingTimer.reset();

              ++it;

              continue;
            }

            logging::log << logging::LvlE::Info << "Polling RS485 slave.."
                        << stl::endl;

            const stl::ErrorE error = (*it)->poll();
            if (stl::ErrorE::Ok != error) {
              // TODO ...
            }

            (*it)->m_pollingTimer.reset();
          } 

          ++it;
        }
      }
    }

    stl::thread::sleepFor(50);
  }

  {
    LockGuard<Mutex> lock(instance->m_lock);

    instance->detachAllSlaveInternal();
  }
}

RS485SlaveBase::RS485SlaveBase(RS485& bus,
                               const RS485::PollingTimeT& pollingTime)
  : m_bus(&bus)
  , m_pollingTimer(pollingTime)
  , m_detach(false)
  , m_attached(false)
{}

RS485SlaveBase::~RS485SlaveBase() {}

bool
RS485SlaveBase::isOpen() const
{
  stl::LockGuard<stl::Mutex> lock(m_lock);

  Assert(nullptr != m_bus);
  if (!m_bus) {
    return false;
  }

  return true;
}

SizeT
RS485SlaveBase::write(const UInt8T* const data, SizeT size)
{
  stl::LockGuard<stl::Mutex> lock(m_lock);

  Assert(nullptr != m_bus);
  if (!m_bus) {
    return 0U; // TODO
  }

  if (m_bus->isOpen()) {
    return m_bus->write(data, size);
  } else {
#if !defined CI_PLATFORM_PC_TEST
    Assert(m_bus->isOpen());
#endif
  }

  return 0U;
}

SizeT
RS485SlaveBase::read(UInt8T* data, SizeT maxSize)
{
  stl::LockGuard<stl::Mutex> lock(m_lock);

  Assert(nullptr != m_bus);

  if (!m_bus) {
    return 0U; // TODO
  }

  if (m_bus->isOpen()) {
    return m_bus->read(data, maxSize);
  } else {
#if !defined CI_PLATFORM_PC_TEST
    Assert(m_bus->isOpen());
#endif
  }

  return 0U;
}

void
RS485SlaveBase::flush()
{
  stl::LockGuard<stl::Mutex> lock(m_lock);

  Assert(nullptr != m_bus);

  if (!m_bus) {
    return; // TODO
  }

  m_bus->flush();
}
}
