#include "stl/compile_info.hpp"
#include "stl/stl_config.hpp"

#if defined(STL_ENABLE_UART) && (STL_ENABLE_UART == 1) &&                      \
  defined(CI_OS_LINUX) && (CI_OS_LINUX == 1)

#include "drivers/board/gpio.hpp"
#include "drivers/board/uart.hpp"

#include "stl/fstream.hpp"
#include "stl/stddef.hpp"
#include "stl/stdint.hpp"
#include "stl/stream_base.hpp"
#include "stl/thread.hpp"

#include "stl/private/fstream_linux.hpp"

#include <fcntl.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <termios.h>
#include <unistd.h>

static int
baudrateTranslate(stl::UInt32T baud);

namespace stl::driver::board {
UART::UART() {}

UART::UART(filesystem::Path path,
           const stream::FileFlagsT flags,
           const UARTConfig& config)
{
  static_cast<void>(open(move(path), flags, config));
}

UART::~UART()
{
  if (isOpen()) {
    close();
  }
}

bool
UART::open(filesystem::Path path,
           const stream::FileFlagsT flags,
           const UARTConfig& config)
{
  Assert(m_impl);

  logging::log << logging::LvlE::Info << "Opening UART " << path << stl::endl;

  if (!transmitterInit()) {
    return false;
  }

  if (!FStream::open(move(path), flags)) {
    return false;
  }

  struct termios options
  {};

  if (-1 == fcntl(m_impl->fd(), F_SETFL, O_RDWR)) {
    return false;
  }

  // Get and modify current options:

  if (0 != tcgetattr(m_impl->fd(), &options)) {
    return false;
  }

  const speed_t baudRate = baudrateTranslate(config.baudrate);

  cfmakeraw(&options); // Void

  if (0 != cfsetispeed(&options, baudRate)) {
    return false;
  }

  if (0 != cfsetospeed(&options, baudRate)) {
    return false;
  }

  // CLOCAL - Ignore modem status lines.
  // CREAD - Enable receiver
  // CSIZE - Character size
  // ICANON - Canonical input (erase and kill processing).
  // ECHO - Enable echo.
  // ECHOE - Echo erase character as error-correcting backspace.
  // ISIG - Enable signals.

  // Clear char size
  options.c_cflag &= ~CSIZE;
  // Set char size
  switch (config.charSize) {
    case UARTCharSizeE::C5Bits:
      options.c_cflag |= CS5;
      break;
    case UARTCharSizeE::C6Bits:
      options.c_cflag |= CS6;
      break;
    case UARTCharSizeE::C7Bits:
      options.c_cflag |= CS7;
      break;
    case UARTCharSizeE::C8Bits:
      options.c_cflag |= CS8;
      break;
    default:
      Assert(false);
      return false; // TODO: Deinit previous calls
  }

  // Clear parity flag
  options.c_cflag &= ~PARENB;
  // Set parity flag
  switch (config.parity) {
    case UARTParityE::None:
      // Do nothing
      break;
    case UARTParityE::Odd:
      options.c_cflag |= PARODD;
      break;
    case UARTParityE::Even:
      options.c_cflag |= PARENB; // Even by default
      break;
    default:
      Assert(false);
      return false; // TODO: Deinit previous calls
  }

  // Clear stop bits flag
  options.c_cflag &= ~CSTOPB;
  // Set stop bits flag
  switch (config.stopBits) {
    case UARTStopBitsE::One:
      break; // One by default
    case UARTStopBitsE::Two:
      options.c_cflag |= CSTOPB;
      break;
    default:
      Assert(false);
      return false;
  }

  options.c_cflag |= (CLOCAL | CREAD);                // TODO
  options.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG); // TODO
  options.c_oflag &= ~OPOST;                          // TODO

  options.c_cc[VMIN] = 0;
  options.c_cc[VTIME] = 100; // Ten seconds (100 deciseconds)

  if (0 != tcsetattr(m_impl->fd(), TCSANOW, &options)) // TODO
  {
    return false;
  }

  int status = 0;
  if (0 != ioctl(m_impl->fd(), TIOCMGET, &status)) // TODO
  {
    return false;
  }

  status |= TIOCM_DTR; // TODO
  status |= TIOCM_RTS; // TODO

  if (0 != ioctl(m_impl->fd(), TIOCMSET, &status)) // TODO
  {
    return false;
  }

  stl::thread::sleepFor(1);

  return -1 != m_impl->fd();
}

SizeT
UART::available() const
{
  Assert(m_impl);
  Assert(-1 != m_impl->fd());

  int result;

  if (ioctl(m_impl->fd(), FIONREAD, &result) == -1) {
    return 0U;
  }

  return result < 0 ? 0U : static_cast<SizeT>(result);
}

void
UART::flush()
{
  Assert(m_impl);
  Assert(-1 != m_impl->fd());

  tcflush(m_impl->fd(), TCIOFLUSH);
}

static GPIOT g_mode = 4;

bool
UART::transmitterInit()
{
  GPIO_setDir(g_mode, GPIODirE::Output);
  stl::thread::sleepFor(1);

  return true;
}

void
UART::setModeReceive()
{
  GPIO_setLvl(g_mode, GPIOLvlE::High);
  stl::thread::sleepFor(1);
}

void
UART::setModeTransmit()
{
  GPIO_setLvl(g_mode, GPIOLvlE::Low);
  stl::thread::sleepFor(1);
}
}

static int
baudrateTranslate(stl::UInt32T baud)
{
  switch (baud) {
    case 50:
      return B50;
    case 75:
      return B75;
    case 110:
      return B110;
    case 134:
      return B134;
    case 150:
      return B150;
    case 200:
      return B200;
    case 300:
      return B300;
    case 600:
      return B600;
    case 1200:
      return B1200;
    case 1800:
      return B1800;
    case 2400:
      return B2400;
    case 4800:
      return B4800;
    case 9600:
      return B9600;
    case 19200:
      return B19200;
    case 38400:
      return B38400;
    case 57600:
      return B57600;
    case 115200:
      return B115200;
    case 230400:
      return B230400;
    case 460800:
      return B460800;
    case 500000:
      return B500000;
    case 576000:
      return B576000;
    case 921600:
      return B921600;
    case 1000000:
      return B1000000;
    case 1152000:
      return B1152000;
    case 1500000:
      return B1500000;
    case 2000000:
      return B2000000;
    case 2500000:
      return B2500000;
    case 3000000:
      return B3000000;
    case 3500000:
      return B3500000;
    case 4000000:
      return B4000000;
    default:
      break;
  }

  return 0;
}

#endif // defined(STL_ENABLE_UART) && (STL_ENABLE_UART == 1)
