#include "drivers/energymeter/shelly_3em.hpp"

#include "httplib/httplib.h"

#include "log/logging.hpp"
#include "parsers/json.hpp"

#include "stl/assert.hpp"
#include "stl/compiler.hpp"
#include "stl/error.hpp"
#include "stl/network.hpp"

namespace stl::driver::energymeter {
const EnergymeterDeviceBase::ConnectionTypeListT
  Shelly3EMBase::m_connectionTypeList = { ConnectionTypeE::Network };

const EnergymeterDeviceBase::UsageListT Shelly3EMBase::m_usageList = {
  UsageE::ElectricalNetworkMonitor
};

Shelly3EM_Network::Shelly3EM_Network(const UsageE usage,
                                     network::NetworkAddress address,
                                     PollingTimeT pollingTime)
  : Shelly3EMBase(usage, ConnectionTypeE::Network)
  , EnergymeterNetworkBase(stl::move(address), stl::move(pollingTime))
{
}

stl::ErrorE
Shelly3EM_Network::configureHardware()
{
  // TODO: Connection status update if needed
  return stl::ErrorE::Ok;
}

stl::ErrorE
Shelly3EM_Network::update()
{
  try {
    // logging::log << logging::LvlE::Debug << "Setting Shelly communication params .. " << address().getIPStr().c_str() << ':' << address().getPort() << stl::endl;

    httplib::Client client(address().getIPStr().c_str(), address().getPort());

    client.set_write_timeout(10, 0);
    client.set_read_timeout(10, 0);
    client.set_connection_timeout(10, 0);

    // logging::log << logging::LvlE::Info << "Fetching Shelly status .." << stl::endl;

    httplib::Result result{};

    for(stl::SizeT i = 0U; i < 3U; ++i)
    {
      result = client.Get("/status");
      if(result)
      {
        break;
      }
      else
      {
        LOGEXT << "[energymeter] Retry.." << stl::endl;
      }
    }

    if (result) {
      // logging::log << logging::LvlE::Debug << "PARSING SHELLY /status RESPONSE: \"" << result->body.c_str() << "\"" << stl::endl << stl::endl;

      stl::parser::JSON response = stl::parser::JSON::parse(result->body);

      stl::parser::JSON energymeterListJSON = response["emeters"];

      Assert(energymeterListJSON.size() == 3U);

      Measurement measurement{};

      if (energymeterListJSON.size() <= 3U) {
        for (stl::SizeT i = 0U; i < energymeterListJSON.size(); ++i) {
          measurement.lineMeasurement[i].current =
            energymeterListJSON[i]["current"].get<float>();
          measurement.lineMeasurement[i].voltage =
            energymeterListJSON[i]["voltage"].get<float>();
          measurement.lineMeasurement[i].powerFactor =
            energymeterListJSON[i]["pf"].get<float>();

          /* logging::log << logging::LvlE::Info << "Shelly measurement L"
                       << (i + 1)
                       << " - I: " << measurement.lineMeasurement[i].current
                       << " U: " << measurement.lineMeasurement[i].voltage
                       << " f: " << measurement.lineMeasurement[i].frequency
                       << " pf: " << measurement.lineMeasurement[i].powerFactor
                       << stl::endl; */
        }
      }

      measurement.power = response["total_power"].get<float>();

      /* logging::log << logging::LvlE::Info
                   << "SHelly measurement - power: " << measurement.power
                   << stl::endl; */

      if (m_statusCb) {
        m_statusCb(stl::move(measurement),
                   stl::driver::energymeter::ConnectionStatusE::Online,
                   m_arg);
      }

      // logging::log << logging::LvlE::Info << "Shelly status updated" << stl::endl;
    } else {
      // logging::log << logging::LvlE::Warning << "NO RESPONSE FROM SHELLY" << stl::endl;

      if (m_statusCb) {
        m_statusCb(
          {}, stl::driver::energymeter::ConnectionStatusE::Offline, m_arg);
      }

      // logging::log << logging::LvlE::Warning << "Cannot get status from Shelly " << CU32(result.error()) << stl::endl;
      return stl::ErrorE::ConnectionError;
    }
  } catch (const std::exception& e) {
    logging::log << logging::LvlE::Exception << '[' << __FILENAME__ << "] "
                 << e.what() << stl::endl;

    if (m_statusCb) {
      m_statusCb(
        {}, stl::driver::energymeter::ConnectionStatusE::Offline, m_arg);
    }

    return stl::ErrorE::FormatError;
  } catch (...) {
    logging::log << logging::LvlE::Exception << '[' << __FILENAME__
                 << "] Unknown exception caught" << stl::endl;

    if (m_statusCb) {
      m_statusCb(
        {}, stl::driver::energymeter::ConnectionStatusE::Offline, m_arg);
    }

    return stl::ErrorE::FormatError;
  }

  return stl::ErrorE::Ok;
}
}
