#include "parsers/json.hpp"

#include "log/logging.hpp"

#include "stl/filesystem.hpp"
#include "stl/fstream.hpp"
#include "stl/string.hpp"
#include "stl/vector.hpp"

namespace stl::parser {
JSONError
JSON_load_internal(const filesystem::Path& path, JSON& json)
{
  Vector<char> jsonData;

  stream::FStream stream(move(path), stream::FF_R);
  if (!stream.isOpen()) {
    logging::log << logging::LvlE::Error << "Config \"" << path
                 << "\" could not be open." << stl::endl;
    return JSONError::DoesNotExist;
  }

  jsonData.resize(stream.size());

  const stl::SizeT readSize = stream.read(
    reinterpret_cast<stl::UInt8T*>(jsonData.data()), jsonData.size());

  if ((jsonData.size() != readSize) || (jsonData.size() == 0U)) {
    logging::log << logging::LvlE::Error << "Config \"" << path
                 << "\" could not be read. [read size: " << readSize << "]"
                 << stl::endl;
    return JSONError::ReadError;
  }

  try {
    json = JSON::parse(std::string(jsonData.data(), jsonData.size()));
  } catch (...) {
    logging::log << logging::LvlE::Error << "Config \"" << path
                 << "\" could not be parsed." << stl::endl;
    return JSONError::ParserError;
  }

  return JSONError::Ok;
}

JSONError
JSON_load(const filesystem::Path& path, JSON& json)
{
  JSONError retv = JSON_load_internal(path, json);
  if (JSONError::Ok != retv) {
    JSONError retvTemp = JSON_load_internal(path + ".backup", json);
    if (JSONError::Ok != retvTemp) {
      return retv; // Return original file error
    }
  }

  return JSONError::Ok;
}

JSONError
JSON_backup_create(const filesystem::Path& path)
{
  Vector<char> jsonData;

  stream::FStream stream(move(path), stream::FF_R);
  if (!stream.isOpen()) {
    logging::log << logging::LvlE::Error << "Config \"" << path
                 << "\" could not be open." << stl::endl;
    return JSONError::DoesNotExist;
  }

  jsonData.resize(stream.size());

  const stl::SizeT readSize = stream.read(
    reinterpret_cast<stl::UInt8T*>(jsonData.data()), jsonData.size());

  if (jsonData.size() != readSize) {
    logging::log << logging::LvlE::Error << "Config \"" << path
                 << "\" could not be read. [read size: " << readSize << "]"
                 << stl::endl;
    return JSONError::ReadError;
  }

  stream.close();

  const filesystem::Path backupPath = path + ".backup";

  stream.open(backupPath, stream::FF_W | stream::FF_Trunc | stream::FF_Create);
  if (!stream.isOpen()) {
    logging::log << logging::LvlE::Error << "File \"" << backupPath
                 << "\" not open!" << stl::endl;
    return JSONError::NoAccess;
  }

  stream.write(reinterpret_cast<const stl::UInt8T*>(jsonData.data()),
               jsonData.size()); // TODO: Check retv

  return JSONError::Ok;
}

JSONError
JSON_save(const filesystem::Path& path, const JSON& json)
{
  JSONError retv = JSON_backup_create(path);
  if ((JSONError::Ok != retv) && (JSONError::DoesNotExist != retv)) {
    return retv;
  }

  stream::FStream stream(path,
                         stream::FF_W | stream::FF_Trunc | stream::FF_Create);
  if (!stream.isOpen()) {
    logging::log << logging::LvlE::Error << "File \"" << path << "\" not open!"
                 << stl::endl;
    return JSONError::NoAccess;
  }

  std::string jsonData;

  try {
    jsonData = json.dump(4);
  } catch (...) {
    logging::log << logging::LvlE::Error << "JSON for file \"" << path
                 << "\" could not be dumped!" << stl::endl;
    return JSONError::DumpError;
  }

  stream.write(reinterpret_cast<const stl::UInt8T*>(jsonData.c_str()),
               jsonData.length()); // TODO: Check retv

  return JSONError::Ok;
}
}