#include "stl/fstream.hpp"

#include "stl/private/fstream_linux.hpp"

namespace stl::stream {
FStreamBase::FStreamBase() {}

FStreamBase::FStreamBase(filesystem::Path path, const FileFlagsT flags)
  : m_path(move(path))
  , m_flags(flags)
{
}

FStreamBase::~FStreamBase()
{
  for (ObserverMapT::iterator it = m_observerMap.begin();
       it != m_observerMap.end();
       ++it) {
    it->second.get().m_alive = false;
  }
}

FStream::FStream()
  : m_impl(makeUnique<FStreamImpl>())
{
}

FStream::~FStream()
{
  if (isOpen()) {
    close();
  }
}

FStream::FStream(filesystem::Path path, const FileFlagsT flags)
  : m_impl(makeUnique<FStreamImpl>(move(path), flags))
{
}

FStream::FStream(FStream&& other)
{
  *this = move(other);
}

FStream&
FStream::operator=(FStream&& other)
{
  static_cast<FStreamBase&>(*this) =
    stl::move(static_cast<FStreamBase&>(other));

  m_impl = move(other.m_impl);

  return *this;
}

bool
FStream::open(filesystem::Path path, const FileFlagsT flags)
{
  Assert(m_impl);
  return m_impl->open(move(path), flags);
}

bool
FStream::isOpen() const
{
  Assert(m_impl);
  return m_impl->isOpen();
}

SizeT
FStream::write(const UInt8T* const data, SizeT size)
{
  Assert(m_impl);
  return m_impl->write(data, size);
}

SizeT
FStream::read(UInt8T* data, SizeT maxSize)
{
  Assert(m_impl);
  return m_impl->read(data, maxSize);
}

bool
FStream::seek(const SizeT offset, const SeekModeE seekMode)
{
  Assert(m_impl);
  return m_impl->seek(offset, seekMode);
}

SizeT
FStream::position() const
{
  Assert(m_impl);
  return m_impl->position();
}

void
FStream::flush()
{
  Assert(m_impl);
  m_impl->flush();
}

void
FStream::close()
{
  Assert(m_impl);
  m_impl->close();
}

SizeT
FStream::size() const
{
  Assert(m_impl);
  return m_impl->size();
}

FStream::operator bool() const
{
  Assert(m_impl);
  return isOpen();
}

FStreamObserver::FStreamObserver(stream::FStreamBase& stream)
  : m_alive(true)
  , m_file(stream)
{
  stream.registerObserver(*this);
}

FStreamObserver::FStreamObserver(FStreamObserver&& other)
  : m_alive(other.m_alive)
  , m_file(move(other.m_file))
{
}

FStreamObserver::~FStreamObserver()
{
  if (m_alive) {
    m_file.get().unregisterObserver(*this);
  }
}

FStreamObserver&
FStreamObserver::operator=(FStreamObserver&& other)
{
  m_alive = other.m_alive;
  m_file = move(other.m_file);

  return *this;
}

bool
FStreamObserver::alive() const
{
  return m_alive;
}

stream::FStreamBase&
FStreamObserver::file()
{
  Assert(m_alive);
  return m_file.get();
}

FStreamObserver::operator bool() const
{
  return alive();
}

FStream::DataContainerT
load(FStreamBase& stream)
{
  const SizeT fileSize = stream.size();

  FStream::DataContainerT buffer;

  buffer.resize(fileSize);

  if (fileSize != stream.read(buffer.data(), fileSize)) {
    return {};
  }

  return buffer;
}

String
loadString(FStreamBase& stream)
{
  const SizeT fileSize = stream.size();

  String buffer;

  buffer.resize(fileSize);

  if (fileSize != stream.read(reinterpret_cast<ByteT*>(&buffer[0]), fileSize)) {
    return {};
  }

  return buffer;
}
}
