#include "stl/compile_info.hpp"
#include "stl/stl_config.hpp"

#include "stl/private/fstream_linux.hpp"

#if 1 // defined(CI_SUPPORT_STD) && (CI_SUPPORT_STD == 1) && (CI_OS_LINUX == 1)

#include "log/logging.hpp"

#include "stl/assert.hpp"
#include "stl/compiler.hpp"
#include "stl/fstream.hpp"
#include "stl/stddef.hpp"

#include <fcntl.h>
#include <sys/stat.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

namespace stl::stream {
FStream::FStreamImpl::FStreamImpl()
  : m_fd(-1)
{
}

FStream::FStreamImpl::FStreamImpl(FStreamImpl&& other)
{
  *this = move(other);
}

FStream::FStreamImpl&
FStream::FStreamImpl::operator=(FStreamImpl&& other)
{
  m_fd = ::stl::move(other.m_fd);
  other.m_fd = -1;

  return *this;
}

FStream::FStreamImpl::FStreamImpl(filesystem::Path path, const FileFlagsT flags)
  : FStreamBase(move(path), flags)
{
  openInternal();
}

FStream::FStreamImpl::~FStreamImpl()
{
  if (isOpen()) {
    close();
  }
}

void
FStream::FStreamImpl::flush()
{ /* DO NOTHING: write/read are not buffered */
}

bool
FStream::FStreamImpl::open(filesystem::Path path, const FileFlagsT flags)
{
  m_path = move(path); // TODO: Refactor a little bit
  m_flags = flags;

  return openInternal(); // TODO: Reimplement with internal open method in the
                         // parent to move and copy parameters and then pass
                         // them into the internal open method as const refs
}

bool
FStream::FStreamImpl::isOpen() const
{
  return -1 != m_fd;
}

SizeT
FStream::FStreamImpl::read(UInt8T* data, SizeT maxSize)
{
  Assert(-1 != m_fd);
  const int retv = ::read(m_fd, reinterpret_cast<char*>(data), maxSize);
  if (retv < 0) {
    logging::log << logging::LvlE::Error << "Reading file failed! " << retv << " error: " << errno
                 << stl::endl;
    return 0; // TODO: Save error
  }

  return retv;
}

SizeT
FStream::FStreamImpl::write(const UInt8T* const data, SizeT size)
{
  Assert(-1 != m_fd);
  const int retv = ::write(m_fd, reinterpret_cast<const char*>(data), size);
  if (retv < 0) {
    return 0; // TODO: Save error
  }

  return retv;
}

bool
FStream::FStreamImpl::seek(const SizeT offset, const SeekModeE seekMode)
{
  Assert(-1 != m_fd);

  int origin = SEEK_SET;

  switch (seekMode) { // TODO: Are origins correct?
    case SeekModeE::Beg:
      origin = SEEK_SET;
      return ::lseek(m_fd, offset, origin) == offset;
    case SeekModeE::Set:
      origin = SEEK_CUR;
      ::lseek(m_fd, offset, origin);
      return true;
    case SeekModeE::End:
      origin = SEEK_END;
      ::lseek(m_fd, offset, origin);
      return true;
    default:
      Assert(false);
      return false;
  };
}

SizeT
FStream::FStreamImpl::position() const
{
  Assert(-1 != m_fd);
  return ::lseek(m_fd, 0, SEEK_CUR);
}

void
FStream::FStreamImpl::close()
{
  Assert(-1 != m_fd);
  ::close(m_fd);
  m_fd = -1;
}

SizeT
FStream::FStreamImpl::size() const
{
  Assert(-1 != m_fd);
  const stl::SizeT oldPosition = position();
  const stl::SizeT fileSize = ::lseek(m_fd, 0, SEEK_END);
  ::lseek(m_fd, oldPosition, SEEK_SET);
  return fileSize;
}

FStream::FStreamImpl::operator bool() const
{
  return isOpen();
}

int
FStream::FStreamImpl::fd() const
{
  return m_fd;
}

bool
FStream::FStreamImpl::openInternal()
{
  int openMode = 0;

  if (((m_flags & FF_R) == FF_R) && ((m_flags & FF_W) == FF_W)) {
    openMode |= O_RDWR;
  } else if (m_flags & FF_R) {
    openMode |= O_RDONLY;
  } else if (m_flags & FF_W) {
    openMode |= O_WRONLY;
  }

  if (m_flags & FF_Trunc) {
    openMode |= O_TRUNC;
  }

  if (m_flags & FF_App) {
    openMode |= O_APPEND;
  }

  if (m_flags & FF_NoBlock) {
    openMode |= O_NONBLOCK;
  }

  if (m_flags & FF_NoDelay) {
    openMode |= O_NDELAY;
  }

  if (m_flags & FF_NoCTTY) {
    openMode |= O_NOCTTY;
  }

  if (m_flags & FF_Create) {
    openMode |= O_CREAT;
  }

  if(openMode & O_CREAT)
  {
    m_fd = ::open(m_path.c_str(), openMode, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH);
  }
  else
  {
    m_fd = ::open(m_path.c_str(), openMode);
  }

  if (-1 == m_fd) {
    logging::log << logging::LvlE::Error << "File \"" << m_path << "\" not open. Error " << errno
                 << stl::endl;
  } else {
    logging::log << logging::LvlE::Info << "File open with following flags: " << CU64(openMode) << "["
                 << m_flags << "]" << stl::endl;
  }

  return isOpen(); // TODO: Read errno and save the value into internal
                   // structure
}
}
#endif // defined(CI_SUPPORT_STD) && (CI_SUPPORT_STD == 1) && (CI_OS_LINUX == 1)
