#include "stl/http_parser.hpp"

#include "stl/compiler.hpp"
#include "stl/assert.hpp"

// Reserved:   !#$&'()*+,/:;=?@[]
// Unreserved: A-Za-z0-9 - _.~

#if defined WS_DDEBUG
#define WS_CHECK_RETURN(retv) do { Assert(ReturnE::ReturnE_Fail != (retv)); return (retv); } while(false)
#else
#define WS_CHECK_RETURN(retv) return (retv)
#endif

namespace http {
    bool HTTPURL::isReservedChar(const char c) {
        return
            (c == '!')
            || ((c >= '&') && (c <= '+'))
            || ((c >= ',') && (c <= '/'))
            || (c == '#') || (c == '$')
            || (c == ':') || (c == ';')
            || (c == '?') || (c == '=')
            || (c == '@') || (c == '[')
            || (c == ']');
    }

    bool HTTPURL::isUnreservedChar(const char c) {
        return
            ((c >= 'A') && (c <= 'Z'))
            || ((c >= 'a') && (c <= 'z'))
            || ((c >= '0') && (c <= '9'))
            || (c == '-') || (c == '_')
            || (c == '.') || (c == '~');
    }

    bool HTTPURL::isAllowedChar(const char c) {
        return
            ((c >= '?') && (c <= 'Z'))
            || ((c >= 'a') && (c <= 'z'))
            || ((c >= '-') && (c <= ';'))
            || (c == '#') || (c == '$')
            || ((c >= '&') && (c <= '+'))
            || (c == '!') || (c == '=')
            || (c == '[') || (c == ']')
            || (c == '~');
    }

    HTTPHeader::HTTPHeader(stl::String name, stl::String value)
        : m_name(stl::move(name))
        , m_value(stl::move(value)) {}

    const stl::String& HTTPRequest::versionMajor() const {
        return m_versionMajor;
    }

    const stl::String& HTTPRequest::versionMinor() const {
        return m_versionMinor;
    }

    void HTTPRequest::addHeader(stl::String&& name, stl::String&& value) {
        m_headerList.emplaceBack(stl::move(name), stl::move(value));
    }

    const stl::String& HTTPRequest::uri() const {
        return m_uri;
    }

    const stl::String& HTTPRequest::content() const {
        return m_content;
    }

    HTTPRequestParser::ReturnE HTTPRequestParser::consume(const char* const data, const stl::SizeT length) {
        Assert(nullptr != data);

        const char* dataCurr = data;
        const char* const dataEnd = data + length;

        if (m_resultReady) {
            WS_CHECK_RETURN(ReturnE::ReturnE_ResultReady);
        }

        if(0 == length) {
            return ReturnE::ReturnE_Continue;
        }

        if (!m_result) {
            m_result = stl::makeUnique<HTTPRequest>();
        }

        ReturnE loc_Return = ReturnE::ReturnE_Fail;

        while (dataEnd != dataCurr) {
            loc_Return = m_parserFunc(*this, dataCurr, dataEnd);

            if (ReturnE::ReturnE_Continue != loc_Return) {
                break;
            }
        }

        WS_CHECK_RETURN(loc_Return);
    }

    void HTTPRequestParser::reset() {
        m_result.reset();
        m_parserFunc = parseRequestLineMethod;
        m_resultReady = false;
        m_stringInProgress = false;
    }

    bool HTTPRequestParser::hasResult() const {
        return m_resultReady && m_result;
    }

    stl::UniquePtr<HTTPRequest> HTTPRequestParser::result() {
        if (!m_resultReady) {
            return nullptr;
        }

        return move(m_result);
    }

    HTTPRequestParser::ReturnE HTTPRequestParser::parseSecondLF(
        HTTPRequestParser& instance, const char*& dataCurr
        , const char* const dataEnd) {
        Assert(nullptr != dataCurr);
        Assert(nullptr != dataEnd);
        Assert(dataCurr < dataEnd);
        Assert(instance.m_result);

        if ('\n' != *dataCurr) {
            WS_CHECK_RETURN(ReturnE::ReturnE_Fail);
        }

        instance.m_resultReady = true;

        WS_CHECK_RETURN(ReturnE::ReturnE_ResultReady);
    }

    HTTPRequestParser::ReturnE HTTPRequestParser::parseSecondCROrHeader(
        HTTPRequestParser& instance, const char*& dataCurr
        , const char* const dataEnd) {
        Assert(nullptr != dataCurr);
        Assert(nullptr != dataEnd);
        Assert(dataCurr < dataEnd);
        Assert(instance.m_result);

        if ('\r' == *dataCurr) {
            ++dataCurr;

            instance.m_parserFunc = parseSecondLF;
        } else {
            instance.m_parserFunc = parseHeaderName;
        }

        WS_CHECK_RETURN(ReturnE::ReturnE_Continue);
    }

    HTTPRequestParser::ReturnE HTTPRequestParser::parseFirstLF(
        HTTPRequestParser& instance, const char*& dataCurr
        , const char* const dataEnd) {
        Assert(nullptr != dataCurr);
        Assert(nullptr != dataEnd);
        Assert(dataCurr < dataEnd);
        Assert(instance.m_result);

        if ('\n' != *dataCurr) {
            WS_CHECK_RETURN(ReturnE::ReturnE_Fail);
        }

        ++dataCurr;

        instance.m_parserFunc = parseSecondCROrHeader;

        WS_CHECK_RETURN(ReturnE::ReturnE_Continue);
    }

    HTTPRequestParser::ReturnE HTTPRequestParser::parseHeaderValue(
        HTTPRequestParser& instance, const char*& dataCurr
        , const char* const dataEnd) {
        Assert(nullptr != dataCurr);
        Assert(nullptr != dataEnd);
        Assert(dataCurr < dataEnd);
        Assert(instance.m_result);

        while (dataEnd != dataCurr) {
            Assert(dataEnd > dataCurr);

            if ('\r' == *dataCurr) {
                if (instance.m_result->m_headerList.back()->item().m_value.empty()) {
                    WS_CHECK_RETURN(ReturnE::ReturnE_Fail);
                } else {
                    if (instance.m_result->m_headerList.back()->item().m_name
                        .ciCompare("Connection")) {
                        if (instance.m_result->m_headerList.back()->item().m_value
                            .ciCompare("keep-alive")) {
                            instance.m_result->m_keepAlive = true;
                        } else {
                            instance.m_result->m_keepAlive = false;
                        }
                    }

                    instance.m_parserFunc = parseFirstLF;
                    ++dataCurr;

                    WS_CHECK_RETURN(ReturnE::ReturnE_Continue);
                }
            } else {
                instance.m_result->m_headerList.back()->item().m_value += *dataCurr;
            }

            ++dataCurr;
        }

        WS_CHECK_RETURN(ReturnE::ReturnE_Continue);
    }

    HTTPRequestParser::ReturnE HTTPRequestParser::skipSpacesBeforeHeaderValue(
        HTTPRequestParser& instance, const char*& dataCurr
        , const char* const dataEnd) {
        Assert(nullptr != dataCurr);
        Assert(nullptr != dataEnd);
        Assert(dataCurr < dataEnd);
        Assert(instance.m_result);

        while (dataEnd != dataCurr) {
            Assert(dataEnd > dataCurr);

            if (!stl::string::isSpace(*dataCurr)) {
                instance.m_parserFunc = parseHeaderValue;
                WS_CHECK_RETURN(ReturnE::ReturnE_Continue);
            }

            ++dataCurr;
        }

        WS_CHECK_RETURN(ReturnE::ReturnE_Continue);
    }

    HTTPRequestParser::ReturnE HTTPRequestParser::parseDoubleDotAfterHeaderName(
        HTTPRequestParser& instance, const char*& dataCurr
        , const char* const dataEnd) {
        Assert(nullptr != dataCurr);
        Assert(nullptr != dataEnd);
        Assert(dataCurr < dataEnd);
        Assert(instance.m_result);

        if (':' != *dataCurr) {
            WS_CHECK_RETURN(ReturnE::ReturnE_Fail);
        }

        ++dataCurr;

        WS_CHECK_RETURN(ReturnE::ReturnE_Continue);
    }

    HTTPRequestParser::ReturnE HTTPRequestParser::parseHeaderName(
        HTTPRequestParser& instance, const char*& dataCurr
        , const char* const dataEnd) {
        Assert(nullptr != dataCurr);
        Assert(nullptr != dataEnd);
        Assert(dataCurr < dataEnd);
        Assert(instance.m_result);

        while (dataEnd != dataCurr) {
            Assert(dataEnd > dataCurr);

            if (!instance.m_stringInProgress) {
                instance.m_result->m_headerList.emplaceBack();
                instance.m_stringInProgress = true;
            }

            if (':' == *dataCurr) {
                if (instance.m_result->m_headerList.back()->item().m_name.empty()) {
                    WS_CHECK_RETURN(ReturnE::ReturnE_Fail);
                }

                instance.m_stringInProgress = false;
                instance.m_parserFunc = skipSpacesBeforeHeaderValue;
                ++dataCurr;

                WS_CHECK_RETURN(ReturnE::ReturnE_Continue);
            } else if (stl::string::isSpace(*dataCurr)) {
                if (instance.m_result->m_headerList.back()->item().m_name.empty()) {
                    WS_CHECK_RETURN(ReturnE::ReturnE_Fail);
                }

                instance.m_stringInProgress = false;
                instance.m_parserFunc = parseDoubleDotAfterHeaderName;
                ++dataCurr;

                WS_CHECK_RETURN(ReturnE::ReturnE_Continue);
            } else if (stl::string::isAlpha(*dataCurr) || ('-' == *dataCurr)) {
                instance.m_result->m_headerList.back()->item().m_name += *dataCurr;
            } else {
                WS_CHECK_RETURN(ReturnE::ReturnE_Fail);
            }

            ++dataCurr;
        }

        WS_CHECK_RETURN(ReturnE::ReturnE_Continue);
    }

    HTTPRequestParser::ReturnE HTTPRequestParser::parseRequestLineVerMinor(
        HTTPRequestParser& instance, const char*& dataCurr
        , const char* const dataEnd) {
        Assert(nullptr != dataCurr);
        Assert(nullptr != dataEnd);
        Assert(dataCurr < dataEnd);
        Assert(instance.m_result);

        while (dataEnd != dataCurr) {
            if ('\r' == *dataCurr) {
                if (instance.m_result->m_versionMinor.empty()) {
                    WS_CHECK_RETURN(ReturnE::ReturnE_Fail);
                } else {
                    instance.m_parserFunc = parseFirstLF;

                    ++dataCurr;

                    WS_CHECK_RETURN(ReturnE::ReturnE_Continue);
                }
            } else if (stl::string::isDigit(*dataCurr)) {
                instance.m_result->m_versionMinor += *dataCurr;
            } else {
                WS_CHECK_RETURN(ReturnE::ReturnE_Fail);
            }

            ++dataCurr;
        }

        WS_CHECK_RETURN(ReturnE::ReturnE_Continue);
    }

    HTTPRequestParser::ReturnE HTTPRequestParser::parseRequestLineVerMajor(
        HTTPRequestParser& instance, const char*& dataCurr
        , const char* const dataEnd) {
        Assert(nullptr != dataCurr);
        Assert(nullptr != dataEnd);
        Assert(dataCurr < dataEnd);
        Assert(instance.m_result);

        while (dataEnd != dataCurr) {
            if ('.' == *dataCurr) {
                if (instance.m_result->m_versionMajor.empty()) {
                    WS_CHECK_RETURN(ReturnE::ReturnE_Fail);
                } else {
                    instance.m_parserFunc = parseRequestLineVerMinor;

                    ++dataCurr;

                    WS_CHECK_RETURN(ReturnE::ReturnE_Continue);
                }
            } else if (stl::string::isDigit(*dataCurr)) {
                instance.m_result->m_versionMajor += *dataCurr;
            } else {
                WS_CHECK_RETURN(ReturnE::ReturnE_Fail);
            }

            ++dataCurr;
        }

        WS_CHECK_RETURN(ReturnE::ReturnE_Continue);
    }

    HTTPRequestParser::ReturnE HTTPRequestParser::parseRequestLineSlashAfterHTTP(
        HTTPRequestParser& instance, const char*& dataCurr
        , const char* const dataEnd) {
        Assert(nullptr != dataCurr);
        Assert(nullptr != dataEnd);
        Assert(dataCurr < dataEnd);
        Assert(instance.m_result);

        if ('/' != *dataCurr) {
            WS_CHECK_RETURN(ReturnE::ReturnE_Fail);
        }

        ++dataCurr;
        instance.m_parserFunc = parseRequestLineVerMajor;

        WS_CHECK_RETURN(ReturnE::ReturnE_Continue);
    }

    HTTPRequestParser::ReturnE HTTPRequestParser::parseRequestLineHTTP(
        HTTPRequestParser& instance, const char*& dataCurr
        , const char* const dataEnd) {
        Assert(nullptr != dataCurr);
        Assert(nullptr != dataEnd);
        Assert(dataCurr < dataEnd);
        Assert(instance.m_result);

        if ('P' != *dataCurr) {
            WS_CHECK_RETURN(ReturnE::ReturnE_Fail);
        }

        ++dataCurr;
        instance.m_parserFunc = parseRequestLineSlashAfterHTTP;

        WS_CHECK_RETURN(ReturnE::ReturnE_Continue);
    }

    HTTPRequestParser::ReturnE HTTPRequestParser::parseRequestLineHTT(
        HTTPRequestParser& instance, const char*& dataCurr
        , const char* const dataEnd) {
        Assert(nullptr != dataCurr);
        Assert(nullptr != dataEnd);
        Assert(dataCurr < dataEnd);
        Assert(instance.m_result);

        if ('T' != *dataCurr) {
            WS_CHECK_RETURN(ReturnE::ReturnE_Fail);
        }

        ++dataCurr;
        instance.m_parserFunc = parseRequestLineHTTP;

        WS_CHECK_RETURN(ReturnE::ReturnE_Continue);
    }

    HTTPRequestParser::ReturnE HTTPRequestParser::parseRequestLineHT(
        HTTPRequestParser& instance, const char*& dataCurr
        , const char* const dataEnd) {
        Assert(nullptr != dataCurr);
        Assert(nullptr != dataEnd);
        Assert(dataCurr < dataEnd);
        Assert(instance.m_result);

        if ('T' != *dataCurr) {
            WS_CHECK_RETURN(ReturnE::ReturnE_Fail);
        }

        ++dataCurr;
        instance.m_parserFunc = parseRequestLineHTT;

        WS_CHECK_RETURN(ReturnE::ReturnE_Continue);
    }

    HTTPRequestParser::ReturnE HTTPRequestParser::parseRequestLineH(
        HTTPRequestParser& instance, const char*& dataCurr
        , const char* const dataEnd) {
        Assert(nullptr != dataCurr);
        Assert(nullptr != dataEnd);
        Assert(dataCurr < dataEnd);
        Assert(instance.m_result);

        if ('H' != *dataCurr) {
            WS_CHECK_RETURN(ReturnE::ReturnE_Fail);
        }

        ++dataCurr;
        instance.m_parserFunc = parseRequestLineHT;

        WS_CHECK_RETURN(ReturnE::ReturnE_Continue);
    }

    HTTPRequestParser::ReturnE HTTPRequestParser::parseRequestLineWSBeforeH(
        HTTPRequestParser& instance, const char*& dataCurr
        , const char* const dataEnd) {
        Assert(nullptr != dataCurr);
        Assert(nullptr != dataEnd);
        Assert(dataCurr < dataEnd);
        Assert(instance.m_result);

        if (' ' != *dataCurr) {
            instance.m_parserFunc = parseRequestLineH;
        } else {
            ++dataCurr;
        }

        WS_CHECK_RETURN(ReturnE::ReturnE_Continue);
    }

    HTTPRequestParser::ReturnE HTTPRequestParser::parseRequestLineURI(
        HTTPRequestParser& instance, const char*& dataCurr
        , const char* const dataEnd) {
        Assert(nullptr != dataCurr);
        Assert(nullptr != dataEnd);
        Assert(dataCurr < dataEnd);
        Assert(instance.m_result);

        while (dataEnd != dataCurr) {
            Assert(dataCurr < dataEnd);

            if (' ' == *dataCurr) {
                if (instance.m_result->m_uri.empty()) {
                    WS_CHECK_RETURN(ReturnE::ReturnE_Fail);
                } else {
                    instance.m_parserFunc = parseRequestLineWSBeforeH;

                    ++dataCurr;

                    WS_CHECK_RETURN(ReturnE::ReturnE_Continue);
                }
            } else if (HTTPURL::isAllowedChar(*dataCurr)) {
                instance.m_result->m_uri += *dataCurr;
            } else {
                WS_CHECK_RETURN(ReturnE::ReturnE_Fail);
            }

            ++dataCurr;
        }

        WS_CHECK_RETURN(ReturnE::ReturnE_Continue);
    }

    HTTPRequestParser::ReturnE HTTPRequestParser::parseRequestLineWSBeforeURI(
        HTTPRequestParser& instance, const char*& dataCurr
        , const char* const dataEnd) {
        Assert(nullptr != dataCurr);
        Assert(nullptr != dataEnd);
        Assert(dataCurr < dataEnd);
        Assert(instance.m_result);

        if (' ' != *dataCurr) {
            instance.m_parserFunc = parseRequestLineURI;
        } else {
            ++dataCurr;
        }

        WS_CHECK_RETURN(ReturnE::ReturnE_Continue);
    }

    HTTPRequestParser::ReturnE HTTPRequestParser::parseRequestLineMethod(
        HTTPRequestParser& instance, const char*& dataCurr
        , const char* const dataEnd) {
        Assert(nullptr != dataCurr);
        Assert(nullptr != dataEnd);
        Assert(dataCurr < dataEnd);
        Assert(instance.m_result);

        instance.m_result = stl::makeUnique<HTTPRequest>();
        if (!instance.m_result) {
            WS_CHECK_RETURN(ReturnE::ReturnE_Fail);
        }

        while (dataEnd != dataCurr) {
            if (' ' == *dataCurr) {
                if (instance.m_result->m_method.empty()) {
                    WS_CHECK_RETURN(ReturnE::ReturnE_Fail);
                } else {
                    instance.m_parserFunc = parseRequestLineWSBeforeURI;

                    ++dataCurr;

                    WS_CHECK_RETURN(ReturnE::ReturnE_Continue);
                }
            } else if (stl::string::isAlpha(*dataCurr)) {
                instance.m_result->m_method += *dataCurr;
            } else {
                WS_CHECK_RETURN(ReturnE::ReturnE_Fail);
            }

            ++dataCurr;
        }

        WS_CHECK_RETURN(ReturnE::ReturnE_Continue);
    }

    int HTTPResponse::consume(const char* const data
        , const stl::SizeT length) {
        UNUSED_PARAM(data);
        UNUSED_PARAM(length);

        return -1;
    }
}