#include "stl/compile_info.hpp"
#include "stl/stl_config.hpp"

#if defined STL_ENABLE_NETWORKING && (STL_ENABLE_NETWORKING == 1)

#include "stl/network.hpp"

#include "log/logging.hpp"

#include "stl/assert.hpp"
#include "stl/compiler.hpp"
#include "stl/endian.hpp"
#include "stl/iostream.hpp"
#include "stl/limits.hpp"
#include "stl/memory.hpp"
#include "stl/new.hpp"
#include "stl/stddef.hpp"
#include "stl/stdint.hpp"
#include "stl/string.hpp"
#include "stl/thread.hpp"
#include "stl/timer.hpp"

namespace stl {
template<>
stream::OStreamBase&
stream::OStreamBase::operator<<(const ::network::NetworkAddress& address)
{
  switch (address.getIPVer()) {
    case ::network::IPVerE::IPv4:
    case ::network::IPVerE::IPv6:
      *this << address.getIPStr() << ':' << address.getPort();
      break;
    default:
      *this << "<unknown address version>";
      break;
  }

  return *this;
}

template<>
stream::OStreamBase&
stream::OStreamBase::operator<<(const ::network::IPv4& ip)
{
  *this << ip.ipStr();

  return *this;
}

template<>
stream::OStreamBase&
stream::OStreamBase::operator<<(const ::network::IPv6& ip)
{
  *this << ip.ipStr();

  return *this;
}

template<>
stream::OStreamBase&
stream::OStreamBase::operator<<(const ::network::IOStatusE& status)
{
  *this << ::network::SocketBase::ioStatusString(status);

  return *this;
}
} // namespace stl

stl::stream::OStreamBase&
operator<<(stl::stream::OStreamBase& stream, const ::network::NetworkAddress& address)
{
  switch (address.getIPVer()) {
    case ::network::IPVerE::IPv4:
    case ::network::IPVerE::IPv6:
      stl::cout << address.getIPStr() << ':' << address.getPort();
      break;
    default:
      stl::cout << "<unknown address version>";
      break;
  }

  return stream;
}

stl::stream::OStreamBase&
operator<<(stl::stream::OStreamBase& stream, const ::network::IPv4& ip)
{
  stl::cout << ip.ipStr();

  return stream;
}

stl::stream::OStreamBase&
operator<<(stl::stream::OStreamBase& stream, const ::network::IPv6& ip)
{
  stl::cout << ip.ipStr();

  return stream;
}

stl::stream::OStreamBase&
operator<<(stl::stream::OStreamBase& stream, const ::network::IOStatusE& status)
{
  stl::cout << ::network::SocketBase::ioStatusString(status);

  return stream;
}

#if (defined STL_ENABLE_LOGGING && (STL_ENABLE_LOGGING == 1))
namespace logging {
template<>
Logger&
Logger::operator<<(const ::network::IOStatusE& status)
{
  stl::cout << ::network::SocketBase::ioStatusString(status);

  return logging::log;
}

template<>
Logger&
Logger::operator<<(const ::network::NetworkAddress& address)
{
  switch (address.getIPVer()) {
    case ::network::IPVerE::IPv4:
    case ::network::IPVerE::IPv6:
      stl::cout << address.getIPStr() << ':' << address.getPort();
      break;
    default:
      stl::cout << "<unknown address version>";
      break;
  }

  return logging::log;
}

template<>
Logger&
Logger::operator<<(const ::network::IPv4& ip)
{
  stl::cout << ip.ipStr();

  return logging::log;
}

template<>
Logger&
Logger::operator<<(const ::network::IPv6& ip)
{
  stl::cout << ip.ipStr();

  return logging::log;
}

template<>
Logger&
Logger::operator<<(const ::network::IPVerE& ver)
{
  switch (ver) {
    case ::network::IPVerE::IPv4:
      stl::cout << "IPv4";
      break;
    case ::network::IPVerE::IPv6:
      stl::cout << "IPv6";
      break;
    default:
      stl::cout << "<unknown IP version>";
      break;
  }

  return logging::log;
}
} // namespace logging

logging::Logger&
operator<<(logging::Logger& stream, const ::network::IOStatusE& status)
{
  stl::cout << ::network::SocketBase::ioStatusString(status);

  return stream;
}

logging::Logger&
operator<<(logging::Logger& stream, const ::network::NetworkAddress& address)
{
  switch (address.getIPVer()) {
    case ::network::IPVerE::IPv4:
    case ::network::IPVerE::IPv6:
      stl::cout << address.getIPStr() << ':' << address.getPort();
      break;
    default:
      stl::cout << "<unknown address version>";
      break;
  }

  return stream;
}

logging::Logger&
operator<<(logging::Logger& stream, const ::network::IPv4& ip)
{
  stl::cout << ip;

  return stream;
}

logging::Logger&
operator<<(logging::Logger& stream, const ::network::IPv6& ip)
{
  stl::cout << ip;

  return stream;
}

logging::Logger&
operator<<(logging::Logger& stream, const ::network::IPVerE& ver)
{
  switch (ver) {
    case ::network::IPVerE::IPv4:
      stl::cout << "IPv4";
      break;
    case ::network::IPVerE::IPv6:
      stl::cout << "IPv6";
      break;
    default:
      stl::cout << "<unknown IP version>";
      break;
  }

  return stream;
}
#endif // (defined STL_ENABLE_LOGGING && (STL_ENABLE_LOGGING == 1))

namespace network {
extern const int _typeLUT[NETWORK_COMTYPE_LUT_SIZE];
extern const int _protoLUT[NETWORK_IPPROTO_LUT_SIZE];
extern const int _dirLUT[NETWORK_SHDIR_LUT_SIZE];
extern stl::UInt16T _afLUT[NETWORK_AF_LUT_SIZE];

IPv4::IPv4()
{
  reset();
}

IPv4::IPv4(const IPv4& other)
  : m_netAddr(other.m_netAddr)
  , m_ipStrLen(other.m_ipStrLen)
{
  stl::memory::copy(other.m_ipStr, other.m_ipStr + sizeof(m_ipStr), m_ipStr);
}

IPv4&
IPv4::operator=(const IPv4& other)
{
  m_netAddr = other.m_netAddr;
  m_ipStrLen = other.m_ipStrLen;
  stl::memory::copy(other.m_ipStr, other.m_ipStr + sizeof(m_ipStr), m_ipStr);

  return *this;
}

IPv4::IPv4(IPv4&& other) noexcept
  : m_netAddr(other.m_netAddr)
  , m_ipStrLen(other.m_ipStrLen)
{
  stl::memory::copy(other.m_ipStr, other.m_ipStr + sizeof(m_ipStr), m_ipStr);
}

IPv4&
IPv4::operator=(IPv4&& other) noexcept
{
  m_netAddr = other.m_netAddr;
  m_ipStrLen = other.m_ipStrLen;
  stl::memory::copy(other.m_ipStr, other.m_ipStr + sizeof(m_ipStr), m_ipStr);

  return *this;
}

IPv4::IPv4(stl::UInt32T ip)
{
  set(ip);
}

stl::UInt32T
IPv4::netAddr() const
{
  return m_netAddr;
}

const char*
IPv4::ipStr() const
{
  return m_ipStr;
}

stl::SizeT
IPv4::ipStrLen() const
{
  return m_ipStrLen;
}

bool
IPv4::parse(const char* ipStr)
{
  if (!parseInternal(ipStr)) {
    reset();
    return false;
  }

  return true;
}

bool
IPv4::parse(const stl::String& ipStr)
{
  return parse(ipStr.c_str());
}

void
IPv4::set(const stl::UInt32T netAddr)
{
  stl::SizeT offset = 0U;

  stl::String octet[4] = {
    stl::string::toString(
      *(reinterpret_cast<const stl::UInt8T*>(&netAddr) + 0)),
    stl::string::toString(
      *(reinterpret_cast<const stl::UInt8T*>(&netAddr) + 1)),
    stl::string::toString(
      *(reinterpret_cast<const stl::UInt8T*>(&netAddr) + 2)),
    stl::string::toString(*(reinterpret_cast<const stl::UInt8T*>(&netAddr) + 3))
  };

  stl::memory::copy(
    octet[0].c_str(), octet[0].c_str() + octet[0].length(), m_ipStr);
  offset += octet[0].length();
  m_ipStr[offset++] = '.';
  stl::memory::copy(
    octet[1].c_str(), octet[1].c_str() + octet[1].length(), m_ipStr + offset);
  offset += octet[1].length();
  m_ipStr[offset++] = '.';
  stl::memory::copy(
    octet[2].c_str(), octet[2].c_str() + octet[2].length(), m_ipStr + offset);
  offset += octet[2].length();
  m_ipStr[offset++] = '.';
  stl::memory::copy(
    octet[3].c_str(), octet[3].c_str() + octet[3].length(), m_ipStr + offset);
  offset += octet[3].length();

  m_ipStr[offset] = '\0';

  m_ipStrLen = offset;
  m_netAddr = netAddr;
}

bool
IPv4::parseInternal(const char* ipStr)
{
  Assert(nullptr != ipStr);

  const char* const ipStrBegin = ipStr;
  const char* numberBegin = ipStr;
  stl::SizeT octetOffset = 0U;
  stl::SizeT numberLength = 0U;
  stl::UInt8T octet;

  m_ipStrLen = 0U;

  bool conversionSuccessful = false;

  while (true) {
    if ('.' == *ipStr) {
      if (octetOffset > 3U) {
        Assert(false);
        return false;
      }

      const stl::SizeT length = static_cast<stl::SizeT>(ipStr - numberBegin);

      if ((0U == length) || (length > 3U)) {
        Assert(false);
        return false;
      }

      octet = stl::string::stringTo<stl::UInt8T>(
        numberBegin, length, conversionSuccessful, numberLength);
      if (!conversionSuccessful) {
        Assert(false);
        return false;
      }

      Assert(numberLength == length);

      *(reinterpret_cast<stl::UInt8T*>(&m_netAddr) + (octetOffset++)) =
        static_cast<stl::UInt8T>(octet);

      numberBegin = ++ipStr;
    } else if ('\0' == *ipStr) {
      if (octetOffset != 3U) {
        Assert(false);
        return false;
      }

      const stl::SizeT length = static_cast<stl::SizeT>(ipStr - numberBegin);

      if ((0U == length) || (length > 3U)) {
        Assert(false);
        return false;
      }

      octet = stl::string::stringTo<stl::UInt8T>(
        numberBegin, length, conversionSuccessful, numberLength);
      if (!conversionSuccessful) {
        Assert(false);
        return false;
      }

      Assert(numberLength == length);

      *(reinterpret_cast<stl::UInt8T*>(&m_netAddr) + octetOffset) =
        static_cast<stl::UInt8T>(octet);

      stl::memory::copy(ipStrBegin, ipStr, m_ipStr);

      m_ipStrLen = static_cast<stl::SizeT>(ipStr - ipStrBegin);
      m_ipStr[m_ipStrLen] = '\0';

      return true;
    } else {
      ++ipStr;
    }
  }

  Assert(false);

  return false;
}

void
IPv4::reset()
{
  m_netAddr = 0U;
  m_ipStr[0] = '0';
  m_ipStr[1] = '.';
  m_ipStr[2] = '0';
  m_ipStr[3] = '.';
  m_ipStr[4] = '0';
  m_ipStr[5] = '.';
  m_ipStr[6] = '0';
  m_ipStr[7] = '\0';
  m_ipStrLen = 7U;
}

IPv6::IPv6()
{
  reset();
}

IPv6::IPv6(const IPv6& other)
  : m_netAddr(other.m_netAddr)
  , m_ipStrLen(other.m_ipStrLen)
{
  stl::memory::copy(other.m_ipStr, other.m_ipStr + sizeof(m_ipStr), m_ipStr);
}

IPv6&
IPv6::operator=(const IPv6& other)
{
  m_netAddr = other.m_netAddr;
  m_ipStrLen = other.m_ipStrLen;
  stl::memory::copy(other.m_ipStr, other.m_ipStr + sizeof(m_ipStr), m_ipStr);

  return *this;
}

IPv6::IPv6(IPv6&& other) noexcept
  : m_netAddr(other.m_netAddr)
  , m_ipStrLen(other.m_ipStrLen)
{
  stl::memory::copy(other.m_ipStr, other.m_ipStr + sizeof(m_ipStr), m_ipStr);
}

IPv6&
IPv6::operator=(IPv6&& other) noexcept
{
  m_netAddr = other.m_netAddr;
  m_ipStrLen = other.m_ipStrLen;
  stl::memory::copy(other.m_ipStr, other.m_ipStr + sizeof(m_ipStr), m_ipStr);

  return *this;
}

IPv6::IPv6(const stl::UInt8T* const ip)
{
  set(ip);
}

const stl::UInt8T*
IPv6::netAddr() const
{
  return nullptr;
}

const char*
IPv6::ipStr() const
{
  return m_ipStr;
}

stl::SizeT
IPv6::ipStrLen() const
{
  return m_ipStrLen;
}

bool
IPv6::parse(const char* ipStr)
{
  UNUSED_PARAM(ipStr);
  return false;
}

void
IPv6::set(const stl::UInt8T* const netAddr)
{
  UNUSED_PARAM(netAddr);
}

/*
    The following are examples of valid IPv6 (normal) addresses:
    2001:db8:3333:4444:5555:6666:7777:8888
    2001:db8:3333:4444:CCCC:DDDD:EEEE:FFFF
    :: (implies all 8 segments are zero)
    2001:db8:: (implies that the last six segments are zero)
    ::1234:5678 (implies that the first six segments are zero)
    2001:db8::1234:5678 (implies that the middle four segments are zero)
    2001:0db8:0001:0000:0000:0ab9:C0A8:0102 (This can be compressed to
   eliminate leading zeros, as follows: 2001:db8:1::ab9:C0A8:102 )

    The following are examples of valid IPv6 (dual) addresses:
    2001:db8:3333:4444:5555:6666:1.2.3.4
    ::11.22.33.44 (implies all six IPv6 segments are zero)
    2001:db8::123.123.123.123 (implies that the last four IPv6 segments are
   zero)
    ::1234:5678:91.123.4.56 (implies that the first four IPv6 segments are
   zero)
    ::1234:5678:1.2.3.4 (implies that the first four IPv6 segments are zero)
    2001:db8::1234:5678:5.6.7.8 (implies that the middle two IPv6 segments
   are zero)

*/

void
IPv6::reset()
{
  m_netAddr = 0U;
  m_ipStr[0] = ':';
  m_ipStr[1] = ':';
  m_ipStr[2] = '0';
  m_ipStr[3] = '\0';
  m_ipStrLen = 3U;
}

// NetworkAddress

NetworkAddress::NetworkAddress()
  : m_sockAddr()
  , m_valid(true) // 0.0.0.0 = OK
{
}

NetworkAddress::NetworkAddress(const char* const ipStr, const PortT port)
{
  stl::memory::fill<stl::UInt8T>(
    0x00U, sizeof(m_sockAddr), reinterpret_cast<stl::UInt8T*>(&m_sockAddr));
  static_cast<void>(set(ipStr, port));
}

NetworkAddress::NetworkAddress(NetworkAddress&& other) noexcept
{
  stl::memory::copy(
    other.m_sockAddr, other.m_sockAddr + sizeof(other.m_sockAddr), m_sockAddr);
}

NetworkAddress&
NetworkAddress::operator=(NetworkAddress&& other) noexcept
{
  stl::memory::copy(
    other.m_sockAddr, other.m_sockAddr + sizeof(other.m_sockAddr), m_sockAddr);

  return *this;
}

NetworkAddress::NetworkAddress(const NetworkAddress& other)
{
  stl::memory::copy(
    other.m_sockAddr, other.m_sockAddr + sizeof(other.m_sockAddr), m_sockAddr);
}

NetworkAddress&
NetworkAddress::operator=(const NetworkAddress& other)
{
  stl::memory::copy(
    other.m_sockAddr, other.m_sockAddr + sizeof(other.m_sockAddr), m_sockAddr);

  return *this;
}

const stl::UInt8T*
NetworkAddress::sockAddr() const
{
  return m_sockAddr;
}

stl::UInt8T*
NetworkAddress::sockAddr()
{
  return m_sockAddr;
}

bool
NetworkAddress::set(const char* ipStr, const stl::UInt16T port)
{
  IPv4 ipV4;

  if (!ipV4.parse(ipStr)) {
    logging::log << logging::LvlE::Warning << "Could not parse IPv4 address \""
                 << ipStr << "\"." << stl::endl;
    // TODO: Try IPv6
    return m_valid = false;
  }

  return set(ipV4, port);
}

bool
NetworkAddress::set(const IPv4& ipV4, const stl::UInt16T port)
{
  *reinterpret_cast<stl::UInt16T*>(&m_sockAddr[0]) =
    _afLUT[static_cast<stl::SizeT>(IPVerE::IPv4)];
  *reinterpret_cast<stl::UInt16T*>(&m_sockAddr[2]) =
    stl::Endian::hostSync(stl::EndianE::EndianE_Big, port);
  *reinterpret_cast<stl::UInt32T*>(&m_sockAddr[4]) = ipV4.netAddr();

  stl::memory::fill(static_cast<stl::UInt8T>(0U), 20U, &m_sockAddr[8]);

  return m_valid = true;
}

bool
NetworkAddress::set(const IPv6& ipV6, const stl::UInt16T port)
{
  UNUSED_PARAM(ipV6);
  UNUSED_PARAM(port);
  // *reinterpret_cast<stl::UInt32T*>(&m_sockAddr[4]) = flowInfo;
  // stl::memory::copy(netAddr, netAddr + 16, &m_sockAddr[8]);
  // *reinterpret_cast<stl::UInt32T*>(&m_sockAddr[24]) = scopeId;
  // TODO

  return m_valid = false;
}

IPVerE
NetworkAddress::getIPVer() const
{
  const stl::UInt16T addressFamily =
    *reinterpret_cast<const stl::UInt16T*>(&m_sockAddr[0]);
  if (_afLUT[0] == addressFamily) {
    return IPVerE::IPv4;
  } else if (_afLUT[1] == addressFamily) {
    return IPVerE::IPv6;
  }

  return IPVerE::_Count;
}

stl::SizeT
NetworkAddress::size() const
{
  switch (getIPVer()) {
    case IPVerE::IPv4:
      return 16U;
    case IPVerE::IPv6:
      return 28U;
    default:
      return 0U;
  }
}

void
NetworkAddress::getAsIPv4(IPv4& ipV4) const
{
  ipV4.set(getIPv4NetworkAddr());
}

IPv4
NetworkAddress::getAsIPv4() const
{
  return IPv4(*reinterpret_cast<const stl::UInt32T*>(&m_sockAddr[4]));
}

void
NetworkAddress::getAsIPv6(IPv6& ipV6) const
{
  ipV6.set(getIPv6NetworkAddr());
}

IPv6
NetworkAddress::getAsIPv6() const
{
  return IPv6(reinterpret_cast<const stl::UInt8T*>(
    &m_sockAddr[4])); // TODO: Incorrect offset
}

stl::UInt32T
NetworkAddress::getIPv4NetworkAddr() const
{
  return *reinterpret_cast<const stl::UInt32T*>(&m_sockAddr[4]);
}

const stl::UInt8T*
NetworkAddress::getIPv6NetworkAddr() const
{
  return reinterpret_cast<const stl::UInt8T*>(&m_sockAddr[8]);
}

stl::String
NetworkAddress::getIPStr() const
{
  switch (getIPVer()) {
    case IPVerE::IPv4:
      return getAsIPv4().ipStr();
    case IPVerE::IPv6:
      return getAsIPv6().ipStr();
    default:
      break;
  }

  return "";
}

PortT
NetworkAddress::getPort() const
{
  switch (getIPVer()) {
    case IPVerE::IPv4:
      return stl::Endian::hostSync(
        stl::EndianE::EndianE_Big,
        *reinterpret_cast<const PortT*>(&m_sockAddr[2]));
    case IPVerE::IPv6:
      return stl::Endian::hostSync(stl::EndianE::EndianE_Big,
                                   *reinterpret_cast<const PortT*>(
                                     &m_sockAddr[2])); // TODO: Correct offset?
    default:
      break;
  }

  return 0U;
}

stl::SizeT
NetworkAddress::sizeMax()
{
  return sizeof(m_sockAddr);
}

// SocketBase

const SocketBase::ActiveDirT SocketBase::m_dirBitLUT[3] = { 0x01U,
                                                            0x02U,
                                                            0x03U }; // Masks

const char*
SocketBase::ioStatusString(const IOStatusE status)
{
  static const char* statusString[] = {
#ifdef ENTRY
#undef ENTRY
#endif

#define ENTRY(errorName) #errorName,
    NETWORK_IO_STATUS_LIST
#undef ENTRY
  };

  return statusString[static_cast<stl::SizeT>(status)];
}

IOStatusE
SocketBase::shutdown(const SockDirE dir)
{
  const stl::SizeT index = static_cast<stl::SizeT>(dir);
  if (index >= sizeof(m_dirBitLUT)) {
    return IOStatusE::IOStatusE_InvalidParam;
  }

  m_activeDirections &= ~m_dirBitLUT[index];

  return shutdownInternal(dir);
}

IOStatusE
SocketBase::close()
{
  const IOStatusE status = closeInternal();

  return status;
}

IOStatusE
SocketBase::bind(const NetworkAddress& networkAddress)
{
  m_boundAddress = networkAddress;

  return bindInternal();
}

SysErrorT
SocketBase::getLastSysError() const
{
  return m_lastSysError;
}

IOStatusE
SocketBase::select()
{
  return IOStatusE::IOStatusE_UnsupportedFunction;
} // TODO: Implement

IOStatusE
SocketBase::poll()
{
  return IOStatusE::IOStatusE_UnsupportedFunction;
} // TODO: Emulate

SocketOpenerBase::SocketOpenerBase(SocketOpenerBase&& other) noexcept
  : SocketBase(stl::move(static_cast<SocketBase&>(other)))
{   
}

SocketOpenerBase&
SocketOpenerBase::operator=(SocketOpenerBase&& other) noexcept
{
  SocketBase::operator=(stl::move(static_cast<SocketBase&>(other)));
  return *this;
}

SocketOpenerBase::SocketOpenerBase(SocketContainerT&& socketContainer)
  : SocketBase(stl::move(socketContainer))
{
}

SocketOpenerBase&
SocketOpenerBase::operator=(SocketContainerT&& socketContainer)
{
  SocketBase::operator=(stl::move(socketContainer)); 
  return *this;
}

IOStatusE
SocketOpenerBase::open(const IPVerE ipVer,
                       const ComTypeE comType,
                       const IPProtoE proto)
{
  if (isOpen()) {
    static_cast<void>(close());
  }

  m_ipVer = ipVer;
  m_comType = comType;
  m_ipProto = proto;
  m_activeDirections = m_dirBitLUT[static_cast<stl::SizeT>(SockDirE::Both)];

  return openInternal();
}

ClientSocketBase::ClientSocketBase(SocketContainerT&& socketContainer)
  : SocketBase(stl::move(socketContainer))
{ 
}

ClientSocketBase::~ClientSocketBase() {}

ClientSocketBase::ClientSocketBase(ClientSocketBase&& other) noexcept
{
  SocketBase::operator=(stl::move(static_cast<SocketBase&>(other))); // TODO: Why cannot use move constructor? Basic ctor is called instead 
}

ClientSocketBase&
ClientSocketBase::operator=(ClientSocketBase&& other) noexcept
{
  SocketBase::operator=(stl::move(static_cast<SocketBase&>(other)));
  return *this;
}

IOStatusE
ClientSocketBase::postConnectAction()
{
  return IOStatusE::IOStatusE_OK;
}

ClientSocket::ClientSocket(ClientSocket&& other) noexcept
  : ClientSocketBase(stl::move(static_cast<ClientSocketBase&>(other)))
  , m_info(stl::move(other.m_info))
  // , SocketOpenerBase(stl::move(static_cast<SocketOpenerBase&>(other))) // TODO: Troublemaker?
{
  
}

ClientSocket&
ClientSocket::operator=(ClientSocket&& other) noexcept
{
  SocketBase::operator=(stl::move(static_cast<SocketBase&>(other)));
  m_info = stl::move(other.m_info);
  CALL_REPORT();
  return *this;
}

Session::Session(stl::UniquePtr<ClientSocketBase>&& clientSocket,
                 NetworkAddress&& clientAddress)
  : m_clientSocket(stl::move(clientSocket))
  , m_clientAddress(stl::move(clientAddress))
  , m_timeout(stl::time::Millis(15000))
{
  Assert(m_clientSocket);
}

Session::Session(Session&& other) noexcept
  : m_clientSocket(stl::move(other.m_clientSocket))
  , m_clientAddress(stl::move(other.m_clientAddress))
  , m_timeout(stl::move(other.m_timeout))
{
}

Session&
Session::operator=(Session&& other) noexcept
{
  m_clientSocket = stl::move(other.m_clientSocket);
  m_clientAddress = stl::move(other.m_clientAddress);
  m_timeout = stl::move(other.m_timeout);

  return *this;
}

stl::timer::Countdown&
Session::timeout()
{
  return m_timeout;
}

ClientSocketBase&
Session::socket()
{
  Assert(m_clientSocket);
  return *m_clientSocket;
}

const NetworkAddress&
Session::remoteAddress() const
{
  return m_clientAddress;
}

ServerSocket::~ServerSocket() {}

ServerSocket::ServerSocket(ServerSocket&& other) noexcept
  : SocketOpenerBase(stl::move(static_cast<SocketOpenerBase&>(other)))
{
  
}

ServerSocket&
ServerSocket::operator=(ServerSocket&& other) noexcept
{
  SocketOpenerBase::operator=(stl::move(static_cast<SocketOpenerBase&>(other)));

  return *this;
}

ServerSocket::ServerSocket(SocketContainerT&& socketContainer)
  : SocketOpenerBase(stl::move(socketContainer))
{
}

ServerSocket&
ServerSocket::operator=(SocketContainerT&& socketContainer)
{
  SocketOpenerBase::operator=(stl::move(socketContainer));

  return *this;
}

IOStatusE
ServerSocket::postAcceptAction(Session& newSocketContainer)
{
  UNUSED_PARAM(newSocketContainer);

  return IOStatusE::IOStatusE_OK;
}
} // namespace network

namespace network {
TCPServer::TCPServer()
  : m_socketPtr(nullptr)
{
}

IOStatusE
TCPServer::begin(const NetworkAddress& bindAddress)
{
  Assert(m_socketPtr);

  logging::log << logging::LvlE::Info
               << "Opening TCP socket: " << bindAddress.getIPVer() << stl::endl;

  ::network::IOStatusE status = m_socketPtr->open(bindAddress.getIPVer(),
                                                  ::network::ComTypeE::Stream,
                                                  ::network::IPProtoE::TCP);

  logging::log << logging::LvlE::Info << "Starting TCP server -> " << stl::endl;

  if (::network::IOStatusE::IOStatusE_OK != status) {
    logging::log << logging::StatE::Fail << stl::endl;
    logging::log << logging::LvlE::Error << "Could not open socket {" << status
                 << "}" << stl::endl;
    return status;
  }

  logging::log << logging::LvlE::Info << "Binding address .. ";

  status = m_socketPtr->bind(bindAddress);

  if (::network::IOStatusE::IOStatusE_OK != status) {
    logging::log << logging::StatE::Fail << stl::endl;
    logging::log << logging::LvlE::Error << "Could not bind address "
                 << bindAddress << " {" << status << "}" << stl::endl;
    return status;
  }

  logging::log << logging::StatE::Success << stl::endl;
  logging::log << logging::LvlE::Info
               << "Opening server socket for listening .. ";

  status = m_socketPtr->listen(255U);

  if (::network::IOStatusE::IOStatusE_OK != status) {
    logging::log << logging::StatE::Fail << stl::endl;
    logging::log << logging::LvlE::Error
                 << "Could not enable listening on socket {" << status << "}"
                 << stl::endl;
    return status;
  }

  logging::log << logging::StatE::Success << stl::endl;
  logging::log << logging::LvlE::Info
               << "Setting server socket as non-blocking .. ";

  status = m_socketPtr->blocking(false);

  if (::network::IOStatusE::IOStatusE_OK != status) {
    logging::log << logging::StatE::Fail << stl::endl;
    logging::log << logging::LvlE::Error
                 << "Could not set socket as non-blocking {" << status << "}"
                 << stl::endl;
    return status;
  }

  logging::log << logging::StatE::Success << stl::endl;
  logging::log << logging::LvlE::Success << "TCP server ready!" << stl::endl;

  return IOStatusE::IOStatusE_OK;
}

IOStatusE
TCPServer::update()
{
  Assert(m_socketPtr);

  Session clientSession;
  IOStatusE status = m_socketPtr->accept(clientSession);

  if (status == ::network::IOStatusE::IOStatusE_OK) {
    LOGEXT << logging::LvlE::Info << "Client " << clientSession.remoteAddress() << " accepted!" << stl::endl;
    
    handleConnection(stl::move(clientSession));
  } else if (status != ::network::IOStatusE::IOStatusE_OpWouldBlock) {
    logging::log << logging::LvlE::Error << "Accept failed {" << status << ":"
                 << m_socketPtr->getLastSysError() << "}" << stl::endl;
    stl::thread::sleepFor(1000);
  }

  return IOStatusE::IOStatusE_OK;
}

IOStatusE
TCPServer::end()
{
  Assert(m_socketPtr);

  return IOStatusE::IOStatusE_OK;
}

void
TCPServer::handleConnection(Session&& session)
{
  Assert(false);
  session.socket().close();
}

void
TCPServer::setEndpoint(ServerSocket& endPoint)
{
  m_socketPtr = &endPoint;
}
} // namespace network

#endif // defined STL_ENABLE_NETWORKING && (STL_ENABLE_NETWORKING == 1)
