#include "stl/compile_info.hpp"
#include "stl/stl_config.hpp"

#if defined STL_ENABLE_NETWORKING && (STL_ENABLE_NETWORKING == 1)

#if defined CI_PLATFORM_MKRWIFI1010 && (CI_PLATFORM_MKRWIFI1010 == 1)

#include "stl/assert.hpp"
#include "stl/compiler.hpp"
#include "stl/endian.hpp"
#include "stl/iostream.hpp"
#include "stl/limits.hpp"
#include "stl/memory.hpp"
#include "stl/network.hpp"
#include "stl/stddef.hpp"
#include "stl/stdint.hpp"
#include "stl/string.hpp"
#include "stl/thread.hpp"
#include "stl/time.hpp"
#include "stl/timer.hpp"

#include "stl/new.hpp"

#include "utility/server_drv.h"
#include "utility/wifi_drv.h"

#include "FreeRTOS.h"
#include "portable.h"
#include "projdefs.h"
#include "task.h"

// #define IOSTATUS_DEBUG

#if defined IOSTATUS_DEBUG
network::IOStatusE
checkAndReturn(const char* file,
               stl::UInt32T line,
               const network::IOStatusE status)
{
  if (network::IOStatusE::IOStatusE_OK != status) {
    stl::cout << "[checkAndReturn] "
              << network::SocketBase::ioStatusString(status) << ": " << file
              << " #" << line << stl::endl;
  }

  return status;
}

#define NETWORK_DEBUG_STATEMENT(stat) stat

#else
#define checkAndReturn(file, line, status) (status)
#define NETWORK_DEBUG_STATEMENT(stat)
#endif

/*
    WL_CONNECTED: assigned when connected to a WiFi network;
    WL_AP_CONNECTED : assigned when a device is connected in Access Point mode;
    WL_AP_LISTENING : assigned when the listening for connections in Access
   Point mode; WL_NO_SHIELD: assigned when no WiFi shield is present;
    WL_IDLE_STATUS: it is a temporary status assigned when WiFi.begin() is
   called and remains active until the number of attempts expires (resulting in
   WL_CONNECT_FAILED) or a connection is established (resulting in
   WL_CONNECTED); WL_NO_SSID_AVAIL: assigned when no SSID are available;
    WL_SCAN_COMPLETED: assigned when the scan networks is completed;
    WL_CONNECT_FAILED: assigned when the connection fails for all the attempts;
    WL_CONNECTION_LOST: assigned when the connection is lost;
    WL_DISCONNECTED: assigned when disconnected from a network;
*/

// TODO: Guard for SSL context and SSL to prevent memory leaks (guard with
// custom destructor template)

#define INVALID_SOCKET static_cast<stl::UInt8T>(NO_SOCKET_AVAIL)
#define AF_INET 2
#define AF_INET6 10

// TODO: Guard for SSL context and SSL to prevent memory leaks (guard with
// custom destructor template)

namespace network {
int _typeLUT[NETWORK_COMTYPE_LUT_SIZE]{ -1, -1, -1 };

int _protoLUT[NETWORK_IPPROTO_LUT_SIZE]{ static_cast<int>(eProtMode::TCP_MODE),
                                         static_cast<int>(eProtMode::UDP_MODE),
                                         -1 };

int _dirLUT[NETWORK_SHDIR_LUT_SIZE]{ -1, -1, -1 };

stl::UInt16T _afLUT[NETWORK_AF_LUT_SIZE]{ AF_INET, AF_INET6 };

using SocketT = stl::UInt8T;

struct SocketContainer
{
  SocketContainer(const bool isServerSocket)
    : m_socket(INVALID_SOCKET)
    , m_isServerSocket(isServerSocket)
    , m_isBlocking(true)
    , m_lastExpectedCode(0)
  {
  }

  SocketContainer(SocketBase::SocketContainerT&& socketContainer)
  {
    m_socket = stl::exchange(
      reinterpret_cast<SocketContainer*>(socketContainer->m_data)->m_socket,
      INVALID_SOCKET);
    m_isServerSocket = false;
    m_isBlocking = true;
    m_lastExpectedCode = 0;
  }

  SocketContainer(SocketContainer&& other) noexcept
  {
    m_socket = stl::exchange(other.m_socket, INVALID_SOCKET);
    m_isServerSocket = other.m_isServerSocket;
    m_isBlocking = other.m_isBlocking;
    m_lastExpectedCode = other.m_lastExpectedCode;
  }

  SocketContainer& operator=(SocketContainer&& other) noexcept
  {
    m_socket = stl::exchange(other.m_socket, INVALID_SOCKET);
    m_isServerSocket = other.m_isServerSocket;
    m_isBlocking = other.m_isBlocking;
    m_lastExpectedCode = other.m_lastExpectedCode;

    return *this;
  }

  SocketContainer(const SocketContainer&) = delete;
  SocketContainer& operator=(const SocketContainer) = delete;

  SocketContainer(const SocketT socket)
    : m_socket(socket)
  {
  }

  static SocketContainer* cast(void* const ptr)
  {
    return reinterpret_cast<SocketContainer*>(ptr);
  }

  static const SocketContainer* cast(const void* const ptr)
  {
    return reinterpret_cast<const SocketContainer*>(ptr);
  }

  SocketT m_socket;
  bool m_isServerSocket;
  bool m_isBlocking;
  int m_lastExpectedCode;
};

IOStatusE
init()
{
#if (defined STL_ENABLE_SSL) && (STL_ENABLE_SSL != 0) &&                       \
  ((defined STL_ENABLE_SSL_SERVER && (STL_ENABLE_SSL_SERVER == 1)) ||          \
   (defined STL_ENABLE_SSL_CLIENT && (STL_ENABLE_SSL_CLIENT == 1)))
  return initSecurityLayer();
#else
  return checkAndReturn(__FILENAME__, __LINE__, IOStatusE::IOStatusE_OK);
#endif // #if (defined STL_ENABLE_SSL) && (STL_ENABLE_SSL != 0)
}

IOStatusE
deinit()
{
  return checkAndReturn(__FILENAME__, __LINE__, IOStatusE::IOStatusE_OK);
}

//////////////////////////////// SocketBaseT ///////////////////////////////

SocketBase::SocketBase()
  : m_socketContainer(nullptr) // Initialized under child classes
  , m_ipVer(IPVerE::_Count)
  , m_comType(ComTypeE::_Count)
  , m_ipProto(IPProtoE::_Count)
  , m_activeDirections(0U)
  , m_lastSysError(0)
  , m_boundAddress()
  , m_blocking(true)
{
}

SocketBase::SocketBase(SocketContainerT&& socketContainer)
  : m_socketContainer(stl::move(socketContainer))
  , m_ipVer(IPVerE::_Count)
  , m_comType(ComTypeE::_Count)
  , m_ipProto(IPProtoE::_Count)
  , m_activeDirections(0U)
  , m_lastSysError(0)
  , m_blocking(true)
{
}

SocketBase::SocketBase(SocketBase&& other) noexcept
{
  
  operator=(stl::move(other));
  
}

SocketBase&
SocketBase::operator=(SocketBase&& other) noexcept
{
  m_socketContainer = stl::move(other.m_socketContainer);
  m_ipVer = other.m_ipVer;
  m_comType = other.m_comType;
  m_ipProto = other.m_ipProto;
  m_activeDirections = other.m_activeDirections;
  m_lastSysError = other.m_lastSysError;
  m_blocking = other.m_blocking;

  return *this;
}

SharedSocketT
SocketBase::fd() const
{
  if (m_socketContainer) {
    return static_cast<SharedSocketT>(
      SocketContainer::cast(m_socketContainer.get())->m_socket);
  } else {
    return static_cast<SharedSocketT>(INVALID_SOCKET);
  }
}

IOStatusE
SocketBase::shutdownInternal(const SockDirE dir)
{
  Assert(m_socketContainer);

  if (!m_socketContainer) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_NotInitialized);
  }

  Assert(INVALID_SOCKET !=
         SocketContainer::cast(m_socketContainer.get())->m_socket);

  return checkAndReturn(__FILENAME__,
                        __LINE__,
                        IOStatusE::IOStatusE_OK); // TODO: How to perform
                                                  // shutdown on this platform?
}

IOStatusE
SocketBase::closeInternal()
{
  Assert(m_socketContainer);

  LOGEXT << logging::LvlE::Warning << "Socket closed" << stl::endl;

  if (!m_socketContainer) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_NotInitialized);
  }

  if (SocketContainer::cast(m_socketContainer.get())->m_isServerSocket) {
    // TODO: What to do?
  } else {
    ServerDrv::stopClient(
      SocketContainer::cast(m_socketContainer.get())->m_socket);
  }

  stl::SizeT delayCounter = 0U;
  while (ServerDrv::getClientState(
           SocketContainer::cast(m_socketContainer.get())->m_socket) !=
           CLOSED &&
         ++delayCounter < 50U) {
    stl::thread::sleepFor(100);
  }

  SocketContainer::cast(m_socketContainer.get())->m_socket = INVALID_SOCKET;
  SocketContainer::cast(m_socketContainer.get())->m_lastExpectedCode = CLOSED;
  // m_lastSysError = 0; // TODO: Workaround

  setSysError();

  return checkAndReturn(__FILENAME__, __LINE__, IOStatusE::IOStatusE_OK);
}

IOStatusE
SocketBase::bindInternal()
{
  Assert(m_socketContainer);

  if (!m_socketContainer) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_NotInitialized);
  }

  Assert(INVALID_SOCKET !=
         SocketContainer::cast(m_socketContainer.get())->m_socket);

  if (m_ipVer != m_boundAddress.getIPVer()) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_InconsistentAddress);
  }

  Assert(m_boundAddress.size() <
         static_cast<stl::SizeT>(stl::Limits<int>::max()));

  return checkAndReturn(__FILENAME__, __LINE__, IOStatusE::IOStatusE_OK);
}

IOStatusE
SocketBase::blocking(const bool blocking)
{
  Assert(m_socketContainer);

  if (!m_socketContainer) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_NotInitialized);
  }

  Assert(INVALID_SOCKET !=
         SocketContainer::cast(m_socketContainer.get())->m_socket);

  SocketContainer::cast(m_socketContainer.get())->m_isBlocking = blocking;

  m_blocking = blocking;

  return checkAndReturn(__FILENAME__, __LINE__, IOStatusE::IOStatusE_OK);
}

static stl::String
wifiStatusStr()
{
  const stl::UInt8T status = WiFiDrv::getConnectionStatus();

  switch (status) {
    case WL_NO_SHIELD:
      return "{255:WL_NO_SHIELD/WL_NO_MODULE}";
    case WL_IDLE_STATUS:
      return "{0:WL_IDLE_STATUS}";
    case WL_NO_SSID_AVAIL:
      return "{1:WL_NO_SSID_AVAIL}";
    case WL_SCAN_COMPLETED:
      return "{2:WL_SCAN_COMPLETED}";
    case WL_CONNECTED:
      return "{3:WL_CONNECTED}";
    case WL_CONNECT_FAILED:
      return "{4:WL_CONNECT_FAILED}";
    case WL_CONNECTION_LOST:
      return "{5:WL_CONNECTION_LOST}";
    case WL_DISCONNECTED:
      return "{6:WL_DISCONNECTED}";
    case WL_AP_LISTENING:
      return "{7:WL_AP_LISTENING}";
    case WL_AP_CONNECTED:
      return "{8:WL_AP_CONNECTED}";
    case WL_AP_FAILED:
      return "{9:WL_AP_FAILED}";
    default: {
      return stl::String("{") + stl::string::toString(status) + '}';
    }
  }
}

void
SocketBase::setSysError()
{
  if (INVALID_SOCKET ==
      SocketContainer::cast(m_socketContainer.get())->m_socket) {
    m_lastSysError = -1;
  } else {
    if (SocketContainer::cast(m_socketContainer.get())->m_isServerSocket) {
      m_lastSysError = static_cast<int>(ServerDrv::getServerState(
        SocketContainer::cast(m_socketContainer.get())->m_socket));
    } else {
      m_lastSysError = static_cast<int>(ServerDrv::getClientState(
        SocketContainer::cast(m_socketContainer.get())->m_socket));
    }
  }

  if (ESTABLISHED != m_lastSysError) {
    NETWORK_DEBUG_STATEMENT(stl::cout << "[LastSocketError] " << m_lastSysError
                                      << ':' << translateSysError()
                                      << " WiFiStatus: " << wifiStatusStr()
                                      << stl::endl);
  }
}

IOStatusE
SocketBase::translateSysError() const
{
  switch (m_lastSysError) {
    case -1:
      return checkAndReturn(
        __FILENAME__, __LINE__, IOStatusE::IOStatusE_NoSocket);
    case CLOSED:
      return checkAndReturn(
        __FILENAME__, __LINE__, IOStatusE::IOStatusE_ConnectionClosed);
    case ESTABLISHED:
    case LISTEN:
      return checkAndReturn(__FILENAME__, __LINE__, IOStatusE::IOStatusE_OK);
    case SYN_SENT:
    case SYN_RCVD:
    case FIN_WAIT_1:
    case FIN_WAIT_2:
    case CLOSE_WAIT:
    case CLOSING:
    case LAST_ACK:
    case TIME_WAIT:
      return checkAndReturn(
        __FILENAME__, __LINE__, IOStatusE::IOStatusE_OpWouldBlock);
    default:
      break;
  }

  return checkAndReturn(__FILENAME__, __LINE__, IOStatusE::IOStatusE_Error);
}

IOStatusE
SocketBase::readReady()
{
  Assert(m_socketContainer);

  if (!m_socketContainer) {
    return IOStatusE::IOStatusE_NotInitialized;
  }

  const uint16_t availData = ServerDrv::availData(
    SocketContainer::cast(m_socketContainer.get())->m_socket);

  setSysError();

  const IOStatusE status = translateSysError();
  switch (status) {
    case IOStatusE::IOStatusE_OK:
      if (availData > 0U) {
        return IOStatusE::IOStatusE_OK;
      } else {
        return IOStatusE::IOStatusE_OpTimeout;
      }
    default:
      break;
  }

  return status;
}

bool
SocketBase::isOpen() const
{
  if (!m_socketContainer) {
    return false;
  }

  return SocketContainer::cast(m_socketContainer.get())->m_socket !=
         INVALID_SOCKET;
}

IOStatusE
SocketBase::openInternal()
{
  Assert(static_cast<stl::SizeT>(m_ipVer) <
         (sizeof(_afLUT) / sizeof(_afLUT[0])));
  Assert(static_cast<stl::SizeT>(m_comType) <
         (sizeof(_typeLUT) / sizeof(_typeLUT[0])));
  Assert(static_cast<stl::SizeT>(m_ipProto) <
         (sizeof(_protoLUT) / sizeof(_protoLUT[0])));

  if (!m_socketContainer) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_NotInitialized);
  }

  if (INVALID_SOCKET !=
      SocketContainer::cast(m_socketContainer.get())->m_socket) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_SocketAlreadyExist);
  }

  SocketContainer::cast(m_socketContainer.get())->m_socket =
    ServerDrv::getSocket();

  // setSysError();

  if (INVALID_SOCKET ==
      SocketContainer::cast(m_socketContainer.get())->m_socket) {
    return checkAndReturn(__FILENAME__, __LINE__, IOStatusE::IOStatusE_Error);
  }

  return checkAndReturn(__FILENAME__, __LINE__, postOpenAction());
}

ClientSocketBase::ClientSocketBase()
  : SocketBase()
{
  m_socketContainer = reinterpret_cast<SocketDataContainer*>(
    stl::memory::MemoryManager<SocketContainer>::allocate(false));
}

stl::SizeT
ClientSocketBase::available()
{
  Assert(m_socketContainer);

  if (!m_socketContainer) {
    return 0U;
  }

  Assert(INVALID_SOCKET !=
         SocketContainer::cast(m_socketContainer.get())->m_socket);

  const uint16_t bytesAvailable = ServerDrv::availData(
    SocketContainer::cast(m_socketContainer.get())->m_socket);

  if (bytesAvailable > stl::Limits<stl::SizeT>::max()) {
    return stl::Limits<stl::SizeT>::max();
  }

  return static_cast<stl::SizeT>(bytesAvailable);
}

IOStatusE
ClientSocketBase::send(const stl::UInt8T* const data,
                       const stl::SizeT size,
                       stl::SizeT& sent)
{
  Assert(nullptr != data);

  Assert(m_socketContainer);
  if (!m_socketContainer) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_NotInitialized);
  }

  Assert(INVALID_SOCKET !=
         SocketContainer::cast(m_socketContainer.get())->m_socket);
  if (INVALID_SOCKET ==
      SocketContainer::cast(m_socketContainer.get())->m_socket) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_InvalidSocket);
  }

  if (size > static_cast<stl::SizeT>(stl::Limits<stl::UInt16T>::max())) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_InvalidParam);
  }

  stl::UInt16T retv = 0U;

  for (stl::SizeT attempt = 0U; attempt < 5U; ++attempt) {
    retv = ServerDrv::sendData(
      SocketContainer::cast(m_socketContainer.get())->m_socket,
      data,
      static_cast<stl::UInt16T>(size));

    if (0U == retv) {
      stl::thread::sleepFor(100);
    } else {
      break;
    }
  }

  if (0U == retv) {
    close();
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_ConnectionClosed);
  }

  SocketContainer::cast(m_socketContainer.get())->m_lastExpectedCode = CLOSED;
  // m_lastSysError = 0; // TODO: Workaround

  setSysError();

  const IOStatusE translatedError = translateSysError();

  if (IOStatusE::IOStatusE_OK == translatedError) {
    sent = static_cast<stl::SizeT>(retv);
  }

  return checkAndReturn(__FILENAME__, __LINE__, translatedError);
}

IOStatusE
ClientSocketBase::receive(stl::UInt8T* const data,
                          const stl::SizeT maxSize,
                          stl::SizeT& received)
{
  Assert(nullptr != data);

  Assert(m_socketContainer);
  if (!m_socketContainer) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_NotInitialized);
  }

  Assert(INVALID_SOCKET !=
         SocketContainer::cast(m_socketContainer.get())->m_socket);
  if (INVALID_SOCKET ==
      SocketContainer::cast(m_socketContainer.get())->m_socket) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_InvalidSocket);
  }

  if (maxSize > static_cast<stl::SizeT>(stl::Limits<stl::UInt16T>::max())) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_InvalidParam);
  }

  stl::UInt16T bufferSize = static_cast<stl::UInt16T>(maxSize);

  if (SocketContainer::cast(m_socketContainer.get())->m_isBlocking) {
    stl::timer::Countdown timeout(
      stl::time::Millis(15000)); // TODO: Configurable wait time

    timeout.reset();

    while (0U == ServerDrv::availData( // TODO: Timeout
                   SocketContainer::cast(m_socketContainer.get())->m_socket)) {
      if (timeout) {
        return checkAndReturn(
          __FILENAME__, __LINE__, IOStatusE::IOStatusE_OpTimeout);
      }

      stl::thread::sleepFor(100);
    }
  }

  NETWORK_DEBUG_STATEMENT(stl::cout << "[network] Data to receive "
                                    << bufferSize << stl::endl);

  const bool dataReceived = ServerDrv::getDataBuf(
    SocketContainer::cast(m_socketContainer.get())->m_socket,
    data,
    &bufferSize); // TODO: Check retv? It depends on if at least one byte
                  // received

  NETWORK_DEBUG_STATEMENT(stl::cout << "[network] Received data " << bufferSize
                                    << stl::endl);
  NETWORK_DEBUG_STATEMENT(stl::cout << "[network] getDataBuf retv "
                                    << dataReceived << stl::endl);

  SocketContainer::cast(m_socketContainer.get())->m_lastExpectedCode = CLOSED;
  // m_lastSysError = 0; // TODO: Workaround

  setSysError();

  const IOStatusE translatedError = translateSysError();

  if (IOStatusE::IOStatusE_OK == translatedError) {
    received = static_cast<stl::SizeT>(bufferSize);
  }

  return checkAndReturn(__FILENAME__, __LINE__, translatedError);
}

IOStatusE
ClientSocketBase::sendTo(const NetworkAddress& networkAddress,
                         const stl::UInt8T* const data,
                         const stl::SizeT size,
                         stl::SizeT& sent)
{
  Assert(nullptr != data);

  Assert(m_socketContainer);
  if (!m_socketContainer) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_NotInitialized);
  }

  Assert(INVALID_SOCKET !=
         SocketContainer::cast(m_socketContainer.get())->m_socket);
  if (INVALID_SOCKET ==
      SocketContainer::cast(m_socketContainer.get())->m_socket) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_InvalidSocket);
  }

  if (size > static_cast<stl::SizeT>(stl::Limits<stl::UInt16T>::max())) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_InvalidParam);
  }

  if (networkAddress.getIPVer() != IPVerE::IPv4) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_UnsupportedParam);
  }

  ServerDrv::startClient(
    networkAddress.getIPv4NetworkAddr(),
    networkAddress.getPort(),
    SocketContainer::cast(m_socketContainer.get())->m_socket,
    eProtMode::UDP_MODE);

  static_cast<void>(ServerDrv::insertDataBuf(
    SocketContainer::cast(m_socketContainer.get())->m_socket,
    data,
    static_cast<stl::UInt16T>(size))); // False returned in every case TODO:
                                       // Edit ServerDrv source?

  static_cast<void>(
    ServerDrv::sendUdpData(SocketContainer::cast(m_socketContainer.get())
                             ->m_socket)); // False returned in every case
                                           // TODO: Edit ServerDrv source?

  SocketContainer::cast(m_socketContainer.get())->m_lastExpectedCode = CLOSED;
  // m_lastSysError = 0; // TODO: Workaround

  setSysError();

  const IOStatusE translatedError = translateSysError();

  if (IOStatusE::IOStatusE_OK == translatedError) {
    sent = static_cast<stl::SizeT>(size);
  }

  return checkAndReturn(__FILENAME__, __LINE__, translatedError);
}

IOStatusE
ClientSocketBase::receiveFrom(stl::UInt8T* const data,
                              const stl::SizeT maxSize,
                              NetworkAddress& networkAddress,
                              stl::SizeT& received)
{
  Assert(nullptr != data);

  Assert(m_socketContainer);
  if (!m_socketContainer) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_NotInitialized);
  }

  Assert(INVALID_SOCKET !=
         SocketContainer::cast(m_socketContainer.get())->m_socket);
  if (INVALID_SOCKET ==
      SocketContainer::cast(m_socketContainer.get())->m_socket) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_InvalidSocket);
  }

  if (maxSize > static_cast<stl::SizeT>(stl::Limits<stl::UInt16T>::max())) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_InvalidParam);
  }

  if (networkAddress.getIPVer() != IPVerE::IPv4) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_UnsupportedParam);
  }

  ServerDrv::startClient(
    networkAddress.getIPv4NetworkAddr(),
    networkAddress.getPort(),
    SocketContainer::cast(m_socketContainer.get())->m_socket,
    eProtMode::UDP_MODE);

  stl::UInt16T bufferSize = static_cast<stl::UInt16T>(maxSize);

  if (SocketContainer::cast(m_socketContainer.get())->m_isBlocking) {
    stl::timer::Countdown timeout(
      stl::time::Millis(5000)); // TODO: Configurable wait time

    timeout.reset();

    while (0U == ServerDrv::availData( // TODO: Timeout
                   SocketContainer::cast(m_socketContainer.get())->m_socket)) {
      if (timeout) {
        return checkAndReturn(
          __FILENAME__, __LINE__, IOStatusE::IOStatusE_OpTimeout);
      }

      stl::thread::sleepFor(100);
    }
  }

  static_cast<void>(ServerDrv::getDataBuf(
    SocketContainer::cast(m_socketContainer.get())->m_socket,
    data,
    &bufferSize)); // False returned in every case TODO: Edit ServerDrv
                   // source?

  SocketContainer::cast(m_socketContainer.get())->m_lastExpectedCode = CLOSED;
  // m_lastSysError = 0; // TODO: Workaround

  setSysError();

  const IOStatusE translatedError = translateSysError();

  if (IOStatusE::IOStatusE_OK == translatedError) {
    received = static_cast<stl::SizeT>(bufferSize);
  }

  return checkAndReturn(__FILENAME__, __LINE__, translatedError);
}

ClientSocket::ClientSocket()
  : SocketBase()
{
  m_socketContainer = reinterpret_cast<SocketDataContainer*>(
    stl::memory::MemoryManager<SocketContainer>::allocate(false));
}

IOStatusE
ClientSocket::connect(const NetworkAddress& networkAddress)
{
  switch (networkAddress.getIPVer()) {
    case IPVerE::IPv4:
      return connect(networkAddress.getAsIPv4(), networkAddress.getPort());
    case IPVerE::IPv6:
      return connect(networkAddress.getAsIPv6(), networkAddress.getPort());
    default:
      break;
  }

  return checkAndReturn(
    __FILENAME__, __LINE__, IOStatusE::IOStatusE_InvalidParam);
}

static bool
connected(const SocketT socket)
{
  if (socket == INVALID_SOCKET) {
    return false;
  } else if (0U != ServerDrv::availData(socket)) {
    return true;
  }

  const stl::UInt8T state = ServerDrv::getClientState(socket);

  return (state != LISTEN) && (state != CLOSED) && (state != FIN_WAIT_1) &&
         (state != FIN_WAIT_2) && (state != TIME_WAIT) && (state != SYN_SENT) &&
         (state != SYN_RCVD) && (state != CLOSE_WAIT);
}

IOStatusE
ClientSocket::connect(const IPv4& ip, const PortT port)
{
  Assert(m_socketContainer);
  Assert(INVALID_SOCKET !=
         SocketContainer::cast(m_socketContainer.get())->m_socket);

  if (!m_socketContainer) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_NotInitialized);
  }

  if (INVALID_SOCKET ==
      SocketContainer::cast(m_socketContainer.get())->m_socket) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_InvalidSocket);
  }

  NETWORK_DEBUG_STATEMENT(stl::cout << "[network] Connecting to " << ip.ipStr()
                                    << ':' << port << stl::endl);

  ServerDrv::startClient(
    ip.netAddr(),
    port,
    SocketContainer::cast(m_socketContainer.get())->m_socket,
    eProtMode::TCP_MODE);

  NETWORK_DEBUG_STATEMENT(stl::cout << "[network] Waiting.." << stl::endl);

  stl::thread::sleepFor(50);

  SocketContainer::cast(m_socketContainer.get())->m_lastExpectedCode = CLOSED;
  // m_lastSysError = 0; // TODO: Workaround

  setSysError();

  if (connected(SocketContainer::cast(m_socketContainer.get())->m_socket)) {
    NETWORK_DEBUG_STATEMENT(stl::cout << "[network] Connected." << stl::endl);
    return checkAndReturn(__FILENAME__, __LINE__, IOStatusE::IOStatusE_OK);
  }

  const IOStatusE status = translateSysError();

  NETWORK_DEBUG_STATEMENT(stl::cout << "[network] Error: " << status
                                    << stl::endl);

  return status;
}

IOStatusE
ClientSocket::connect(const IPv6& ip, const PortT port)
{
  Assert(m_socketContainer);
  Assert(INVALID_SOCKET !=
         SocketContainer::cast(m_socketContainer.get())->m_socket);

  if (!m_socketContainer) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_NotInitialized);
  }

  if (INVALID_SOCKET ==
      SocketContainer::cast(m_socketContainer.get())->m_socket) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_InvalidSocket);
  }

  return checkAndReturn(
    __FILENAME__, __LINE__, IOStatusE::IOStatusE_UnsupportedFunction);
}

IOStatusE
ClientSocket::connect(const char* host, const PortT port)
{
  Assert(host);
  Assert(m_socketContainer);
  Assert(INVALID_SOCKET !=
         SocketContainer::cast(m_socketContainer.get())->m_socket);

  if (!host) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_InvalidParam);
  }

  if (!m_socketContainer) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_NotInitialized);
  }

  if (INVALID_SOCKET ==
      SocketContainer::cast(m_socketContainer.get())->m_socket) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_InvalidSocket);
  }

  NETWORK_DEBUG_STATEMENT(stl::cout << "[network] Connecting to " << host << ':'
                                    << port << stl::endl);

  ServerDrv::startClient(
    host,
    stl::string::length(host),
    port,
    SocketContainer::cast(m_socketContainer.get())->m_socket,
    eProtMode::TCP_MODE);

  NETWORK_DEBUG_STATEMENT(stl::cout << "[network] Waiting.." << stl::endl);

  stl::thread::sleepFor(50);

  SocketContainer::cast(m_socketContainer.get())->m_lastExpectedCode = CLOSED;
  // m_lastSysError = 0; // TODO: Workaround

  setSysError();

  if (connected(SocketContainer::cast(m_socketContainer.get())->m_socket)) {
    NETWORK_DEBUG_STATEMENT(stl::cout << "[network] Connected." << stl::endl);
    return checkAndReturn(__FILENAME__, __LINE__, IOStatusE::IOStatusE_OK);
  }

  const IOStatusE status = translateSysError();

  NETWORK_DEBUG_STATEMENT(stl::cout << "[network] Error: " << status
                                    << stl::endl);

  return status;
}

ServerSocket::ServerSocket()
{
  m_socketContainer = reinterpret_cast<SocketDataContainer*>(
    stl::memory::MemoryManager<SocketContainer>::allocate(true));
}

IOStatusE
ServerSocket::listen(const stl::SizeT backlog)
{
  UNUSED_PARAM(backlog);

  Assert(backlog < static_cast<stl::SizeT>(255U));
  Assert(backlog <= stl::Limits<stl::SizeT>::max());

  Assert(m_socketContainer);
  if (!m_socketContainer) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_NotInitialized);
  }

  Assert(INVALID_SOCKET !=
         SocketContainer::cast(m_socketContainer.get())->m_socket);
  if (INVALID_SOCKET ==
      SocketContainer::cast(m_socketContainer.get())->m_socket) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_InvalidSocket);
  }

  return checkAndReturn(__FILENAME__, __LINE__, IOStatusE::IOStatusE_OK);
}

IOStatusE
ServerSocket::accept(Session& newSession)
{
  Assert(m_socketContainer);
  if (!m_socketContainer) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_NotInitialized);
  }

  Assert(INVALID_SOCKET !=
         SocketContainer::cast(m_socketContainer.get())->m_socket);
  if (INVALID_SOCKET ==
      SocketContainer::cast(m_socketContainer.get())->m_socket) {
    return checkAndReturn(
      __FILENAME__, __LINE__, IOStatusE::IOStatusE_InvalidSocket);
  }

  const SocketT newSocket = ServerDrv::availServer(
    SocketContainer::cast(m_socketContainer.get())->m_socket);

  SocketContainer::cast(m_socketContainer.get())->m_lastExpectedCode = CLOSED;
  // m_lastSysError = 0; // TODO: Workaround

  setSysError();

  if (INVALID_SOCKET == newSocket) {
    return translateSysError();
  }

  SocketContainerT newClientSocket = reinterpret_cast<SocketDataContainer*>(
    stl::memory::MemoryManager<SocketContainer>::allocate(newSocket));

  if (!newClientSocket) {
    SocketBase::shutdown(SockDirE::Both);
    SocketBase::close();
    return IOStatusE::IOStatusE_AllocError;
  }

  stl::UniquePtr<ClientSocketBase> clientSocketBase =
    stl::makeUnique<ClientSocketBase>(
      ClientSocketBase(stl::move(newClientSocket)));
  if (!clientSocketBase) {
    SocketBase::shutdown(SockDirE::Both);
    SocketBase::close();
    return IOStatusE::IOStatusE_AllocError;
  }

  NetworkAddress networkAddress; // TODO: How to get remote address?

  newSession = Session(stl::move(clientSocketBase), stl::move(networkAddress));

  return IOStatusE::IOStatusE_OK;
}
} // namespace network

#endif // defined CI_OS_WINDOWS && (CI_OS_WINDOWS == 1)

#endif // defined STL_ENABLE_NETWORKING
