#include "stl/compile_info.hpp"
#include "stl/stl_config.hpp"

#if defined STL_ENABLE_NETWORKING && (STL_ENABLE_NETWORKING == 1)

#if (defined STL_ENABLE_SSL) && (STL_ENABLE_SSL != 0)

#if (defined CI_OS_WINDOWS && (CI_OS_WINDOWS == 1)) ||                         \
  (defined CI_OS_LINUX && (CI_OS_LINUX == 1))

#if (defined STL_ENABLE_SSL) && (STL_ENABLE_SSL != 0) &&                       \
  ((defined STL_ENABLE_SSL_SERVER && (STL_ENABLE_SSL_SERVER == 1)) ||          \
   (defined STL_ENABLE_SSL_CLIENT && STL_ENABLE_SSL_CLIENT == 1))

#if (defined CI_OS_WINDOWS && (CI_OS_WINDOWS == 1))
#pragma comment(lib, "./thirdparty/openssl/lib/libcrypto.lib")
// #pragma comment(lib, "lib64/vs14/libcrypto.lib")
#pragma comment(lib, "./thirdparty/openssl/lib/libssl.lib")
// #pragma comment(lib, "lib64/vs14/libssl.lib")
#endif // defined CI_OS_WINDOWS && (CI_OS_WINDOWS == 1)

#include "stl/assert.hpp"
#include "stl/compiler.hpp"
#include "stl/endian.hpp"
#include "stl/iostream.hpp"
#include "stl/limits.hpp"
#include "stl/memory.hpp"
#include "stl/network.hpp"
#include "stl/new.hpp"
#include "stl/stddef.hpp"
#include "stl/stdint.hpp"
#include "stl/string.hpp"
#include "stl/thread.hpp"

#define OPENSSL_NO_DEPRECATED

#include "openssl/crypto.h"
#include "openssl/err.h"
#include "openssl/ssl.h"

static int
certVerifyCB(X509_STORE_CTX* ctx, void* arg)
{ // TODO: Remove
  UNUSED_PARAM(ctx);
  UNUSED_PARAM(arg);
  return 1;
}

static int
testErrPrint(const char* str, size_t len, void* u)
{
  UNUSED_PARAM(u);

  stl::cout.write(reinterpret_cast<const stl::UInt8T*>(str), len);

  return 1;
}

static int
pkPassCB(char* buf, int size, int rwflag, void* userdata)
{
  UNUSED_PARAM(rwflag);
  UNUSED_PARAM(userdata);

  static const char pass[] = "test";

  if (sizeof(pass) > static_cast<stl::SizeT>(size)) {
    return 0;
  } else {
    stl::memory::copy(pass, pass + sizeof(pass), buf);

    return sizeof(pass) - 1;
  }
}

namespace network {
class SSLDeleter
{
public:
  SSLDeleter() noexcept = default;

  SSLDeleter(const SSLDeleter&) {}

  void operator()(SSL* const ptr) const noexcept { SSL_free(ptr); }
};

class SSLCTXDeleter
{
public:
  SSLCTXDeleter() noexcept = default;

  SSLCTXDeleter(const SSLCTXDeleter&) {}

  void operator()(SSL_CTX* const ptr) const noexcept { SSL_CTX_free(ptr); }
};

class X509Deleter
{
public:
  X509Deleter() noexcept = default;

  X509Deleter(const X509Deleter&) {}

  void operator()(X509* const ptr) const noexcept { X509_free(ptr); }
};

using SSLPtr = stl::UniquePtr<SSL, SSLDeleter>;

using X509Ptr = stl::UniquePtr<X509, X509Deleter>;

struct SSLContainer
{
  SSLContainer()
    : m_lastSSLError(-1)
  {
  }

  SSLContainer(SSLPtr&& ssl)
    : m_ssl(stl::move(ssl))
    , m_lastSSLError(-1)
  {
    Assert(m_ssl);
  }

  SSLContainer(const SSLContainer&) = delete;
  SSLContainer& operator=(const SSLContainer&) = delete;

  SSLContainer(SSLContainer&& other) noexcept
    : m_ssl(stl::move(other.m_ssl))
    , m_lastSSLError(other.m_lastSSLError) {
      Assert(m_ssl);
    }

  SSLContainer& operator=(SSLContainer&& other) noexcept
  {
    m_ssl = stl::move(other.m_ssl);
    m_lastSSLError = stl::exchange(other.m_lastSSLError, -1);

    Assert(m_ssl);

    return *this;
  }

  static SSLContainer* cast(void* const ptr)
  {
    return reinterpret_cast<SSLContainer*>(ptr);
  }

  static const SSLContainer* cast(const void* const ptr)
  {
    return reinterpret_cast<const SSLContainer*>(ptr);
  }

  void setLastSSLError(const int error,
                       const char* filename,
                       const unsigned int line)
  {
    if (m_ssl) {
      m_lastSSLError = SSL_get_error(m_ssl.get(), error);

      switch(m_lastSSLError)
      {
        case SSL_ERROR_NONE:
        case SSL_ERROR_WANT_READ:
        case SSL_ERROR_WANT_WRITE:
        case SSL_ERROR_WANT_ACCEPT:
        case SSL_ERROR_WANT_CONNECT:
        case SSL_ERROR_WANT_ASYNC:
        case SSL_ERROR_WANT_ASYNC_JOB:
        case SSL_ERROR_WANT_CLIENT_HELLO_CB:
        case SSL_ERROR_WANT_RETRY_VERIFY:
        case SSL_ERROR_WANT_X509_LOOKUP:
        break;
        default:
        {
          stl::cout << "[SSLError:" << filename << ':' << line << "] " << m_lastSSLError << ": ";
          ERR_print_errors_cb(testErrPrint, nullptr);
        }
      }
    }
  }
public: // TODO: Remove
  SSLPtr m_ssl;
  int m_lastSSLError;
};

IOStatusE
initSecurityLayer()
{
#if (defined STL_ENABLE_SSL) && (STL_ENABLE_SSL != 0)
  int status = OPENSSL_init_ssl(0, nullptr);
  if (1 != status) {
    return IOStatusE::IOStatusE_SSLInitError;
  }

  status = OPENSSL_init_crypto(OPENSSL_INIT_LOAD_CONFIG, nullptr);
  if (1 != status) {
    return IOStatusE::IOStatusE_SSLInitError;
  }
#endif
  return IOStatusE::IOStatusE_OK;
}

using SSLCtxPtr = stl::UniquePtr<SSL_CTX, SSLCTXDeleter>;

struct SSLCtxContainer
{
  SSLCtxContainer()
    : m_sslMethod(nullptr)
    , m_sslCtx(nullptr)
    , m_lastSSLCtxError(1)
  {
  }

  SSLCtxContainer(const SSLCtxContainer&) = delete;
  SSLCtxContainer& operator=(const SSLCtxContainer&) = delete;

  SSLCtxContainer(SSLCtxContainer&& other) noexcept
  {
    operator=(stl::move(other));
  }

  SSLCtxContainer& operator=(SSLCtxContainer&& other) noexcept
  {
    m_sslMethod = other.m_sslMethod;
    m_sslCtx = stl::move(other.m_sslCtx);
    m_lastSSLCtxError = stl::exchange(other.m_lastSSLCtxError, 1);

    return *this;
  }

  static SSLCtxContainer* cast(void* const ptr)
  {
    return reinterpret_cast<SSLCtxContainer*>(ptr);
  }

  const static SSLCtxContainer* cast(const void* const ptr)
  {
    return reinterpret_cast<const SSLCtxContainer*>(ptr);
  }

  void setLastSSLError(int error, const char* filename, const unsigned int line)
  {
    if (m_sslCtx) {
      m_lastSSLCtxError = error;

      if (1 != m_lastSSLCtxError) {
        // TODO: Remove and adjust
        stl::cout << "[SSLError:" << filename << ':' << line << "] ";
        ERR_print_errors_cb(testErrPrint, nullptr);
      }
    }
  }

  const SSL_METHOD* m_sslMethod;
  SSLCtxPtr m_sslCtx;
  int m_lastSSLCtxError;
};

SSLSocketBase::SSLSocketBase()
  : m_sslCtxContainer(reinterpret_cast<SSLDataContainer*>(
      stl::memory::MemoryManager<SSLCtxContainer>::allocate()))
{
}

SSLSocketBase::SSLSocketBase(SSLSocketBase&& other) noexcept
  : m_sslCtxContainer(stl::move(other.m_sslCtxContainer))
{
}

SSLSocketBase&
SSLSocketBase::operator=(SSLSocketBase&& other) noexcept
{
  m_sslCtxContainer = stl::move(other.m_sslCtxContainer);

  return *this;
}

int
SSLSocketBase::getLastSSLCtxError() const
{
  return SSLCtxContainer::cast(m_sslCtxContainer.get())->m_lastSSLCtxError;
}

void
SSLSocketBase::setLastSSLCtxError(const int error,
                                  const char* filename,
                                  const unsigned int line)
{
  SSLCtxContainer::cast(m_sslCtxContainer.get())
    ->setLastSSLError(error, filename, line);
}
}

#endif

#if (defined STL_ENABLE_SSL) && (STL_ENABLE_SSL != 0) &&                       \
  ((defined STL_ENABLE_SSL_SERVER && (STL_ENABLE_SSL_SERVER == 1)) ||          \
   (defined STL_ENABLE_SSL_CLIENT && STL_ENABLE_SSL_CLIENT == 1))

namespace network {
SSLClientSocketBase::SSLClientSocketBase()
  : SSLSocketBase()
{
  m_sslContainer = reinterpret_cast<SSLDataContainer*>(
    stl::memory::MemoryManager<SSLContainer>::allocate());
}

SSLClientSocketBase::SSLClientSocketBase(
  SocketContainerT&& socketContainer,
  SSLStructContainerT&& sslStructContainer) noexcept
  : ClientSocketBase(stl::move(socketContainer))
  , SSLSocketBase(stl::move(sslStructContainer))
{
  m_sslContainer = reinterpret_cast<SSLDataContainer*>(
    stl::memory::MemoryManager<SSLContainer>::allocate());

  
}

SSLClientSocketBase::SSLClientSocketBase(
  ClientSocketBase&& clientSocketBase,
  SSLStructContainerT&& sslStructContainer) noexcept
  : ClientSocketBase(stl::move(clientSocketBase))
  , SSLSocketBase()
{
  m_sslContainer = stl::move(sslStructContainer);

  
}

SSLClientSocketBase::SSLClientSocketBase(SSLClientSocketBase&& other) noexcept
{
  
  SSLClientSocketBase::operator=(stl::move(other));
  
}

SSLClientSocketBase&
SSLClientSocketBase::operator=(SSLClientSocketBase&& other) noexcept
{
  
  SSLSocketBase::operator=(stl::move(static_cast<SSLSocketBase&>(other)));
  m_sslContainer = stl::move(other.m_sslContainer);
  m_networkAddr = stl::move(other.m_networkAddr);
  

  Assert(nullptr != m_sslContainer.get());

  return *this;
}

IOStatusE
SSLClientSocketBase::shutdown(const SockDirE dir)
{
  if (!m_sslContainer) {
    static_cast<void>(ClientSocketBase::shutdown(SockDirE::Both));
    return IOStatusE::IOStatusE_SSLNotReady;
  }

  if (!SSLContainer::cast(m_sslContainer.get())->m_ssl) {
    return IOStatusE::IOStatusE_SSLNotReady;
  }

  const int status =
    SSL_shutdown(SSLContainer::cast(m_sslContainer.get())->m_ssl.get());

  setLastSSLError(status, __FILENAME__, __LINE__);

  return ClientSocketBase::shutdown(dir);
}

IOStatusE
SSLClientSocketBase::postConnectAction()
{
  Assert(m_socketContainer);

  if (!m_sslContainer) {
    static_cast<void>(SocketBase::shutdown(SockDirE::Both));
    return IOStatusE::IOStatusE_SSLNotReady;
  }

  if (!SSLContainer::cast(m_sslContainer.get())->m_ssl) {
    return IOStatusE::IOStatusE_SSLNotReady;
  }

  int status = SSL_set_fd(SSLContainer::cast(m_sslContainer.get())->m_ssl.get(),
                          static_cast<int>(fd()));

  setLastSSLError(status, __FILENAME__, __LINE__);

  if (1 != status) {
    return IOStatusE::IOStatusE_SSLFDError;
  }

  // TODO: Remove
  const char* cipherList = SSL_get_cipher_list(
    SSLContainer::cast(m_sslContainer.get())->m_ssl.get(), 0);
  stl::cout << "[Client] SSL Version: "
            << SSL_get_version(
                 SSLContainer::cast(m_sslContainer.get())->m_ssl.get())
            << stl::endl;
  stl::cout << "[Client] CipherList: " << cipherList << stl::endl;
  stl::cout << "[Client] SSL Client Version"
            << SSL_client_version(
                 SSLContainer::cast(m_sslContainer.get())->m_ssl.get())
            << stl::endl;

  status = SSL_connect(SSLContainer::cast(m_sslContainer.get())->m_ssl.get());

  setLastSSLError(status, __FILENAME__, __LINE__);
  setSysError(__FILENAME__, __LINE__);

  if (status == 1) {
    return IOStatusE::IOStatusE_OK;
  } else if ((SSL_ERROR_WANT_READ == status) ||
             (SSL_ERROR_WANT_WRITE == status)) {
    return IOStatusE::IOStatusE_OpWouldBlock;
  } else {
    ERR_print_errors_cb(testErrPrint, nullptr);
  }

  return IOStatusE::IOStatusE_SSLError;
}

IOStatusE SSLClientSocketBase::readReady()
{
  // 
  // 
  // const int retv = SSL_pending(SSLContainer::cast(m_sslContainer.get())->m_ssl.get());
  // 
  // if(0 == retv)
  // {
  //   return IOStatusE::IOStatusE_OpWouldBlock;
  // }

  return IOStatusE::IOStatusE_OK;
}

IOStatusE
SSLClientSocketBase::send(const stl::UInt8T* const data,
                          const stl::SizeT size,
                          stl::SizeT& sent)
{
  sent = 0U;

  Assert(nullptr != data);

  if (!m_sslContainer) {
    static_cast<void>(ClientSocketBase::shutdown(SockDirE::Both));
    return IOStatusE::IOStatusE_SSLNotReady;
  }

  if (!SSLContainer::cast(m_sslContainer.get())->m_ssl) {
    return IOStatusE::IOStatusE_SSLNotReady;
  }

  if (size > static_cast<stl::SizeT>(stl::Limits<int>::max())) {
    return IOStatusE::IOStatusE_InvalidParam;
  }

  const int retv =
    SSL_write(SSLContainer::cast(m_sslContainer.get())->m_ssl.get(),
              data,
              static_cast<int>(size));

  SSLContainer::cast(m_sslContainer.get())
    ->setLastSSLError(retv, __FILENAME__, __LINE__);
  setSysError(__FILENAME__, __LINE__);

  if (retv > 0) {
    sent = static_cast<stl::SizeT>(retv);
    return IOStatusE::IOStatusE_OK;
  } else if ((SSL_ERROR_WANT_READ == retv) || (SSL_ERROR_WANT_WRITE == retv)) {
    return IOStatusE::IOStatusE_OpWouldBlock;
  }

  return IOStatusE::IOStatusE_SSLError;
}

IOStatusE
SSLClientSocketBase::receive(stl::UInt8T* const data,
                             const stl::SizeT maxSize,
                             stl::SizeT& received)
{
  received = 0U;

  Assert(nullptr != data);

  if (!m_sslContainer) {
    LOGEXT << logging::LvlE::Error << "SSL container not created" << stl::endl;
    static_cast<void>(ClientSocketBase::shutdown(SockDirE::Both));
    return IOStatusE::IOStatusE_SSLNotReady;
  }

  if (!SSLContainer::cast(m_sslContainer.get())->m_ssl) {
    LOGEXT << logging::LvlE::Error << "SSL handle not created" << stl::endl;
    return IOStatusE::IOStatusE_SSLNotReady;
  }

  if (maxSize > static_cast<stl::SizeT>(stl::Limits<int>::max())) {
    LOGEXT << logging::LvlE::Error << "Invalid max size " << maxSize << stl::endl;
    return IOStatusE::IOStatusE_InvalidParam;
  }

  /* stl::thread::sleepFor(1000);

  char buf[1024]{};
  const int a = SSL_read(::network::SSLContainer::cast(m_sslContainer.get())->m_ssl.get(), buf, sizeof(buf));

  // stl::SizeT a = 0U;
  // network::IOStatusE stat = instance->m_session.socket().receive(reinterpret_cast<stl::UInt8T*>(buf), sizeof(buf), a);

  LOGEXT << logging::LvlE::Debug << "TEST " << maxSize << " x " << static_cast<int>(a) << " = " << SSL_get_error(::network::SSLContainer::cast(m_sslContainer.get())->m_ssl.get(), a) << stl::endl;

  if(a > 0)
  {
    logging::log.dump((const stl::UInt8T*)buf, a, 20, true);
  }

  exit(0); */

  int retv
    = SSL_read(SSLContainer::cast(m_sslContainer.get())->m_ssl.get(),
            data,
            static_cast<int>(maxSize));

  SSLContainer::cast(m_sslContainer.get())
    ->setLastSSLError(retv, __FILENAME__, __LINE__);
  setSysError(__FILENAME__, __LINE__);

  if (retv > 0) {
    received = static_cast<stl::SizeT>(retv);
    return IOStatusE::IOStatusE_OK;
  } else if ((SSL_ERROR_WANT_READ == SSLContainer::cast(m_sslContainer.get())->m_lastSSLError) || (SSL_ERROR_WANT_WRITE == SSLContainer::cast(m_sslContainer.get())->m_lastSSLError)) {
    return IOStatusE::IOStatusE_OpWouldBlock;
  }

  LOGEXT << logging::LvlE::Error << "SSL read operation failed " << retv << " " << SSLContainer::cast(m_sslContainer.get())->m_lastSSLError << stl::endl;

  return IOStatusE::IOStatusE_SSLError;
}

int
SSLClientSocketBase::getLastSSLError() const
{
  if (m_sslContainer) {
    return SSLContainer::cast(m_sslContainer.get())->m_lastSSLError;
  }

  return -1;
}

void
SSLClientSocketBase::setLastSSLError(const int error,
                                     const char* filename,
                                     const unsigned int line)
{
  if (m_sslContainer && SSLContainer::cast(m_sslContainer.get())->m_ssl) {
    SSLContainer::cast(m_sslContainer.get())->m_lastSSLError = SSL_get_error(
      SSLContainer::cast(m_sslContainer.get())->m_ssl.get(), error);

    if (SSL_ERROR_NONE !=
        SSLContainer::cast(m_sslContainer.get())->m_lastSSLError) {
      stl::cout << "[SSLError:" << filename << ':' << line << "] ";
      ERR_print_errors_cb(testErrPrint, nullptr);
    }
  } else {
    SSLContainer::cast(m_sslContainer.get())->m_lastSSLError = -1;
  }
}
}

#endif

#if (defined STL_ENABLE_SSL) && (STL_ENABLE_SSL != 0) &&                       \
  (defined STL_ENABLE_SSL_CLIENT && (STL_ENABLE_SSL_CLIENT == 1))

namespace network {
SSLClientSocket::SSLClientSocket()
  : SSLClientSocketBase()
{
}

IOStatusE
SSLClientSocket::postOpenAction()
{
  stl::cout << "[OPENSSL]: " << OpenSSL_version(OPENSSL_FULL_VERSION_STRING)
            << stl::endl;

  if (!m_sslContainer) {
    static_cast<void>(SocketBase::shutdown(SockDirE::Both));
    LOGEXT << logging::LvlE::Error << "SSL container not created" << stl::endl;
    return IOStatusE::IOStatusE_SSLNotReady;
  }

  if (!m_sslCtxContainer) {
    static_cast<void>(SocketBase::shutdown(SockDirE::Both));
    LOGEXT << logging::LvlE::Error << "SSL CTX container not created"
           << stl::endl;
    return IOStatusE::IOStatusE_SSLNotReady;
  }

  SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslMethod =
    TLS_client_method();

  SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx =
    SSL_CTX_new(SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslMethod);

  if (!SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx) {
    LOGEXT << logging::LvlE::Error << "SSL CTX not created" << stl::endl;
    return IOStatusE::IOStatusE_SSLError;
  }

  int status = SSL_CTX_set_ciphersuites(
    SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx.get(),
    "TLS_AES_256_GCM_SHA384:TLS_AES_128_GCM_SHA256:"
    "TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256:"
    "TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384:"
    "TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA256:"
    "TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA384:"
    "TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256:"
    "TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384:"
    "TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA256:"
    "TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA384:"
    "TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA256:"
    "TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA384:"
    "TLS_DHE_RSA_WITH_AES_128_GCM_SHA256:"
    "TLS_DHE_RSA_WITH_AES_256_GCM_SHA384:"
    "TLS_DHE_RSA_WITH_AES_128_CBC_SHA:"
    "TLS_DHE_RSA_WITH_AES_256_CBC_SHA:"
    "TLS_DHE_RSA_WITH_AES_128_CBC_SHA256:"
    "TLS_DHE_RSA_WITH_AES_256_CBC_SHA256:"
    "TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305_SHA256:"
    "TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305:"
    "TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305_SHA256:"
    "TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305:"
    "TLS_AES_256_GCM_SHA384:"
    "TLS_CHACHA20_POLY1305_SHA256:"
    "TLS_AES_128_CCM_SHA256:"
    "TLS_AES_128_CCM_8_SHA256:"
    "TLS_AES_128_GCM_SHA256");
  if (1 != status) {
    LOGEXT << logging::LvlE::Error << "SSL CTX ciphersuite not set"
           << stl::endl;
    return IOStatusE::IOStatusE_SSLCipherError;
  }

  status = SSL_CTX_set_cipher_list(
    SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx.get(),
    "TLSv1.3:kRSA:!eNULL:!aNULL");
  if (1 != status) {
    LOGEXT << logging::LvlE::Error << "SSL CTX cipher list not set"
           << stl::endl;
    return IOStatusE::IOStatusE_SSLCipherError;
  }

  // long statusl = SSL_CTX_set_min_proto_version(
  //     sslBaseContextContainer->m_sslCtx, TLS1_2_VERSION);
  // if (1 != statusl) {
  //     return IOStatusE::IOStatusE_SSLContextError;
  // }

  // SSL_CTX_set_options(sslBaseContextContainer->m_sslCtx, SSL_OP_NO_SSLv2
  //     | SSL_OP_NO_SSLv3); // SSL_OP_NO_COMPRESSION

  // SSL_CTX_set_verify(sslBaseContextContainer->m_sslCtx, SSL_VERIFY_NONE
  //     , nullptr);
  // SSL_CTX_set_cert_verify_callback(sslBaseContextContainer->m_sslCtx
  //     , certVerifyCB, NULL);
  // ERR_print_errors_cb(testErrPrint, nullptr);

  // SSL_CTX_set_verify_depth(sslBaseContextContainer->m_sslCtx, 2);

  // int retv = SSL_CTX_load_verify_locations(
  //     sslBaseContextContainer->m_sslCtx, "random-org-chain.pem", NULL);

  // const char PREFERRED_CIPHERS[] = "HIGH:!aNULL:!kRSA:!PSK:!SRP:!MD5:!RC4";
  // retv = SSL_set_cipher_list(sslBaseContextContainer->m_sslCtx,
  // PREFERRED_CIPHERS);

  // X509* cert = SSL_get_peer_certificate(sslBaseContextContainer->m_sslCtx);
  // X509_free(cert);

  /* Step 2: verify the result of chain verification */
  /* Verification performed according to RFC 4158    */
  // retv = SSL_get_verify_result(ssl);

  status = SSL_CTX_load_verify_locations(
    SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx.get(),
    nullptr,
    "../cert/ca/IONTtech_DS.ca-chain.cert.pem");
  if (1 != status) {
    LOGEXT << logging::LvlE::Error << "Could not add verify location"
           << stl::endl;
    return IOStatusE::IOStatusE_SSLError;
  }

  const stl::String2 certificatePath = (stl::String2("../cert/client/") + m_info + ".cert.pem");
  status = SSL_CTX_use_certificate_file(
    SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx.get(),
    certificatePath.c_str(),
    SSL_FILETYPE_PEM);
  if (1 != status) {
    LOGEXT << logging::LvlE::Error << "Could not add certificate file " << certificatePath
           << stl::endl;
    return IOStatusE::IOStatusE_SSLError;
  }

  const stl::String2 keyPath = (stl::String2("../cert/client/") + m_info + ".key.pem");
  status = SSL_CTX_use_PrivateKey_file(
    SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx.get(),
    keyPath.c_str(),
    SSL_FILETYPE_PEM);
  if (1 != status) {
    LOGEXT << logging::LvlE::Error << "Could not use private kez given " << keyPath
           << stl::endl;
    return IOStatusE::IOStatusE_SSLError;
  }

  if (1 != SSL_CTX_check_private_key(
             SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx.get())) {
    LOGEXT << logging::LvlE::Error << "Private key check failed" << stl::endl;
    return IOStatusE::IOStatusE_SSLError;
  }

  LOGEXT << logging::LvlE::Success << "SSL socket ready!" << stl::endl;

  return IOStatusE::IOStatusE_OK;
}

SSLClientSocket::SSLClientSocket(SSLClientSocket&& other) noexcept
{
  operator=(stl::move(other));
}

SSLClientSocket&
SSLClientSocket::operator=(SSLClientSocket&& other) noexcept
{
  SSLClientSocketBase::operator=(stl::move(other));

  return *this;
}

SSLClientSocket::~SSLClientSocket()
{
  // TODO
}

IOStatusE
SSLClientSocket::postConnectAction()
{
  return SSLClientSocketBase::postConnectAction();
}

IOStatusE
SSLClientSocket::connectInternal()
{
  IOStatusE status = ClientSocket::connect(m_networkAddr);

  if (IOStatusE::IOStatusE_OK != status) {
    return status;
  }

  SSLContainer::cast(m_sslContainer.get())->m_ssl =
    SSL_new(SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx.get());

  if (!SSLContainer::cast(m_sslContainer.get())->m_ssl) {
    SocketBase::shutdown(SockDirE::Both);
    SocketBase::close();
    return IOStatusE::IOStatusE_SSLError;
  }

  // TODO: Remove
  const char* cipherList = SSL_get_cipher_list(
    SSLContainer::cast(m_sslContainer.get())->m_ssl.get(), 0);
  stl::cout << "[SSL Version]: "
            << SSL_get_version(
                 SSLContainer::cast(m_sslContainer.get())->m_ssl.get())
            << stl::endl;
  stl::cout << "CipherList: " << cipherList << stl::endl;

  return IOStatusE::IOStatusE_OK;
}

IOStatusE
SSLClientSocket::connect(const NetworkAddress& networkAddress)
{
  switch (networkAddress.getIPVer()) {
    case IPVerE::IPv4:
      return connect(networkAddress.getAsIPv4(), networkAddress.getPort());
    case IPVerE::IPv6:
      return connect(networkAddress.getAsIPv6(), networkAddress.getPort());
    default:
      break;
  }

  return IOStatusE::IOStatusE_InvalidParam;
}

IOStatusE
SSLClientSocket::connect(const IPv4& ip, const PortT port)
{
  static_cast<void>(
    m_networkAddr.set(ip, port)); // It can fail only if parsing string

  return connectInternal();
}

IOStatusE
SSLClientSocket::connect(const IPv6& ip, const PortT port)
{
  static_cast<void>(
    m_networkAddr.set(ip, port)); // It can fail only if parsing string

  return connectInternal();
}

IOStatusE
SSLClientSocket::connect(const char* host, const PortT port)
{
  UNUSED_PARAM(host);
  UNUSED_PARAM(port);

  return IOStatusE::IOStatusE_UnsupportedFunction;
}
}

#endif

#if (defined STL_ENABLE_SSL) && (STL_ENABLE_SSL != 0) &&                       \
  (defined STL_ENABLE_SSL_SERVER && (STL_ENABLE_SSL_SERVER == 1))

namespace network {
SSLServerSocket::SSLServerSocket()
  : SSLSocketBase()
{
  m_sslCtxContainer = reinterpret_cast<SSLDataContainer*>(
    stl::memory::MemoryManager<SSLCtxContainer>::allocate());
}

IOStatusE
SSLServerSocket::postOpenAction()
{
  if (!m_sslCtxContainer) {
    static_cast<void>(SocketBase::shutdown(SockDirE::Both));
    return IOStatusE::IOStatusE_SSLNotReady;
  }

  int status = 0;

  stl::cout << "[OPENSSL]: " << OpenSSL_version(OPENSSL_FULL_VERSION_STRING)
            << stl::endl;

  if (nullptr == m_sslCtxContainer) {
    return IOStatusE::IOStatusE_NotInitialized;
  }

  SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslMethod =
    TLS_server_method();

  if (!SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslMethod) {
    return IOStatusE::IOStatusE_SSLMethodError;
  }

  SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx =
    SSL_CTX_new(SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslMethod);

  if (!SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx) {
    return IOStatusE::IOStatusE_SSLContextError;
  }

  /* // SSL_CTX_set_default_passwd_cb_userdata
  SSL_CTX_set_default_passwd_cb(
    SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx.get()
    , pkPassCB); */

  // // , "./cert/iont_rootcert.pem", SSL_FILETYPE_PEM) <= 0) {
  // status = SSL_CTX_use_certificate_file(
  status = SSL_CTX_use_certificate_chain_file(
    SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx.get(),
    "./certs/IONTtech_DSServer.bundle.cert.pem");

  setLastSSLCtxError(status, __FILENAME__, __LINE__);

  if (status <= 0) {
    return IOStatusE::IOStatusE_SSLError;
  }

  // // , "./cert/iont_rootkey.pem", SSL_FILETYPE_PEM) <= 0) {
  status = SSL_CTX_use_PrivateKey_file(
    SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx.get(),
    "./certs/IONTtech_DSServer.key.pem",
    SSL_FILETYPE_PEM);

  setLastSSLCtxError(status, __FILENAME__, __LINE__);

  if (status <= 0) {
    return IOStatusE::IOStatusE_SSLError;
  }

  if (1 != SSL_CTX_check_private_key(
             SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx.get())) {
    LOGEXT << logging::LvlE::Error << "Private key check failed" << stl::endl;
    return IOStatusE::IOStatusE_SSLError;
  }

  SSL_CTX_set_cert_verify_callback(
    SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx.get(),
    certVerifyCB,
    nullptr);

  /* TODO: !!!!
It looks like you are missing the call to SSL_CTX_use_certificate_chain_file and SSL_CTX_set_client_CA_list on the server. I believe SSL_CTX_set_client_CA_list triggers the machinery to perform client authentication (i.e., it elicits the ClientCertificate message from the RFC in section 7.4.6 during the exchange).
If SSL_CTX_set_client_CA_list causes the failure you are looking for, then I'm inclined to believe its a bug in the OpenSSL library. You should get a failure when specifying SSL_VERIFY_PEER and SSL_VERIFY_FAIL_IF_NO_PEER_CERT at the server because that's what the docs say.
Also see Is verification supposed to fail with SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT without SSL_CTX_set_client_CA_list on the OpenSSL Users mailing list. Its in response to this question.
Also see Testing SSL/TLS Client Authentication with OpenSSL and OpenSSL client not sending client certificate.
  */

  SSL_CTX_set_verify(
    SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx.get(),
    SSL_VERIFY_NONE // TODO: Verify
    // SSL_VERIFY_PEER // | SSL_VERIFY_FAIL_IF_NO_PEER_CERT
    ,
    nullptr); // Global SSL_set_verify

  // TLS 1.2
  // status =
  // SSL_CTX_set_ciphersuites(SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx.get()
  //     , "TLS_AES_256_GCM_SHA384:TLS_AES_128_GCM_SHA256:"
  //         "TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256:"
  //         "TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384:"
  //         "TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA256:"
  //         "TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA384:"
  //         "TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256:"
  //         "TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384:"
  //         "TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA256:"
  //         "TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA384:"
  //         "TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA256:"
  //         "TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA384:"
  //         "TLS_DHE_RSA_WITH_AES_128_GCM_SHA256:"
  //         "TLS_DHE_RSA_WITH_AES_256_GCM_SHA384:"
  //         "TLS_DHE_RSA_WITH_AES_128_CBC_SHA:"
  //         "TLS_DHE_RSA_WITH_AES_256_CBC_SHA:"
  //         "TLS_DHE_RSA_WITH_AES_128_CBC_SHA256:"
  //         "TLS_DHE_RSA_WITH_AES_256_CBC_SHA256:"
  //         "TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305_SHA256:"
  //         "TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305:"
  //         "TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305_SHA256:"
  //         "TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305");

  // status =
  // SSL_CTX_set_ciphersuites(SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx
  //     , "TLS_AES_256_GCM_SHA384:"
  //         "TLS_CHACHA20_POLY1305_SHA256:"
  //         "TLS_AES_128_CCM_SHA256:"
  //         "TLS_AES_128_CCM_8_SHA256:"
  //         "TLS_AES_128_GCM_SHA256");

  status = SSL_CTX_set_ciphersuites(
    SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx.get(),
    "TLS_AES_256_GCM_SHA384:TLS_AES_128_GCM_SHA256:"
    "TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256:"
    "TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384:"
    "TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA256:"
    "TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA384:"
    "TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256:"
    "TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384:"
    "TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA256:"
    "TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA384:"
    "TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA256:"
    "TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA384:"
    "TLS_DHE_RSA_WITH_AES_128_GCM_SHA256:"
    "TLS_DHE_RSA_WITH_AES_256_GCM_SHA384:"
    "TLS_DHE_RSA_WITH_AES_128_CBC_SHA:"
    "TLS_DHE_RSA_WITH_AES_256_CBC_SHA:"
    "TLS_DHE_RSA_WITH_AES_128_CBC_SHA256:"
    "TLS_DHE_RSA_WITH_AES_256_CBC_SHA256:"
    "TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305_SHA256:"
    "TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305:"
    "TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305_SHA256:"
    "TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305:"
    "TLS_AES_256_GCM_SHA384:"
    "TLS_CHACHA20_POLY1305_SHA256:"
    "TLS_AES_128_CCM_SHA256:"
    "TLS_AES_128_CCM_8_SHA256:"
    "TLS_AES_128_GCM_SHA256");

  setLastSSLCtxError(status, __FILENAME__, __LINE__);

  // SSL_CTX_set1_groups_list(sslBaseContextContainer->m_sslCtx, "X25519:X448");

  if (1 != status) {
    return IOStatusE::IOStatusE_SSLCipherError;
  }

  // status = SSL_set_cipher_list(clientSSL,
  // "HIGH:!aNULL:!kRSA:!PSK:!SRP:!MD5:!RC4");
  status = SSL_CTX_set_cipher_list(
    SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx.get(),
    "TLSv1.3:kRSA:!eNULL:!aNULL");

  setLastSSLCtxError(status, __FILENAME__, __LINE__);

  // SSL_set_cipher_list(clientSSL,
  // "TLS_AES_256_GCM_SHA384:TLS_CHACHA20_POLY1305_SHA256:TLS_AES_128_CCM_SHA256:TLS_AES_128_CCM_8_SHA256:TLS_AES_128_GCM_SHA256");
  // int status = SSL_set_ciphersuites(clientSSL,
  // "TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384");
  // //ADH-AES128-SHA:ADH-AES128-SHA256:ADH-AES256-SHA256
  if (1 != status) {
    return IOStatusE::IOStatusE_SSLCipherError;
  }

  // long statusl = SSL_CTX_set_min_proto_version(
  //     SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx.get(),
  //     TLS1_2_VERSION);
  // if(1 != statusl) {
  //     return IOStatusE::IOStatusE_SSLContextError;
  // }

  // int status = SSL_CTX_set_cipher_list(
  //     SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx.get()
  //     ,
  //     "ECDHE_ECDSA_WITH_AES_256_GCM_SHA384:ECDHE_ECDSA_WITH_AES_128_GCM_SHA256");
  //
  // if(1 != status) {
  //     return IOStatusE::IOStatusE_SSLCipherError;
  // }

  // SSL_CTX_set_options(SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx.get(),
  // SSL_OP_NO_SSLv2
  //     | SSL_OP_NO_SSLv3 | SSL_OP_NO_SSLv3 | SSL_OP_NO_TLSv1
  //     | SSL_OP_NO_TLSv1_1); // SSL_OP_NO_COMPRESSION

  SSL_CTX_set_options(
    SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx.get(),
    SSL_OP_NO_SSLv2 | SSL_OP_NO_SSLv3 |
      SSL_OP_NO_SSLv3); // SSL_OP_NO_COMPRESSION

  // const int status = SSL_CTX_set_cipher_list(
  //     sslBaseContextContainer->m_sslCtx
  //     , "ECDH:!aNULL:!eNULL:!ADH:!SHA:@STRENGTH");
  // if(1 != status) {
  //     IOStatusE::IOStatusE_SSLError;
  // }

  // SSL_CTX_set_verify_depth(sslBaseContextContainer->m_sslCtx, 2);

  // TODO: Add cert support?
  // /* Set the key and cert */
  // if
  // (SSL_CTX_use_certificate_file(SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx.get(),
  // "cert.pem", SSL_FILETYPE_PEM) <= 0) {
  //     SSL_CTX_free(sslCtx);
  //     clientSocket.close();
  //     m_stop.store(true);
  //     return IOStatusE::IOStatusE_CertError;
  // }
  //
  // if
  // (SSL_CTX_use_PrivateKey_file(SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx.get(),
  // "key.pem", SSL_FILETYPE_PEM) <= 0) {
  //     SSL_CTX_free(sslCtx);
  //     clientSocket.close();
  //     m_stop.store(true);
  //     return IOStatusE::IOStatusE_CertError;
  // }

  return IOStatusE::IOStatusE_OK;
}

SSLServerSocket::SSLServerSocket(SocketContainerT&& socketContainer)
  : ServerSocket(stl::move(socketContainer))
{
}

SSLServerSocket::SSLServerSocket(SSLServerSocket&& other) noexcept
  : SSLSocketBase(stl::move(static_cast<SSLSocketBase&>(other)))
{
}

SSLServerSocket&
SSLServerSocket::operator=(SSLServerSocket&& other) noexcept
{
  SSLSocketBase::operator=(stl::move(static_cast<SSLSocketBase&>(other)));

  return *this;
}

IOStatusE
SSLServerSocket::postAcceptAction(Session& session)
{
  if (!m_sslCtxContainer) {
    static_cast<void>(SocketBase::shutdown(SockDirE::Both));
    LOGEXT << logging::LvlE::Error << "No SSL context container" << stl::endl;
    return IOStatusE::IOStatusE_SSLNotReady;
  }

  if (!SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx) {
    LOGEXT << logging::LvlE::Error << "No SSL context" << stl::endl;
    return IOStatusE::IOStatusE_SSLNotReady;
  }

  SSLPtr clientSSL =
    SSL_new(SSLCtxContainer::cast(m_sslCtxContainer.get())->m_sslCtx.get());
  if (!clientSSL) {
    SocketBase::shutdown(SockDirE::Both);
    SocketBase::close();
    LOGEXT << logging::LvlE::Error << "SSL instance not created" << stl::endl;
    return IOStatusE::IOStatusE_SSLError;
  }

  SSLStructContainerT sslContainer = reinterpret_cast<SSLDataContainer*>(
    stl::memory::MemoryManager<SSLContainer>::allocate(stl::move(clientSSL)));

  if (!sslContainer) {
    SocketBase::shutdown(SockDirE::Both);
    SocketBase::close();
    LOGEXT << logging::LvlE::Error << "SSL container not created" << stl::endl;
    return IOStatusE::IOStatusE_AllocError;
  }

  int status = SSL_set_fd(SSLContainer::cast(sslContainer.get())->m_ssl.get(), session.socket().fd());
  if(1 != status)
  {
    LOGEXT << logging::LvlE::Error << "Could not set fd for SSL" << stl::endl;
    return IOStatusE::IOStatusE_SSLError;
  }
  
  // openssl s_client -tls1_3 -cipher
  // "ECDHE-ECDSA-AES256-GCM-SHA384:@SECLEVEL=0" -connect 127.0.0.1:25434
  status = SSL_set_ciphersuites(
    SSLContainer::cast(sslContainer.get())->m_ssl.get(),
    "TLS_AES_256_GCM_SHA384:"
    "TLS_CHACHA20_POLY1305_SHA256:"
    "TLS_AES_128_CCM_SHA256:"
    "TLS_AES_128_CCM_8_SHA256:"
    "TLS_AES_128_GCM_SHA256"); // ADH-AES128-SHA:ADH-AES128-SHA256:ADH-AES256-SHA256
  // int status =
  // SSL_set_ciphersuites(SSLContainer::cast(sslContainer.get())->m_ssl.get(),
  // "TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256:TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384:TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA256:TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA384:TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256:TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384:TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA256:TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA384:TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA256:TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA384:TLS_DHE_RSA_WITH_AES_128_GCM_SHA256:TLS_DHE_RSA_WITH_AES_256_GCM_SHA384:TLS_DHE_RSA_WITH_AES_128_CBC_SHA:TLS_DHE_RSA_WITH_AES_256_CBC_SHA:TLS_DHE_RSA_WITH_AES_128_CBC_SHA256:TLS_DHE_RSA_WITH_AES_256_CBC_SHA256:TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305_SHA256:TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305:TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305_SHA256:TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305");
  // //ADH-AES128-SHA:ADH-AES128-SHA256:ADH-AES256-SHA256

  SSLContainer::cast(sslContainer.get())
    ->setLastSSLError(status, __FILENAME__, __LINE__);
  if (1 != status) {
    SSL_shutdown(SSLContainer::cast(sslContainer.get())->m_ssl.get());
    SocketBase::shutdown(SockDirE::Both);
    SocketBase::close();
    LOGEXT << logging::LvlE::Error << "Could not set cipher suites" << stl::endl;
    return IOStatusE::IOStatusE_SSLCipherError;
  }

  // TODO: Remove
  const char* cipherList =
    SSL_get_cipher_list(SSLContainer::cast(sslContainer.get())->m_ssl.get(), 0);
  stl::cout << "[SSL Version]: "
            << SSL_get_version(
                 SSLContainer::cast(sslContainer.get())->m_ssl.get())
            << stl::endl;
  stl::cout << "[SSL Client Version] "
            << SSL_client_version(
                 SSLContainer::cast(sslContainer.get())->m_ssl.get())
            << stl::endl;
  stl::cout << "CpiherList: " << cipherList << stl::endl;

  /* X509Ptr cert =
    SSL_get_certificate(SSLContainer::cast(sslContainer.get())->m_ssl.get());
  if (cert) {
    X509_print_fp(stdout, cert.get());
  } */

  status = SSL_accept(SSLContainer::cast(sslContainer.get())->m_ssl.get());

  SSLContainer::cast(sslContainer.get())
    ->setLastSSLError(status, __FILENAME__, __LINE__);
  setSysError(__FILENAME__, __LINE__);

  if (status != 1) {
    // TODO: SSLContainer::cast(m_sslContainer.get())->m_lastSysError ? For
    // example 10054 connreset
    LOGEXT << logging::LvlE::Error << "SSL accept not successful" << stl::endl;
    return IOStatusE::IOStatusE_ConnectionClosed;
  }

  stl::UniquePtr<ClientSocketBase> clientSocketBase =
    stl::UniquePtr<ClientSocketBase>(
      stl::memory::MemoryManager<SSLClientSocketBase>::allocate(
        stl::move(session.socket()), stl::move(sslContainer)));

  if (!clientSocketBase) {
    SocketBase::shutdown(SockDirE::Both);
    SocketBase::close();
    LOGEXT << logging::LvlE::Error << "Client socket base not created" << stl::endl;
    return IOStatusE::IOStatusE_AllocError;
  }

  Session sslSession(stl::move(clientSocketBase),
                     NetworkAddress(session.remoteAddress()));

  session = stl::move(sslSession);
  LOGEXT << logging::LvlE::Success << "SSL handshake done" << stl::endl;

  return IOStatusE::IOStatusE_OK;
}
}

#endif

#endif // (defined CI_OS_WINDOWS && (CI_OS_WINDOWS == 1)) || (defined
       // CI_OS_LINUX && (CI_OS_LINUX == 1))
#endif // #if (defined STL_ENABLE_SSL) && (STL_ENABLE_SSL != 0)

#endif // defined STL_ENABLE_NETWORKING
