#include "stl/compile_info.hpp"
#include "stl/stl_config.hpp"

#if defined STL_ENABLE_NETWORKING && (STL_ENABLE_NETWORKING == 1)

#if defined CI_OS_WINDOWS && (CI_OS_WINDOWS == 1)

#pragma comment(lib, "ws2_32.lib")

#include "stl/assert.hpp"
#include "stl/compiler.hpp"
#include "stl/endian.hpp"
#include "stl/iostream.hpp"
#include "stl/limits.hpp"
#include "stl/memory.hpp"
#include "stl/network.hpp"
#include "stl/new.hpp"
#include "stl/stddef.hpp"
#include "stl/stdint.hpp"
#include "stl/string.hpp"

#include <WS2tcpip.h>
#include <Winsock2.h>

// TODO: Guard for SSL context and SSL to prevent memory leaks (guard with
// custom destructor template)

namespace network {
int _typeLUT[NETWORK_COMTYPE_LUT_SIZE]{ SOCK_STREAM, SOCK_DGRAM, SOCK_RAW };

int _protoLUT[NETWORK_IPPROTO_LUT_SIZE]{ IPPROTO_TCP,
                                         IPPROTO_UDP,
                                         IPPROTO_RAW };

int _dirLUT[NETWORK_SHDIR_LUT_SIZE]{ SD_RECEIVE, SD_SEND, SD_BOTH };

stl::UInt16T _afLUT[NETWORK_AF_LUT_SIZE]{ AF_INET, AF_INET6 };

using SocketT = SOCKET;

struct SocketContainer
{
  SocketContainer()
    : m_socket(INVALID_SOCKET)
  {
  }

  SocketContainer(SocketBase::SocketContainerT&& socketContainer)
  {
    m_socket = stl::exchange(
      reinterpret_cast<SocketContainer*>(socketContainer->m_data)->m_socket,
      INVALID_SOCKET);
  }

  SocketContainer(SocketContainer&& other) noexcept
  {
    m_socket = stl::exchange(other.m_socket, INVALID_SOCKET);
  }

  SocketContainer& operator=(SocketContainer&& other) noexcept
  {
    m_socket = stl::exchange(other.m_socket, INVALID_SOCKET);
    return *this;
  }

  SocketContainer(const SocketContainer&) = delete;
  SocketContainer& operator=(const SocketContainer) = delete;

  SocketContainer(const SocketT socket)
    : m_socket(socket)
  {
  }

  static SocketContainer* cast(void* const ptr)
  {
    return reinterpret_cast<SocketContainer*>(ptr);
  }

  static const SocketContainer* cast(const void* const ptr)
  {
    return reinterpret_cast<const SocketContainer*>(ptr);
  }

  SocketT m_socket;
};

static WSAData wsad = { 0 };

IOStatusE
init()
{
  if (0 != WSAStartup(MAKEWORD(2, 2), &wsad)) {
    // m_sysError = getNetworkSysError();

    return IOStatusE::IOStatusE_Error;
  }

#if (defined STL_ENABLE_SSL) && (STL_ENABLE_SSL != 0) &&                       \
  ((defined STL_ENABLE_SSL_SERVER && (STL_ENABLE_SSL_SERVER == 1)) ||          \
   (defined STL_ENABLE_SSL_CLIENT && (STL_ENABLE_SSL_CLIENT == 1)))
  return initSecurityLayer();
#else
  return IOStatusE::IOStatusE_OK;
#endif // #if (defined STL_ENABLE_SSL) && (STL_ENABLE_SSL != 0)
}

IOStatusE
deinit()
{
  if (0 != WSACleanup()) {
    // m_sysError = getNetworkSysError();

    return IOStatusE::IOStatusE_Error;
  }

  return IOStatusE::IOStatusE_OK;
}

//////////////////////////////// SocketBaseT ///////////////////////////////

SocketBase::SocketBase()
  : m_socketContainer(reinterpret_cast<SocketDataContainer*>(
      stl::memory::MemoryManager<SocketContainer>::allocate()))
  , m_ipVer(IPVerE::_Count)
  , m_comType(ComTypeE::_Count)
  , m_ipProto(IPProtoE::_Count)
  , m_activeDirections(0U)
  , m_lastSysError(0)
  , m_blocking(true)
{
}

SocketBase::SocketBase(SocketContainerT&& socketContainer)
  : m_socketContainer(stl::move(socketContainer))
  , m_ipVer(IPVerE::_Count)
  , m_comType(ComTypeE::_Count)
  , m_ipProto(IPProtoE::_Count)
  , m_activeDirections(0U)
  , m_lastSysError(0)
  , m_blocking(true)
{
  
}

SocketBase::SocketBase(SocketBase&& other) noexcept
{
  
  operator=(stl::move(other));
  
}

SocketBase&
SocketBase::operator=(SocketBase&& other) noexcept
{
  m_ipVer = other.m_ipVer;
  m_socketContainer = stl::move(other.m_socketContainer);
  m_lastSysError = other.m_lastSysError;
  m_comType = other.m_comType;
  m_ipProto = other.m_ipProto;
  m_activeDirections = other.m_activeDirections;
  m_blocking = other.m_blocking;

  return *this;
}

SharedSocketT
SocketBase::fd() const
{
  if (m_socketContainer) {
    return static_cast<SharedSocketT>(
      SocketContainer::cast(m_socketContainer.get())->m_socket);
  } else {
    return static_cast<SharedSocketT>(INVALID_SOCKET);
  }
}

IOStatusE
SocketBase::shutdownInternal(const SockDirE dir)
{
  Assert(m_socketContainer);

  if (!m_socketContainer) {
    return IOStatusE::IOStatusE_NotInitialized;
  }

  Assert(INVALID_SOCKET !=
         SocketContainer::cast(m_socketContainer.get())->m_socket);

  const int retv =
    ::shutdown(SocketContainer::cast(m_socketContainer.get())->m_socket,
               _dirLUT[static_cast<stl::SizeT>(dir)]);

  setSysError(__FILENAME__, __LINE__);

  if (0 == retv) {
    return IOStatusE::IOStatusE_OK;
  }

  return translateSysError();
}

IOStatusE
SocketBase::closeInternal()
{
  Assert(m_socketContainer);

  LOGEXT << logging::LvlE::Warning << "Socket closed" << stl::endl;

  if (!m_socketContainer) {
    return IOStatusE::IOStatusE_NotInitialized;
  }

  const int retv =
    ::closesocket(SocketContainer::cast(m_socketContainer.get())->m_socket);

  SocketContainer::cast(m_socketContainer.get())->m_socket = INVALID_SOCKET;

  setSysError(__FILENAME__, __LINE__);

  if (0 != retv) {
    return translateSysError();
  }

  return IOStatusE::IOStatusE_OK;
}

IOStatusE
SocketBase::bindInternal()
{
  Assert(m_socketContainer);

  if (!m_socketContainer) {
    return IOStatusE::IOStatusE_NotInitialized;
  }

  Assert(INVALID_SOCKET !=
         SocketContainer::cast(m_socketContainer.get())->m_socket);

  if (m_ipVer != m_boundAddress.getIPVer()) {
    return IOStatusE::IOStatusE_InconsistentAddress;
  }

  Assert(m_boundAddress.size() <
         static_cast<stl::SizeT>(stl::Limits<int>::max()));
  const int retv =
    ::bind(SocketContainer::cast(m_socketContainer.get())->m_socket,
           reinterpret_cast<const sockaddr*>(&m_boundAddress),
           static_cast<int>(m_boundAddress.size()));

  setSysError(__FILENAME__, __LINE__);

  if (0 == retv) {
    return IOStatusE::IOStatusE_OK;
  }

  return translateSysError();
}

IOStatusE
SocketBase::blocking(const bool blocking)
{
  Assert(m_socketContainer);

  if (!m_socketContainer) {
    return IOStatusE::IOStatusE_NotInitialized;
  }

  Assert(INVALID_SOCKET !=
         SocketContainer::cast(m_socketContainer.get())->m_socket);

  unsigned long arg = blocking ? 0 : 1;

  const int retv = ioctlsocket(
    SocketContainer::cast(m_socketContainer.get())->m_socket, FIONBIO, &arg);

  if (0 == retv) {
    m_blocking = blocking;
  }

  setSysError(__FILENAME__, __LINE__);

  return (retv != INVALID_SOCKET) ? translateSysError()
                                  : IOStatusE::IOStatusE_Error;
}

void
SocketBase::setSysError(const char* filename, const stl::UInt32T line)
{
  m_lastSysError = WSAGetLastError();

  if ((0 != m_lastSysError) && (WSAEWOULDBLOCK != m_lastSysError)) {
    char* s = NULL;
    FormatMessageA(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM |
                     FORMAT_MESSAGE_IGNORE_INSERTS,
                   NULL,
                   m_lastSysError,
                   MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                   (LPSTR)&s,
                   0,
                   NULL);

    stl::cout << s << stl::endl;

    LocalFree(s);
  }
}

IOStatusE
SocketBase::translateSysError() const
{
  switch (m_lastSysError) {
    case 0:
      return IOStatusE::IOStatusE_OK;
    case WSAEWOULDBLOCK:
      return IOStatusE::IOStatusE_OpWouldBlock;
    case WSAECONNRESET:
      return IOStatusE::IOStatusE_ConnectionForciblyClosed;
    default:
      break;
  }

  return IOStatusE::IOStatusE_Error;
}

bool
SocketBase::isOpen() const
{
  if (!m_socketContainer) {
    return false;
  }

  return SocketContainer::cast(m_socketContainer.get())->m_socket !=
         INVALID_SOCKET;
}

IOStatusE
SocketBase::readReady()
{
  Assert(m_socketContainer);

  if (!m_socketContainer) {
    return IOStatusE::IOStatusE_NotInitialized;
  }

  FD_SET readFDs;
  // FD_SET writeFDs;
  FD_SET exceptFDs;

  stl::memory::zero(&readFDs);
  // stl::memory::zero(&writeFDs);
  stl::memory::zero(&exceptFDs);

  FD_SET(SocketContainer::cast(m_socketContainer.get())->m_socket, &readFDs);
  // FD_SET(SocketContainer::cast(m_socketContainer.get())->m_socket
  //   , &writeFDs);
  FD_SET(SocketContainer::cast(m_socketContainer.get())->m_socket, &exceptFDs);

  timeval tm;

  tm.tv_sec = 0;
  tm.tv_usec = 100;

  const int retv = ::select(0, &readFDs, nullptr, &exceptFDs, &tm);

  setSysError(__FILENAME__, __LINE__);

  if (retv < 0) {
    return translateSysError();
  } else if (retv == 0) {
    return IOStatusE::IOStatusE_OpTimeout;
  } else {
    if (FD_ISSET(SocketContainer::cast(m_socketContainer.get())->m_socket,
                 &exceptFDs)) {
      // socket has inbound OOB data, or non-blocking connect() has failed, or
      // some other socket error, handle as needed...
    }

    if (FD_ISSET(SocketContainer::cast(m_socketContainer.get())->m_socket,
                 &readFDs)) {
      // socket has inbound data, or has disconnected gracefully, handle as
      // needed...
      return IOStatusE::IOStatusE_OK;
    }

    // if(FD_ISSET(SocketContainer::cast(m_socketContainer.get())->m_socket,
    // &writeFDs)) {
    //   // socket is writable, or non-blocking connect() has succeeded, handle
    //   as needed...
    // }

    return IOStatusE::IOStatusE_OpTimeout;
  }
}

IOStatusE
SocketBase::openInternal()
{
  Assert(static_cast<int>(m_ipVer) < (sizeof(_afLUT) / sizeof(_afLUT[0])));
  Assert(static_cast<int>(m_comType) <
         (sizeof(_typeLUT) / sizeof(_typeLUT[0])));
  Assert(static_cast<int>(m_ipProto) <
         (sizeof(_protoLUT) / sizeof(_protoLUT[0])));

  if (!m_socketContainer) {
    return IOStatusE::IOStatusE_NotInitialized;
  }

  if (INVALID_SOCKET !=
      SocketContainer::cast(m_socketContainer.get())->m_socket) {
    return IOStatusE::IOStatusE_SocketAlreadyExist;
  }

  SocketContainer::cast(m_socketContainer.get())->m_socket =
    ::socket(_afLUT[static_cast<int>(m_ipVer)],
             _typeLUT[static_cast<int>(m_comType)],
             _protoLUT[static_cast<int>(m_ipProto)]);

  setSysError(__FILENAME__, __LINE__);

  if (INVALID_SOCKET ==
      SocketContainer::cast(m_socketContainer.get())->m_socket) {
    return translateSysError();
  }

  return postOpenAction();
}

ClientSocketBase::ClientSocketBase()
  : SocketBase(){};

stl::SizeT
ClientSocketBase::available()
{
  Assert(m_socketContainer);

  if (!m_socketContainer) {
    return 0U;
  }

  Assert(INVALID_SOCKET !=
         SocketContainer::cast(m_socketContainer.get())->m_socket);

  // ioctl(fd, FIONREAD, &bytes_available)

  u_long bytesAvailable = 0U;

  const int status =
    ioctlsocket(SocketContainer::cast(m_socketContainer.get())->m_socket,
                FIONREAD,
                &bytesAvailable);

  setSysError(__FILENAME__, __LINE__);

  if (0 != status) {
    return 0U;
  }

  const IOStatusE translatedError = translateSysError();

  if (IOStatusE::IOStatusE_OK == translatedError) {
    if (bytesAvailable > stl::Limits<stl::SizeT>::max()) {
      return stl::Limits<stl::SizeT>::max();
    }

    return static_cast<stl::SizeT>(bytesAvailable);
  }

  return 0U;
}

IOStatusE
ClientSocketBase::send(const stl::UInt8T* const data,
                       const stl::SizeT size,
                       stl::SizeT& sent)
{
  Assert(nullptr != data);
  Assert(m_socketContainer);

  if (!m_socketContainer) {
    return IOStatusE::IOStatusE_NotInitialized;
  }

  Assert(INVALID_SOCKET !=
         SocketContainer::cast(m_socketContainer.get())->m_socket);

  if (size > static_cast<stl::SizeT>(stl::Limits<int>::max())) {
    return IOStatusE::IOStatusE_InvalidParam;
  }

  const int retv =
    ::send(SocketContainer::cast(m_socketContainer.get())->m_socket,
           reinterpret_cast<const char*>(data),
           static_cast<int>(size),
           0);

  setSysError(__FILENAME__, __LINE__);

  const IOStatusE translatedError = translateSysError();

  if (IOStatusE::IOStatusE_OK == translatedError) {
    sent = static_cast<stl::SizeT>(retv);
  }

  return IOStatusE::IOStatusE_OK;
}

IOStatusE
ClientSocketBase::receive(stl::UInt8T* const data,
                          const stl::SizeT maxSize,
                          stl::SizeT& received)
{
  Assert(nullptr != data);
  Assert(m_socketContainer);

  LOGEXT << logging::LvlE::Debug << "NOOOOOOOOOOOOOO" << stl::endl;

  if (!m_socketContainer) {
    return IOStatusE::IOStatusE_NotInitialized;
  }

  Assert(INVALID_SOCKET !=
         SocketContainer::cast(m_socketContainer.get())->m_socket);

  if (maxSize > static_cast<stl::SizeT>(stl::Limits<int>::max())) {
    return IOStatusE::IOStatusE_InvalidParam;
  }

  const int retv =
    ::recv(SocketContainer::cast(m_socketContainer.get())->m_socket,
           reinterpret_cast<char*>(data),
           static_cast<int>(maxSize),
           0);

  setSysError(__FILENAME__, __LINE__);

  const IOStatusE translatedError = translateSysError();

  if (0 == retv) {
    return IOStatusE::IOStatusE_ConnectionClosed;
  } else if (IOStatusE::IOStatusE_OK == translatedError) {
    received = static_cast<stl::SizeT>(retv);
  }

  return translatedError;
}

IOStatusE
ClientSocketBase::sendTo(const NetworkAddress& networkAddress,
                         const stl::UInt8T* const data,
                         const stl::SizeT size,
                         stl::SizeT& sent)
{
  Assert(nullptr != data);
  Assert(m_socketContainer);

  if (!m_socketContainer) {
    return IOStatusE::IOStatusE_NotInitialized;
  }

  Assert(INVALID_SOCKET !=
         SocketContainer::cast(m_socketContainer.get())->m_socket);

  if (size > static_cast<stl::SizeT>(stl::Limits<int>::max())) {
    return IOStatusE::IOStatusE_InvalidParam;
  }

  const int retv =
    ::sendto(SocketContainer::cast(m_socketContainer.get())->m_socket,
             reinterpret_cast<const char*>(data),
             static_cast<int>(size),
             MSG_DONTROUTE,
             reinterpret_cast<const sockaddr*>(networkAddress.sockAddr()),
             static_cast<int>(networkAddress.size()));

  setSysError(__FILENAME__, __LINE__);

  const IOStatusE translatedError = translateSysError();

  if (IOStatusE::IOStatusE_OK == translatedError) {
    sent = static_cast<stl::SizeT>(retv);
  }

  return IOStatusE::IOStatusE_OK;
}

IOStatusE
ClientSocketBase::receiveFrom(stl::UInt8T* const data,
                              const stl::SizeT maxSize,
                              NetworkAddress& networkAddress,
                              stl::SizeT& received)
{
  Assert(nullptr != data);
  Assert(m_socketContainer);

  if (!m_socketContainer) {
    return IOStatusE::IOStatusE_NotInitialized;
  }

  Assert(INVALID_SOCKET !=
         SocketContainer::cast(m_socketContainer.get())->m_socket);

  if (maxSize > static_cast<stl::SizeT>(stl::Limits<int>::max())) {
    return IOStatusE::IOStatusE_InvalidParam;
  }

  int addrLen = static_cast<int>(NetworkAddress::sizeMax());

  const int retv =
    ::recvfrom(SocketContainer::cast(m_socketContainer.get())->m_socket,
               reinterpret_cast<char*>(data),
               static_cast<int>(maxSize),
               0,
               reinterpret_cast<sockaddr*>(networkAddress.sockAddr()),
               &addrLen);

  setSysError(__FILENAME__, __LINE__);

  const IOStatusE translatedError = translateSysError();

  if (0 == retv) {
    return IOStatusE::IOStatusE_ConnectionClosed;
  } else if (IOStatusE::IOStatusE_OK == translatedError) {
    received = static_cast<stl::SizeT>(retv);
  }

  return translatedError;
}

IOStatusE
ClientSocket::connect(const NetworkAddress& networkAddress)
{
  switch (networkAddress.getIPVer()) {
    case IPVerE::IPv4:
      return connect(networkAddress.getAsIPv4(), networkAddress.getPort());
    case IPVerE::IPv6:
      return connect(networkAddress.getAsIPv6(), networkAddress.getPort());
    default:
      break;
  }

  return IOStatusE::IOStatusE_InvalidParam;
}

IOStatusE
ClientSocket::connect(const IPv4& ip, const PortT port)
{
  Assert(m_socketContainer);

  if (!m_socketContainer) {
    return IOStatusE::IOStatusE_NotInitialized;
  }

  Assert(INVALID_SOCKET !=
         SocketContainer::cast(m_socketContainer.get())->m_socket);

  NetworkAddress addr;

  static_cast<void>(addr.set(ip, port)); // It can fail only if parsing string

  const int retv =
    ::connect(SocketContainer::cast(m_socketContainer.get())->m_socket,
              reinterpret_cast<const sockaddr*>(addr.sockAddr()),
              static_cast<int>(addr.size()));

  setSysError(__FILENAME__, __LINE__);

  if (0 == retv) {
    return IOStatusE::IOStatusE_OK;
  }

  return translateSysError();
}

IOStatusE
ClientSocket::connect(const IPv6& ip, const PortT port)
{
  Assert(m_socketContainer);

  if (!m_socketContainer) {
    return IOStatusE::IOStatusE_NotInitialized;
  }

  Assert(INVALID_SOCKET !=
         SocketContainer::cast(m_socketContainer.get())->m_socket);

  NetworkAddress addr;

  static_cast<void>(addr.set(ip, port)); // It can fail only if parsing string

  const int retv =
    ::connect(SocketContainer::cast(m_socketContainer.get())->m_socket,
              reinterpret_cast<const sockaddr*>(addr.sockAddr()),
              static_cast<int>(addr.size()));

  setSysError(__FILENAME__, __LINE__);

  if (0 == retv) {
    return IOStatusE::IOStatusE_OK;
  }

  return translateSysError();
}

IOStatusE
ClientSocket::connect(const char* host, const PortT port)
{
  UNUSED_PARAM(host);
  UNUSED_PARAM(port);

  return IOStatusE::IOStatusE_UnsupportedFunction;
}

ServerSocket::ServerSocket()
{
}

IOStatusE
ServerSocket::listen(const stl::SizeT backlog)
{
  Assert(backlog <= static_cast<stl::SizeT>(SOMAXCONN));
  Assert(backlog <= stl::Limits<stl::SizeT>::max());
  Assert(m_socketContainer);

  if (!m_socketContainer) {
    return IOStatusE::IOStatusE_NotInitialized;
  }

  Assert(INVALID_SOCKET !=
         SocketContainer::cast(m_socketContainer.get())->m_socket);

  const int retv =
    ::listen(SocketContainer::cast(m_socketContainer.get())->m_socket,
             static_cast<int>(backlog));

  setSysError(__FILENAME__, __LINE__);

  if (0 == retv) {
    return IOStatusE::IOStatusE_OK;
  }

  return translateSysError();
}

IOStatusE
ServerSocket::accept(Session& newSession)
{
  Assert(m_socketContainer);

  NetworkAddress networkAddress;

  socklen_t addrLength = static_cast<socklen_t>(networkAddress.sizeMax());

  if (!m_socketContainer) {
    return IOStatusE::IOStatusE_NotInitialized;
  }

  const SocketT newSocket =
    ::accept(SocketContainer::cast(m_socketContainer.get())->m_socket,
             reinterpret_cast<sockaddr*>(&networkAddress),
             &addrLength);

  setSysError(__FILENAME__, __LINE__);

  if (INVALID_SOCKET == newSocket) {
    return translateSysError();
  }

  SocketContainerT newClientSocket = reinterpret_cast<SocketDataContainer*>(
    stl::memory::MemoryManager<SocketContainer>::allocate(newSocket));

  if (!newClientSocket) {
    SocketBase::shutdown(SockDirE::Both);
    SocketBase::close();
    return IOStatusE::IOStatusE_AllocError;
  }

  stl::UniquePtr<ClientSocketBase> clientSocketBase =
    stl::makeUnique<ClientSocketBase>(
      ClientSocketBase(stl::move(newClientSocket)));
  if (!clientSocketBase) {
    SocketBase::shutdown(SockDirE::Both);
    SocketBase::close();
    return IOStatusE::IOStatusE_AllocError;
  }

  newSession = Session(stl::move(clientSocketBase), stl::move(networkAddress));

  return IOStatusE::IOStatusE_OK;
}
}

#endif // defined CI_OS_WINDOWS && (CI_OS_WINDOWS == 1)

#endif // defined STL_ENABLE_NETWORKING
