#include "stl/ostream.hpp"

#include "stl/string.hpp"
#include "stl/string_conv.hpp"
#include "stl/string_utils.hpp"

namespace stl::stream {
static const char g_nullString[] = "<null>";
static const char g_trueString[] = "true";
static const char g_falseString[] = "false";
static const char g_eol[] = "\r\n";

template<>
OStreamBase&
OStreamBase::operator<<(const HexT&)
{
  m_integerMode = IntegerModeE::Hex;

  return *this;
}

template<>
OStreamBase&
OStreamBase::operator<<(const DecT&)
{
  m_integerMode = IntegerModeE::Dec;

  return *this;
}

template<>
OStreamBase&
OStreamBase::operator<<(const OctT&)
{
  m_integerMode = IntegerModeE::Oct;

  return *this;
}

template<>
OStreamBase&
OStreamBase::operator<<(const Precision& floatPrecision)
{
  m_floatPrecision = floatPrecision.m_floatPrecision;

  return *this;
}

template <>
OStreamBase& OStreamBase::operator<<(const EndLineT&)
{
    this->write(reinterpret_cast<const UInt8T*>(g_eol), string::length(g_eol));

    m_floatPrecision = FLOAT_PRECISION_DEFAULT;
    m_integerMode = IntegerModeE::Dec;

    return *this;
}

template <>
OStreamBase& OStreamBase::operator<<(const std::string& string)
{
    this->write(reinterpret_cast<const UInt8T*>(string.c_str()), string.length());

    return *this;
}

OStreamBase& OStreamBase::write(const char* const str, SizeT length)
{
    Assert(nullptr != str);

    this->write(reinterpret_cast<const UInt8T*>(str), length);

    return *this;
}

OStreamBase& OStreamBase::write(char* const str, SizeT length)
{
    Assert(nullptr != str);

    this->write(reinterpret_cast<const UInt8T*>(str), length);

    return *this;
}

#if defined STL_ENABLE_FLOAT && (STL_ENABLE_FLOAT == 1)
template <>
OStreamBase& OStreamBase::operator<<(const float& value)
{
    const std::string valueString = std::to_string(value);

    write(valueString.c_str(), valueString.length());

    return *this;
}
#endif

#if defined STL_ENABLE_DOUBLE && (STL_ENABLE_DOUBLE == 1)
template <>
OStreamBase& OStreamBase::operator<<(const double& value)
{
    const std::string valueString = std::to_string(value);

    write(valueString.c_str(), valueString.length());

    return *this;
}
#endif

#if defined STL_ENABLE_LONG_DOUBLE && (STL_ENABLE_LONG_DOUBLE == 1)
template <>
OStreamBase& OStreamBase::operator<<(const long double& value)
{
    const std::string valueString = std::to_string(value);

    write(valueString.c_str(), valueString.length());

    return *this;
}
#endif

template<>
OStreamBase&
OStreamBase::operator<<(char* const value)
{
  if (nullptr == value) {
    write(reinterpret_cast<const UInt8T*>(g_nullString),
          static_cast<SizeT>(sizeof(g_nullString)));
  } else {
    write(reinterpret_cast<const UInt8T*>(value), string::length(value));
  }

  return *this;
}

template<>
OStreamBase&
OStreamBase::operator<<(const char* const value)
{
  if (nullptr == value) {
    write(reinterpret_cast<const UInt8T*>(g_nullString),
          static_cast<SizeT>(sizeof(g_nullString)));
  } else {
    write(reinterpret_cast<const UInt8T*>(value), string::length(value));
  }

  return *this;
}

template<>
OStreamBase&
OStreamBase::operator<<(const char& value)
{
  write(reinterpret_cast<const UInt8T*>(&value), static_cast<SizeT>(1U));

  return *this;
}

template<>
OStreamBase&
OStreamBase::operator<<(const bool& value)
{
  if (value) {
    write(reinterpret_cast<const UInt8T*>(g_trueString),
          static_cast<SizeT>(sizeof(g_trueString) - 1U));
  } else {
    write(reinterpret_cast<const UInt8T*>(g_falseString),
          static_cast<SizeT>(sizeof(g_falseString) - 1U));
  }

  return *this;
}

template<>
OStreamBase&
OStreamBase::operator<<(const unsigned char& value)
{
  String valueString = string::toString(value);

  write(reinterpret_cast<const UInt8T*>(valueString.c_str()),
        valueString.length());

  return *this;
}

template<>
OStreamBase&
OStreamBase::operator<<(const short& value)
{
  String valueString = string::toString(value);

  write(reinterpret_cast<const UInt8T*>(valueString.c_str()),
        valueString.length());

  return *this;
}

template<>
OStreamBase&
OStreamBase::operator<<(const unsigned short& value)
{
  String valueString = string::toString(value);

  write(reinterpret_cast<const UInt8T*>(valueString.c_str()),
        valueString.length());

  return *this;
}

template<>
OStreamBase&
OStreamBase::operator<<(const int& value)
{
  String valueString = string::toString(value);

  write(reinterpret_cast<const UInt8T*>(valueString.c_str()),
        valueString.length());

  return *this;
}

template<>
OStreamBase&
OStreamBase::operator<<(const unsigned int& value)
{
  String valueString = string::toString(value);

  write(reinterpret_cast<const UInt8T*>(valueString.c_str()),
        valueString.length());

  return *this;
}

template<>
OStreamBase&
OStreamBase::operator<<(const long& value)
{
  String valueString = string::toString(value);

  write(reinterpret_cast<const UInt8T*>(valueString.c_str()),
        valueString.length());

  return *this;
}

template<>
OStreamBase&
OStreamBase::operator<<(const unsigned long& value)
{
  String valueString = string::toString(value);

  write(reinterpret_cast<const UInt8T*>(valueString.c_str()),
        valueString.length());

  return *this;
}

#if defined STL_ENABLE_INT64 && (STL_ENABLE_INT64 == 1)
template<>
OStreamBase&
OStreamBase::operator<<(const long long& value)
{
  String valueString = string::toString(value);

  write(reinterpret_cast<const UInt8T*>(valueString.c_str()),
        valueString.length());

  return *this;
}

template<>
OStreamBase&
OStreamBase::operator<<(const unsigned long long& value)
{
  String valueString = string::toString(value);

  write(reinterpret_cast<const UInt8T*>(valueString.c_str()),
        valueString.length());

  return *this;
}
#endif

void
OStreamBase::dump(const UInt8T* data,
                  const SizeT size,
                  const SizeT bytesPerLine,
                  const bool upperCase)
{
  *this << "Data dump [" << size << "]:" << endl;

  for (SizeT i = 0U; i < size;) {
    const SizeT toPrint = math::min(bytesPerLine, size - i);

    for (SizeT h = 0U; h < toPrint; ++h) {
      *this << string::nibbleToHex(data[h] >> 4, upperCase)
            << string::nibbleToHex(data[h], upperCase);

      if ((h + 1U) != toPrint) {
        *this << ' ';
      }
    }

    fill(' ', (bytesPerLine - toPrint) * 3U);

    *this << " | ";

    for (SizeT a = 0U; a < toPrint; ++a) {
      *this << (string::isPrint(data[a]) ? static_cast<char>(data[a]) : '.');
    }

    *this << endl;

    i += toPrint;
    data += toPrint;
  }
}

void
OStreamBase::fill(const char c, const SizeT count)
{
  for (SizeT i = 0U; i < count; ++i) { // TODO: Optimize
    logging::log << c;
  }
}
}