#include "stl/string.hpp"

#include "stl/iostream.hpp"

#ifdef STL_ENABLE_IOSTREAM
namespace stl {
template<>
stream::OStreamBase&
stream::OStreamBase::operator<<(const stl::String& string)
{
  write(string.c_str(), string.length());

  return *this;
}
}

stl::stream::OStreamBase&
operator<<(stl::stream::OStreamBase& stream, const stl::String& string)
{
  stream << string.c_str();

  return stream;
}
#endif

namespace stl {
String::String()
  : m_ptr(makeUnique<char[]>(1U))
  , m_size(0U)
  , m_reserved(0U)
{
  Assert(m_ptr);
  m_ptr.get()[0] = '\0';
}

String::String(const String& other)
{
  *this = other;
}

String::String(String&& other) noexcept
{
  *this = move(other);

  Assert(other.m_ptr);
  Assert(m_ptr);
}

String::String(const char* const str)
  : m_ptr(nullptr)
  , m_size(0U)
  , m_reserved(0U)
{
  *this = str;
}

String::String(const char* const str, const SizeT length)
  : m_ptr(nullptr)
  , m_size(0U)
  , m_reserved(0U)
{
  Assert(nullptr != str);
  m_ptr = makeUnique<char[]>(length + 1U);
  Assert(nullptr != m_ptr.get());
  static_cast<void>(memory::copy(str, str + length, m_ptr.get()));
  m_ptr.get()[length] = '\0';
  m_size = length;
  m_reserved = m_size;
}

String::String(const char* const str, const SizeT offset, const SizeT length)
  : m_ptr(nullptr)
  , m_size(0U)
  , m_reserved(0U)
{
  Assert(nullptr != str);
  m_ptr = makeUnique<char[]>(length + 1U);
  Assert(nullptr != m_ptr.get());
  static_cast<void>(
    memory::copy(str + offset, str + offset + length, m_ptr.get()));
  m_ptr.get()[length] = '\0';
  m_size = length;
  m_reserved = m_size;
}

String::String(const char c, const SizeT count)
  : m_ptr(nullptr)
  , m_size(0U)
  , m_reserved(0U)
{
  m_ptr = makeUnique<char[]>(count + 1U);
  Assert(nullptr != m_ptr.get());
  static_cast<void>(memory::fill(c, count, m_ptr.get()));
  m_ptr.get()[count] = '\0';
  m_size = count;
  m_reserved = m_size;
}

// Getters

const SizeT
String::reserved() const noexcept
{
  return m_reserved;
}

const SizeT
String::size() const noexcept
{
  return m_size;
}

const SizeT
String::length() const noexcept
{
  return m_size;
}

bool
String::empty() const noexcept
{
  return m_size == 0U;
}

const char*
String::c_str() const
{
  return m_ptr.get();
}

// Allocation methods

void
String::resize(const SizeT newSize)
{
  UniquePtr<char[]> temp = makeUnique<char[]>(newSize + 1U);

  if (newSize <= m_size) {
    memory::copy(m_ptr.get(), m_ptr.get() + newSize, temp.get());
    temp.get()[newSize] = '\0';
    m_ptr = move(temp);
    m_size = newSize;
    m_reserved = m_size;
  } else {
    memory::copy(m_ptr.get(), m_ptr.get() + m_size, temp.get());
    memory::fill('\0', newSize - m_size, temp.get());
    m_ptr = move(temp);
    m_size = newSize;
    m_reserved = m_size;
  }

  m_ptr = move(temp);
}

void
String::reserve(const SizeT toReserve)
{
  if (toReserve > m_reserved) {
    UniquePtr<char[]> temp = makeUnique<char[]>(toReserve + 1U);

    memory::copy(m_ptr.get(), m_ptr.get() + m_size, temp.get());

    m_ptr = move(temp);

    m_reserved = toReserve;
  }
}

// Growth methods

// Main implementation of string growth
void
String::insert(const SizeT lOffset,
               const char* what,
               const SizeT rOffset,
               const SizeT length)
{
  Assert(nullptr != what);
  Assert(lOffset <= m_size);
  if (length > 0U) {
    const SizeT newSize = m_size + length;

    UniquePtr<char[]> temp = makeUnique<char[]>(newSize + 1U);

    memory::copy(m_ptr.get(), m_ptr.get() + lOffset, temp.get());
    memory::copy(what + rOffset, what + rOffset + length, temp.get() + lOffset);
    memory::copy(m_ptr.get() + lOffset,
                 m_ptr.get() + m_size,
                 temp.get() + lOffset + length);
    temp.get()[newSize] = '\0';
    m_ptr = move(temp);
    m_size = newSize;
    m_reserved = m_size;
  }
}

void
String::insert(const char* what, const SizeT rOffset, const SizeT length)
{
  Assert(nullptr != what);
  insert(0U, what, rOffset, length);
}

void
String::insert(const char* what, const SizeT length)
{
  Assert(nullptr != what);
  insert(0U, what, 0U, length);
}

void
String::insert(const char* what)
{
  const SizeT length = string::length(what);
  insert(0U, what, 0U, length);
}

void
String::insert(const SizeT lOffset,
               const String& what,
               const SizeT rOffset,
               const SizeT length)
{
  Assert((rOffset + length) <= what.length());

  insert(lOffset, what.c_str(), rOffset, length);
}

void
String::insert(const String& what, const SizeT rOffset, const SizeT length)
{
  insert(0U, what, rOffset, length);
}

void
String::insert(const String& what, const SizeT length)
{
  insert(0U, what, 0U, length);
}

void
String::insert(const String& what)
{
  insert(0U, what, 0U, what.length());
}

void
String::append(const char* what, const SizeT rOffset, const SizeT length)
{
  Assert(nullptr != what);
  insert(m_size, what, rOffset, length);
}

void
String::append(const char* what, const SizeT length)
{
  Assert(nullptr != what);
  append(what, 0U, length);
}

void
String::append(const char* what)
{
  Assert(nullptr != what);
  const SizeT length = string::length(what);
  append(what, 0U, length);
}

void
String::append(const char what)
{
  append(&what, 0U, 1U);
}

void
String::append(const String& what, const SizeT rOffset, const SizeT length)
{
  insert(m_size, what, rOffset, length);
}

void
String::append(const String& what, const SizeT length)
{
  append(what, 0U, length);
}

void
String::append(const String& what)
{
  append(what, 0U, what.length());
}

void
String::append(String&& what)
{
  append(what, 0U, what.length());

  Assert(what.m_ptr);
  Assert(m_ptr);
}

// Shrink methods

void
String::remove(const SizeT offset, const SizeT length)
{
  Assert((offset + length) <= m_size);

  const SizeT newSize = m_size - length;

  UniquePtr<char[]> temp = makeUnique<char[]>(newSize + 1U);
  memory::copy(m_ptr.get(), m_ptr.get() + offset, temp.get());
  memory::copy(
    m_ptr.get() + offset + length, m_ptr.get() + m_size, temp.get() + offset);
  temp.get()[newSize] = '\0';
  m_ptr = move(temp);
  m_size = newSize;
  m_reserved = m_size;
}

bool
String::clear()
{
  m_size = 0U;
  m_reserved = 0U;

  m_ptr = makeUnique<char[]>(1U);
  if (m_ptr) {
    m_ptr.get()[0] = '\0';
    return true;
  }
  return false;
}

// Operators

String&
String::operator=(const String& other)
{
  m_ptr = makeUnique<char[]>(other.m_size + 1U);
  m_size = other.m_size;
  m_reserved = other.m_reserved;
  Assert(m_ptr);
  memory::copy(
    other.m_ptr.get(), other.m_ptr.get() + other.m_size + 1U, m_ptr.get());
  return *this;
}

String&
String::operator=(String&& other)
{
  m_ptr = move(other.m_ptr);
  m_size = other.m_size;
  m_reserved = other.m_reserved;

  other.clear();

  Assert(other.m_ptr);
  Assert(m_ptr);

  return *this;
}

String&
String::operator=(const char* const str)
{
  Assert(nullptr != str);
  const SizeT length = string::length(str);
  m_ptr = makeUnique<char[]>(length + 1U);
  static_cast<void>(memory::copy(str, str + length, m_ptr.get()));
  m_ptr.get()[length] = '\0';
  m_size = length;
  m_reserved = m_size;

  return *this;
}

String
String::operator+(const char c) const
{
  String output;
  output.reserve(m_size + 1U);
  output.append(m_ptr.get(), m_size);
  output += c;

  return output;
}

String&
String::operator+=(const char c)
{
  append(c);
  return *this;
}

String
String::operator+(const char* str) const
{
  const SizeT length = string::length(str);

  String output;
  output.reserve(m_size + length);
  output.append(m_ptr.get(), m_size);
  output += str;

  return output;
}

String&
String::operator+=(const char* str)
{
  Assert(nullptr != str);
  append(str);
  return *this;
}

String
String::operator+(const String& string) const
{
  String output;
  output.reserve(m_size + string.length() + 1U);
  output += m_ptr.get();
  output += string;

  return output;
}

String&
String::operator+=(const String& string)
{
  append(string);
  return *this;
}

String
String::operator+(String&& string) const
{
  String output;
  output.reserve(m_size + string.length());
  output += m_ptr.get();
  output += move(string);

  Assert(string.m_ptr);
  Assert(m_ptr);

  return output;
}

String&
String::operator+=(String&& string)
{
  append(move(string));

  Assert(string.m_ptr);
  Assert(m_ptr);

  return *this;
}

char&
String::operator[](const SizeT index)
{
  Assert(index < m_size);
  Assert(m_ptr.get());
  return m_ptr.get()[index];
}

const char&
String::operator[](const SizeT index) const
{
  Assert(index < m_size);
  Assert(m_ptr.get());
  return m_ptr.get()[index];
}

// Utility methods

int
String::csCompare(const char* str) const
{
  Assert(nullptr != str);

  const char* localPtr = m_ptr.get();

  do {
    if (*str == '\0' || *localPtr == '\0' || *str != *localPtr) {
      return *localPtr - *str;
    }

    ++localPtr, ++str;
  } while (true);

  return 0;
}

int
String::csCompare(const SizeT offset, const char* str) const
{
  Assert(nullptr != str);
  Assert(offset <= length());

  const char* localPtr = m_ptr.get() + offset;

  do {
    if (*str == '\0' || *localPtr == '\0' || *str != *localPtr) {
      return *localPtr - *str;
    }

    ++localPtr, ++str;
  } while (true);

  return 0;
}

int
String::csCompare(const SizeT offset, const char* str, SizeT toCompare) const
{
  Assert(nullptr != str);
  Assert(offset <= length());

  const char* localPtr = m_ptr.get() + offset;

  while (toCompare--) {
    if (*str == '\0' || *localPtr == '\0' || *str != *localPtr) {
      return *localPtr - *str;
    }

    ++localPtr, ++str;
  }

  return 0;
}

int
String::ciCompare(const char* str) const
{
  Assert(nullptr != str);

  const char* localPtr = m_ptr.get();

  do {
    if (*str == '\0' || *localPtr == '\0' ||
        (string::toLower(*str) != string::toLower(*localPtr))) {
      return *localPtr - *str;
    }

    ++localPtr, ++str;
  } while (true);

  return 0;
}

int
String::ciCompare(const SizeT offset, const char* str) const
{
  Assert(nullptr != str);
  Assert(offset <= length());

  const char* localPtr = m_ptr.get() + offset;

  do {
    if (*str == '\0' || *localPtr == '\0' ||
        (string::toLower(*str) != string::toLower(*localPtr))) {
      return *localPtr - *str;
    }

    ++localPtr, ++str;
  } while (true);

  return 0;
}

int
String::ciCompare(const SizeT offset, const char* str, SizeT toCompare) const
{
  Assert(nullptr != str);
  Assert(offset <= length());

  const char* localPtr = m_ptr.get() + offset;

  while (toCompare--) {
    if (*str == '\0' || *localPtr == '\0' ||
        (string::toLower(*str) != string::toLower(*localPtr))) {
      return *localPtr - *str;
    }

    ++localPtr, ++str;
  }

  return 0;
}

int
String::csCompare(const String& str) const
{
  return csCompare(str.c_str());
}
int
String::csCompare(const SizeT offset, const String& str) const
{
  return csCompare(offset, str.c_str());
}
int
String::csCompare(const SizeT offset,
                  const String& str,
                  const SizeT toCompare) const
{
  return csCompare(offset, str.c_str(), toCompare);
}

int
String::ciCompare(const String& str) const
{
  return ciCompare(str.c_str());
}
int
String::ciCompare(const SizeT offset, const String& str) const
{
  return ciCompare(offset, str.c_str());
}
int
String::ciCompare(const SizeT offset,
                  const String& str,
                  const SizeT toCompare) const
{
  return ciCompare(offset, str.c_str(), toCompare);
}

String
String::substr(const SizeT offset, const SizeT length) const
{
  Assert((offset + length) <= m_size);
  return String(c_str(), offset, length);
}

String
String::substr(const SizeT offset) const
{
  Assert(offset <= m_size);
  return String(c_str(), offset, length() - offset);
}

SizeT
String::find(const char c) const
{
  for (stl::SizeT i = 0U; i < m_size; ++i) {
    if (m_ptr.get()[i] == c) {
      return i;
    }
  }

  return nPos;
}

SizeT
String::find(const char* str) const
{
  return csFind(str);
}

SizeT
String::csFind(const char* str, const SizeT length) const
{
  Assert(nullptr != str);

  const char* localPtr = m_ptr.get();
  const char* localEndPtr = m_ptr.get() + m_size;
  const char* foundStrBeginPtr = nullptr;
  const char* strPtr = str;

  if (0U == length) {
    return nPos;
  }

  while (localPtr <= localEndPtr) {
    if (*strPtr == *localPtr) {
      if (nullptr == foundStrBeginPtr) {
        foundStrBeginPtr = localPtr;
      }

      ++strPtr;

      if (static_cast<SizeT>(strPtr - str) == length) {
        return foundStrBeginPtr - m_ptr.get();
      }
    } else {
      if (nullptr != foundStrBeginPtr) {
        if ('\0' == *strPtr) {
          return foundStrBeginPtr - m_ptr.get();
        }

        localPtr = foundStrBeginPtr + 1;
        strPtr = str;
        foundStrBeginPtr = nullptr;
      }
    }

    ++localPtr;
  }

  return nPos;
}

SizeT
String::ciFind(const char* str, const SizeT length) const
{
  Assert(nullptr != str);

  const char* localPtr = m_ptr.get();
  const char* localEndPtr = m_ptr.get() + m_size;
  const char* foundStrBeginPtr = nullptr;
  const char* strPtr = str;

  if (0U == length) {
    return nPos;
  }

  while (localPtr <= localEndPtr) {
    if (string::toLower(*strPtr) == string::toLower(*localPtr)) {
      if (nullptr == foundStrBeginPtr) {
        foundStrBeginPtr = localPtr;
      }

      ++strPtr;

      if (static_cast<SizeT>(strPtr - str) == length) {
        return foundStrBeginPtr - m_ptr.get();
      }
    } else {
      if (nullptr != foundStrBeginPtr) {
        if ('\0' == *strPtr) {
          return foundStrBeginPtr - m_ptr.get();
        }

        localPtr = foundStrBeginPtr + 1;
        strPtr = str;
        foundStrBeginPtr = nullptr;
      }
    }

    ++localPtr;
  }

  return nPos;
}

SizeT
String::csFind(const char* str) const
{
  Assert(nullptr != str);

  const char* localPtr = m_ptr.get();
  const char* localEndPtr = m_ptr.get() + m_size;
  const char* foundStrBeginPtr = nullptr;
  const char* strPtr = str;

  while (localPtr <= localEndPtr) {
    if (*strPtr == *localPtr) {
      if (nullptr == foundStrBeginPtr) {
        foundStrBeginPtr = localPtr;
      }

      ++strPtr;
    } else {
      if (nullptr != foundStrBeginPtr) {
        if ('\0' == *strPtr) {
          return foundStrBeginPtr - m_ptr.get();
        }

        localPtr = foundStrBeginPtr + 1;
        strPtr = str;
        foundStrBeginPtr = nullptr;
      }
    }

    ++localPtr;
  }

  return nPos;
}

SizeT
String::ciFind(const char* str) const
{
  Assert(nullptr != str);

  const char* localPtr = m_ptr.get();
  const char* localEndPtr = m_ptr.get() + m_size;
  const char* foundStrBeginPtr = nullptr;
  const char* strPtr = str;

  while (localPtr <= localEndPtr) {
    if (string::toLower(*strPtr) == string::toLower(*localPtr)) {
      if (nullptr == foundStrBeginPtr) {
        foundStrBeginPtr = localPtr;
      }

      ++strPtr;
    } else {
      if (nullptr != foundStrBeginPtr) {
        if ('\0' == *strPtr) {
          return foundStrBeginPtr - m_ptr.get();
        }

        localPtr = foundStrBeginPtr + 1;
        strPtr = str;
        foundStrBeginPtr = nullptr;
      }
    }

    ++localPtr;
  }

  return nPos;
}

SizeT
String::ciFind(const String& str, const SizeT length) const
{
  return ciFind(str.c_str(), length);
}

SizeT
String::ciFind(const String& str) const
{
  return ciFind(str.c_str(), str.length());
}

SizeT
String::replace(const char what, const char with)
{
  SizeT count = 0U;

  for (SizeT i = 0U; i < m_size; ++i) {
    if (m_ptr.get()[i] == what) {
      m_ptr.get()[i] = with;
      ++count;
    }
  }

  return count;
}

bool
String::operator==(const String& other) const
{
  return 0 == csCompare(other);
}

bool
String::operator!=(const String& other) const
{
  return !operator==(other);
}

bool
String::operator>(const String& other) const
{
  const char* ptrThis = m_ptr.get();
  const char* ptrOther = other.m_ptr.get();

  while (*ptrThis && *ptrOther) {
    if (*ptrThis > *ptrOther) {
      return true;
    } else if (*ptrThis < *ptrOther) {
      return false;
    }

    ++ptrThis, ++ptrOther;
  }

  return false;
}

bool
String::operator>=(const String& other) const
{
  const char* ptrThis = m_ptr.get();
  const char* ptrOther = other.m_ptr.get();

  while (*ptrThis && *ptrOther) {
    if (*ptrThis > *ptrOther) {
      return true;
    } else if (*ptrThis < *ptrOther) {
      return false;
    }

    ++ptrThis, ++ptrOther;
  }

  return length() == other.length();
}

bool
String::operator<(const String& other) const
{
  const char* ptrThis = m_ptr.get();
  const char* ptrOther = other.m_ptr.get();

  while (*ptrThis && *ptrOther) {
    if (*ptrThis < *ptrOther) {
      return true;
    } else if (*ptrThis > *ptrOther) {
      return false;
    }

    ++ptrThis, ++ptrOther;
  }

  return false;
}

bool
String::operator<=(const String& other) const
{
  const char* ptrThis = m_ptr.get();
  const char* ptrOther = other.m_ptr.get();

  while (*ptrThis && *ptrOther) {
    if (*ptrThis < *ptrOther) {
      return true;
    } else if (*ptrThis > *ptrOther) {
      return false;
    }

    ++ptrThis, ++ptrOther;
  }

  return length() == other.length();
}
}
