#include "stl/time.hpp"

#include "stl/assert.hpp"
#include "stl/limits.hpp"
#include "stl/string_conv.hpp"

#include "log/logging.hpp"

#include <time.h>

#define LEAPOCH (946684800LL + 86400 * (31 + 29))

#define DAYS_PER_400Y (365 * 400 + 97)
#define DAYS_PER_100Y (365 * 100 + 24)
#define DAYS_PER_4Y (365 * 4 + 1)

namespace stl::time {
TimeValue::TimeValue()
  : m_timeValue(static_cast<STimeT>(0))
{
}

TimeValue::TimeValue(const TimeValue& other)
  : m_timeValue(other.m_timeValue)
{
}

TimeValue&
TimeValue::operator=(const TimeValue& other)
{
  m_timeValue = other.m_timeValue;

  return *this;
}

TimeValue::TimeValue(TimeValue&& other)
  : m_timeValue(other.m_timeValue)
{
}

TimeValue&
TimeValue::operator=(TimeValue&& other)
{
  m_timeValue = other.m_timeValue;

  return *this;
}

TimeValue&
TimeValue::operator=(const Nanos& value)
{
  m_timeValue = value.m_timeValue;
  return *this;
}

TimeValue&
TimeValue::operator=(const Micros& value)
{
  m_timeValue = value.m_timeValue;
  return *this;
}

TimeValue&
TimeValue::operator=(const Millis& value)
{
  m_timeValue = value.m_timeValue;
  return *this;
}

TimeValue&
TimeValue::operator=(const Sec& value)
{
  m_timeValue = value.m_timeValue;
  return *this;
}

TimeValue&
TimeValue::operator=(const Mins& value)
{
  m_timeValue = value.m_timeValue;
  return *this;
}

TimeValue&
TimeValue::operator=(const Hours& value)
{
  m_timeValue = value.m_timeValue;
  return *this;
}

TimeValue::TimeValue(STimeT value)
  : m_timeValue(value)
{
}

TimeValue
TimeValue::operator+(const TimeValue& other) const
{
  return m_timeValue + other.m_timeValue;
}

TimeValue
TimeValue::operator-(const TimeValue& other) const
{
  return m_timeValue - other.m_timeValue;
}

bool
TimeValue::operator==(const TimeValue& other) const
{
  return m_timeValue == other.m_timeValue;
}

bool
TimeValue::operator!=(const TimeValue& other) const
{
  return m_timeValue != other.m_timeValue;
}

bool
TimeValue::operator>(const TimeValue& other) const
{
  return m_timeValue > other.m_timeValue;
}

bool
TimeValue::operator<(const TimeValue& other) const
{
  return m_timeValue < other.m_timeValue;
}

bool
TimeValue::operator>=(const TimeValue& other) const
{
  return m_timeValue >= other.m_timeValue;
}

bool
TimeValue::operator<=(const TimeValue& other) const
{
  return m_timeValue <= other.m_timeValue;
}

Nanos::Nanos(STimeT value)
  : TimeValue(value)
{
}

Nanos::Nanos(const TimeValue& timeValue)
  : TimeValue(timeValue)
{
}

STimeT
Nanos::value() const
{
  return m_timeValue / MULTIPLIER;
}

Micros::Micros(const TimeValue& timeValue)
  : TimeValue(timeValue)
{
}

Micros::Micros(STimeT value)
  : TimeValue(value * MULTIPLIER)
{
}

STimeT
Micros::value() const
{
  return m_timeValue / MULTIPLIER;
}

Millis::Millis(const TimeValue& timeValue)
  : TimeValue(timeValue)
{
}

Millis::Millis(STimeT value)
  : TimeValue(value * MULTIPLIER)
{
}

STimeT
Millis::value() const
{
  return m_timeValue / MULTIPLIER;
}

Sec::Sec(const TimeValue& timeValue)
  : TimeValue(timeValue)
{
}

Sec::Sec(STimeT value)
  : TimeValue(value * MULTIPLIER)
{
}

STimeT
Sec::value() const
{
  return m_timeValue / MULTIPLIER;
}

Mins::Mins(const TimeValue& timeValue)
  : TimeValue(timeValue)
{
}

Mins::Mins(STimeT value)
  : TimeValue(value * MULTIPLIER)
{
}

STimeT
Mins::value() const
{
  return m_timeValue / MULTIPLIER;
}

Hours::Hours(const TimeValue& timeValue)
  : TimeValue(timeValue)
{
}

Hours::Hours(STimeT value)
  : TimeValue(value * MULTIPLIER)
{
}

STimeT
Hours::value() const
{
  return m_timeValue / MULTIPLIER;
}

DateTime::DateTime(Millis utcTimestamp, const UInt32T timeZone)
  : m_utcTimestamp(utcTimestamp.value())
  , m_timeZone(timeZone)
{
  const STimeT utcTimeMs = utcTimeZoneToMs(timeZone);

  Assert((utcTimestamp.value() == 0) ||
         (utcTimestamp.value() >= math::abs(utcTimeMs)));

  if ((static_cast<STimeT>(0) == utcTimestamp.value()) ||
      (utcTimeMs < static_cast<STimeT>(-43200)) ||
      (utcTimeMs > static_cast<STimeT>(50400))) {
    m_valid = false;
  } else {
    m_valid = utcTimestampToTime(utcTimestamp.value() + utcTimeMs, *this);
  }
}

STimeT
DateTime::utcTimeZoneToMs(const UInt32T timeZone)
{
  static STimeT lut[] = {
    -43200, // 12:00
    -39600, // 11:00
    -36000, // 10:00
    -34200, // 09:30
    -32400, // 09:00
    -28800, // 08:00
    -25200, // 07:00
    -21600, // 06:00
    -18000, // 05:00
    -14400, // 04:00
    -12600, // 03:30
    -9800,  // 03:00
    -6200,  // 02:00
    -3600,  // 01:00
    0,      // 00:00
    3600,   // 01:00
    7200,   // 02:00
    10800,  // 03:00
    12600,  // 03:30
    14400,  // 04:00
    16200,  // 04:30
    18000,  // 05:00
    19800,  // 05:30
    20700,  // 05:45
    21600,  // 06:00
    23400,  // 06:30
    25200,  // 07:00
    28800,  // 08:00
    31500,  // 08:45
    32400,  // 09:00
    34200,  // 09:30
    36000,  // 10:00
    37800,  // 10:30
    39600,  // 11:00
    43200,  // 12:00
    45900,  // 12:45
    46800,  // 13:00
    50400   // 14:00
  };

  static const SizeT lutItemCount = sizeof(lut) / sizeof(time::STimeT);

  if (timeZone > lutItemCount) {
    return 0;
  }

  return lut[timeZone];
}

// TODO: int __secs_to_tm(long long t, struct tm *tm)
bool
DateTime::utcTimestampToTime(const STimeT utcTimestamp, DateTime& destination)
{
  time_t unix_timestamp = utcTimestamp;
  struct tm tm_info;

  tm_info = *gmtime(&unix_timestamp);

  destination.m_hour = tm_info.tm_hour;
  destination.m_isdst = tm_info.tm_isdst != 0;
  destination.m_min = tm_info.tm_min;
  destination.m_month = tm_info.tm_mon + 1;
  destination.m_monthDay = tm_info.tm_mday;
  destination.m_sec = tm_info.tm_sec;
  destination.m_utcTimestamp = utcTimestamp;
  destination.m_valid = true;
  destination.m_weekDay = tm_info.tm_wday;
  destination.m_year = tm_info.tm_year + 1900;
  destination.m_yearDay = tm_info.tm_yday;

  /* STimeT days, secs;
  STimeT remdays, remsecs, remyears;
  STimeT qc_cycles, c_cycles, q_cycles;
  STimeT years, months;
  STimeT wday, yday, leap;

  static const char days_in_month[] = { 31, 30, 31, 30, 31, 31,
                                        30, 31, 30, 31, 31, 29 };

  // Reject time_t values whose year would overflow int
  // if (t < INT_MIN * 31622400LL || t > INT_MAX * 31622400LL)
  //     return false;

  secs = utcTimestamp - LEAPOCH;
  days = secs / 86400;
  remsecs = secs % 86400;
  if (remsecs < 0) {
    remsecs += 86400;
    days--;
  }

  wday = (3 + days) % 7;
  if (wday < 0)
    wday += 7;

  qc_cycles = days / DAYS_PER_400Y;
  remdays = days % DAYS_PER_400Y;
  if (remdays < 0) {
    remdays += DAYS_PER_400Y;
    qc_cycles--;
  }

  c_cycles = remdays / DAYS_PER_100Y;
  if (c_cycles == 4)
    c_cycles--;
  remdays -= c_cycles * DAYS_PER_100Y;

  q_cycles = remdays / DAYS_PER_4Y;
  if (q_cycles == 25)
    q_cycles--;
  remdays -= q_cycles * DAYS_PER_4Y;

  remyears = remdays / 365;
  if (remyears == 4)
    remyears--;
  remdays -= remyears * 365;

  leap = !remyears && (q_cycles || !c_cycles);
  yday = remdays + 31 + 28 + leap;
  if (yday >= 365 + leap)
    yday -= 365 + leap;

  years = remyears + 4 * q_cycles + 100 * c_cycles + 400 * qc_cycles;

  for (months = 0; days_in_month[months] <= remdays; months++)
    remdays -= days_in_month[months];

  if (years + 100 > INT_MAX || years + 100 < INT_MIN)
    return false;

  destination.m_year = years + 100;
  destination.m_month = months + 2;

  if (destination.m_month >= 12) {
    destination.m_month -= 12;
    ++destination.m_year;
  }

  destination.m_monthDay = remdays + 1;
  destination.m_weekDay = wday;
  destination.m_yearDay = yday;
  destination.m_year += 1900;

  destination.m_hour = remsecs / 3600;
  destination.m_min = remsecs / 60 % 60;
  destination.m_sec = remsecs % 60; */

  return true;
}
} // namespace stl::time

stl::stream::OStreamBase&
operator<<(stl::stream::OStreamBase& stream, const stl::time::TimeValue& value)
{
  stream << value.baseValue();
  return stream;
}

stl::stream::OStreamBase&
operator<<(stl::stream::OStreamBase& stream, const stl::time::Nanos& value)
{
  stream << value.value() << " ns";
  return stream;
}

stl::stream::OStreamBase&
operator<<(stl::stream::OStreamBase& stream, const stl::time::Micros& value)
{
  stream << value.value() << " us";
  return stream;
}

stl::stream::OStreamBase&
operator<<(stl::stream::OStreamBase& stream, const stl::time::Millis& value)
{
  stream << value.value() << " ms";
  return stream;
}

stl::stream::OStreamBase&
operator<<(stl::stream::OStreamBase& stream, const stl::time::Sec& value)
{
  stream << value.value() << " s";
  return stream;
}

stl::stream::OStreamBase&
operator<<(stl::stream::OStreamBase& stream, const stl::time::Mins& value)
{
  stream << value.value() << " m";
  return stream;
}

stl::stream::OStreamBase&
operator<<(stl::stream::OStreamBase& stream, const stl::time::Hours& value)
{
  stream << value.value() << " h";
  return stream;
}

stl::stream::OStreamBase&
operator<<(stl::stream::OStreamBase& stream, const stl::time::DateTime& value)
{
  stream << value.monthDay() << '.' << value.month() << ". " << value.year()
         << ' ' << value.hours() << ':' << value.minutes();
  return stream;
}

template<>
stl::stream::OStreamBase&
stl::stream::OStreamBase::operator<<(const stl::time::TimeValue& value)
{
  *this << value.baseValue();
  return *this;
}

template<>
stl::stream::OStreamBase&
stl::stream::OStreamBase::operator<<(const stl::time::Nanos& value)
{
  *this << value.value() << " ns";
  return *this;
}

template<>
stl::stream::OStreamBase&
stl::stream::OStreamBase::operator<<(const stl::time::Micros& value)
{
  *this << value.value() << " us";
  return *this;
}

template<>
stl::stream::OStreamBase&
stl::stream::OStreamBase::operator<<(const stl::time::Millis& value)
{
  *this << value.value() << " ms";
  return *this;
}

template<>
stl::stream::OStreamBase&
stl::stream::OStreamBase::operator<<(const stl::time::Sec& value)
{
  *this << value.value() << " s";
  return *this;
}

template<>
stl::stream::OStreamBase&
stl::stream::OStreamBase::operator<<(const stl::time::Mins& value)
{
  *this << value.value() << " m";
  return *this;
}

template<>
stl::stream::OStreamBase&
stl::stream::OStreamBase::operator<<(const stl::time::Hours& value)
{
  *this << value.value() << " h";
  return *this;
}

template<>
stl::stream::OStreamBase&
stl::stream::OStreamBase::operator<<(const stl::time::DateTime& value)
{
  stl::String minutes = stl::string::toString(value.minutes());
  if (minutes.length() == 0U) {
    minutes.insert("0", 0U, 1U);
  }

  stl::String hours = stl::string::toString(value.hours());
  if (hours.length() == 0U) {
    hours.insert("0", 0U, 1U);
  }

  *this << value.monthDay() << '.' << value.month() << ". " << value.year()
        << ' ' << value.hours() << ':' << value.minutes();
  return *this;
}
