#include "stl/version.hpp"

#include "stl/iostream.hpp"
#include "stl/string.hpp"
#include "stl/string_conv.hpp"

namespace stl {
template<>
stream::OStreamBase&
stream::OStreamBase::operator<<(const stl::Version& version)
{
  *this << version.major() << '.' << version.minor() << '.'
        << version.revision();

  return *this;
}
}

namespace stl {
Version::Version()
  : m_major(1U)
  , m_minor(0U)
  , m_revision(0U)
{
}

Version::Version(const String& version)
{
  const bool parseRetv = parse(version);

  Assert(parseRetv);

  UNUSED_PARAM(parseRetv);
}

Version::Version(const MajorT major, const MinorT minor, const RevT revision)
  : m_major(major)
  , m_minor(minor)
  , m_revision(revision)
{
}

Version::Version(const Version& other)
  : m_major(other.m_major)
  , m_minor(other.m_minor)
  , m_revision(other.m_revision)
{
}

Version&
Version::operator=(const Version& other)
{
  m_major = other.m_major;
  m_minor = other.m_minor;
  m_revision = other.m_revision;

  return *this;
}

Version::Version(Version&& other) noexcept
  : m_major(other.m_major)
  , m_minor(other.m_minor)
  , m_revision(other.m_revision)
{
}

Version&
Version::operator=(Version&& other) noexcept
{
  m_major = other.m_major;
  m_minor = other.m_minor;
  m_revision = other.m_revision;

  return *this;
}

Version::MajorT
Version::major() const
{
  return m_major;
}

Version::MinorT
Version::minor() const
{
  return m_minor;
}

Version::RevT
Version::revision() const
{
  return m_revision;
}

Version::MajorT&
Version::major()
{
  return m_major;
}

Version::MinorT&
Version::minor()
{
  return m_minor;
}

Version::RevT&
Version::revision()
{
  return m_revision;
}

bool
Version::parse(const stl::String& string)
{
  SizeT numberLength = 0;
  SizeT offset = 0;
  bool result = false;

  const MajorT major =
    string::stringTo<MajorT>(offset, string, result, numberLength);
  offset += numberLength;

  Assert(numberLength <= string.length());
  Assert(((string.length() - numberLength) != 0U));
  Assert(string[offset] == '.');
  if ((numberLength == 0U) || ((string.length() - numberLength) == 0U) ||
      string[offset] != '.') {
    return false;
  }

  const MinorT minor =
    string::stringTo<MinorT>(++offset, string, result, numberLength);
  offset += numberLength;

  Assert(numberLength <= string.length());
  Assert(((string.length() - numberLength) != 0U));
  Assert(string[offset] == '.');
  if ((numberLength == 0U) || ((string.length() - numberLength) == 0U) ||
      string[offset] != '.') {
    return false;
  }

  const RevT rev =
    string::stringTo<RevT>(++offset, string, result, numberLength);
  if (numberLength == 0U) {
    return false;
  }

  m_major = major;
  m_minor = minor;
  m_revision = rev;

  return true;
}

String
Version::toString() const
{
  String output;

  output += string::toString(m_major);
  output += '.';
  output += string::toString(m_minor);
  output += '.';
  output += string::toString(m_revision);

  return output;
}

bool
Version::operator<(const Version& other) const
{
  if (m_major < other.m_major) {
    return true;
  } else if (m_major == other.m_major) {
    if (m_minor < other.m_minor) {
      return true;
    } else if (m_minor == other.m_minor) {
      return m_revision < other.m_revision;
    }
  }

  return false;
}

bool
Version::operator<=(const Version& other) const
{
  if (m_major < other.m_major) {
    return true;
  } else if (m_major == other.m_major) {
    if (m_minor < other.m_minor) {
      return true;
    } else if (m_minor == other.m_minor) {
      return m_revision <= other.m_revision;
    }
  }

  return false;
}

bool
Version::operator>(const Version& other) const
{
  if (m_major > other.m_major) {
    return true;
  } else if (m_major == other.m_major) {
    if (m_minor > other.m_minor) {
      return true;
    } else if (m_minor == other.m_minor) {
      return m_revision > other.m_revision;
    }
  }

  return false;
}

bool
Version::operator>=(const Version& other) const
{
  if (m_major > other.m_major) {
    return true;
  } else if (m_major == other.m_major) {
    if (m_minor > other.m_minor) {
      return true;
    } else if (m_minor == other.m_minor) {
      return m_revision >= other.m_revision;
    }
  }

  return false;
}

bool
Version::operator==(const Version& other) const
{
  return (m_major == other.m_major) && (m_minor == other.m_minor) &&
         (m_revision == other.m_revision);
}

bool
Version::operator!=(const Version& other) const
{
  return !operator==(other);
}
}